/*
#    $Id: ArgUtils.cpp,v 1.5 2018/02/25 06:44:52 gps Exp $
#
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <iostream>
#include <fstream>
#include <string.h>
#include <strings.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <map>
#include <stdlib.h>
#include <getopt.h>

#include "gpscommon.h"
#include "ArgUtils.h"

extern char *optarg;
extern int optind;

#if USE_STREAM
#include "sharcrt.h"
extern int numwriters;
extern writerinfo_t *writerThreads[MAXTHREADS];
#endif

int SetReceiverType(SOps *pOps, string argument);

string CONFIGFILE = "sharc.cfg";
#define WHITESPACE "\t "
#define TIMEUNIQUE "/"
#define TIMEOK "0123456789/-:"
#define INDEXOK "0123456789-,"

//  ===================================================================
//  NAME:		ShowMan
//
//  PURPOSE:	Print out options in Man Page format
//
//  RETURNS:	nothing
//
//  HISTORY:
//
//  ===================================================================
void ShowMan(){

	int i;
	debug(D_FLOW,"ShowMan: Entering...\n");
	cout << ".TH SHARC 1 \"August 2001\" USGS \"User Manuals\"";
	cout << endl << ".SH NAME";
	cout << endl << "sharc \\- Download GPS Receivers";

	cout << endl << ".SH SYNOPSIS";
	cout << endl << ".B sharc ";
	for (i = 0; myoptions[i].param[0] != NULL; i++) {
		cout << endl << ".B [--";
		for (int k = 0; (k < MAXPARAM) && (myoptions[i].param[k] != NULL); k++) {
			if (k != 0) cout << "|";
			cout << myoptions[i].param[k];
		}
		if (myoptions[i].description != NULL) {
			cout << endl <<  ".I " << myoptions[i].description;
		}
		cout << endl << ".B ]";
	}
	cout << endl << ".SH DESCRIPTION";

	cout << endl << ".SH OPTIONS";
	cout << endl;
	cout << endl << "\tAny argument that has a space or special character in it";
	cout << endl << "\tshould be enclosed in double quotes. It is a good idea to";
	cout << endl << "\tdo that to almost all arguments";
	cout << endl;

	for (i = 0; myoptions[i].param[0] != NULL; i++) {
		cout << endl << ".IP \"--";
		for (int k = 0; (k < MAXPARAM) && (myoptions[i].param[k] != NULL); k++) {
			if (k != 0) cout << "|";
			cout << myoptions[i].param[k];
		}
		if (myoptions[i].description != NULL) {
			cout << " " << myoptions[i].description << "\"";
		}
		if (myoptions[i].argumentdesc != NULL) {
			cout << endl << myoptions[i].argumentdesc;
		}
	}
	cout << endl;
	debug(D_FLOW,"ShowUsage: Exiting\n");

}

//  ===================================================================
//  NAME:		ShowUsage
//
//  PURPOSE:	print out help message from --help switch
//
//  RETURNS:	nothing
//
//  HISTORY:
//
//  ===================================================================
void ShowUsage(){

	int i;
	debug(D_FLOW,"ShowUsage: Entering...\n");
	cerr << "\tWelcome to Sharc!!! " << VERSION << endl;
	cerr << "\tAny argument that has a space or special character in it" << endl;
	cerr << "\tshould be enclosed in double quotes. It is a good idea to" << endl;
	cerr << "\tdo that to almost all arguments" << endl;
	cerr << "" << endl;

	for (i = 0; myoptions[i].param[0] != NULL; i++) {
		cerr << endl << "--";
		for (int k = 0; (k < MAXPARAM) && (myoptions[i].param[k] != NULL); k++) {
			if (k != 0) cerr << " | ";
			cerr << myoptions[i].param[k];
		}
		if (myoptions[i].description != NULL) {
			cerr << " " << myoptions[i].description;
		}
		cerr << endl;
		if (myoptions[i].argumentdesc != NULL) {
			int count = 0;
			for (unsigned k = 0; k < strlen(myoptions[i].argumentdesc); k++) {
				if (((count > 55) && (myoptions[i].argumentdesc[k] == ' ')) ||
					(myoptions[i].argumentdesc[k] == '\n')) {
					cerr << endl;
					count = 0;
					continue;
				}
				if (count == 0) cerr << "\t";
				cerr << myoptions[i].argumentdesc[k];
				count++;
			}
			cerr << endl;
		}
		cerr << endl;
	}
	debug(D_FLOW,"ShowUsage: Exiting\n");

}


//  ===================================================================
//  NAME:		DecodeArguments
//
//  PURPOSE:	Decode command line switches in argv array and put 
//				results in pOps and pComm structs
//
//  RETURNS:	returns 0 if all options were successfully read
//				exits with return code 1 if any option did not read correctly
//				It will print out all errors before exiting
//
//  HISTORY:
//
//  ===================================================================
int DecodeArguments(int argc, char **argv, SOps *pOps, SComm *pComm) {

	debug(D_FLOW,"DecodeArguments: Entering...\n");

	static struct option long_options[500];
	int i;
	int entry = 0;

#define MAXSWITCHES 1024
	int index[MAXSWITCHES];
	string myarg[MAXSWITCHES];
	for (i = 0; myoptions[i].param[0] != NULL; i++) {
		for (int k = 0; (k < MAXPARAM) && (myoptions[i].param[k] != NULL); k++) {
			long_options[entry].name = myoptions[i].param[k];
			long_options[entry].val = myoptions[i].val;
			long_options[entry].has_arg = myoptions[i].has_arg;
			long_options[entry].flag = NULL;
			entry++;
		}
	}
	long_options[entry].name = 0;
	long_options[entry].val = 0;
	long_options[entry].has_arg = 0;
	long_options[entry].flag = 0;

	int count = 0;
	for (count = 0; count < MAXSWITCHES; )
	{
		int option_index = 0;

		int c = getopt_long(argc, argv, "", long_options, &option_index);

		if (c == -1) break;
		if (c == SWITCH_configfile) {
			SetValues(c, optarg, pOps, pComm);
			continue;
		}
		if (c == SWITCH_datadir) {
			SetValues(c, optarg, pOps, pComm);
			continue;
		}
#if USE_STREAM
		if (c == SWITCH_stream) {
			SetValues(c, optarg, pOps, pComm);
			continue;
		}
#endif
		index[count] = c;
		if (optarg != NULL) myarg[count] = optarg;
		count++;
	}
	ReadConfig(pComm, pOps);
	for (int i = 0; i < count; i++) {
		SetValues(index[i], (char *) myarg[i].c_str(), pOps, pComm);
	}
	if (ArgError(NULL) != 0) exit(FAILURE);
	debug(D_FLOW,"DecodeArguments: Exiting\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		ReadConfig
//
//  PURPOSE:	Read config file to get command line arguments from file
//				rather than from command line. There should be one switch
//				per line in file. Exits if ArgError is set.
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
void ReadConfig(SComm *pComm, SOps *pOps) {

	string tmp, tmp2, keyword, argument;
	ifstream pfd;
	char buffer[256];
	int i, index;

	debug(D_FLOW,"ReadConfig: Entering...\n");
	pfd.open(CONFIGFILE.c_str());
	if (!pfd) return;

	while (pfd.getline(buffer, 255)) {
		if (pfd.gcount() == 0) continue;
		tmp = buffer;
		// remove trailing and leading WHITESPACE
		size_t pos = tmp.find_first_not_of(WHITESPACE);
		if (pos != string::npos) tmp.erase(0,pos);

		pos = tmp.find_last_not_of(WHITESPACE);
		if (pos != string::npos) tmp.erase(pos + 1,tmp.length()-pos);

		// loop on comment line
		if (tmp[0] == '#') continue;

		// pull out and delete keyword
		keyword = tmp.substr(0, tmp.find_first_of(WHITESPACE));
		tmp.erase(0, tmp.find_first_of(WHITESPACE));
		tmp.erase(0, tmp.find_first_not_of(WHITESPACE));

		// get argument to keyword
		if (tmp[0] == '\"') {
			argument = tmp.substr(1,tmp.find_last_of("\"")-1);
		} else {
			argument = tmp.substr(0,tmp.find_first_of(WHITESPACE));
		}

		// check keyword against option list
		index = -1;
		for (i = 0; myoptions[i].param[0] != NULL; i++) {
			for (int k = 0; (k < MAXPARAM) && (myoptions[i].param[k] != NULL); k++) {
				tmp2 = myoptions[i].param[k];
				if (keyword == tmp2) {
					index = myoptions[i].val;
					break;
				}
			}
		}
		if (index == -1) {
			debug(D_WARN,"keyword = %s is not found\n",keyword.c_str());
			ArgError("Problem with config file entries\n");
			continue;
		}
		SetValues(index, (char *) argument.c_str(), pOps, pComm);
	}
	if (ArgError(NULL) != 0) exit(FAILURE);

	debug(D_FLOW,"ReadConfig: Exiting\n");
	return;
}


//  ===================================================================
//  NAME:		SetValues
//
//  PURPOSE:	Set the values given from the command line switches
//				or from the ReadConfig subroutine
//
//  RETURNS:	Always returns 0. However, it will set ArgError with a message
//				for each problem switch encountered. This allows all of 
//				the problems to be related back to the user instead of 
//				just the first problems encountered.
//
//  HISTORY:
//
//  ===================================================================
int SetValues(int c, char *argument, SOps *pOps, SComm *pComm) {
	string tmp;
	int i;
	int tmpint;
	int status;
	int start, numindex;
	char *pchar;
	string tokens[MAXTEMPLATESTRINGS];
	int numtokens;
#if USE_STREAM
	double tmpdouble;
	ifstream pfd;
#endif
	debug(D_FLOW,"SetValues: Entering...\n");
try {
	switch (c)
	{
		case SWITCH_debug:
			DEBUGLEVEL = strtol(argument,&pchar,10);
			if (*pchar != '\0') throw "Invalid Debug Level\n";
			debug(D_VALUES, "SetValues: Debug level set to: %d\n", DEBUGLEVEL);
			break;

		case SWITCH_debugfile:
			DEBUGFILE = argument;
			if (DEBUGFILE.compare("-") == 0) {
				DEBUGFD = STDOUT_FILENO;
			} else {
				DEBUGFD = -1;
				status = WriteFile(NULL, DEBUGFILE.c_str(), (unsigned char *) "", 0);
				if (status == FAILURE) DEBUGFD = 2; //Unable to open file, set to STDERR
			}
			debug(D_VALUES, "SetValues: DEBUGFILE is now %s\n", DEBUGFILE.c_str());
			break;

		case SWITCH_configfile:
			CONFIGFILE = argument;
			debug(D_VALUES, "SetValues: CONFIGFILE is now %s\n", CONFIGFILE.c_str());
			break;

		case SWITCH_port:
			if (pComm->isIP) break;
			numtokens = stringtokenize(argument,":",tokens,3);
			debug(D_VALUES, "SetValues: Port is now %s with %d tokens\n", argument,numtokens);
			if (numtokens == 2) {
				pComm->isIP = TRUE;
				pComm->ipHost = tokens[0];
				pComm->ipPort = atoi(tokens[1].c_str());
			} else {
#ifndef __CYGWIN32__
				struct stat buf;
				int portret;
				portret = stat(argument, &buf);
				if (portret == -1)
					throw "port does not exist\n";
#endif
			}
			pOps->getID = TRUE;
			pComm->port = argument;
			debug(D_VALUES, "SetValues: port is now %s\n", pComm->port.c_str());
			break;

		case SWITCH_databits:
			tmpint = strtol(argument,&pchar,10);
			if (*pchar != '\0')                      throw "Invalid argument to --databits\n";
			if (! ((tmpint == 7) || (tmpint == 8)) ) throw "Invalid dataBits (7 or 8 only)\n";
			pComm->dataBits = tmpint;
			debug(D_VALUES, "SetValues: dataBits is now %d\n", pComm->dataBits);
			break;

		case SWITCH_stopbits:
			tmpint = strtol(argument,&pchar,10);
			if (*pchar != '\0') throw "Invalid argument to --stopbits\n";
			if (! ((tmpint == 0) || (tmpint == 1)) ) throw "Invalid stopBits (0 or 1 only)\n";
			pComm->stopBits = tmpint;
			debug(D_VALUES, "SetValues: stopBits is now %d\n", pComm->stopBits);
			break;

		case SWITCH_parity:
			if (strlen(argument) > 1) throw "Invalid argument to --parity \n";
			switch (toupper(argument[0])) {
				case 'N': case 'E': case 'O': case 'S':
					pComm->parity = toupper(argument[0]);
					debug(D_VALUES, "SetValues: parity is now %c\n", toupper(pComm->parity));
					break;
				default:
					throw "Invalid argument to --parity\n";
			}
			break;

		case SWITCH_flow:
			tmp = "";
			pComm->serial_flow = -1;
			for (i = 0; i < (int) strlen(argument); i++)
				tmp += tolower(argument[i]);

			if (tmp.compare("none") == 0) pComm->serial_flow = SERIAL_FLOW_NONE;
			if (tmp.compare("hard") == 0) pComm->serial_flow = SERIAL_FLOW_HARD;
			if (tmp.compare("soft") == 0) pComm->serial_flow = SERIAL_FLOW_SOFT;
			if (tmp.compare("both") == 0) pComm->serial_flow = SERIAL_FLOW_BOTH;
			if (pComm->serial_flow < 0) throw "Invalid argument to --flow\n";

			debug(D_VALUES, "SetValues: serial flow is now %s\n",tmp.c_str());
			break;

		case SWITCH_baud:
			pComm->baud = strtol(argument,&pchar,10);
			if (*pchar != '\0') throw "Invalid argument to --baud\n";
			switch(pComm->baud) {
				case 115200:
				case 57600:
				case 38400:
				case 19200:
				case 9600:
				case 4800:
				case 2400:
				case 1200:
				case 600:
				case 300:
				case 110:
					break;
				default:
					throw "Invalid baud rate";
					break;
			}
			debug(D_VALUES, "SetValues: baud is now %d\n", pComm->baud);
			break;

		case SWITCH_login:
			tmp = argument;
			numtokens = stringtokenize(tmp, ",", tokens, 2);
			if (numtokens != 2)
				throw "Invalid argument to --login";
			pComm->login = tokens[0];
			pComm->password =  tokens[1];
			debug(D_VALUES, "SetValues: login is %s, password is %s\n", pComm->login.c_str(), pComm->password.c_str());
			break;

		case SWITCH_minbps:
			tmp = argument;
			numtokens = stringtokenize(tmp, ",", tokens, 2);
			switch (numtokens) {
				case 2:
					pComm->minbps_secs = strtol(tokens[1].c_str(),&pchar,10);
					debug(100,"minbps_secs is now %d\n",pComm->minbps_secs);
					if (*pchar != '\0') throw "Invalid argument to --minbps\n";
				case 1:
					pComm->minbps = strtol(tokens[0].c_str(),&pchar,10);
					debug(100,"minbps is now %d\n",pComm->minbps);
					if (*pchar != '\0') throw "Invalid argument to --minbps\n";
					break;
				default:
					throw "Invalid argument to --minbps\n";
					break;
			}
			break;

		case SWITCH_connectvia:
			tmp = "";
			for (i = 0; i < (int) strlen(argument); i++)
				tmp += tolower(argument[i]);
			pComm->isFW = pComm->isMC = pComm->isDC = FALSE;
			if (tmp == "freewave") pComm->isFW = TRUE;
			if (tmp == "modem")    pComm->isMC = TRUE;
			if (tmp == "direct")   pComm->isDC = TRUE;
			if (! (pComm->isFW || pComm->isMC || pComm->isDC))
				throw "Invalid argument to connectvia (direct, modem, freewave)\n";

			debug(D_VALUES, "SetValues: connectvia is set to %s\n", tmp.c_str());
			break;

		case SWITCH_freewaveslave:
			pComm->isFW = TRUE;
			pComm->isMC = FALSE;
			pComm->isDC = FALSE;
			pComm->modemDial = argument;
			if (pComm->modemDial.find_first_not_of("0123456789") != string::npos)
				throw "Freewave number must only contain digits\n";

			debug(D_VALUES, "SetValues: modemDial is set to %s\n", pComm->modemDial.c_str());
			break;

		case SWITCH_phonenumber:
			numtokens = stringtokenize(argument,":",tokens,3);
			if (numtokens == 2) {
				pComm->isIP = TRUE;
				pComm->ipHost = tokens[0];
				pComm->ipPort = atoi(tokens[1].c_str());
				pOps->getID = TRUE;
				pComm->port = argument;
				debug(D_VALUES, "SetValues: Found IP address, Overriding port to %s:%d\n", pComm->ipHost.c_str(), pComm->ipPort);
			} else {
				if (pComm->isDC) {
					pComm->isDC = FALSE;
					pComm->isMC = TRUE;
					pComm->isFW = FALSE;
				}
				pComm->modemDial = argument;
				if (pComm->modemDial.find_first_not_of("-,#*0123456789") != string::npos)
					throw "Phone number must only contain digits and -,#*\n";

				debug(D_VALUES, "SetValues: modemDial is set to %s\n", pComm->modemDial.c_str());
			}
			break;

		case SWITCH_repeater:
			pComm->fwUsesRepeater = TRUE;
			tmp = argument;
			if (tmp.length() == 1) {
				pComm->fwRepeaterCBEntry = tmp;
			} else if (tmp.length() == 7) {
				pComm->fwRepeater1 = tmp.substr(0,7);
			} else if ((tmp.length() == 15) && (tmp[7] == ',')) {
				pComm->fwRepeater1 = tmp.substr(0,7);
				pComm->fwRepeater2 = tmp.substr(8,7);
			} else
				throw "Freewave repeater argument makes no sense\n";
			debug(D_VALUES, "SetValues: freewave callbook entry is set to %s\n", pComm->fwRepeaterCBEntry.c_str());
			debug(D_VALUES, "SetValues: freewave repeater1 is set to %s\n", pComm->fwRepeater1.c_str());
			debug(D_VALUES, "SetValues: freewave repeater1 is set to %s\n", pComm->fwRepeater2.c_str());
			break;

		case SWITCH_initstring:
			pComm->modemInit = argument;
			debug(D_VALUES, "SetValues: modemInit is set to %s\n", pComm->modemInit.c_str());
			break;

		case SWITCH_verifycode:
			pOps->verifySiteCode = TRUE;
			pOps->siteCodeToVerify = argument;
			pOps->getFAT = TRUE;
			pOps->getID = TRUE;
			DEBUGOUT = argument;
			DEBUGOUT += ":";
			debug(D_VALUES, "SetValues: sitecode is set to %s\n", pOps->siteCodeToVerify.c_str());
			break;

		case SWITCH_fatfile:
			pOps->FATFileName = argument;
			pOps->getFAT = TRUE;
			pOps->getID = TRUE;
			pOps->writeFAT = TRUE;
			unlink(pOps->FATFileName.c_str());
			debug(D_VALUES, "SetValues: FATFileName is set to %s\n", pOps->FATFileName.c_str());
			break;

		case SWITCH_sumfile:
			pOps->SUMFileName = argument;
			SUMFILE = argument;
			WriteFile(NULL, SUMFILE.c_str(), (unsigned char *) "", 0);
			debug(D_VALUES, "SetValues: SUMFILE is set to %s\n", pOps->SUMFileName.c_str());
			break;

		case SWITCH_getid:
			pOps->getID = TRUE;
			break;

		case SWITCH_retries:
			pOps->retries = strtol(argument,&pchar,10);
			if (*pchar != '\0') throw "Invalid argument to --retries\n";
			break;

		case SWITCH_download:
		case SWITCH_purge:

			pOps->getFAT = TRUE;
			pOps->getID = TRUE;

			tmp = "";
			for (i = 0; i < (int) strlen(argument); i++)
				tmp += tolower(argument[i]);

			if (tmp.compare("all") == 0) {
				if (c == SWITCH_download) pOps->downLoadAll = TRUE;
				if (c == SWITCH_purge)    pOps->purgeAll    = TRUE;
				break;
			}

			pOps->fileOps[pOps->numFileOps] = new SFileOps;
			InitFileOpsStruct(pOps->fileOps[pOps->numFileOps]);
			if (c == SWITCH_download)
				pOps->fileOps[pOps->numFileOps]->downLoad = TRUE;
			else
				pOps->fileOps[pOps->numFileOps]->purge = TRUE;

			numindex = 0;
			status = DecodeIndexTimeRange(pOps, argument,
						pOps->fileOps[pOps->numFileOps]->indexlist,
						&numindex,
						pOps->fileOps[pOps->numFileOps]->fileStart,
						pOps->fileOps[pOps->numFileOps]->fileEnd,
						&pOps->fileOps[pOps->numFileOps]->session);

			if (status == FAILURE) throw "Invalid return from DecodeIndexTimeRange\n";

			status = SUCCESS;
			pOps->fileOps[pOps->numFileOps]->indexlistnum = numindex;
			if (numindex == -1) status = FAILURE;
			if (pOps->fileOps[pOps->numFileOps]->fileStart->set() !=
					pOps->fileOps[pOps->numFileOps]->fileEnd->set()) {
							 status = FAILURE; 
			}

			if (status == FAILURE) {
				delete pOps->fileOps[pOps->numFileOps]->fileStart;
				delete pOps->fileOps[pOps->numFileOps]->fileEnd;
				delete pOps->fileOps[pOps->numFileOps];
				throw "Cannot decode download arguments\n";
			}

			// time ranges supplied break;
			// indexlist should already be filled in.
			if (pOps->fileOps[pOps->numFileOps]->fileStart->set() == 1) {
				pOps->numFileOps++;
				break;
			}

			// No time range was specified.
			// created class instances for each index requested.
			start = pOps->numFileOps;
			pOps->fileOps[pOps->numFileOps]->index = pOps->fileOps[start]->indexlist[0] - 1;
			pOps->numFileOps++;

			for (i = 1; i < numindex; i++) {

				pOps->fileOps[pOps->numFileOps] = new SFileOps;
				InitFileOpsStruct(pOps->fileOps[pOps->numFileOps]);
				if (c == SWITCH_download)
					pOps->fileOps[pOps->numFileOps]->downLoad = TRUE;
				else
					pOps->fileOps[pOps->numFileOps]->purge = TRUE;

				pOps->fileOps[pOps->numFileOps]->index = pOps->fileOps[start]->indexlist[i] - 1;
				pOps->fileOps[pOps->numFileOps]->session = pOps->fileOps[start]->session;
				pOps->numFileOps++;

			}
			break;

		case SWITCH_joinfiles:
			pOps->joinFiles = TRUE;
			break;

		case SWITCH_session:
			int hours;
			long start,stop;
			float interval;
			int elev;
			int minsv;
			elev = 0;
			minsv = 1;
			tmpint = sscanf(argument,"%ld,%ld,%f,%d,%d",&start,&stop,&interval,&elev,&minsv);
			if (tmpint == 5) {
				pOps->setSessions = TRUE;
				Receiver::AddSession(start,stop,interval,elev,minsv);
				tmpint = Receiver::GetNumSessions();
				debug(D_VALUES,"SetValues: There are now %d sessions\n",Receiver::GetNumSessions());
			} else {
				// only one of this type should be specified
				if (pOps->setSessions) throw "Trying to set too many sessions\n";

				tmpint = 0;
				tmp = argument;
				numtokens = stringtokenize(tmp, ",", tokens, 4);
				switch (numtokens) {
					case 4:
						minsv = strtol(tokens[3].c_str(),&pchar,10);
						if (*pchar != '\0') throw "Do not understand 4 argument of --session\n";
					case 3:
						elev = strtol(tokens[2].c_str(),&pchar,10);
						if (*pchar != '\0') throw "Do not understand 3 argument of --session\n";
					case 2:
						interval = strtod(tokens[1].c_str(),&pchar);
						if (*pchar != '\0') throw "Do not understand 2 argument of --session\n";

						hours = strtol(tokens[0].c_str(),&pchar,10);
						if (*pchar != '\0') throw "Do not understand 1 argument of --session\n";
						break;
					default:
						throw "Incorrect format for --sessions switch\n";
				}

				if ((24 % hours) != 0) throw "hours does not divide into 24 without remainder\n";

				pOps->setSessions = TRUE;
				for (int i = 0; i < 24/hours; i++) {
					start = i*hours*3600;
					stop = ((i+1)*hours*3600)-1;
					if (start != 0) {
						start += pOps->sessionOffset;
						if (start < 0) start = 0;
					}
					if (stop != 86399) {
						stop += pOps->sessionOffset;
						if (stop > 86399) stop = 86399;
					}
					Receiver::AddSession(start,stop,interval,elev,minsv);
					debug(D_VALUES,"SetValues: Setting Session for %d %d %f %d %d\n", start,stop,interval,elev,minsv);
					debug(D_VALUES,"SetValues: There are now %d sessions\n",Receiver::GetNumSessions());
				}
			}
			break;

		case SWITCH_offset:
			if (pOps->setSessions) throw "--offset should be used before session programming switch\n";
			pOps->sessionOffset = strtol(argument,&pchar,10);
			if (*pchar != '\0') throw "Offset is not an integer\n";
			break;

		case SWITCH_receivertype:
			if (SetReceiverType(pOps, argument) != 0) throw "SetValues: Could not set receiver type\n";
			debug(D_VALUES, "SetValues: ReceiverType is set to %s\n", argument);
			break;

		case SWITCH_filetemplate:
			pOps->FileTemplate = argument;
			debug(D_VALUES, "SetValues: FileTemplate is set to %s\n", pOps->FileTemplate.c_str());
			break;

		case SWITCH_filetimefrom:
			tmp = argument;
			tmp = "";
			for (i = 0; i < (int) strlen(argument); i++)
				tmp += tolower(argument[i]);

			numtokens = stringtokenize(tmp, " \t", tokens, MAXTEMPLATESTRINGS);
			if ((numtokens != 1) && (numtokens != 2)) throw "incorrect arguments to filetimefrom\n";

			if (numtokens == 1) tokens[1] = "start";

			i = -1;
			if (tokens[0].compare("request") == 0) {
				if (tokens[1].compare("start") == 0) {
					debug(D_VALUES, "SetValues: FileTimeFrom is REQUESTSTART\n");
					i = REQUESTSTART;
				}
				if (tokens[1].compare("stop") == 0) {
					debug(D_VALUES, "SetValues: FileTimeFrom is REQUESTSTOP\n");
					i = REQUESTSTOP;
				}
			}
			if (tokens[0].compare("actual") == 0) {
				if (tokens[1].compare("start") == 0) {
					debug(D_VALUES, "SetValues: FileTimeFrom is ACTUALSTART\n");
					i = ACTUALSTART;
				}
				if (tokens[1].compare("stop") == 0) {
					debug(D_VALUES, "SetValues: FileTimeFrom is ACTUALSTOP\n");
					i = ACTUALSTOP;
				}
			}

			if (i == -1) throw "something wrong with arguments to filetimefrom\n";

			pOps->FileTimeFrom = i;
			break;

		case SWITCH_strings:
			tmp = argument;
			numtokens = stringtokenize(tmp, " \t", tokens, MAXTEMPLATESTRINGS);
			for (i = 0; i < numtokens; i++) {
				if (pOps->numTemplateStrings > MAXTEMPLATESTRINGS) break;
				pOps->TemplateStrings[pOps->numTemplateStrings++] = tokens[i];
			}
			debug(D_VALUES,"SetValues: There are now %d strings\n", pOps->numTemplateStrings);
			break;

		case SWITCH_datadir:
			pOps->DataDir = argument;
			WriteFile(pOps->DataDir.c_str(), NULL, NULL, 0);
			break;

		case SWITCH_resumedir:
			pComm->resumeDir = argument;
			char buf[100];
			if (TestDirectory(pComm->resumeDir.c_str(), buf) != SUCCESS) {
				debug(D_ERR,"SetValues: TestDirectory returned: %s\n",buf);
				throw "Resume directory does not exist\n";
			}
			debug(D_VALUES,"SetValues: Setting ZModem Resume dir to %s\n", pComm->resumeDir.c_str());
			break;

		case SWITCH_forcesave:
			if (pComm->resumeDir == "") throw "Must specify --resumedir before --forcesave\n";
			pComm->forceSave = TRUE;
			debug(D_VALUES,"SetValues: Setting Force Save for all ZMODEM downloads\n");
			break;

		case SWITCH_forceterminal:
			pOps->forceTerminal = true;
		case SWITCH_terminal:
			pOps->useTerminal = true;
			break;

		case SWITCH_multtimeout:
			tmpint = strtol(argument,&pchar,10);
			if (*pchar != '\0') throw "Invalid argument to --multtimeout\n";
			pComm->timeoutMultiplier = tmpint;
			break;

		case SWITCH_maxerror:
			tmpint = strtol(argument,&pchar,10);
			if (*pchar != '\0') throw "Invalid argument to --maxerror\n";
			pComm->maxTransferErrors = tmpint;
			break;

		case SWITCH_help:
			ShowUsage();
			exit(SUCCESS);
			break;

		case SWITCH_manpage:
			ShowMan();
			exit(SUCCESS);
			break;

// Real time stuff
#if USE_STREAM
		case SWITCH_stream:
			tmp = argument;
			numtokens = stringtokenize(tmp,",",tokens,4);
			if (numtokens != 2) throw "Problem with --stream switch\n";

			pOps->streamSample = rnd(strtod(tokens[0].c_str(),NULL), -1);
			pOps->streamPort = tokens[1][0];
			pOps->stream = TRUE;
			pOps->streamType = DATATYPE_BINEX;
			debug(D_VALUES,"SetValues: Sample set to %f\n",pOps->streamSample);
			break;

		case SWITCH_mount:
			tmp = argument;
			numtokens = stringtokenize(tmp,":",tokens,3);
			if (numtokens < 2) throw "Problem with --mount switch\n";
			if (numtokens > 3) throw "Problem with --mount switch\n";
			pOps->isNTRIP = TRUE;
			pOps->ntripMount = tokens[0];
			// if 2 args, will be set to Mount. If 3 args, will be set to middle arg.
			pOps->ntripUser = tokens[numtokens-2];
			// Always last arg
			pOps->ntripPassword = tokens[numtokens-1];
			pOps->forceStream = TRUE;
			debug(D_VALUES,"SetValues: Connecting to %s:%s:%s\n",pOps->ntripMount.c_str(), pOps->ntripUser.c_str(), pOps->ntripPassword.c_str());
			break;

		case SWITCH_datatype:
			tmp = "";
			for (i = 0; i < (int) strlen(argument); i++)
				tmp += tolower(argument[i]);
			pOps->streamType = 0;
			if (tmp.compare("raw") == 0) pOps->streamType = DATATYPE_RAW;
			if (tmp.compare("dben") == 0) pOps->streamType = DATATYPE_DBEN;
			if (tmp.compare("binex") == 0) pOps->streamType = DATATYPE_BINEX;
			if (pOps->streamType == DATATYPE_UNKNOWN) throw "datatype not set properly for real time stream\n";
			break;

		case SWITCH_file:
			tmp = argument;
			numtokens = stringtokenize(tmp,",",tokens,3);
			if ((numtokens != 2) && (numtokens != 3)) throw "Problem with --file switch\n";
			writerThreads[numwriters] = InitWriterInfo(pOps);
			writerThreads[numwriters]->file_template = tokens[0];
			writerThreads[numwriters]->seconds_in_file = (int) strtol(tokens[1].c_str(),NULL,10);
			if (writerThreads[numwriters]->seconds_in_file == 0 && errno == EINVAL) {
				throw "Unable to resolve integer as second argument to --file\n";
			}
			writerThreads[numwriters]->sample_rate = 0;
			writerThreads[numwriters]->writernum = 1 << numwriters;
			writerThreads[numwriters]->output_type = WRITE_FILE;
			if (numtokens == 3) {
				tmpdouble = rnd(strtod(tokens[2].c_str(),NULL), -1);
				writerThreads[numwriters]->sample_rate = static_cast<long>(tmpdouble * 1000);
			}
			numwriters++;
			break;

		case SWITCH_net:
			tmp = argument;
			numtokens = stringtokenize(tmp,":,",tokens,3);
			if ((numtokens != 2) && (numtokens != 3)) throw "--net host:port[,sample]\n";
			//create new writerthread
			writerThreads[numwriters] = InitWriterInfo(pOps);
			writerThreads[numwriters]->writernum = 1 << numwriters;
			writerThreads[numwriters]->output_type = WRITE_NET;
			writerThreads[numwriters]->hostname = tokens[0];
			writerThreads[numwriters]->port = (int)strtol(tokens[1].c_str(),NULL,10);
			writerThreads[numwriters]->sample_rate = 0;

			// sample rate passed
			if (numtokens == 2) {
				tmpdouble = rnd(strtod(tokens[2].c_str(),NULL), -1);
				writerThreads[numwriters]->sample_rate = static_cast<long>(tmpdouble * 1000);
			}

			numwriters++;
			break;

		case SWITCH_ntrip:
			tmp = argument;
			numtokens = stringtokenize(tmp,":@,",tokens,3);
			if ((numtokens != 4) && (numtokens != 5)) throw "--ntrip mount:password@host:port[,sample]\n";
			//create new writerthread
			writerThreads[numwriters] = InitWriterInfo(pOps);
			writerThreads[numwriters]->writernum = 1 << numwriters;
			writerThreads[numwriters]->output_type = WRITE_NTRIP;
			writerThreads[numwriters]->mount    = tokens[0];
			writerThreads[numwriters]->password = tokens[1];
			writerThreads[numwriters]->hostname = tokens[2];
			writerThreads[numwriters]->port = (int)strtol(tokens[3].c_str(),NULL,10);
			writerThreads[numwriters]->sample_rate = 0;

			// sample rate passed
			if (numtokens == 5) {
				tmpdouble = rnd(strtod(tokens[5].c_str(),NULL), -1);
				writerThreads[numwriters]->sample_rate = static_cast<long>(tmpdouble * 1000);
			}

			numwriters++;
			break;

		case SWITCH_stats:
			tmp = argument;
			numtokens = stringtokenize(tmp,":,",tokens,2);
			if (numtokens > 2) throw "--stats frequency,leapseconds\n";
			pOps->stats = strtol(tokens[0].c_str(),NULL,10);
			pOps->leapsec = strtol(tokens[1].c_str(),NULL,10);
			debug(D_VALUES,"SetValues: Setting Statistics output to %d seconds (leap = %d)\n",pOps->stats,pOps->leapsec);
			break;

		case SWITCH_forcestream:
			pOps->forceStream = TRUE;
			debug(D_VALUES,"SetValues: Setting Force Stream\n");
			break;

#ifdef RTCM_ENABLE
		case SWITCH_rtcm:
			pfd.open(argument);
			if (!pfd) throw "Unable to open RTCM input file\n";
			pfd.close();
			pOps->RTCMantennafile = argument;
			pOps->translateRTCM = TRUE;
			pOps->RTCMVersion = DATATYPE_UNKNOWN;
			break;

		case SWITCH_rtcmv22:
			if (! pOps->translateRTCM) {
				throw "Need to use --rtcm <antennafile> swtich prior to this switch\n";
			}
			pOps->RTCMVersion = DATATYPE_RTCM22;
			break;
		case SWITCH_rtcmv24:
			if (! pOps->translateRTCM) {
				throw "Need to use --rtcm <antennafile> swtich prior to this switch\n";
			}
			pOps->RTCMVersion = DATATYPE_RTCM24;
			break;
		case SWITCH_rtcmv30:
			if (! pOps->translateRTCM) {
				throw "Need to use --rtcm <antennafile> swtich prior to this switch\n";
			}
			pOps->RTCMVersion = DATATYPE_RTCM30;
			break;
#endif


#if 0
		case SWITCH_exec:
		case SWITCH_pipe:
			throw "--exec,--pipe: These would be cool if they worked :)\n";
			break;
#endif
#endif
		case SWITCH_version:
		  cerr << VERSION << endl;
			exit(SUCCESS);
			break;

		default:
			throw "I do not understand one of the switches\n";
			break;
	} // end switch 
	}
	catch (const char *pmessage)
	{
		ArgError(pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"SetValues: Exiting\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		DecodeIndexTimeRange
//
//  PURPOSE:	Decode the argument to the --download and --purge switches.
//				Each switch will accept up to three arguments which include
//				the list of indexes to download and/or the time range to download
//				followed by a session code of 1 character.
//
//  RETURNS:	-1 on error
//
//  HISTORY:
//
//  ===================================================================
int DecodeIndexTimeRange(SOps *pOps, char *argument, int indexes[], int *numindexes, gpstime *filestart, gpstime *fileend, char *session) {

	string decode = argument;
	size_t tmp;
	string tokens[4];
	int status = SUCCESS;

	debug(D_FLOW,"DecodeIndexTimeRange: Entering...\n");

	// remove whitespaces at beginning or end of line
	tmp = decode.find_first_not_of(WHITESPACE);
	if (tmp != string::npos) decode.erase(0,tmp);

	tmp = decode.find_last_not_of(WHITESPACE);
	if (tmp != string::npos) decode.erase(tmp + 1,decode.length()-tmp);

	try
	{
		int numtokens = 0;
		if (decode.length() == 0) throw "No Argument supplied to download switch\n";

		numtokens = stringtokenize(decode, WHITESPACE, tokens, 4);
		if (numtokens > 3) throw "More than three arguments supplied to Download switch\n";
		if (numtokens < 1) throw "No Argument supplied to download switch\n";

		switch (numtokens) {

			// simplest case
			//	Index list in first position
			//	Time range in second position
			//	Session code in third position
			case 3:
				debug(D_STD_V3,"Using three (3) argument version of DecodeIndexTimeRange\n");
				if (GetIndexList(tokens[0], indexes, numindexes) != SUCCESS)
					throw "Index list for 3 argument switch returned incorrect\n";
				if (GetTimeRange(tokens[1], filestart, fileend) != SUCCESS)
					throw "Time Range for 3 argument switch returned incorrect\n";
				if (GetSession(tokens[2], session) != SUCCESS)
					throw "something wrong with session parameter\n";
				break;

			// worst case:
			//	possibilities:
			//		index_list time_range
			//		index_list session
			//		time_range session
			case 2:
				debug(D_STD_V3,"Using two (2) argument version of DecodeIndexTimeRange\n");
				// check for time range in first place
				// this means "time_range session" supplied
				if (GetTimeRange(tokens[0], filestart, fileend) == SUCCESS) {
					if (GetSession(tokens[1], session) != SUCCESS)
						throw "looking for \"time_range session\", something wrong\n";
					break;
				}

				// check for time range in second place
				// this means "index_list time_range" supplied
				if (GetTimeRange(tokens[1], filestart, fileend) == SUCCESS) {
					if (GetIndexList(tokens[0], indexes, numindexes) != SUCCESS)
						throw "looking for \"index_list time_range\", something wrong\n";
					break;
				}

				// must be "index_list session"
				if (GetIndexList(tokens[0], indexes, numindexes) != SUCCESS)
					throw "looking for \"index_list session\", something wrong\n";
				if (GetSession(tokens[1], session) != SUCCESS)
					throw "looking for \"index_list session\", something wrong\n";
				break;


			// Either index list or time range was specified
			case 1:
				debug(D_STD_V3,"Using one (1) argument version of DecodeIndexTimeRange\n");
				if (GetTimeRange(tokens[0], filestart, fileend) != SUCCESS) {
					if (GetIndexList(tokens[0], indexes, numindexes) != SUCCESS) 
						throw "Looking for index_list or time_range, something wrong\n";
				}
				break;
		}
	}
	catch (const char *pmessage)
	{
		ArgError(pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"DecodeIndexTimeRange: Exiting...\n");
	return(status);
}

//  ===================================================================
//  NAME:		GetTimeRange
//
//  PURPOSE:	decode time range argument to --download switch
//
//  RETURNS:	SUCCESS or FAILURE. On success, the filestart and fileend
//				variables should be set correctly
//
//  HISTORY:
//
//  ===================================================================
int GetTimeRange(string token, gpstime *filestart, gpstime *fileend) {

	debug(D_FLOW,"GetTimeRange: Entering...\n");
	int status = SUCCESS;
	try
	{
		if (token.length() == 0)
			throw "Time range token is 0 length\n";

		// unique time element does not exist, can not be time string
		if (token.find_first_of(TIMEUNIQUE) == string::npos)
			throw "Unique time range character not found\n";

		// must only contain characters from TIMEOK define
		if (token.find_first_not_of(TIMEOK) != string::npos)
			throw "Incorrect character in time range string\n";
	}
	catch (const char *pmessage)
	{
		debug(D_FLOW,pmessage);
		return(FAILURE);
	}

	try {

		string time1 = token.substr(0,token.find('-'));
		string time2 = token.substr(token.find('-')+1);

		int start_hour = 0; int start_min  = 0; int start_sec  = 0;
		int stop_hour = 23; int stop_min  = 59; int stop_sec  = 59;

		int t11,t12,t13;
		int t21,t22,t23;
		int num_start = sscanf(time1.c_str(),"%d/%d/%d:%d:%d:%d",
									&t11,&t12,&t13,
									&start_hour,
									&start_min,
									&start_sec);
		int num_stop = sscanf(time2.c_str(),"%d/%d/%d:%d:%d:%d",
									&t21,&t22,&t23,
									&stop_hour,
									&stop_min,
									&stop_sec);

		if ((num_start < 2) || (num_stop < 2))
			throw "Time Argument format seems problematic\n";

		if ((num_start == 2) && (num_stop == 2)) {
			filestart->set_time(t11, (double)t12);
			fileend->set_time(t21, (double)t22);
		} else {
			if ((num_start < 3) || (num_stop < 3))
				throw "One side of time range has less than other\n";

			if ((t11 < 1930) || (t21 < 1930))
				throw "Year needs to be 4 digits\n";

			if ((t12 == 0) || (t22 == 0)) {					// using julian day
				double sday = t13+((double)((start_hour*3600)+(start_min*60)+start_sec)/86400.0);
				double eday = t23+((double)((stop_hour*3600)+(stop_min*60)+stop_sec)/86400.0);
				filestart->set_time(t11, sday, 'n');
				fileend->set_time(t21, eday, 'n');
			} else {
				//must be times in yyyy/mm/dd
				filestart->set_time(t11, t12, t13, start_hour, start_min, (double)start_sec);
				fileend->set_time(t21, t22, t23, stop_hour, stop_min, (double)stop_sec);
			}
		}
		if (*filestart > *fileend) 
			throw "End time is less than begin time\n";
	}
	catch (const char *pmessage)
	{
		ArgError(pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"GetTimeRange: Exiting...\n");
	return(status);
}


//  ===================================================================
//  NAME:		GetIndexList
//
//  PURPOSE:	decode index list argument to --download/--purge command
//				line switch.
//
//  RETURNS:	SUCCESS or FAILURE, indexes and numindexes should be set on
//				SUCCESS
//
//  HISTORY:
//
//  ===================================================================
int GetIndexList(string token, int indexes[], int *numindexes) {

	debug(D_FLOW,"GetIndexList: Entering...\n");
	int status = SUCCESS;

	if (token.find_first_not_of(INDEXOK) != string::npos) {
		debug(D_FLOW, "Unknown character in index range\n");
		return(FAILURE);
	}

	try
	{
		string tmp[100];
		int numtokens = stringtokenize(token,",", tmp, 100);
		for (int i = 0; i < numtokens; i++) {
			int istart, istop;
			if (tmp[i].find("-") != string::npos) {
				int scanned = sscanf(tmp[i].c_str(),"%d-%d",&istart,&istop);
				if (scanned != 2) throw "Problem with index range\n";
			} else {
				int scanned = sscanf(tmp[i].c_str(),"%d",&istart);
				if (scanned == 1) istop = istart;
			}

			if (istop < istart) 
				throw "End of index range is less than start\n";

			for (int i = istart; i<=istop; i++) {
				indexes[(*numindexes)++] = i;
			}
		}
	}
	catch (const char *pmessage)
	{
		ArgError(pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"GetIndexList: Exiting...\n");
	return(status);
}


//  ===================================================================
//  NAME:		GetSession
//
//  PURPOSE:	decode session argument to --download/--purge switch
//
//  RETURNS:	SUCCESS/FAILURE, sets session on SUCCESS
//
//  HISTORY:
//
//  ===================================================================
int GetSession(string token, char *session) {
	debug(D_FLOW,"GetSession: Entering...\n");
	if (token.length() > 1) {
		*session = ' ';
		return(FAILURE);
	}

	if (!isalnum(token[0])) {
		*session = ' ';
		return(FAILURE);
	}

	*session = token[0];
	debug(D_FLOW,"GetSession: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		ArgError
//
//  PURPOSE:	Store up errors during command line parsing
//
//  RETURNS:	returns 1 when setting an error. Passing null as error
//				string will cause it to return FAILURE if error exist or SUCCESS
//				if no error exists.
//
//  HISTORY:
//
//  ===================================================================
int ArgError(const char *error) {
	static string Error;
	if (error == NULL) {
		if (Error.length() > 0) {
			debug(D_FATALERROR,"%s",Error.c_str());
			return(FAILURE);
		} else {
			return(SUCCESS);
		}
	} else {
		Error += error;
	}
	return(FAILURE);
}

//  ===================================================================
//  NAME:		InitFileOpsStruct
//
//  PURPOSE:	Initialize FileOpsStruct
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
void InitFileOpsStruct(SFileOps *pFileOps) {
	debug(D_FLOW,"InitFileOpsStruct: Entering...\n");
	pFileOps->pReceiver = NULL;
	pFileOps->downLoad = FALSE;
	pFileOps->downLoadOK = -99;
	pFileOps->purge = FALSE;
	pFileOps->index = -1;
	pFileOps->fileStart = new gpstime();
	pFileOps->fileEnd = new gpstime();
	pFileOps->session = ' ';
	memset(pFileOps->indexlist, 0, sizeof(pFileOps->indexlist));
	pFileOps->indexlist[0]=-1;
	pFileOps->indexlistnum = 0;
	debug(D_FLOW,"InitFileOpsStruct: Exiting\n");
}

#if USE_STREAM
writerinfo_t * InitWriterInfo(SOps *pOps) {
// Globals for thread creation
//        pthread_t thread;
	writerinfo_t * wi = new writerinfo_t;
	wi->output_type = WRITE_UNKNOWN;
	wi->streamType = DATATYPE_UNKNOWN;
	wi->sample_rate = 0;
	wi->active = 0;
	wi->writernum = 0;
	wi->filedes[0] = -1;
	wi->filedes[1] = -1;
// file stuff
	wi->dir_template = "";
	wi->file_template = "";
	wi->seconds_in_file = 0;
// pipe?
	wi->pipe = FALSE;
// network stuff
	wi->hostname = "";
	wi->port = 0;
  // network ntrip stuff
	wi->mount = "";
	wi->password = "";
// exec stuff
	wi->program_name = "";
// translation stuff
#ifdef RTCM_ENABLE
	wi->outputRTCM = pOps->translateRTCM;
	wi->RTCMVersion = pOps->RTCMVersion;
#endif
	return(wi);
}

#endif

//  ===================================================================
//  NAME:		SetReceiverType
//
//  PURPOSE:	Set the receiver type in the Ops struct depending on 
//				the default type or the type passed via --type
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int SetReceiverType(SOps *pOps, string argument) {
	debug(D_FLOW,"SetReceiverType: Entering...\n");
	for (int i = 0; i < NUMRECEIVERTYPES ; i++) {
		if ( strcasecmp(argument.c_str(),rectype[i].name.c_str()) == 0 ) {
			pOps->ReceiverType = rectype[i].type;
			pOps->ReceiverTypeString = rectype[i].name;
			RECEIVERTYPE = rectype[i].type;
			RECEIVERSTRING = rectype[i].name;
			debug(D_STD_V,"SetReceiverType: Set type to %s\n",rectype[i].name.c_str());
			return(SUCCESS);
		}
	}
	debug(D_FLOW,"SetReceiverType: Exiting\n");
	return(FAILURE);
}
