/* 
#    $Id: Ashtech.cpp,v 1.1 2011/04/04 17:29:34 gps Exp $
#
#	 Ashtech.cpp -- This file will be the major interaction with
#                   an ashtech receiver/file
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <iomanip>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>

#include "gpscommon.h"
#include "Ashtech.h"


//  ===================================================================
//  NAME:		Ashtech::Ashtech
//
//  PURPOSE:	Set some default values
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
Ashtech::Ashtech () {

	debug(D_FLOW,"Ashtech::Ashtech: Entering...\n");
	pFATentry = NULL;
	pImageHead = NULL;

	usesFAT = TRUE;
	myIndex = -1;

	// Actual RFile image
	cimageFile = 0;
	imageFile = NULL;
	header = NULL;
	fatEntry = NULL;
	session = ' ';
	for (int i = 0; i < MAX_RCVR_FILES; i++) indexData[i] = NULL;
	debug(D_FLOW,"Ashtech::Ashtech: Exiting\n");

}

#define NOAUTOBAUD 1
int Ashtech::Login(Serial *pSerial, int direct, string login, string password) {

	long baud, currentspeed;
	debug(D_FLOW,"Ashtech::Login: Entering\n");
	if (!direct || NOAUTOBAUD) {
		debug(D_STD_V3,"Ashtech::Login: Not direct, no change to baud rate\n");
	} else {

		currentspeed = pSerial->getSpeed();
		baud = AutoBaud(pSerial);

		debug(D_CHECK,"Ashtech::Login: Checking %d == %d\n",baud, currentspeed);
		if (baud != currentspeed)
			return(SetPortInfo(pSerial, currentspeed));

	}
	debug(D_FLOW,"Ashtech::Login: Exiting\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Ashhtech::AutoBaud
//
//  PURPOSE:	find the current port and baud rate of the receiver
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
long Ashtech::AutoBaud(Serial *pSerial) {

	debug(D_FLOW,"Ashtech::AutoBaud: Entering...\n");

	long BaudRates[7];
	BaudRates[0] = 115200;
	BaudRates[1] = 57600;
	BaudRates[2] = 38400;
	BaudRates[3] = 19200;
	BaudRates[4] = 9600;
	BaudRates[5] = 4800;
	BaudRates[6] = 2400;

	long currentspeed = pSerial->getSpeed();

	char port;
	long speed;
	if (GetPortInfo(pSerial,&port,&speed) == SUCCESS) return(currentspeed);
	for (int i = 0; i <= 6; i++) {
		debug(D_STD_V,"Ashtech::AutoBaud: Trying speed %d\n",BaudRates[i]);
		pSerial->setSpeed(BaudRates[i]);
		if (GetPortInfo(pSerial,&port,&speed) == 0) {
			debug(D_STD_V,"Ashtech::AutoBaud: GetPortInfo succeeded for port %c and speed %d\n",port,speed);
			return(speed);
		}
	}
	debug(D_FLOW,"Ashtech::AutoBaud: Exiting...\n");
	return(-1);
}

//  ===================================================================
//  NAME:		Ashhtech::GetPortInfo
//
//  PURPOSE:	find the current port and baud rate of the receiver
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::GetPortInfo(Serial *pSerial, char *port, long *speed) {

	string PRT;
	size_t ret;
	int status = SUCCESS;
	char buffer[1025];
	debug(D_FLOW,"Ashtech::GetPortInfo: Entering...\n");
	*port = '0';
	*speed = -1;
	try {
		SendQuery(pSerial,"PRT");
		ret = pSerial->Readline(buffer,1024,1000);
		if (ret <= 0)
			throw "Ashtech::GetPortInfo: No Response from port request\n";

		PRT = buffer;
		if ((ret = PRT.find("PRT")) == string::npos)
			throw "Ashtech::GetPortInfo: No PORT String\n";

		string tokens[5];
		int numtokens = stringtokenize(PRT,",",tokens,5);
		if (numtokens != 4)
			throw "Ashtech::GetPortInfo: Incorrect return from Port request\n";

		if (tokens[0].compare("$PASHR") != 0)
			throw "Ashtech::GetPortInfo: Bad response from query\n";

		if (tokens[1].compare("PRT") != 0)
			throw "Ashtech::GetPortInfo: Bad response from query\n";

		*port = tokens[2][0];
		switch (tokens[3][0]) {
			case '0': *speed=300; break;
			case '1': *speed=600; break;
			case '2': *speed=1200; break;
			case '3': *speed=2400; break;
			case '4': *speed=4800; break;
			case '5': *speed=9600; break;
			case '6': *speed=19200; break;
			case '7': *speed=38400; break;
			case '8': *speed=57600; break;
			case '9': *speed=115200; break;
			default: throw "Ashtech::GetPortInfo: Unable to find baudrate\n";
		}
		debug(D_STD_V3,"Ashtech::GetPortInfo: Found PRT String %s",PRT.c_str());
	}
	catch (const char *pmessage)
	{
   		debug(D_WARN, pmessage);
		status = FAILURE;
   	}

	debug(D_FLOW,"Ashtech::GetPortInfo: Exiting...\n");
	return(status);

}

//  ===================================================================
//  NAME:		Ashtech::SetPortInfo
//
//  PURPOSE:	set the current port and baud rate of the receiver
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::SetPortInfo(Serial *pSerial, long baud) {

	char portc;
	long curbaud;
	char baudchar;
	int status = SUCCESS;
	debug(D_FLOW,"Ashtech::SetPortInfo: Entering...\n");

	try {
		if (GetPortInfo(pSerial, &portc, &curbaud) != 0)
			throw "Ashtech::SetPortInfo: Cannot get port info\n";

		debug(D_VALUES,"Ashtech::SetPortInfo: Baud to set = %d\n",baud);
		switch (baud) {
			case 300:    baudchar='0'; break;
			case 600:    baudchar='1'; break;
			case 1200:   baudchar='2'; break;
			case 2400:   baudchar='3'; break;
			case 4800:   baudchar='4'; break;
			case 9600:   baudchar='5'; break;
			case 19200:  baudchar='6'; break;
			case 38400:  baudchar='7'; break;
			case 57600:  baudchar='8'; break;
			case 115200: baudchar='9'; break;
			default: throw "Ashtech::SetPortInfo: Unrecognized baud rate\n";
		}

		char buffer[100];
		sprintf(buffer,"SPD,%c,%c",portc,baudchar);
		SendCommand(pSerial,buffer,FALSE,TRUE);
		pSerial->setSpeed(baud);
		char port;
		long speed;
		if (GetPortInfo(pSerial,&port,&speed) != 0) {
			AutoBaud(pSerial);
			throw "Ashtech::SetPortInfo: The speed has _NOT_ been set\n";
		}
		debug(D_STD_V2,"Ashtech::SetPortInfo: The speed has been set\n");
	}
	catch (const char *pmessage) {
		debug(D_WARN,pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"Ashtech::SetPortInfo: Exiting...\n");
	return(status);

}

//  ===================================================================
//  NAME:		Ashhtech::SetImageFile
//
//  PURPOSE:	set the imagefile contents and size.
//
//  RETURNS:	SUCCESS if image correctly copied. FAILURE if
//				memory cannot be allocated.
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::setImageFile(unsigned char *myImage, int fileSize) {

	debug(D_FLOW,"Ashtech::setImageFile: Entering...\n");

	if (imageFile != NULL)
		delete(imageFile);

	imageFile = new unsigned char[fileSize];
	if (imageFile == NULL) {
		debug(D_FATALERROR,"Unable to allocate memory for imageFile\n");
		return(FAILURE);
	}

	memcpy(imageFile, myImage, fileSize);
	cimageFile = fileSize;
	debug(D_FLOW,"Ashtech::setImageFile: Exiting\n");
	return(SUCCESS);

}

//  ===================================================================
//  NAME:		Ashtech::CompareSessions
//
//  PURPOSE:	Compare sessions found on receiver with sessions 
//				supplied by command line.
//
//  RETURNS:	SUCCESS on complete match
//				FAILURE otherwise
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::CompareSessions(SAshtechSession *tmpsess) {

	addToSumFile("\nSession Programming:\n RefDay = %d, Offset = %d, inuse = %d\n",
				tmpsess->refday, tmpsess->offset, tmpsess->inuse);
	addToSumFile("inuse start      end  interval mask minsv type\n");
	for (int i = 0; i < NUM_SESS; i++) {
		addToSumFile("%2d %c      %02d:%02d    %02d:%02d %03.1f      %03d  %03d    %03d\n",
				i,
				tmpsess->flag[i],
				tmpsess->start[i]/3600, (tmpsess->start[i]%3600)/60,
				tmpsess->end[i]/3600, (tmpsess->end[i]%3600)/60,
				tmpsess->interval[i], (int)tmpsess->elevmask[i],
				(int)tmpsess->minsv[i], (int)tmpsess->type[i]);
	}

	int activeSess = 0; 
	for (int i = 0; i < NUM_SESS; i++) {
		if (tmpsess->flag[i] == 'Y') activeSess++;
		if (tmpsess->flag[i] == 'A') activeSess++;
	}

	bool found = (activeSess == numSessions);
	for (int i = 0; (i < numSessions) && (found); i++) {
		long start    = GetSessionStartTime(i);
		long end      = GetSessionEndTime(i);
		float interval = GetSessionInterval(i);
		int elevmask = GetSessionElevationMask(i);
		int minsv    = GetSessionMinimumSV(i);
		found = FALSE;
		for (int j = 0; (j < NUM_SESS) && (!found); j++) {
			if (tmpsess->flag[j] == 'N') continue;
			if (start != tmpsess->start[j]) continue;
			if (end != tmpsess->end[j]) continue;
			if (interval != tmpsess->interval[j]) continue;
			if (elevmask != tmpsess->elevmask[j]) continue;
			if (minsv != tmpsess->minsv[j]) continue;
			debug(D_STD_V,"Ashtech::CompareSessions: Found Match for Session %d\n",j);
			found = TRUE;
		}
	}
	debug(D_FLOW,"Ashtech::CompareSessions: Exiting...\n");
	if (found) return(SUCCESS);
	return(FAILURE);

}

//  ===================================================================
//  NAME:		Ashtech::WriteFiles
//
//  PURPOSE:	Write the image file to disk. This may not include
//				all of the information the file needs. It will only
//				write the contents of imageFile.
//
//  RETURNS:	SUCCESS on writing the file successfully, FAILURE
//				if write failed.
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::WriteFiles() {

	return(WriteFile(NULL, (char *) GetFileName(myIndex).c_str(), imageFile, cimageFile));

}

//  ===================================================================
//  NAME:		Ashtech::CloseOpenFile
//
//  PURPOSE:	Close the Open file on the ashtech using pash command
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::CloseOpenFile(Serial *pSerial) {

	debug(D_FLOW,"CloseOpenFile: Entering...\n");
	debug(D_STD_V2,"CloseOpenFile:Closing Open File on Receiver\n");
	SendCommand(pSerial, "FIL,C", FALSE, FALSE);
	debug(D_FLOW,"CloseOpenFile: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Ashtech::DeleteFileByTime
//
//  PURPOSE:	Delete all of the file indexes that are completely contained
//				within the specified time range
//
//  RETURNS:	SUCCESS or FAILURE. This really should return the number of 
//				files deleted.
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::DeleteFileByTime(Serial *pSerial, gpstime *start, gpstime *stop) {
	int i,j;
	int numindex, indexes[MAX_RCVR_FILES];
	int deleteOK = FALSE;

	debug(D_FLOW,"Ashtech::DeleteFileByTime: Entering...\n");

	numindex = FindIndexesByTime(start, stop, indexes, TRUE);

	for (j = (MAX_RCVR_FILES-1); j >= 0; j--) {
		for (i = 0; i < numindex; i++) {
			if (indexes[i] != j) continue;
			debug(D_STD,"Ashtech::DeleteFileByTime: Deleteing file at index %d\n", indexes[i]);
			deleteOK = DeleteFileAtIndex(pSerial, indexes[i]);
		}
	}
	debug(D_FLOW,"DeleteFiles: Exiting\n");
	return(deleteOK);
}

//  ===================================================================
//  NAME:		Ashtech::Stream
//
//  PURPOSE:	Set the ashtech to start streaming
//
//  RETURNS:	SUCCESS/FAILURE
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::Stream(Serial *pSerial, int format, double rate, char port) {
#ifdef USE_STREAM
	debug(D_FLOW,"Ashtech::Stream: Entering...\n");
	if (format == DATATYPE_DBEN) {
		return (DBENStream(pSerial, rate, port));
	}
	if (format == DATATYPE_BINEX) {
		return (BinexStream(pSerial, rate, port));
	}
#endif
	return(FAILURE);

}
int Ashtech::DBENStream(Serial *pSerial, double rate, char port) {

	debug(D_FLOW,"Ashtech::DBENStream: Entering...\n");
	int status = SUCCESS;
	int rc;
	if (rate <= 0) {
		char buffer[101];
		sprintf(buffer,"OUT,%c",port);
		for (int i = 0; i < 10; i++) {
			pSerial->ClearBuffers();
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == SUCCESS) break;
		}
		return(rc);
	}

	try {
		char buffer[101];

		sprintf(buffer,"DOI,%0.1f",rate);
		rc = SendCommand(pSerial, buffer, FALSE, TRUE);
//		if (rc == FAILURE) throw "Error setting rate for DBEN\n";
		if (rc == FAILURE) {
			SendQuery(pSerial, "DOI");
			pSerial->Read(buffer, 99, 100);
			debug(1,"Ashthech::DBENStream: No ACK found for DOI\n");
			debug(1,buffer);
		}

		sprintf(buffer,"IDR,PASH,RPC");
		rc = SendCommand(pSerial, buffer, FALSE, TRUE);
		if (rc == FAILURE) {
			SendQuery(pSerial, "IDR");
			pSerial->Read(buffer, 99, 100);
			debug(1,"Ashthech::DBENStream: IDR string\n");
			debug(1,buffer);
		}

		sprintf(buffer,"OUT,%c,DBN,PBN,BIN",port);
		rc = SendCommand(pSerial, buffer, FALSE, TRUE);
		if (rc == FAILURE) {
			SendQuery(pSerial, "DOI");
			pSerial->Read(buffer, 99, 100);
			debug(1,"Ashthech::DBENStream: No ACK found for OUT\n");
			debug(1,buffer);
		}

		debug(D_STD,"Ashtech::DBENStream: Receiver has been set to stream\n");
	}
	catch (const char *pmessage) {
		debug(D_WARN,pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"Ashtech::DBENxStream: Exiting...\n");
	return(status);
}

int Ashtech::BinexStream(Serial *pSerial, double rate, char port) {

	debug(D_FLOW,"Ashtech::BNXStream: Entering...\n");
	int status = SUCCESS;
	int rc;
	if (rate <= 0) {
		char buffer[101];
		sprintf(buffer,"BNX,OFF,%c",port);
		for (int i = 0; i < 10; i++) {
			pSerial->ClearBuffers();
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == SUCCESS) break;
		}
		return(rc);
	}

	try {
		char buffer[101];
		for (char pp = 'A'; pp <= 'B'; pp++) {
			// turn on 0x0 type for port
			sprintf(buffer,"BNX,TYP,%d,%c,%.1f",0x0,pp,rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Error setting streaming for type 0x0\n";

			// turn on 0x1 type for port
			sprintf(buffer,"BNX,TYP,%d,%c,%.1f",0x1,pp,rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Error setting streaming for type 0x1\n";

			// turn on 0x7e type for port
			sprintf(buffer,"BNX,TYP,%d,%c,%.1f",0x7e, pp,rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Error setting streaming for type 0x7e\n";

			// turn on 0x0 type for port
			sprintf(buffer,"BNX,TYP,%d,%c,%.1f",0x7f, pp,rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Error setting streaming for type 0x7f\n";

		}

		if ((RxID.find("CJ12") != string::npos) || (RxID.find("CJ00") != string::npos)) {
			sprintf(buffer,"BNX,INT,%.1f",rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Receiver has _NOT_ been set to stream\n";
		}

		sprintf(buffer,"BNX,ON,%c",port);
		rc = SendCommand(pSerial, buffer, FALSE, TRUE);
		if (rc == FAILURE) throw "Receiver has _NOT_ been set to stream\n";
		debug(D_STD,"Ashtech::BinexStream: Receiver has been set to stream\n");
	}
	catch (const char *pmessage) {
		debug(D_WARN,pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"Ashtech::BinexStream: Exiting...\n");
	return(status);

}

//  ===================================================================
//  NAME:		Ashtech::DownloadReceiverID
//
//  PURPOSE:	Get the ID from the Ashtech Receiver. This will also
//				reset the receiver type depending on the ID returned
//
//  RETURNS:	Receiver ID if found, Empty string if not found
//
//  HISTORY:
//
//  ===================================================================
string Ashtech::DownloadReceiverID(Serial *pSerial) {

	char buffer[4096];
	int ret;
	size_t tmp;

	debug(D_FLOW,"Ashtech::DownloadReceiverID: Entering...\n");

	RxID = "";
	if (! pSerial->isConnected()) {
		debug(D_ERRMAJOR,"Not connected to the receiver\n");;
		return("");
	}

#if 0
	pSerial->Read(buffer, 4095, 100);
	for (int attempt = 0; (attempt < 1000) && (RxID == ""); attempt++, sleep(2)) {

		for (int i = 0; i < 10; i++) {
			pSerial->ClearBuffers();
			pSerial->ClearBuffers();
			pSerial->ClearBuffers();
			int rc;
			int c = pSerial->Read(buffer, 4095, 100);
			sprintf(buffer,"BNX,OFF,B");
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			printf("Read %d chars\n",c);
			if (rc == SUCCESS) break;
		}
		sprintf(buffer,"BNX,OFF,A");
		for (int i = 0; i < 10; i++) {
			pSerial->ClearBuffers();
			pSerial->ClearBuffers();
			pSerial->ClearBuffers();
			int rc;
			int c = pSerial->Read(buffer, 4095, 100);
			sprintf(buffer,"BNX,OFF,A");
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			printf("Read %d chars\n",c);
			if (rc == SUCCESS) break;
		}
	}
	exit(1);
#endif
	for (int attempt = 0; (attempt < 3) && (RxID == ""); attempt++) {
		pSerial->ClearBuffers();
		pSerial->ClearBuffers();
		debug(D_STD,"Querying Receiver ID (attempt %d):\n", attempt);
		SendQuery(pSerial, "RID");

		while ((ret = pSerial->Readline(buffer, 100, 1000)) != 0) {
			RxID += buffer;
		}
	}
	if (RxID == "") {
		debug(D_ERRMAJOR,"Unable to read Receiver ID\n");
		return(RxID);
	}

	if ((tmp = RxID.find("RID")) == string::npos) {
		debug(D_ERRMAJOR,"Incorrect return string\n");
		RxID = "";
		return(RxID);
	}
	addToSumFile(" Receiver ID: %s\n",RxID.c_str());
	RxID.erase(0, RxID.find(",", tmp)+1);

	if (RxID.find("ZM") == 0) {
			RECEIVERTYPE = rectype[RX_Z12].type;
			RECEIVERSTRING = rectype[RX_Z12].name;
			debug(D_STD_V,"Found Receiver type RX_Z12\n");
	}
	if (RxID.find("UZ") == 0) {
			RECEIVERTYPE = rectype[RX_MICROZ].type;
			RECEIVERSTRING = rectype[RX_MICROZ].name;
			debug(D_STD_V,"Found Receiver type RX_MICROZ\n");
	}

	debug(D_STD,"Found: %s\n",RxID.c_str());
	debug(D_FLOW,"Ashtech::DownloadReceiverID: Exiting\n");
	return(RxID);
}

//  ===================================================================
//  NAME:		Ashtech::SendQuery
//
//  PURPOSE:	Send the specified query to the receiver. Do not wait for
//				a return.
//
//  RETURNS:	returns SUCCESS if Write to serial port OK. FAILURE if
//				serial port is not connected.
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::SendQuery(Serial *pSerial, string QueryString) {

	int status = SUCCESS;
	debug(D_FLOW,"Ashtech::SendQuery: Entering...\n");

	try {
		if (! pSerial->isConnected())
			throw "Not connected to the receiver\n";

		string Send = "";
		if (QueryString.find("$PASHQ,") != 0) {
			Send += "$PASHQ,";
		}
		Send += QueryString;

		pSerial->ClearBuffers();
		debug(D_FLOW,"Ashtech::SendQuery: Sending %s\n",Send.c_str());
		Send += "\r\n";
		pSerial->Write((char *)Send.c_str());
	}
	catch (const char *pmessage)
	{
   		debug(D_WARN, pmessage);
		status = FAILURE;
	}

	debug(D_FLOW,"Ashtech::SendQuery: Exiting\n");
	return(status);
}

//  ===================================================================
//  NAME:		Ashtech::SendCommand
//
//  PURPOSE:	Send the specified PASH command to the receiver. Add the checksum
//				to the end if needed and wait for an ack if desired.
//
//  RETURNS:	if WaitAck is requested it will return SUCCESS/FAILURE depending
//				on return from receiver. If WaitAck is not requested it will
//				return SUCCESS if connected to the serial port and Write went OK.
//
//  HISTORY:
//
//  ===================================================================
int Ashtech::SendCommand(Serial *pSerial, string CommandString, int DoCheckSum, int WaitAck) {

	int status = SUCCESS;

	debug(D_FLOW,"Ashtech::SendCommand: Entering...\n");

	try {
		if (! pSerial->isConnected())
			throw "Not connected to the receiver\n";

		char buffer[101];
		string Send = "";
		if (CommandString.find("$PASHS,") != 0) {
			Send += "$PASHS,";
		}
		Send += CommandString;
		int Checksum = 0;

		if (DoCheckSum) {
			for (int i = 1; i < (int)Send.length(); i++) {
				Checksum ^= Send[i];
			}
			sprintf(buffer, "*%02X", Checksum);
			Send += buffer;
		}
		debug(D_FLOW,"Ashtech::SendCommand: Sending %s\n", Send.c_str());
		Send += "\r\n";
		pSerial->ClearBuffers();
		pSerial->Write((char *)Send.c_str());

		if (WaitAck) 
			status = CheckAck(pSerial);
	}
	catch (const char *pmessage)
	{
   		debug(D_WARN, pmessage);
		status = FAILURE;
	}

	debug(D_FLOW,"SendCommand: Exiting\n");
	return(status);
}

int Ashtech::CheckAck(Serial *pSerial) {
		string Check;
		char buffer[101];
		debug(D_FLOW,"Ashtech::WaitAck: Entering...\n");
		for (int ticks = 3000; ticks > 0; ticks -= 500) {
			if (pSerial->Readline(buffer,100,500) != 0) {
				Check += buffer;
				if (Check.find("ACK") != string::npos) return(SUCCESS);
				if (Check.find("NAK") != string::npos) return(FAILURE);
			}
			debug(D_FLOW,"Ashtech::CheckAck: Waiting for response - %d\r",ticks);
			ticks -= 500;
		}
		debug(D_WARN,"Ashtech::CheckAck: Timeout waiting for ACK/NAK\n");
		debug(D_FLOW,"Ashtech::CheckAck: Exiting...\n");
		return(FAILURE);
}

void Ashtech::AddToIndexBuffer(int index, unsigned char *in, size_t size) {
	if (index >= MAX_RCVR_FILES) return;
	if (indexData[index] != NULL) return;

	indexData[index] = new unsigned char[size];
	if (indexData[index] != NULL) return;
	memcpy(indexData[index], in, size);
	return;
}

#define PROMPT "Terminal-%s> "
int Ashtech::TerminalSession(Serial *pSerial) {

	debug(D_FLOW,"Ashtech::TerminalSession: Entering...\n");
	bool quit = false;
	char prompt[100];
	sprintf(prompt,PROMPT,GetSiteName(0).c_str());
	FILE *in = fdopen(1,"r");
	do {
		string myin = "";
		int query = -1;
		std::cout << prompt;
		char input[1024];
		fgets(input,1024,in);
		if (input[0] == '\\') {
			for (int i = 1; i < (int) strlen(input); i++) {
				if ((input[i] == '\r') || (input[i] == '\n')) break;
				myin += input[i];
			}
		} else {
			for (int i = 0; i < (int) strlen(input); i++) {
				if ((input[i] == '\r') || (input[i] == '\n')) break;
				myin += toupper(input[i]);
			}
		}
		printf("Input = %s\n",myin.c_str());
		if (myin.compare("QUIT") == 0) break;
		if (myin.find("$PASHS,") == 0) query = 0;
		if (myin.find("$PASHQ,") == 0) query = 1;
		switch (query) {
			case 0:
				SendCommand(pSerial,myin);
				break;
			case 1:
				SendQuery(pSerial,myin);
				break;
			default:
				printf("Unknown Command or Query: %s\n",myin.c_str());
				continue;
		}
		char buffer[1024];
		int c;
		while ((c = pSerial->Readline(buffer, 1024, 1000)) > 0) {
			std::cout << buffer;
		}
	} while (!quit);
	fclose(in);

	debug(D_FLOW,"Ashtech::TerminalSession: Exiting...\n");
	return(SUCCESS);
}

