/* Copyright 1997, 1998, 1999 University Corporation for Atmospheric Research

	This software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Author: Louis H. Estey <lou@unavco.ucar.edu>, Doug Hunt <dhunt@ucar.edu>
	Date:   July 1999
		part of the BINEX source code release, see also:

	http://www.unavco.ucar.edu/software/binex
*/

/**----------------------------------------------------------------------    
* @name       binex.c
* 
* BINEX file reading and writing utility library
*  
* @author     Lou Estey, Doug Hunt
* @since      1997
* @version    $Revision: 1.5 $
* @cvsinfo    $Id: binex.cpp,v 1.5 2012/02/21 18:52:53 gps Exp gps $
* @log        $Log: binex.cpp,v $
* @log        Revision 1.5  2012/02/21 18:52:53  gps
* @log        added checks for abnormal record lengths that could cause hanging
* @log
* @log        Revision 1.4  2012/02/10 21:52:28  gps
* @log        made comments identical to Keith's version
* @log
* @log        Revision 1.18  2012/02/07 01:55:11  kfstark
* @log        PRN issue
* @log
* @log        Revision 1.17  2012/02/04 02:48:56  kfstark
* @log        Several changes to Topcon Download
* @log        Support for GLONASS binex records to be decoded
* @log        Added delete capability to Topcon
* @log        Fixed a few bugs here and there.
* @log
* @log        Revision 1.16  2010/12/10 17:59:51  kfstark
* @log        Add Topcon download and streaming support.
* @log
* @log        Revision 1.15  2010/04/08 22:32:35  kfstark
* @log        minor fixes plus fix for deleted pointer that would hose things.
* @log
* @log        Revision 1.14  2010/03/05 18:10:30  kfstark
* @log        Added corrections sent in by Magnus Danielson
* @log        Fixed compilation issue with RTCM_ENABLE not included for reference to libgps.
* @log
* @log        Revision 1.13  2008/09/23 20:50:18  kfstark
* @log        Fix problem with RTCM23 vs RTCM24 since 23 doesn't actually exist
* @log
* @log        Revision 1.12  2008/07/17 21:14:58  kfstark
* @log        Major changes to stream handling for conversion to RTCM
* @log
* @log        Revision 1.11  2007/04/27 20:42:37  kfstark 
* @log        Fixed problem with SIGPIPE handling in network send
* @log        removed malloc dependency in sharc and sharcrt
* @log
* @log        Revision 1.10  2007/04/26 01:12:16  kfstark
* @log        Fixed problems with little endian binex decoding
* @log
* @log        Revision 1.9  2007/04/26 00:57:54  kfstark
* @log        More streaming changes dealing with endianness
* @log
* @log        Revision 1.8  2007/04/25 20:35:37  kfstark
* @log        Fixes for streaming and big endian issues
* @log
* @log        Revision 1.7  2007/04/25 05:32:04  kfstark
* @log        Fix problems with compiling on solaris
* @log        Get Big Endian RTCM conversion from binex working properly
* @log
* @log        Revision 1.6  2007/04/24 17:36:16  kfstark
* @log        Major changes to stream handling. Condensed a lot of crap^H^H^H^Hstuff.
* @log        Added the ability to use libgps for RTCM output for real time binex streams.
* @log
* @log        Revision 1.5  2007/02/08 19:17:08  kfstark
* @log        added disable-debug config argument
* @log        other comiplation issues
* @log        Changed signal handling in real time streaming
* @log
* @log        Revision 1.4  2006/04/05 17:10:03  kfstark
* @log        Upgrade time: Add --retries option. Add --minbps option (Thanks Mike).
* @log        This version will allow retries if the bps drops below the specified minbps for a specified amount of time.
* @log        Added exceptions: SerialException/SerialRetryable/SerialFail to allow for the downloads to kill themselves gracefully.
* @log        Also, a bunch of fixes for the ethernet stuff to get it working correctly. Deleted the dedicated read/write routines for the ethernet stuff so that it uses the Serial routines instead. This should simplify things and fix the timeout issues in the Ethernet side of things. Code is currently ifdef'd out.
* @log        Cleaned up a bunch of tab/space issues
* @log
* @log        Revision 1.3  2003/08/29 18:21:22  kfstark
* @log        Changes to make compilation under newer versions of gcc work.
* @log
* @log        Revision 1.2  2003/02/04 00:28:13  kfstark
* @log        Added md5 routines to be local
* @log        Fixed session programming problem
* @log        TIME_MULT for xmodem
* @log
* @log        Revision 1.1  2002/04/02 00:01:20  kfstark
* @log        Adding files for streaming binex support
* @log
* @log        Revision 1.11  2001/12/17 22:09:13  cvs
* @log        Added parse_c function to do fast perl parsing
* @log
* @log        Revision 1.10  2001/12/07 20:05:53  cvs
* @log        changes to Binex.pm to make parse friendlier
* @log
* @log        Revision 1.9  2001/04/27 22:34:21  huntd
* @log        Added format and write routines to the high level perl interface
* @log
* @log        Revision 1.8  2001/02/17 23:34:20  cvs
* @log        updated for lahey f95
* @log
* @log        Revision 1.7  2000/12/11 21:00:46  cvs
* @log        bug in md5 checksum
* @log
* @log        Revision 1.6  2000/12/04 21:39:29  cvs
* @log        added a call to read_body in read_header (!) to allow for check of out of sync records
* @log
* @log        Revision 1.5  2000/12/04 17:32:01  cvs
* @log        added Lous
* @log
* @log        Revision 1.4  2000/12/01 22:21:11  cvs
* @log        More fiddling with uint4 problem
* @log
* @log        Revision 1.3  2000/12/01 18:27:53  cvs
* @log        added routines to deal with unsigned int problems in fortran and perl
* @log
* @log        Revision 1.2  2000/06/14 21:36:36  cvs
* @log        Added binex tell function
* @log
* @log        Revision 1.1  2000/06/02 16:12:36  cvs
* @log        added binex_seek to version with underscores
* @log
* @log        Revision 1.5  2000/04/28 16:05:30  cvs
* @log        fixed bug in binex_append_mGFZI for big endian machines
* @log
* @log        Revision 1.4  2000/03/14 18:33:27  huntd
* @log        Dutil
* @log
* @log        Revision 1.3  2000/01/05 22:28:15  huntd
* @log        cleanup of directory
* @log
* @log        Revision 1.2  1999/12/01 23:12:01  huntd
* @log        changes to allow cosmicDoc to work
* @log
* @log        Revision 1.1.1.1  1999/12/01 22:59:42  huntd
* @log        First import of the binex module into cvs.  Version 0.12.
* @log
* @log        Revision 1.5  1999/09/20 15:56:44  huntd
* @log        Provisional check in to move source dir
* @log
*
* -----------------------------------------------------------------------*/

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "gpscommon.h"
#include "md5.h"


#if USE_STREAM
#define DBENMAXTYPES 5
#define DBEN_L1 0
#define DBEN_C1 1
#define DBEN_L2 2
#define DBEN_P1 3
#define DBEN_P2 4

#define i(code)		((uint1)obs.observation.observable.index[(code)])	/* obs observable index for the given code */
#define ok(code)	((code) < 0xff && i(code) < 0xff)			/* obs observable index has valid value */
#define mi(code)	((uint1)met.observation.observable.index[(code)])	/* met observable index for the given code */
#define mok(code)	((code) < 0xff && mi(code) < 0xff)			/* met observable index has valid value */

#ifdef RTCM_ENABLE
#ifdef HAVE_LIBGPS
extern "C" {
//#define ASHTECH_ENABLE
//#define BINARY_ENABLE
//#define NTPSHM_ENABLE
//#include "../../gpsd/gpsd-2.36/gpsd.h"
#include <gps.h>
//#include "../../gpsd/gpsd-2.36-kfs/gps.h"
}

#endif
#endif

#define PROTOTYPES 1
#include "binex.h"
#include "crc.h"


#include <bitset>
#include <iostream>
#include <limits>
#include <sstream>
#include <string>

template<typename T>
void show_bin(const T& value)
{
const std::bitset<std::numeric_limits<T>::digits + 1> bs(value);
for (unsigned int i = 0; i < (bs.size()-1); i++) {
	cout << bs[i];
}
cout << endl;
}

// Global variables
binexStat binex = { 0 };  // structure for storing binex state information
char system_RINEX_code[3] = { 'T', 'G', 'R' };

#define round(f, n) (floor(0.5 + (f)/(n))*(n)) /* round f to the nearest n */

/**----------------------------------------------------------------------    
* @name       open_binex
* 
* Opens the input file name and returns a binex Handle structure with
* status information and the file handle.  Yses memory passed in for
* the binexHandle.  Use this routine with FORTRAN and pass a sufficiently
* long character variable (say character*128) to use as the binexHandle.
*
* NOTE:  The file name variable must either be:
*        1) Null terminated, or
*        2) Have at least one space ' ' at the end 
*           of the file name.  This is necessary because FORTRAN does not
*           null terminate it's strings.
*
* @parameter  file_name The name of the input binex file
* @           mode      file mode for open:  'r' for reading or 'w' for writing
* @           handle    record full of file information
* @return     Status value:  0 = OK, non-zero = error opening file or 
*                            one file already open
* @example    open_binex ("hirate_YY.DDD_HH.MM.SS_gpsmet", 'r')
* ----------------------------------------------------------------------*/
void open_binex  (char * file_name, char * mode, binexHandle *handle, int *rc) {
  
  char *loc;
  int itmp;

  if (*mode != 'r' && *mode != 'w') {
    *rc = BADMODE;
    return;
  }

  // Check the endian_ness of the machine and determine if
  // the current record matches it.
  itmp = test_little_endian();
  if (itmp == UNKNOWN_TRUTH) {
    fprintf (stderr, "This machine does not support IEEE floating point numbers.  Troubles!\n");
    *rc = BADFLOATS;
    return;
  }
  binex.little_endian = itmp;

  // Default to writing in native endian-ness
  if (*mode == 'w') {
    handle->little_endian  = binex.little_endian;
    handle->bytes_reversed = 0;
  }

  // Add null to end of input string, make sure file name is null terminated
  loc = index (file_name, (int)' ');
  if (loc != NULL) *loc = '\0';
  loc = index (file_name, (int)'\0');
  if (loc == NULL) {
    *rc = BADFILENAME;
    return;
  }

  if ((handle->fh = fopen(file_name, mode)) == NULL) {
    *rc = ERROPEN;
    return;
  } 

  handle->mode = *mode;

  *rc = 0;
  return;

}

/**----------------------------------------------------------------------    
* @name       new_binex
* 
* Opens the input file name and returns the binex Handle structure.  This
* routine is like open_binex but it new's the handle instead of using
* memory passed in.  Use this routine for C or perl callers
*
* NOTE:  The file name variable must be null terminated!
*
* @parameter  file_name The name of the input binex file
* @           mode      file mode for open:  'r' for reading or 'w' for writing
* @return     new'ed and filled in binexHandle structure or NULL on error.
* ----------------------------------------------------------------------*/
binexHandle * new_binex  (int fd, const char * mode) {

  binexHandle *handle;
  int itmp;

  if (*mode != 'r' && *mode != 'w') {
    return NULL;
  }

  // Check the endian_ness of the machine and determine if
  // the current record matches it.
  itmp = test_little_endian();
  if (itmp == UNKNOWN_TRUTH) {
    fprintf (stderr, "This machine does not support IEEE floating point numbers.  Troubles!\n");
    return NULL;
  }
  binex.little_endian = itmp;

  handle = new binexHandle;
  
  if ((handle->fh = fdopen(fd, mode)) == NULL) {
	printf("Cannot open file descriptor\n");
    return NULL;
  } 

  // Default to writing in native endian-ness
  if (*mode == 'w') {
    handle->little_endian  = binex.little_endian;
    handle->bytes_reversed = 0;
  }

  handle->mode = *mode;

  return handle;

}

/**----------------------------------------------------------------------    
* @name       new_binex
* 
* Opens the input file name and returns the binex Handle structure.  This
* routine is like open_binex but it news the handle instead of using
* memory passed in.  Use this routine for C or perl callers
*
* NOTE:  The file name variable must be null terminated!
*
* @parameter  file_name The name of the input binex file
* @           mode      file mode for open:  'r' for reading or 'w' for writing
* @return     new'ed and filled in binexHandle structure or NULL on error.
* ----------------------------------------------------------------------*/
binexHandle * new_binex  (char * file_name, char * mode) {

  binexHandle *handle;
  char *loc;
  int itmp;

  if (*mode != 'r' && *mode != 'w') {
    return NULL;
  }

  // Check the endian_ness of the machine and determine if
  // the current record matches it.
  itmp = test_little_endian();
  if (itmp == UNKNOWN_TRUTH) {
    fprintf (stderr, "This machine does not support IEEE floating point numbers.  Troubles!\n");
    return NULL;
  }
  binex.little_endian = itmp;

  // Add null to end of input string, make sure file name is null terminated
  loc = index (file_name, (int)'\0');
  if (loc == NULL) {
    return NULL;
  }

  handle = new binexHandle;
  
  if ((handle->fh = fopen(file_name, mode)) == NULL) {
	printf("Cannot open file\n");
    return NULL;
  } 

  // Default to writing in native endian-ness
  if (*mode == 'w') {
    handle->little_endian  = binex.little_endian;
    handle->bytes_reversed = 0;
  }

  handle->mode = *mode;

  printf("I am returning after successfully opening the file\n");
  return handle;

}

/**----------------------------------------------------------------------    
* @name       close_binex
* 
* Closes the current binex file, allowing another to be opened
*
* @parameter  handle         The binexHandle handle to close
* @return     Status value:  0 = OK, non-zero = error closing file 
* @example    close_binex (handle, rc)
* ----------------------------------------------------------------------*/
void close_binex (binexHandle *handle, int *rc) {
  *rc = 0;

  if (fclose (handle->fh)) {
    *rc = ERRCLOSE;
  } 

  return;
}

/**----------------------------------------------------------------------    
* @name       free_binex
* 
* Closes the binex file associated with 'handle', freeing up resources
*
* @parameter  handle         The binexHandle handle to close
* @return     rc = 0 on success, non-zero on error
* ----------------------------------------------------------------------*/
int free_binex (binexHandle *handle) {

  int rc = 0;

  if (fclose (handle->fh)) {
    rc = ERRCLOSE;
  } 

  delete handle;

  return (rc);
}


/**----------------------------------------------------------------------    
* @name       set_endian
* 
* Set the endian-ness of a write-mode binex handle.  Use this if you want
* to write a binex record in a non-native endian-ness 
* (i.e. big endian from a little endian machine)
*
* @parameter  handle         The binexHandle handle to change
* @           little_endian  1 to set write to little endian, 0 to set to 
*                            big endian
* @example    close_binex ()
* ----------------------------------------------------------------------*/
void set_endian (binexHandle *handle, int *little_endian) {
  
  if (handle->mode == 'r') {
    printf ("Set_endian has no effect on read file handles\n");
    return;
  }

  handle->little_endian  = (bool)*little_endian;
  handle->bytes_reversed = (handle->little_endian != binex.little_endian);

  return;
}

  
/**----------------------------------------------------------------------    
* @name       seek_reverse
* 
* Seeks backward in the current (open) binex file one binex record.  Only
* works if the previous binex record is reverse-readable.  Returns
* an error and leaves the file handle where it started if this is not the
* case.
*
* @parameter  handle         The binexHandle handle to seek
* @return     Status value:  0 = OK, non-zero = error 
* @example    seek_foward (handle, rc)
* ----------------------------------------------------------------------*/
void seek_reverse (binexHandle *handle, int *rc) {
  unsigned char len_ubnxi[4];
  int len;
  int n;
  unsigned char sync;

  *rc = 0;

  // Read ending sync byte, verify it is from a reversible record
  if (fseek (handle->fh, -1, 1)) {
    *rc = BADSEEK;
    return;
  }

  sync = fgetc(handle->fh);

  if (sync == BINEX_FBetx_LE) {
    handle->little_endian = 1;
    handle->reversible    = 1;
  } else if (sync == BINEX_FBetx_BE) {
    handle->little_endian = 0;
    handle->reversible    = 1;
  } else {
#ifdef DEBUG
    fprintf (stderr, "seek_reverse:  incorrect sync byte = %d found at end of reversible record\n", sync);
#endif
    *rc = NOREVERSE;
    return;
  }

  if (binex.little_endian != handle->little_endian) {
    handle->bytes_reversed = 1;
  } else {
    handle->bytes_reversed = 0;
  }

  // Read ending length field
  if (fseek (handle->fh, -5, 1)) {
    *rc = BADSEEK;
    return;
  }

  if (fread ((void *)len_ubnxi, 4, 1, handle->fh) != 1) {
    *rc = BADREAD;
    return;
  }

  reverse_bytes((unsigned char *)len_ubnxi, 4);

  // Determine number of valid bytes in this UBNXI
  n = 0;
  while (((len_ubnxi[n] & 0x80) == 0x80) && n<4) n++;
  n++;
  
  // Fail if this UBNXI has too many high bits set
  if (n > 4) {
    *rc = BADLEN;
    fseek (handle->fh, 1, 1); // Put file pointer back at end of record
    return;
  }

  // Seek backwards to the beginning of this BINEX record
#ifdef DEBUG
  fprintf (stderr, "seek_reverse:  length of %x:%x:%x:%x read from reversible section (ubnxi)\n", 
	  len_ubnxi[0],
	  len_ubnxi[1],
	  len_ubnxi[2],
	  len_ubnxi[3]);
  fprintf (stderr, "seek_reverse:  length of ubnxi: %d\n", n);
#endif
  len = ubnxi_to_uint4 (handle, len_ubnxi, n);
#ifdef DEBUG
  fprintf (stderr, "seek_reverse:  length of %d read from reversible section\n", len);
#endif
    if (fseek (handle->fh, -1 * (n+len), 1)) {
    *rc = BADSEEK;
    return;
  }

  return;
}

/**----------------------------------------------------------------------    
* @name       seek_end
* 
* Seeks to the end of the current file.
*
* @parameter  handle         The binexHandle handle to seek
* @return     Status value:  0 = OK, non-zero = error 
* @example    seek_end (handle, rc)
* ----------------------------------------------------------------------*/
void seek_end (binexHandle *handle, int *rc) {
  *rc = 0;

  // Seek to end of file
  if (fseek (handle->fh, 0, 2)) {
    *rc = BADSEEK;
    return;
  }

  return;

}

/**----------------------------------------------------------------------    
* @name       seek_binex
* 
* General seeks in a binex file
*
* @parameter  handle         The binexHandle handle to seek
* @           offset         seek value
* @           whence         seek mode: 
* @                          0 = seek to this absolute location
* @                          1 = seek offset bytes from current location
* @                          2 = seek offset bytes from end of file (usually negative)
* @return     Status value:  0 = OK, non-zero = error 
* @example    seek_binex (handle, -20, 2, rc)
* ----------------------------------------------------------------------*/
void seek_binex (binexHandle *handle, int *offset, int *whence, int *rc) {
  *rc = 0;

  // Seek within file
  if (fseek (handle->fh, *offset, *whence)) {
    *rc = BADSEEK;
    return;
  }

  return;

}

/**----------------------------------------------------------------------    
* @name       tell_binex
* 
* Tells file pointer location in binex file
*
* @parameter  handle         The binexHandle handle to seek
* @           where          returned byte offset into file
* @example    tell_binex (handle, &where);
* ----------------------------------------------------------------------*/
void tell_binex (binexHandle *handle, int *where) {

  *where = (long)ftell(handle->fh);

  return;

}

/**----------------------------------------------------------------------    
* @name       read_record
* 
* @parameter  handle The binexHandle file handle to read from
* @           rec_id    The binex record ID read in (out)
* @           rec_len   The length of the record in bytes (out)
* @           reversible  1 = reverse-readable record, 0 = regular record
* @return     Status value:  0 = OK, non-zero = error finding or reading header
* @example    read_header (rec_id, rec_len);
* ----------------------------------------------------------------------*/
void read_header_no_check (binexHandle *handle, unsigned int *rec_id, unsigned int *rec_len, int *reversible, int *little_endian, int *rc) {

  unsigned char sync;
  unsigned char n = 0;   // length of ubnxis

  // Check for correct file handle
  if (fileno(handle->fh) < 0) {
	printf("The file isn't open\n");
    *rc = NOTOPEN;
    return;
  }

  while (1) {  // loop until correct record header/body read in

    // find the next sync byte
    while (1) {  // loop until sync byte found
      sync = fgetc(handle->fh);

      if (feof(handle->fh)) {
        *rc = ENDOFFILE;
        return;
      }

      if (sync == BINEX_F_stx_LE) {
        handle->little_endian = 1;
        handle->reversible    = 0;
        break;
      }
      if (sync == BINEX_F_stx_BE) {
        handle->little_endian = 0;
        handle->reversible    = 0;
        break;
      }
      if (sync == BINEX_FBstx_LE) {
        handle->little_endian = 1;
        handle->reversible    = 1;
        break;
      }
      if (sync == BINEX_FBstx_BE) {
        handle->little_endian = 0;
        handle->reversible    = 1;
        break;
      }

    }

    // set output values
    *reversible    = (int)handle->reversible;
    *little_endian = (int)handle->little_endian;

    if (binex.little_endian != handle->little_endian) {
      handle->bytes_reversed = 1;
    } else {
      handle->bytes_reversed = 0;
    }

    *rec_id  = read_ubnxi (handle, &(handle->hdr_bytes[0]), &n);
    if ( *rec_id != 127 && *rec_id != 0 && *rec_id != 1) {
      debug(D_VALUES, "read_header_no_check: rec_id = %d\n",*rec_id);
      continue;
    }

    handle->n_hdr_bytes = n;
    *rec_len = read_ubnxi (handle, &(handle->hdr_bytes[n]), &n);
    if ( *rec_len > 1000 ) {
      debug(D_VALUES, "read_header_no_check: rec_id = %d rec_len=%d\n",*rec_id, *rec_len);
      continue;
    }
    handle->n_len_bytes  = n;
    handle->n_hdr_bytes += n;

    break;
  }

  *rc = 0;
  return;

}



/**----------------------------------------------------------------------    
* @name       read_header
* 
* Searches for a sync byte in the current open file (must be open for reading)
* and then reads in the binex record header following.  It also verifies that
* a correct trailer exists after the body and that the CRC matches.  If not
* it keeps searching for a correct BINEX record until end-of-file is reached.
* This extra trouble is necessary in that some BINEX streams start in the middle
* of a record.
*
* @parameter  handle The binexHandle file handle to read from
* @           rec_id    The binex record ID read in (out)
* @           rec_len   The length of the record in bytes (out)
* @           reversible  1 = reverse-readable record, 0 = regular record
* @return     Status value:  0 = OK, non-zero = error finding or reading header
* @example    read_header (rec_id, rec_len);
* ----------------------------------------------------------------------*/
void read_header (binexHandle *handle, unsigned int *rec_id, unsigned int *rec_len, int *reversible, int *little_endian, int *rc) {

  unsigned char sync;
//  unsigned char buf[4];  // place for ubnxis
  unsigned char n = 0;   // length of ubnxis
  unsigned long nexttry = 0;   // saved index into file:  one past current sync byte
  unsigned long bodystart = 0; // saved index into file:  start of body
  unsigned char *rec;    // buffer to place binex body for sanity check
  int temprc;            // return code from read body

  // Check for correct file handle
#define NOTFIFO 1
#if NOTFIFO
  if (ftell(handle->fh) < 0) {
	printf("The file isn't open\n");
    *rc = NOTOPEN;
    return;
  }
#endif

  while (1) {  // loop until correct record header/body read in

    // find the next sync byte
    while (1) {  // loop until sync byte found
      sync = fgetc(handle->fh);

      if (feof(handle->fh)) {
        *rc = ENDOFFILE;
        return;
      }

      if (sync == BINEX_F_stx_LE) {
        handle->little_endian = 1;
        handle->reversible    = 0;
        break;
      }
      if (sync == BINEX_F_stx_BE) {
        handle->little_endian = 0;
        handle->reversible    = 0;
        break;
      }
      if (sync == BINEX_FBstx_LE) {
        handle->little_endian = 1;
        handle->reversible    = 1;
        break;
      }
      if (sync == BINEX_FBstx_BE) {
        handle->little_endian = 0;
        handle->reversible    = 1;
        break;
      }

    }

#if NOTFIFO
    nexttry = ftell(handle->fh);
#endif

    // set output values
    *reversible    = (int)handle->reversible;
    *little_endian = (int)handle->little_endian;

    if (binex.little_endian != handle->little_endian) {
      handle->bytes_reversed = 1;
    } else {
      handle->bytes_reversed = 0;
    }

    *rec_id  = read_ubnxi (handle, &(handle->hdr_bytes[0]), &n);
    handle->n_hdr_bytes = n;
    *rec_len = read_ubnxi (handle, &(handle->hdr_bytes[n]), &n);
    handle->n_len_bytes  = n;
    handle->n_hdr_bytes += n;

#if NOTFIFO
    bodystart = ftell(handle->fh);
#endif

    /* Now read the body and verify that the CRC is correct */
    rec = new unsigned char[*rec_len];
    read_body (handle, rec, rec_len, &temprc);
    delete rec;
  
#if NOTFIFO
    if (temprc == 0) {
      fseek(handle->fh, bodystart, SEEK_SET);  // set file pointer to start of body
      break;
    } else {
      fseek(handle->fh, nexttry, SEEK_SET);    // set file pointer past last identified sync byte
    }
#endif

  }

  *rc = 0;
  return;

}


/**----------------------------------------------------------------------    
* @name       read_body
* 
* Reads in the body of the rinex message into a buffer passed in
* by the caller.  Also reads in the checksum/CRC and verifies its 
* correctness.  Finally reads in any trailing bytes and complains if
* they are not correct.  
*
* After this routine is called, the various extract_* routines can be
* called to decode the contents of the buffer, dealing with endian
* issues if necessary.
*
* @parameter  handle The binexHandle file handle to read from
* @           rec   A buffer of the correct size to hold the record (in/out)
* @           len   The length of the buffer (got from read_header) (in)
* @return     Status value:  0 = OK, non-zero = error reading record or 
*                            checksum problem
* @example    read_body (buffer, len)
* ----------------------------------------------------------------------*/
void read_body (binexHandle *handle, unsigned char *rec, unsigned int *len, int *rc) {

  unsigned int  crc_len;       // length of computed CRC
  unsigned char comp_crc[16];  // computed CRC
  unsigned int  crc_offset = 0;
  unsigned char crc[16];
  unsigned char buf[4];
  unsigned int  trailer_len;
//  int i;
  int res;

  // Check for correct file handle
  if (fileno(handle->fh) < 0) {
  	fprintf(stderr, "File Handle is not open\n");
    *rc = NOTOPEN;
    return;
  }

  if (fread (rec, (size_t)*len, 1, handle->fh) != 1) {
  	fprintf(stderr, "Could not read record of size %d\n",*len);
    *rc = BADREAD;
    return;
  }

  crc_len = binex_crc (comp_crc, handle->hdr_bytes, rec, &crc_offset, 
		       handle->n_hdr_bytes, *len, handle->bytes_reversed);
  
  if (fread (crc, (size_t)crc_len, 1, handle->fh) != 1) {
  	fprintf(stderr, "Could not read CRC of size %d\n",crc_len);
    *rc = BADREAD;
    return;
  }
  
  for (unsigned int i=0;i<crc_len;i++){
    if (comp_crc[i] != crc[i]) {
      *rc = BADCRC;
      return;
    }
  }

  // Handle trailing bytes of reversible record
  if (handle->reversible) {
    
    res = fread (buf, (size_t)handle->n_len_bytes, 1, handle->fh);
    if (res != 1) {
  	fprintf(stderr, "Could not read Trailing bytes of reversible record\n");
      *rc = BADREAD;
      return;
    }
    reverse_bytes (buf, handle->n_len_bytes);
    trailer_len = ubnxi_to_uint4 (handle, buf, handle->n_len_bytes);

    // trailer_len is now equal to *len + n_len_bytes + n_hdr_bytes + crc_bytes
    // Since I don't have all that info here, I am just checking for a reasonable value here.
    // D. Hunt 12/1/99
    if (trailer_len <= *len) {
      *rc = BADTRAILER;
      return;
    }
    
    if (fread (buf, (size_t)1, 1, handle->fh) != 1) {
      *rc = BADREAD;
      return;
    }
 
    if (((buf[0] == BINEX_FBetx_LE) && handle->little_endian) ||
	((buf[0] == BINEX_FBetx_BE) && !handle->little_endian)) {
      *rc = 0;
      return;
    } else {
      *rc = BADTRAILER;
      return;
    }

  }

  *rc = 0;
  return;
  
}

/**----------------------------------------------------------------------    
* @name       write_binex
* 
* Writes a binex record, including header and checksum.  If 'reversible'
* is specified, write a reversible record.
*
* @parameter  handle The binexHandle file handle to read from
* @           rec   A buffer holding the record (in)
* @           len   The length of the record (in)
* @           type  The binex record type (in)
* @           reversible 1 = write reversible record, 0 = write regular record (in)
* @           bytesOut The number of bytes successfully written (out)
* @return     rc : 0 = OK, non-zero = error writing record
* @example    write_binex (&rec, &rec_len, &type, &reversible, &bytes_out, &rc)
* ----------------------------------------------------------------------*/
void write_binex (binexHandle *handle, unsigned char *rec, unsigned int *len, unsigned int *type, 
		  int *reversible, unsigned int *bytes_out, int *rc) {

  unsigned int  crc_len;       // length of computed CRC
  unsigned char crc[16];       // computed CRC
  unsigned char buf[8];
  unsigned int  trailer_len, header_len;
  unsigned int  offset;
  unsigned char sync;

  // Check for correct file handle
  if (ftell(handle->fh) < 0) {
    *rc = NOTOPEN;
    return;
  }

  handle->reversible     = *reversible;

  if        ( handle->little_endian &&  handle->reversible) {
    sync = BINEX_FBstx_LE;
  } else if ( handle->little_endian && !handle->reversible) {
    sync = BINEX_F_stx_LE;
  } else if (!handle->little_endian &&  handle->reversible) {
    sync = BINEX_FBstx_BE;
  } else if (!handle->little_endian && !handle->reversible) {
    sync = BINEX_F_stx_BE;
  }

  // Write binex sync byte
  if (fwrite (&sync, 1, 1, handle->fh) != 1) {
    *rc = BADWRITE;
    return;
  }
  (*bytes_out)++;
  
  // Write binex record ID and length
  offset = header_len = 0;
  header_len  = uint4_to_ubnxi(handle, buf, &offset, *type);
  header_len += uint4_to_ubnxi(handle, buf, &offset, *len);
  if (fwrite (buf, header_len, 1, handle->fh) != 1) {
    *rc = BADWRITE;
    return;
  }
  (*bytes_out) += header_len;
  
  // Write binex record 
  if (fwrite (rec, *len, 1, handle->fh) != 1) {
    *rc = BADWRITE;
    return;
  }
  (*bytes_out) += *len;
  
  // Write CRC
  offset  = 0;
  crc_len = binex_crc(crc, buf, (unsigned char *)rec, &offset, header_len, *len, handle->bytes_reversed); 
  if (fwrite (crc, crc_len, 1, handle->fh) != 1) {
    *rc = BADWRITE;
    return;
  }
  (*bytes_out) += crc_len;
  
  // If this is to be a reverse-readable record
  if (handle->reversible) {
    
    // Write message length with bytes reversed
    offset = 0;

    // Length is of:  sync byte + message id + message len + message + crc
    trailer_len = uint4_to_ubnxi(handle, buf, &offset, 1 + header_len + *len + crc_len);
    reverse_bytes (buf, trailer_len);
    if (fwrite (buf, trailer_len, 1, handle->fh) != 1) {
      *rc = BADWRITE;
      return;
    }
    (*bytes_out) += trailer_len;

    if(handle->little_endian) {
      sync = BINEX_FBetx_LE;
    } else {
      sync = BINEX_FBetx_BE;
    } 
    
    // Write binex sync byte
    if (fwrite (&sync, 1, 1, handle->fh) != 1) {
      *rc = BADWRITE;
      return;
    }
    (*bytes_out)++;
  }
  
  *rc = 0;
  return;
  
}

/**----------------------------------------------------------------------    
* @name       write_binex_buf
* 
* Writes a binex record, including header and checksum.  If 'reversible'
* is specified, write a reversible record.
*
* @parameter  handle The binexHandle file handle to read from
* @           rec   A buffer holding the record (in)
* @           len   The length of the record (in)
* @           type  The binex record type (in)
* @           reversible 1 = write reversible record, 0 = write regular record (in)
* @           bytesOut The number of bytes successfully written (out)
* @return     rc : 0 = OK, non-zero = error writing record
* @example    write_binex (&rec, &rec_len, &type, &reversible, &bytes_out, &rc)
* ----------------------------------------------------------------------*/
unsigned char * write_binex_buf (binexHandle *handle, unsigned char *rec, unsigned int *len, unsigned int *type, 
		  int *reversible, unsigned int *bytes_out, int *rc) {

  unsigned int  crc_len;       // length of computed CRC
  unsigned char crc[16];       // computed CRC
  unsigned char buf[8];
  unsigned int  trailer_len, header_len;
  unsigned int  offset;
  unsigned char sync;
  unsigned char * out;

  out = new unsigned char[*len + 32];
  *bytes_out = 0;

  handle->reversible     = *reversible;

  if        ( handle->little_endian &&  handle->reversible) {
    sync = BINEX_FBstx_LE;
  } else if ( handle->little_endian && !handle->reversible) {
    sync = BINEX_F_stx_LE;
  } else if (!handle->little_endian &&  handle->reversible) {
    sync = BINEX_FBstx_BE;
  } else if (!handle->little_endian && !handle->reversible) {
    sync = BINEX_F_stx_BE;
  }

  // Write binex sync byte
  out[*bytes_out] = sync;
  (*bytes_out)++;
  
  // Write binex record ID and length
  offset = header_len = 0;
  header_len  = uint4_to_ubnxi(handle, buf, &offset, *type);
  header_len += uint4_to_ubnxi(handle, buf, &offset, *len);
  memcpy((void *) &out[*bytes_out], buf, header_len);
  (*bytes_out) += header_len;
  
  // Write binex record 
  memcpy((void *) &out[*bytes_out], rec, *len);
  (*bytes_out) += *len;
  
  // Write CRC
  offset  = 0;
  crc_len = binex_crc(crc, buf, (unsigned char *)rec, &offset, header_len, *len, handle->bytes_reversed); 
  memcpy((void *) &out[*bytes_out], crc, crc_len);
  (*bytes_out) += crc_len;
  
  // If this is to be a reverse-readable record
  if (handle->reversible) {
    
    // Write message length with bytes reversed
    offset = 0;

    // Length is of:  sync byte + message id + message len + message + crc
    trailer_len = uint4_to_ubnxi(handle, buf, &offset, 1 + header_len + *len + crc_len);
    reverse_bytes (buf, trailer_len);
    memcpy((void *) &out[*bytes_out], buf, trailer_len);
    (*bytes_out) += trailer_len;

    if(handle->little_endian) {
      sync = BINEX_FBetx_LE;
    } else {
      sync = BINEX_FBetx_BE;
    } 
    
    // Write binex sync byte
    out[*bytes_out] = sync;
    (*bytes_out)++;
  }
  
  *rc = 0;
  return(out);
  
}


/**----------------------------------------------------------------------    
* @name       test_little_endian
* 
* Determines if the current machine is little endian.  If so, returns
* TRUE (1).
* @return     1 = little endian, 0 = big endian
* ----------------------------------------------------------------------*/
#ifdef __BYTE_ORDER

int test_little_endian () {

  if (__BYTE_ORDER == __BIG_ENDIAN) return(0);

  if (__BYTE_ORDER == __LITTLE_ENDIAN) return(1);

  return(UNKNOWN_TRUTH);
}

#else 
int test_little_endian () {

  // determine endian-ness of this processor/compiler

  double          d= 23456789.012;
  
  // IEEE 754 little endian ordering of 23456789.012:
  static unsigned char    uc[8]= { 0xe9, 0x26, 0x31, 0x50, 0xc1, 0x5e, 0x76, 0x41 };
  
  if (!memcmp((void *)uc, (void *)&d, 8))
    return(1);     // little-endian processor ordering:

  reverse_bytes((unsigned char *)&d, 8);
  if (!memcmp((void *)uc, (void *)&d, 8)) {
    return(0);     // big-endian processor ordering:
  }

#if 0
  memcpy((void *)uc, (void *)&d, 8);
  fprintf(stderr, "%.3lf == %02x %02x %02x %02x %02x %02x %02x %02x on this version\n",
          d, uc[0], uc[1], uc[2], uc[3], uc[4], uc[5], uc[6], uc[7]);
#endif
  
#if DEBUG
  fprintf(stderr, "little_endian()= UNKNOWN\n");
#endif
  
  return(UNKNOWN_TRUTH);
  
}
#endif

//-------------------------------------------------------

unsigned int read_ubnxi (binexHandle *handle, unsigned char *rec, unsigned char *n) {
        *n= 0x00;
        do {
                if ((size_t)fread((void *)&rec[*n], (size_t)1, (size_t)1, handle->fh) < 1) {
                        *n= 0x00;
                        return(0L);
                }
                switch (++(*n)) {
                case 0x01:
                case 0x02:
                case 0x03:
                case 0x05:
                        switch (rec[*n - 0x01] & 0x80) {
                        case 0x80:
                                break;
                        default:
                                return(ubnxi_to_uint4(handle, rec, *n));
                        }
                        break;
                default:        /* 4 bytes read */
                        return(ubnxi_to_uint4(handle, rec, *n));
                }
        } while (*n < 0x04);

        *n= 0x00;
        return (0L);
}

//-------------------------------------------------------

void binex_extract_ubnxi (binexHandle *handle, unsigned char *message, 
				      unsigned int *offset, int *rc, int *val) {

  unsigned char rec[4];
  unsigned char n = 0x00;
  *rc = 0;

  do {

    rec[n] = message[*offset + n];

    switch (++n) {
    case 0x01:
    case 0x02:
    case 0x03:
    case 0x05:
      switch (rec[n - 0x01] & 0x80) {
      case 0x80:
	break;
      default:
	/* This casts an unsigned int to an int.  This should not
	   cause overflow as max ubnxi < max signed int. */
	*val = (int)ubnxi_to_uint4(handle, rec, n);
	*offset += n;
	return;
      }
      break;
    default:        /* 4 bytes read */
      *val = (int)ubnxi_to_uint4(handle, rec, n);
      *offset += n;
      return;
    }

  } while (n < 0x04);

  *rc = -1; 
  val = 0;
  return;
}

//-------------------------------------------------------
char *SV_id (SAT_ID *system)
{
        static char     buffer[4];

        sprintf(buffer, "%c%2d", system_RINEX_code[system->sys], (int)system->SVid);
        return(buffer);
}

//-------------------------------------------------------
unsigned int ubnxi_to_uint4 (binexHandle *handle, unsigned char *rec, unsigned char n) {
	unsigned char	i;
	unsigned int	ul;

#ifdef DEBUG
	fprintf (stderr, "rec[0] = %x, rec[1] = %x, rec[2] = %x, rec[3] = %x\n", 
		 rec[0], rec[1], rec[2], rec[3]);

	fprintf (stderr, "len = %u\n", n);

	fprintf (stderr, "le = %d\n", binex.little_endian);
#endif

	for (i= 0, ul= 0L; i < n; i++)
		switch (handle->little_endian) {
		case TRUE:
			ul|= (unsigned int)(rec[i] & (i < 0x03 ? 0x7f : 0xff)) << 7*i;
#ifdef DEBUG
			fprintf (stderr, "ul = %x, i = %u\n", ul, i);
#endif
			break;
		case FALSE:
			ul|= (unsigned int)(rec[i] & (i < 0x03 ? 0x7f : 0xff)) << 7*(n-0x01-i);
			break;
		}

#ifdef DEBUG
	fprintf (stderr, "ul = %u\n", ul);
#endif

	return(ul);
}


//-------------------------------------------------------
void binex_append_time (binexHandle *handle, unsigned char *message, unsigned int *offset, unsigned char sec_resolution, unsigned int minutes, double seconds) {
  unsigned char	s;
  double frac;
  
  /* teqc epochs start at 1.0 Jan 1980, whereas GPS time starts at 6.0 Jan 1980,
     a difference of 5 days = 7200 minutes
     */
  append_uint4(handle, message, offset, minutes - 7200);
  
  /* note: cases 0x05 and 0x06 need to be tested
   */
  switch (sec_resolution) {
  case 0x00:
    /* resolution only to minutes
     */
    break;
  case 0x01:
    /* resolution to 0.25 seconds
     */
    append_uint1(handle, message, offset, (unsigned char)floor(seconds*4. + 0.5));
    break;
  case 0x02:
    /* resolution to 0.001 seconds
     */
    append_uint2(handle, message, offset, (unsigned short)floor(seconds*1000. + 0.5));
    break;
  case 0x04:
    /* resolution to 20 nanoseconds
     */
    append_uint4(handle, message, offset, (unsigned int)floor(seconds*50e6 + 0.5));
    break;
  case 0x05:
    /* resolution to 0.1 nanoseconds:
       
       6 bits used to store integer seconds; "1" = 1 second
       34 bits used to store nearest fractional 0.1 nanoseconds; "1" = 0.1 nanosecond
       */
    s= (unsigned char)floor(round(seconds, 1e-10));
    frac= floor((seconds - (double)s)*1e10 + 0.5);
    switch (binex.little_endian) {
    case TRUE:
      append_uint1(handle, message, offset, s | (0xc0 & (unsigned char)fmod(frac, 4.) << 6));
      append_uint4(handle, message, offset, (unsigned int)(frac/4.));
      break;
    case FALSE:
      append_uint1(handle, message, offset, (s << 2) | (0x03 & (unsigned char)(frac/4294967296.)));
      append_uint4(handle, message, offset, (unsigned int)fmod(frac, 4294967296.));
      break;
    }
    break;
  case 0x06:
    /* resolution to 0.25 picoseconds:
       
       6 bits used to store integer seconds; "1" = 1 second
       42 bits used to store nearest fractional 0.25 picoseconds; "1" = 0.25 picosecond
       */
    s= (unsigned char)floor(round(seconds, 2.5e-13));
    frac= floor((seconds - (double)s)*4e12 + 0.5);
    switch (binex.little_endian) {
    case TRUE:
      append_uint2(handle, message, offset, (unsigned short)s | (0xffc0 & (unsigned short)fmod(frac, 1024.) << 6));
      append_uint4(handle, message, offset, (unsigned int)(frac/1024.));
      break;
    case FALSE:
      append_uint2(handle, message, offset, ((unsigned short)s << 10) | (0x03ff & (unsigned short)(frac/4294967296.)));
      append_uint4(handle, message, offset, (unsigned int)fmod(frac, 4294967296.));
      break;
    }
    break;
  case 0x08:
    /* 8-byte floating-point resolution
     */
    append_real8(handle, message, offset, seconds);
    break;
  default:
    fprintf(stderr, "! Error ! binex_append_time(): case %x not found\n", (int)sec_resolution);
    exit(-1);
  }
}

//-------------------------------------------------------
void binex_append_mGFZi (binexHandle *handle, unsigned char *message, unsigned int *offset, double delta, bool eight_byte) {

	/* value        bits of leading byte
	   -----       -----------------------
			L -> M:  big   endian (absolute value is in the least significant bits)
			M <- L: little endian (absolute value is in the most significant bits)

	   < 64		....00:  +  6-bit int
			....10:  -  6-bit int

	   < 8254	...001:  + 13-bit int
			...101:  - 13-bit int

	   < 268443709	..0011:  + 28-bit int
			..1011:  - 28-bit int

	   else		..0111:  + 60-bit int
			..1111:  - 60-bit int
	*/
	static double	double_limit= BINEX_MAXFLOAT/10.;
	double		abs_delta;
	unsigned char	sgn_delta;
	unsigned char	uc = 0;
	unsigned short	us = 0;
	unsigned int	ul = 0;

	if (delta > double_limit) {
		/* store in uint1: "-0" == 0x02 = 00000010; 0x40 = 01000000
		*/
		switch (binex.little_endian) {
		case TRUE:
			uc= 0x02;
			break;
		case FALSE:
			uc= 0x40;
			break;
		}
		append_uint1(handle, message, offset, uc);
		return;
	}

	abs_delta= floor(fabs(delta) + 0.5);
	sgn_delta= delta < 0. ? 0x01 : 0x00;
	if (abs_delta < 64.) {			/* 2^6 */
		/* store in uint1: 0xfc = 11111100; 0x3f = 00111111
		   note: any "-0" representation is converted to the "+0" representation
		*/
		switch (binex.little_endian) {
		case TRUE:
			uc= (sgn_delta << 1) | ((0xfc & (unsigned char)abs_delta) << 2);
			switch (uc) {
			case 0x02:
				uc= 0x00;
				break;
			}
			break;
		case FALSE:
			uc= (sgn_delta << 6) | (0x3f & (unsigned char)abs_delta);
			switch (uc) {
			case 0x40:
				uc= 0x00;
				break;
			}
			break;
		}
		append_uint1(handle, message, offset, uc);
	}
	else if (abs_delta < 8254.) {		/* 2^13 + 62 */
		/* store in uint2: 0xf9 = 11111001; 0x9f = 10011111
		*/
		abs_delta-= 62.;

		/* note: values of -1, -0, +0, and +1 are reserved for future use
		*/
		switch (binex.little_endian) {
		case TRUE:
			us= ((unsigned short)sgn_delta << 2) | (0xfff9 & (0x0001 | (unsigned short)abs_delta << 3));
			break;
		case FALSE:
			us= ((unsigned short)sgn_delta << 13) | (0x9fff & (0x8000 | (unsigned short)abs_delta));
			break;
		}
		append_uint2(handle, message, offset, us);
	}
	else if (abs_delta < 268443709.) {	/* 2^28 + 8253 */
		/* store in uint4: 0xf3 = 11110011; 0xcf = 11001111
		*/
		abs_delta-= 8253.;

		switch (binex.little_endian) {
		case TRUE:
			ul= ((unsigned int)sgn_delta << 3) | (0xfffffff3 & (0x00000003 | (unsigned int)abs_delta << 4));
			break;
		case FALSE:
			ul= ((unsigned int)sgn_delta << 28) | (0xcfffffff & (0xc0000000 | (unsigned int)abs_delta));
			break;
		}
		append_uint4(handle, message, offset, ul);
	}
	else {
		abs_delta-= 268443708.;

		switch (eight_byte) {
		case FALSE:
			/* store in an uint4 and an uint2: 0xf7 = 11110111; 0xef = 11101111; 2^28 = 268435456; 2^16 = 65536
			*/
			switch (binex.little_endian) {
			case TRUE:
				ul= ((unsigned int)sgn_delta << 3) | (0xfffffff7 & (0x00000007 | (unsigned int)fmod(abs_delta, 268435456.) << 4));
				append_uint4(handle, message, offset, ul);
				us= (unsigned short)(abs_delta/268435456.);
				break;
			case FALSE:
				ul= ((unsigned int)sgn_delta << 28) | (0xefffffff & (0xe0000000 | (unsigned int)(abs_delta/65536.)));
				append_uint4(handle, message, offset, ul);
				us= (unsigned short)fmod(abs_delta, 65536.);
				break;
			}
			append_uint2(handle, message, offset, us);
			break;
		case TRUE:
			/* store in two uint4s: 0xf7 = 11110111; 0xef = 11101111; 2^28 = 268435456; 2^32 = 4294967296
			*/
			switch (binex.little_endian) {
			case TRUE:
				ul= ((unsigned int)sgn_delta << 3) | (0xfffffff7 & (0x00000007 | (unsigned int)fmod(abs_delta, 268435456.) << 4));
				append_uint4(handle, message, offset, ul);
				ul= (unsigned int)(abs_delta/268435456.);
				break;
			case FALSE:
				ul= ((unsigned int)sgn_delta << 28) | (0xefffffff & (0xe0000000 | (unsigned int)(abs_delta/4294967296.)));
				append_uint4(handle, message, offset, ul);
				ul= (unsigned int)fmod(abs_delta, 4294967296.);
				break;
			}
			append_uint4(handle, message, offset, ul);
			break;
		}
	}
}

//-------------------------------------------------------
void binex_append_mGFZI	(binexHandle *handle, unsigned char *message, unsigned int *offset, double delta) {
	/* value        bits of leading byte
	   -----       -----------------------
			L  ->  M:  big   endian (absolute value is in the least significant bits)
			M  <-  L: little endian (absolute value is in the most significant bits)

	   <   2^4	....0000:  +  4-bit int <               16 = 0x0000000000000010 (reserving -0)
			....1000:  -  4-bit int

	   < ~2^12	....0001:  + 12-bit int <             4110 = 0x000000000000100e (reserving +-1, +-0)
			....1001:  - 12-bit int

	   < ~2^20	....0010:  + 20-bit int <          1052685 = 0x000000000010100d (reserving -0)
			....1010:  - 20-bit int

	   < ~2^28	....0011:  + 28-bit int <        269488140 = 0x000000001010100c (reserving -0)
			....1011:  - 28-bit int

	   < ~2^36	....0100:  + 36-bit int <      68988964875 = 0x000000101010100b (reserving -0)
			....1100:  - 36-bit int

	   < ~2^44	....0101:  + 44-bit int <   17661175009290 = 0x000010101010100a (reserving -0)
			....1101:  - 44-bit int

	   < ~2^52	....0110:  + 52-bit int < 4521260802379785 = 0x0010101010101009 (reserving -0)
			....1110:  - 52-bit int

	   < ~2^60	....0111:  + 60-bit int <                    0x1010101010101008 (reserving -0)
			....1111:  - 60-bit int
	*/
	static double	double_limit= BINEX_MAXFLOAT/10.;
	double		abs_delta;
	unsigned char	sgn_delta;
	unsigned char	uc = 0;
	unsigned short	us = 0;
	unsigned int	ul = 0;

	if (delta > double_limit) {
		/* store in uint1: "-0" == 0x08 = 00001000; 0x80 = 10000000
		*/
		switch (binex.little_endian) {
		case TRUE:
			uc= 0x08;
			break;
		case FALSE:
			uc= 0x80;
			break;
		}
		append_uint1(handle, message, offset, uc);
		return;
	}

	abs_delta= floor(fabs(delta) + 0.5);
	sgn_delta= delta < 0. ? 0x01 : 0x00;
	if (abs_delta < 16.) {
		/* store in uint1: 0xf0 = 11110000; 0x0f = 00001111
		   note: any "-0" representation is converted to the "+0" representation
		*/
		switch (binex.little_endian) {
		case TRUE:
			uc= (sgn_delta << 3) | (0xf0 & (unsigned char)abs_delta << 4);
			switch (uc) {
			case 0x08:
				uc= 0x00;
				break;
			}
			break;
		case FALSE:
			uc= (sgn_delta << 7) | (0x0f & (unsigned char)abs_delta);
			switch (uc) {
			case 0x80:
				uc= 0x00;
				break;
			}
			break;
		}
		append_uint1(handle, message, offset, uc);
	}
	else if (abs_delta < 4110.) {
		/* store in uint2: 0xf1 = 11110001; 0x1f = 00011111
		*/
		abs_delta-= 14.;

		/* note: values of -1, -0, +0, and +1 are reserved for future use
		*/
		switch (binex.little_endian) {
		case TRUE:
			us= ((unsigned short)sgn_delta <<  3) | (0xfff1 & (0x0001 | (unsigned short)abs_delta << 4));
			break;
		case FALSE:
			us= ((unsigned short)sgn_delta << 15) | (0x1fff & (0x1000 | (unsigned short)abs_delta));
			break;
		}
		append_uint2(handle, message, offset, us);
	}
	else if (abs_delta < 1052685.) {
		/* store in 3 bytes of uint4: 0xf2 = 11110010; 0x2f = 00101111
		*/
		abs_delta-= 4109.;

		switch (binex.little_endian) {
		case TRUE:
			ul= ((unsigned int)sgn_delta <<  3) | (0x00fffff2 & (0x00000002 | (unsigned int)abs_delta << 4));
			break;
		case FALSE:
                        ul= ((unsigned int)sgn_delta << 23) | (0x002fffff & (0x00200000 | (unsigned int)abs_delta));
			break;
		}
		append_uint3(handle, message, offset, ul);
	}
	else if (abs_delta < 269488140.) {
		/* store in uint4: 0xf3 = 11110011; 0x3f = 00111111
		*/
		abs_delta-= 1052684.;

		switch (binex.little_endian) {
		case TRUE:
			ul= ((unsigned int)sgn_delta <<  3) | (0xfffffff3 & (0x00000003 | (unsigned int)abs_delta << 4));
			break;
		case FALSE:
			ul= ((unsigned int)sgn_delta << 31) | (0x3fffffff & (0x30000000 | (unsigned int)abs_delta));
			break;
		}
		append_uint4(handle, message, offset, ul);
	}
	else if (abs_delta < 68988964875.) {
		/* store in an uint4 and an uint1: 0xf4 = 11110100; 0x4f = 01001111; 2^28 = 268435456; 2^8 = 256
		*/
		abs_delta-= 269488139.;

		switch (binex.little_endian) {
		case TRUE:
			ul= ((unsigned int)sgn_delta <<  3) | (0xfffffff4 & (0x00000004 | (unsigned int)fmod(abs_delta, 268435456.) << 4));
			append_uint4(handle, message, offset, ul);
			uc= (unsigned char)(abs_delta/268435456.);
			break;
		case FALSE:
			ul= ((unsigned int)sgn_delta << 31) | (0x4fffffff & (0x40000000 | (unsigned int)(abs_delta/256.)));
			append_uint4(handle, message, offset, ul);
			uc= (unsigned char)fmod(abs_delta, 256.);
			break;
		}
		append_uint1(handle, message, offset, uc);
	}
	else if (abs_delta < 17661175009290.) {
		/* store in an uint4 and an uint2: 0xf5 = 11110101; 0x5f = 01011111; 2^28 = 268435456; 2^16 = 65536
		*/
		abs_delta-= 68988964874.;

		switch (binex.little_endian) {
		case TRUE:
			ul= ((unsigned int)sgn_delta <<  3) | (0xfffffff5 & (0x00000005 | (unsigned int)fmod(abs_delta, 268435456.) << 4));
			append_uint4(handle, message, offset, ul);
			us= (unsigned short)(abs_delta/268435456.);
			break;
		case FALSE:
			ul= ((unsigned int)sgn_delta << 31) | (0x5fffffff & (0x50000000 | (unsigned int)(abs_delta/65536.)));
			append_uint4(handle, message, offset, ul);
			us= (unsigned short)fmod(abs_delta, 65536.);
			break;
		}
		append_uint2(handle, message, offset, us);
	}
	else if (abs_delta < 4521260802379785.) {
		/* store in an uint4 and 3 bytes of uint4: 0xf6 = 11110110; 0x6f = 01101111; 2^28 = 268435456; 2^24 = 16777216
		*/
		abs_delta-= 17661175009289.;

		switch (binex.little_endian) {
		case TRUE:
			ul= ((unsigned int)sgn_delta <<  3) | (0xfffffff6 & (0x00000006 | (unsigned int)fmod(abs_delta, 268435456.) << 4));
			append_uint4(handle, message, offset, ul);
			ul= (unsigned int)(abs_delta/268435456.);
			break;
		case FALSE:
			ul= ((unsigned int)sgn_delta << 31) | (0x6fffffff & (0x60000000 | (unsigned int)(abs_delta/16777216.)));
			append_uint4(handle, message, offset, ul);
			ul= (unsigned int)fmod(abs_delta, 16777216.);
			break;
		}
		append_uint3(handle, message, offset, ul);
	}
	else {
		/* store in two uint4s: 0xf7 = 11110111; 0x7f = 01111111; 2^28 = 268435456; 2^32 = 4294967296
		*/
		abs_delta-= 4521260802379784.;

		switch (binex.little_endian) {
		case TRUE:
			ul= ((unsigned int)sgn_delta <<  3) | (0xfffffff7 & (0x00000007 | (unsigned int)fmod(abs_delta, 268435456.) << 4));
			append_uint4(handle, message, offset, ul);
			ul= (unsigned int)(abs_delta/268435456.);
			break;
		case FALSE:
			ul= ((unsigned int)sgn_delta << 31) | (0x7fffffff & (0x70000000 | (unsigned int)(abs_delta/4294967296.)));
			append_uint4(handle, message, offset, ul);
			ul= (unsigned int)fmod(abs_delta, 4294967296.);
			break;
		}
		append_uint4(handle, message, offset, ul);
	}
}


//-------------------------------------------------------
double binex_extract_mGFZi (binexHandle *handle, unsigned char *message, unsigned int *offset, bool *valid, bool eight_byte) {

	static double	abs_delta;
	unsigned char	uc, i, flag;
	unsigned short	us;
	unsigned int	ul;
	short		sgn_delta;

	flag=0;
	switch (handle->little_endian) {
	case TRUE:
		flag= message[*offset] & 0x0f;
		break;
	case FALSE:
		/* rotate the 4 flag bits in the byte:
		*/
		for (uc= 0x01, i= 0x04, flag= 0x00; uc < 0x10; uc+= 0x02, i+= 0x01)
			flag|= (unsigned char)(((unsigned int)message[*offset] & (0x01 << i)) >> uc);
		break;
	}

	*valid= (bool)TRUE;
	sgn_delta= 1;
	abs_delta= 0.;
	switch (flag & 0x03) {
	case 0x02:
		sgn_delta= -1;
		/* no break
		*/
	case 0x00:
		/* use 1 byte:
		*/
		extract_uint1(handle, message, offset, &uc);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)((unsigned int)uc >> 2);
			break;
		case 0x00:
			abs_delta= (double)((unsigned int)uc & 0x3f);
			break;
		}
		if (sgn_delta == -1 && abs_delta == 0.) {
			/* "-0" reserved for "no data" indicator
			*/
			*valid= (bool)FALSE;
			return(D_NO_DATA);
		}

		return((double)sgn_delta * abs_delta);
	default:
		switch (flag & 0x07) {
		case 0x05:
			sgn_delta= -1;
			/* no break
			*/
		case 0x01:
			/* use 2 bytes:
			*/
			extract_uint2(handle, message, offset, &us);
			switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
			case 0x01:
				abs_delta= (double)((unsigned int)us >> 3);
				break;
			case 0x00:
				abs_delta= (double)((unsigned int)us & 0x1fff);
				break;
			}
			return((double)sgn_delta * (62. + abs_delta));
		default:
			switch (flag & 0x0f) {
			case 0x0b:
				sgn_delta= -1;
				/* no break
				*/
			case 0x03:
				/* use 4 bytes:
				*/
				extract_uint4(handle, message, offset, &ul);
				switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
				case 0x01:
					abs_delta= (double)(ul >> 4);
					break;
				case 0x00:
					abs_delta= (double)(ul & 0x0fffffff);
					break;
				}
				return((double)sgn_delta * (8253. + abs_delta));
			case 0x0f:
				sgn_delta= -1;
				/* no break
				*/
			case 0x07:
				extract_uint4(handle, message, offset, &ul);
				switch (eight_byte) {
				case FALSE:
					/* use 6 bytes:
					*/
					switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
					case 0x01:
						abs_delta= (double)(ul >> 4);
						extract_uint2(handle, message, offset, &us);
						abs_delta+= (double)us*268435456.;
						break;
					case 0x00:
						abs_delta= (double)(ul & 0x0fffffff)*65536.;
						extract_uint2(handle, message, offset, &us);
						abs_delta+= (double)us;
						break;
					}
					break;
				case TRUE:
					/* use 8 bytes:
					*/
					switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
					case 0x01:
						abs_delta= (double)(ul >> 4);
						extract_uint4(handle, message, offset, &ul);
						abs_delta+= (double)ul*268435456.;
						break;
					case 0x00:
						abs_delta= (double)(ul & 0x0fffffff)*4294967296.;
						extract_uint4(handle, message, offset, &ul);
						abs_delta+= (double)ul;
						break;
					}
					break;
				}
				return((double)sgn_delta * (268443708. + abs_delta));
			default:
				/* something really screwed up!
				*/
				fprintf(stderr, "binex_extract_mGFZi(): flag = %x\n", (int)flag);
				*valid= (bool)FALSE;
				return(0.);
			}
		}
	}
}

//-------------------------------------------------------
// FORTRAN version
void binex_extract_mgfz_littlei (binexHandle *handle, unsigned char *message, unsigned int *offset, int *rc, int *eight_byte, double *val) {

  bool valid;
  *rc = 0;

  *val = binex_extract_mGFZi (handle, message, offset, &valid, *eight_byte);
  if (!valid) *rc = BADGFZI;

  return;
}



//-------------------------------------------------------
double binex_extract_mGFZI (binexHandle *handle, unsigned char *message, unsigned int *offset, bool *valid) {

	double	abs_delta;
//	unsigned char i;
	unsigned char	uc, flag;
	unsigned short	us;
	unsigned int	ul;
	short		sgn_delta;
	flag = 0;

	switch (handle->little_endian) {
	case TRUE:
		flag= message[*offset] & 0x0f;
		break;
	case FALSE:
		flag= message[*offset] >> 4 & 0x0f;
		break;
	}
	//debug(D_VALUES, "binex_extract_mGFZI: Entering offset = %d, byte1 = %x, flag = %d\n",*offset,message[*offset], flag);

	*valid= (bool)TRUE;
	sgn_delta= 1;
	abs_delta= 0.;
	switch (flag) {
	case 0x08:
		sgn_delta= -1;
		/* no break
		*/
	case 0x00:
		/* use 1 byte:
		*/
		extract_uint1(handle, message, offset, &uc);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)((unsigned int)uc >> 4);
			break;
		case 0x00:
			abs_delta= (double)((unsigned int)uc & 0x0f);
			break;
		}
		if (sgn_delta == -1 && abs_delta == 0.) {
			/* "-0" reserved for "no data" indicator
			*/
			*valid= (bool)FALSE;
			return(D_NO_DATA);
		}

		return((double)sgn_delta * abs_delta);

	case 0x09:
		sgn_delta= -1;
		/* no break
		*/
	case 0x01:
		/* use 2 bytes:
		*/
		extract_uint2(handle, message, offset, &us);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)((unsigned int)us >> 4);
			break;
		case 0x00:
			abs_delta= (double)((unsigned int)us & 0x0fff);
			break;
		}
		return((double)sgn_delta * (14. + abs_delta));

	case 0x0a:
		sgn_delta= -1;
		/* no break
		*/
	case 0x02:
		/* use 3 bytes:
		*/
		extract_uint3(handle, message, offset, &ul);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)((unsigned int)ul >> 4);
			break;
		case 0x00:
			abs_delta= (double)((unsigned int)ul >> 8 & 0x000fffff);
			break;
		}
		return((double)sgn_delta * (4109. + abs_delta));

	case 0x0b:
		sgn_delta= -1;
		/* no break
		*/
	case 0x03:
		/* use 4 bytes:
		*/
		extract_uint4(handle, message, offset, &ul);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)(ul >> 4);
			break;
		case 0x00:
			abs_delta= (double)(ul & 0x0fffffff);
			break;
		}
		return((double)sgn_delta * (1052684. + abs_delta));

	case 0x0c:
		sgn_delta= -1;
		/* no break
		*/
	case 0x04:
		/* use 5 bytes:
		*/
		extract_uint4(handle, message, offset, &ul);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)(ul >> 4);
			extract_uint1(handle, message, offset, &uc);
			abs_delta+= (double)uc*268435456.;
			break;
		case 0x00:
			abs_delta= (double)(ul & 0x0fffffff)*256.;
			extract_uint1(handle, message, offset, &uc);
			abs_delta+= (double)uc;
			break;
		}
		return((double)sgn_delta * (269488139. + abs_delta));

	case 0x0d:
		sgn_delta= -1;
		/* no break
		*/
	case 0x05:
		/* use 6 bytes:
		*/
		extract_uint4(handle, message, offset, &ul);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)(ul >> 4);
			extract_uint2(handle, message, offset, &us);
			abs_delta+= (double)us*268435456.;
			break;
		case 0x00:
			abs_delta= (double)(ul & 0x0fffffff)*65536.;
			extract_uint2(handle, message, offset, &us);
			abs_delta+= (double)us;
			break;
		}
		return((double)sgn_delta * (68988964874. + abs_delta));

	case 0x0e:
		sgn_delta= -1;
		/* no break
		*/
	case 0x06:
		/* use 7 bytes:
		*/
		extract_uint4(handle, message, offset, &ul);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)(ul >> 4);
			extract_uint4(handle, message, offset, &ul);
			abs_delta+= (double)ul*268435456.;
			break;
		case 0x00:
			abs_delta= (double)(ul & 0x0fffffff)*16777216.;
			extract_uint4(handle, message, offset, &ul);
			abs_delta+= (double)(ul >> 8);
			break;
		}
		return((double)sgn_delta * (17661175009289. + abs_delta));

	case 0x0f:
		sgn_delta= -1;
		/* no break
		*/
	case 0x07:
		/* use 8 bytes:
		*/
		extract_uint4(handle, message, offset, &ul);
		switch ((handle->bytes_reversed & 0x01) ^ (binex.little_endian & 0x01)) {
		case 0x01:
			abs_delta= (double)(ul >> 4);
			extract_uint4(handle, message, offset, &ul);
			abs_delta+= (double)ul*268435456.;
			break;
		case 0x00:
			abs_delta= (double)(ul & 0x0fffffff)*4294967296.;
			extract_uint4(handle, message, offset, &ul);
			abs_delta+= (double)ul;
			break;
		}
		return((double)sgn_delta * (4521260802379784. + abs_delta));

	default:
		/* something really screwed up!
		*/
		fprintf(stderr, "binex_extract_mGFZI(): flag = %x\n", (int)flag);
		*valid= (bool)FALSE;
		return(0.);
	}
}

//-------------------------------------------------------
// FORTRAN version
void binex_extract_mgfz_capi (binexHandle *handle, unsigned char *message, unsigned int *offset, int *rc, double *val) {

  bool valid;
  
  *rc = 0;
  *val = binex_extract_mGFZI (handle, message, offset, &valid);

  if      (*val == D_NO_DATA) *rc = NODATAGFZI;
  else if (!valid)            *rc = BADGFZI;

  return;
}


//-------------------------------------------------------
void append_uint1 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned char uc) {
  buf[*off]= uc;
  (*off)++;
}

//-------------------------------------------------------
void append_uint2 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned short us) {
  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&us, 2);
  memcpy((void *)(buf + *off), (void *)&us, (size_t)2);
  (*off)+= 2;
}

//-------------------------------------------------------
void append_uint2_as_int4 (binexHandle *handle, unsigned char *buf, unsigned int *off, int us) {
  unsigned short us1 = (unsigned short)us;
  if (us > (2^16)) printf ("Warning, %d truncated to fit uint2", us);
  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&us1, 2);
  memcpy((void *)(buf + *off), (void *)&us1, (size_t)2);
  (*off)+= 2;
}

//-------------------------------------------------------
void append_sint2 (binexHandle *handle, unsigned char *buf, unsigned int *off, short s) {
  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&s, 2);
  memcpy((void *)(buf + *off), (void *)&s, (size_t)2);
  (*off)+= 2;
}

//-------------------------------------------------------
void append_uint3 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned int ul) {
  
  /* get the 'good' 3 bytes out of the ul */
  if (!binex.little_endian) ul <<= 8;
    

  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&ul, 3);
  memcpy((void *)(buf + *off), (void *)&ul, (size_t)3);
  (*off)+= 3;
}

//-------------------------------------------------------
void append_uint4 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned int ul) {
  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&ul, 4);
  memcpy((void *)(buf + *off), (void *)&ul, (size_t)4);
  (*off)+= 4;
}

//-------------------------------------------------------
void append_uint4_as_double (binexHandle *handle, unsigned char *buf, unsigned int *off, double ul) {
  unsigned int ul1 = (unsigned int)ul;
  if (ul > (2^32)) printf ("Warning, %f truncated to fit uint4", ul);
  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&ul1, 4);
  memcpy((void *)(buf + *off), (void *)&ul1, (size_t)4);
  (*off)+= 4;
}

//-------------------------------------------------------
void append_sint4 (binexHandle *handle, unsigned char *buf, unsigned int *off, long sl) {
  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&sl, 4);
  memcpy((void *)(buf + *off), (void *)&sl, (size_t)4);
  (*off)+= 4;
}

//-------------------------------------------------------
void append_real4 (binexHandle *handle, unsigned char *buf, unsigned int *off, float f) {
  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&f, 4);
  memcpy((void *)(buf + *off), (void *)&f, (size_t)4);
  (*off)+= 4;
}

//-------------------------------------------------------
void append_real8 (binexHandle *handle, unsigned char *buf, unsigned int *off, double d) {
  if (handle->bytes_reversed) reverse_bytes((unsigned char *)&d, 8);
  memcpy((void *)(buf + *off), (void *)&d, (size_t)8);
  (*off)+= 8;
}

//-------------------------------------------------------
void extract_uint1 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned char *val) {
	*val= buf[*off];
	(*off)++;
#ifdef DEBUG
	fprintf (stderr, "extract_uint1:  uc = %x, off = %d\n", *val, *off);
#endif
}

//-------------------------------------------------------
void extract_uint2 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned short *us) {
	memcpy((void *)us, (void *)(buf + *off), (size_t)2);
	(*off)+= 2;
	if (handle->bytes_reversed)
		reverse_bytes((unsigned char *)us, 2);
}

//-------------------------------------------------------
void extract_sint2 (binexHandle *handle, unsigned char *buf, unsigned int *off, short *s) {
	memcpy((void *)s, (void *)(buf + *off), (size_t)2);
	(*off)+= 2;
	if (handle->bytes_reversed)
		reverse_bytes((unsigned char *)s, 2);
}

//-------------------------------------------------------
void extract_uint3 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned int *ul) {
	*ul= 0L;
	memcpy((void *)ul, (void *)(buf + *off), (size_t)3);
	(*off)+= 3;
	if (handle->bytes_reversed)
		reverse_bytes((unsigned char *)ul, 4);
}

//-------------------------------------------------------
void extract_uint4 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned int *ul) {
	memcpy((void *)ul, (void *)(buf + *off), (size_t)4);
	(*off)+= 4;
	if (handle->bytes_reversed)
		reverse_bytes((unsigned char *)ul, 4);
#ifdef DEBUG
	fprintf (stderr, "extract_uint4:  ul = %d, off = %d\n", *ul, *off);
#endif
}

//-------------------------------------------------------
void extract_sint4 (binexHandle *handle, unsigned char *buf, unsigned int *off, long *l) {
	memcpy((void *)l, (void *)(buf + *off), (size_t)4);
	(*off)+= 4;
	if (handle->bytes_reversed)
		reverse_bytes((unsigned char *)l, 4);
}

//-------------------------------------------------------
void extract_uint5 (binexHandle *handle, unsigned char *buf, unsigned int *off, unsigned long long *ul) {
	*ul= 0L;
	memcpy((void *)ul, (void *)(buf + *off), (size_t)5);
	(*off)+= 5;
	if (handle->bytes_reversed)
		reverse_bytes((unsigned char *)ul, 8);
}


//-------------------------------------------------------
void extract_real4 (binexHandle *handle, unsigned char *buf, unsigned int *off, float *f) {
	memcpy((void *)f, (void *)(buf + *off), (size_t)4);
	(*off)+= 4;
	if (handle->bytes_reversed)
		reverse_bytes((unsigned char *)f, 4);
}

//-------------------------------------------------------
void extract_real8 (binexHandle *handle, unsigned char *buf, unsigned int *off, double *d) {
	memcpy((void *)d, (void *)(buf + *off), (size_t)8);
	(*off)+= 8;
	if (handle->bytes_reversed)
		reverse_bytes((unsigned char *)d, 8);
}


//-------------------------------------------------------
void extract_2cNb (binexHandle *handle, uint1 *buf, unsigned int *off, uint1 skip, uint1 sN, bool *valid, double *d)
{
        // NOTE: offset is not incremented; looking at bits only, don't know how many bytes
	uint4	Mul= 0L, Lul= 0L;
	uint1	bytes;

        debug(D_FLOW,"extract_2cNb: Entering...\n");
	*valid = true;

	if ((skip + sN)%8) {
		/* the LSBs do not start at a byte boundary;
		   cannot use this algorithm
		*/
		fprintf(stderr, "extract_2cNb(): LSBs do not start at a byte boundary; shift bits to use this algorithm\n");
		exit(-1);
	}

	if (sN > 52) {
		fprintf(stderr, "extract_2cNb(): number of bits cannot exceed 52 (no. of mantissa bits in 8-byte floating point)\n");
		exit(-1);
	}

	bytes= (skip + sN)/8 + ((skip + sN)%8 ? 1 : 0);
	switch (bytes) {
	case 8:
	case 7:
	case 6:
	case 5:
		switch (handle->little_endian) {
		case TRUE:
			memcpy((void *)&Mul, (void *)(buf + *off + 4), (size_t)(bytes-4));
			memcpy((void *)&Lul, (void *)(buf + *off),     (size_t)4);
			break;
		case FALSE:
		default:
			memcpy((void *)&Lul, (void *)(buf + *off + bytes - 4), (size_t)4);
			memcpy((void *)&Mul, (void *)(buf + *off),             (size_t)(bytes-4));
			break;
		}

		switch (handle->bytes_reversed) {
		case TRUE:
			reverse_bytes((uint1 *)&Lul, 4);
			if (bytes > 5)
				reverse_bytes((uint1 *)&Mul, bytes-4);
		}

		switch (handle->little_endian) {
		case FALSE:
			/* for big-endian processors only:
			   (move the most significant bytes to the least significant bytes)
			*/			Mul>>= 8*(8-bytes);		/* shift msb to lsb */
		}

		/* mask off any bits that are not part of this 2cNb int:
		*/
		Mul&= ~(0xffffffff << (sN-32));

		/* check for setting of "invalid" == -2*(n-1), i.e. just the sign bit is set
		*/
		if (Mul == (uint4)(0x01 << (sN-33)) && Lul == 0L) {
#if DEBUG
			fprintf(stderr, "extract_2cNb(): Mul= 0x%08x Lul= 0x%08x signals 'invalid'\n", Mul, Lul);
#endif
			*valid= false;
			*d= 0.;
		}
		else {
#if DEBUG
			fprintf(stderr, "Mul/Lul before sign-completion= %08x %08x ...\n", Mul, Lul);
#endif
			if (Mul & 1 << (sN-33)) {
			  Mul|= 0xffffffff << (sN-32);	/* negative: complete 4-byte 2's complement */
				*d= 4294967296.*(1.+(real8)(sint4)Mul) + (real8)Lul - 4294967296.;
			}
			else {
				*d= 4294967296.*(real8)Mul + (real8)Lul;
			}
#if DEBUG
			fprintf(stderr, " after sign-completion= %08x %08x  d= %14.0lf\n", Mul, Lul, d);
#endif
		}
		break;
	default:
	  memcpy((void *)&Lul, (void *)(buf + *off), (size_t)bytes);
#if DEBUG
		fprintf(stderr, "extract_2cNb(): Lul= 0x%08x\n", Lul);
#endif

		if (handle->bytes_reversed && bytes > 1)
			reverse_bytes((uint1 *)&Lul, bytes);
#if DEBUG
		fprintf(stderr, "extract_2cNb(): Lul= 0x%08x\n", Lul);
#endif

		switch (handle->little_endian) {
		case FALSE:
			/* for big-endian processors only:
			   (move the most significant bytes to the least significant bytes)
			*/
			Lul>>= 8*(4-bytes);		/* shift msb to lsb */
#if DEBUG
			fprintf(stderr, "extract_2cNb(): Lul= 0x%08x  (big endian representation)\n", Lul);
#endif
		}

		/* mask off any bits that are not part of this 2cNb int:
		*/
		Lul&= ~(0xffffffff << sN);
#if DEBUG
		fprintf(stderr, "extract_2cNb(): Lul= 0x%08x  (after masking)\n", Lul);
#endif

		/* check for setting of "invalid" == -2*(n-1), i.e. just the sign bit is set
		*/
		if (Lul == (uint4)(0x01 << (sN-1))) {
#if DEBUG
			fprintf(stderr, "extract_2cNb(): Lul= 0x%08x signals 'invalid'\n", Lul);
#endif
			*valid= false;
			*d= 0.;
		} else {
#if DEBUG
			fprintf(stderr, "Lul before sign-completion= %08x ...", Lul);
#endif
			if (Lul & 1 << (sN-1))
				Lul|= 0xffffffff << sN;		/* negative: complete 4-byte 2's complement */

			*d= (real8)(sint4)Lul;
#if DEBUG
			fprintf(stderr, " after sign-completion= %08x   d= %11.0lf\n", Lul, d);
#endif
		}
	}

	(*off)+= bytes;

	debug(D_FLOW,"extract_2cNb: Exiting (%d bytes)...\n", bytes);
	return;
}

//-------------------------------------------------------
void
extract_xNb (binexHandle *handle, uint1 *buf, unsigned int *off, uint1 skip, uint1 sN, bool sign, bool *valid, double *d)
{
	// NOTE: offset is not incremented; looking at bits only, don't know how many bytes
	uint4	Mul= 0L, Lul= 0L;
	uint1	bytes, s_mask, mask, s= 0x00;

#if DEBUG_EXTRACT || DEBUG
	fprintf(err.fp, "extract_xNb()\n");
#endif
	*valid = true;

	if ((skip + sN)%8) {
		/* LSBs do not start at a byte boundary;
		   cannot use this algorithm
		*/
		fprintf(stderr, "extract_xNb(): LSBs do not start at a byte boundary; cannot use this algorithm\n");
		exit(-1);
	}

	if (sN > 52) {
		fprintf(stderr, "extract_xNb(): number of bits cannot exceed 52 (no. of mantissa bits in 8-byte floating point)\n");
		exit(-1);
	}

	switch (skip) {
	case 0: mask= 0x00; s_mask= 0x80; break;
	case 1: mask= 0x80; s_mask= 0x40; break;
	case 2: mask= 0xc0; s_mask= 0x20; break;
	case 3: mask= 0xe0; s_mask= 0x10; break;
	case 4: mask= 0xf0; s_mask= 0x08; break;
	case 5: mask= 0xf8; s_mask= 0x04; break;
	case 6: mask= 0xfc; s_mask= 0x02; break;
	case 7: mask= 0xfe; s_mask= 0x01; break;
	}

	bytes= (skip + sN)/8 + ((skip + sN)%8 ? 1 : 0);
	switch (bytes) {
	case 8:
	case 7:
	case 6:
	case 5:
	  switch (handle->little_endian) {
		case TRUE:
			memcpy((void *)&Mul, (void *)(buf + *off + 4), (size_t)(bytes-4));
			memcpy((void *)&Lul, (void *)(buf + *off),     (size_t)4);
			break;
		default:
			memcpy((void *)&Lul, (void *)(buf + *off + bytes - 4), (size_t)4);
			memcpy((void *)&Mul, (void *)(buf + *off),             (size_t)(bytes-4));
		}
#if DEBUG
		fprintf(err.fp, "memcpy:  Mul= 0x%08x  Lul= 0x%08x\n", Mul, Lul);
#endif

		switch (handle->bytes_reversed) {
		case TRUE:
			reverse_bytes((uint1 *)&Lul, 4);
			if (bytes > 5)
				reverse_bytes((uint1 *)&Mul, bytes-4);
#if DEBUG
			fprintf(err.fp, "reverse: Mul= 0x%08x  Lul= 0x%08x\n", Mul, Lul);
#endif
		}

		switch (handle->little_endian) {
		case FALSE:
			/* for big-endian processors only:
			   (move the most significant bytes to the least significant bytes)
			*/
#if DEBUG
			fprintf(err.fp, "before>>:Mul= 0x%08x  Lul= 0x%08x bytes= %d shift= %d\n", Mul, Lul, bytes, 8*((8-bytes)%4));
#endif
			Mul>>= 8*(4 < bytes ? 8-bytes : 4-bytes);		/* shift msb to lsb */
#if DEBUG
			fprintf(err.fp, "shift:   Mul= 0x%08x  Lul= 0x%08x\n", Mul, Lul);
#endif
		}

		if (sign)
			s= s_mask & Mul >> 8*(bytes-1); 			/* extract sign bit */
#if DEBUG
		fprintf(err.fp, "before:  Mul= 0x%08x  ~(0x%08x << %d)\n", Mul, mask | s, 8*((bytes-1)%4));
#endif
		Mul&= ~((mask | s) << 8*(4 < bytes ? bytes-5 : bytes-1));	/* mask out higher bits */
#if DEBUG
		fprintf(err.fp, "mask:    Mul= 0x%08x\n", Mul);
#endif

		*d= (real8)Lul + 4294967296.*(real8)Mul;
		break;
	default:
		memcpy((void *)&Lul, (void *)(buf + *off), (size_t)bytes);
#if DEBUG
		fprintf(err.fp, "memcpy:  Lul= 0x%08x\n", Lul);
#endif

		if (handle->bytes_reversed && bytes > 1) {
			reverse_bytes((uint1 *)&Lul, bytes);
#if DEBUG
			fprintf(err.fp, "reverse: Lul= 0x%08x\n", Lul);
#endif
		}

		switch (handle->little_endian) {
		case FALSE:
			/* for big-endian processors only:
			   (move the most significant bytes to the least significant bytes)
			*/
#if DEBUG
			fprintf(err.fp, "before>>:Lul= 0x%08x bytes= %d shift= %d\n", Lul, bytes, 8*((8-bytes)%4));
#endif
			Lul>>= 8*(4 < bytes ? 8-bytes : 4-bytes);		/* shift msb to lsb */
#if DEBUG
			fprintf(err.fp, "shift:   Lul= 0x%08x\n", Lul);
#endif
		}

		if (sign)
			s= s_mask & Lul >> 8*(bytes-1); 			/* extract sign bit */
#if DEBUG
		fprintf(err.fp, "before:  Lul= 0x%08x  ~(0x%08x << %d)\n", Lul, mask | s, 8*((bytes-1)%4));
#endif
		Lul&= ~((mask | s) << 8*(4 < bytes ? bytes-5 : bytes-1));	/* mask out higher bits */
#if DEBUG
		fprintf(err.fp, "mask:    Lul= 0x%08x\n", Lul);
#endif

		*d = (real8)Lul;
	}

	switch (sign) {
	case TRUE:
		if (s && !d)
			*valid = false;
	}
#if DEBUG
	fprintf(err.fp, "buf=");
	for (mask= 0; mask < bytes; mask++)
		fprintf(err.fp, " %02x", buf[*off + mask]);
	fprintf(err.fp, "\n");
	fprintf(err.fp, "Mul= %08x Lul= %08x  s= 0x%02x d= %.0lf\n", Mul, Lul, s, s ? -d : d);
#endif

	(*off)+= bytes;

#if DEBUG
	fprintf(err.fp, "\td= %.9lf\n\n", s ? -d : d);
#endif
	if (s) {
	  *d *= -1;
	}

	return;
}


//-------------------------------------------------------
void extract_uint2_as_int4 (binexHandle *handle, unsigned char *buf, unsigned int *off, int *us) {
  unsigned short us1;
  memcpy((void *)&us1, (void *)(buf + *off), (size_t)2);
  (*off)+= 2;
  if (handle->bytes_reversed)
    reverse_bytes((unsigned char *)&us1, 2);
  *us = (int)us1;
}

//-------------------------------------------------------
void extract_uint4_as_double (binexHandle *handle, unsigned char *buf, unsigned int *off, double *ul) {
  unsigned int ul1;
  memcpy((void *)&ul1, (void *)(buf + *off), (size_t)4);
  (*off)+= 4;
  if (handle->bytes_reversed)
    reverse_bytes((unsigned char *)&ul1, 4);
  *ul = (double)ul1;
#ifdef DEBUG
  printf (stderr, "extract_uint4_as_double:  ul = %f, off = %d\n", *ul, *off);
#endif

}

//-------------------------------------------------------
unsigned char binex_build_SV_id (SAT_ID *system)
{
	unsigned char	id;

	if (0x00 < system->SVid && system->SVid < 0x21)
		/* number must be 1-32 for GPS, 1-24 for GLONASS
		*/
		id= system->SVid - 0x01;
	else {
		fprintf(stderr, "! Error ! binex_build_SV_id(): invalid satellite number (%s)\n", SV_id(system));
		exit(-1);
	}

	switch (system->sys) {
	case NAVSTAR_GPS:
		id|= 0x00 << 5;
		break;
	case GLONASS:
		id|= 0x01 << 5;
		break;
	default:
		fprintf(stderr, "! Error ! binex_build_SV_id(): unsupported satellite constellation (= 0x%02x)\n", (int)system->sys);
		exit(-1);
	}

	return(id);
}

//-------------------------------------------------------
char binex_extract_SV_id (unsigned char id, SAT_ID *system)
{
	switch (id >> 5 & 0x03) {
	case 0x00:
		system->sys= (unsigned char)NAVSTAR_GPS;
		break;
	case 0x01:
		system->sys= (unsigned char)GLONASS;
		break;
	default:
		return ((bool)FALSE);
	}
	system->SVid = (id & 0x1f) + 0x01;

	return ((bool)TRUE);
}

//-------------------------------------------------------
unsigned int uint4_to_ubnxi (binexHandle *handle, unsigned char *buf, unsigned int *offset, unsigned int value) {

  /* convert an uint4 which is < 536870912 (2^29)
     to the standard BINEX 1 to 4 byte sequence
     */
  unsigned char	uc;
  unsigned short	us, us1;
  unsigned int	ul, ul1, ul2, ul3, bytes;
  
  if (value < 128) {
    /* one byte needed:
     */
    bytes= 1;
    uc= (unsigned char)value;
    append_uint1 (handle, buf, offset, uc);
  }
  else if (value < 16384) {
    /* two bytes needed:
     */
    bytes= 2;
    us= (unsigned short)value;
    us1= 0x7f00 & us << 1;
    us&= 0x007f;
    us|= us1;
    switch (handle->little_endian) {
    case TRUE:
      us|= 0x0080;
      break;
    case FALSE:
      us|= 0x8000;
      break;
    }
    append_uint2 (handle, buf, offset, us);
  }
  else if (value < 2097152) {
    /* three bytes needed:
     */
    bytes= 3;
    ul= (unsigned int)value;
    ul2= 0x007f0000 & ul << 2;
    ul1= 0x00007f00 & ul << 1;
    ul&= 0x0000007f;
    ul|= ul1 | ul2;
    switch (handle->little_endian) {
    case TRUE:
      ul|= 0x00008080;
      break;
    case FALSE:
      ul|= 0x00808000;
      break;
    }
    append_uint3 (handle, buf, offset, ul);
  }
  else if (value < 536870912) {
    /* four bytes needed:
     */
    bytes= 4;
    ul= (unsigned int)value;
    ul3= 0xff000000 & ul << 3;
    ul2= 0x007f0000 & ul << 2;
    ul1= 0x00007f00 & ul << 1;
    ul&= 0x0000007f;
    ul|= ul1 | ul2 | ul3;
    switch (handle->little_endian) {
    case TRUE:
      ul|= 0x00808080;
      break;
    case FALSE:
      ul|= 0x80808000;
      break;
    }
    append_uint4 (handle, buf, offset, ul);
  }
  else {
    bytes= 0;
    fprintf(stderr, "! Error ! uint4_to_ubnxi(): value is %d >= 536870912\n", value);
    exit(-1);
  }
  
  return(bytes);
}

//-------------------------------------------------------
void binex_append_ubnxi (binexHandle *handle, unsigned char *buf, 
	                                                unsigned int *offset, int *rc, unsigned int val) {
  *rc = uint4_to_ubnxi (handle, buf, offset, val);

  return;

}



//-------------------------------------------------------
unsigned int binex_crc (unsigned char *crc, unsigned char *buff, unsigned char *mess,
	 unsigned int *crc_offset, unsigned int buff_length, unsigned int mess_length, bool reverse)
{
	unsigned char	uc, bytes;
	unsigned short	us;
	unsigned int	ul, length;
        unsigned char md5[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};    
//	int rc;

	/* compute BINEX CRC for this message length
	*/
	length= buff_length + mess_length;
	if (length < 128) {
		/* 1-byte checksum
		*/
		bytes= 1;
		uc= cks08(buff, mess, buff_length, mess_length);
		memcpy((void *)(crc + *crc_offset), (void *)&uc, (size_t)bytes);
	}
	else if (length < 4096) {
		/* 2-byte CRC
		*/
		bytes= 2;
		us= crc16(buff, mess, buff_length, mess_length);
		if (reverse)
			reverse_bytes((unsigned char *)&us, 2);
		memcpy((void *)(crc + *crc_offset), (void *)&us, (size_t)bytes);
	}
	else if (length < 1048576) {
		/* 4-byte CRC
		*/
		bytes= 4;
		ul= crc32(buff, mess, buff_length, mess_length);
		if (reverse)
			reverse_bytes((unsigned char *)&ul, 4);
		memcpy((void *)(crc + *crc_offset), (void *)&ul, (size_t)bytes);

		//printf ("ul = %u\n", ul);

	}
	else {
		/* 16-byte MD5
		*/
		bytes= 16;
#if 0
	        rc = md5_encode(md5, buff, mess, buff_length, mess_length);
#else 

		MD5 *pCTX = new MD5;
		pCTX->update(buff, buff_length);
		pCTX->finalize();
		memcpy((unsigned char *) md5, (unsigned char *) pCTX, 16);

#endif
       		if (reverse) reverse_bytes((unsigned char *)md5, 16);
		memcpy((void *)(crc + *crc_offset), (void *)md5, (size_t)bytes);
	}
	(*crc_offset)+= (unsigned int)bytes;

	return ((int)bytes);
}

//-------------------------------------------------------
void reverse_bytes (unsigned char *rec, int n)
{
	switch (n) {
	case 1:
	        return;
	case 2:
		swap_uc(&rec[0], &rec[1]);
		return;
	case 3:
		swap_uc(&rec[0], &rec[2]);
		return;
	case 4:
		swap_uc(&rec[0], &rec[3]);
		swap_uc(&rec[1], &rec[2]);
		return;
	case 6:
		swap_uc(&rec[0], &rec[1]);
		swap_uc(&rec[2], &rec[5]);
		swap_uc(&rec[3], &rec[4]);
		return;
	case 8:
		swap_uc(&rec[0], &rec[7]);
		swap_uc(&rec[1], &rec[6]);
		swap_uc(&rec[2], &rec[5]);
		swap_uc(&rec[3], &rec[4]);
		return;
	default:
		fprintf(stderr, "reverse_bytes(): uncoded case of %d bytes", n);
		exit(-1);
	}
}

//-------------------------------------------------------
void swap_pp (void **a, void **b)
{
	void	*tmp;
	tmp= *a;
	*a= *b;
	*b= tmp;
}

//-------------------------------------------------------
void swap_c (char *a, char *b)
{
	char	tmp;
	tmp= *a;
	*a= *b;
	*b= tmp;
}

//-------------------------------------------------------
void swap_uc (unsigned char *a, unsigned char *b)
{
	unsigned char	tmp;
	tmp= *a;
	*a= *b;
	*b= tmp;
}

//-------------------------------------------------------
void swap_i (int *a, int *b)
{
	int	tmp;
	tmp= *a;
	*a= *b;
	*b= tmp;
}

//-------------------------------------------------------
void swap_d (double *a, double *b)
{
	double	tmp;
	tmp= *a;
	*a= *b;
	*b= tmp;
}

//-------------------------------------------------------
// Parses the most recent record read, returning an array of doubles.
// This must be called after a binex record has been read in.
// These code values are recognized:
// 
// u = ubnxi value
// I = mGFZI value
// i = mGFZi value
// 1 = uint1
// 2 = uint2
// 3 = uint3
// 4 = uint4
// f = float
// d = double
// s = sint2
// S = sint4        
// 
// @parameter  string   The above string.
// @           handle   The binex handle used to read in the record
// @           rec      A pointer to the binex record (sans header/trailer)
// @           offset   The current offset into the record
// @           vals     An array of doubles into which to place the values parsed from the record
// @           rc       Return code:  0 = success, negative = error
// 
void parse_c (char * string, binexHandle *handle, unsigned char *rec, unsigned int *offset, char *vals, int *rc) {
  
  int i;
  char code;
  int len = strlen(string);
  double        *data       = (double *)vals;
  double         val_double = 0.0;
  float          val_float  = 0.0;
  int            val_sint4  = 0;
  long           val_long   = 0;
  unsigned char  val_uint1  = 0;
  unsigned short val_uint2  = 0;
  unsigned int   val_uint4  = 0;
  int            eight_byte = 1;
  short val_sint2  = 0;
	   
  // Assume no errors
  *rc = 0;

  for (i=0;i<len;i++) {
    code = string[i];
    switch (code) {
      
    case 'u' :
      binex_extract_ubnxi (handle, rec, offset, rc, &val_sint4);
      if (rc) { fprintf (stderr, "Cannot read UBNXI(u) value at %d\n", *offset); return; }
      data[i] = (double)val_sint4;
      break;
    case 'I' :
      binex_extract_mgfz_capi (handle, rec, offset, rc, &val_double);
      if      (*rc == -17) {
	val_double = 0.0;  // return 0.0 on 'no value' GFZI
      } else if (*rc) {
	fprintf (stderr, "Cannot read MGFZI(I) value at %d\n", *offset); return;
      }
      data[i] = val_double;
      break;
    case 'i' :
      binex_extract_mgfz_littlei (handle, rec, offset, rc, &eight_byte, &val_double);
      if (*rc) { fprintf (stderr, "Cannot read MGFZi(i) value at %d\n",*offset); return; }
      data[i] = val_double;
      break;
    case '1' :
      extract_uint1 (handle, rec, offset, &val_uint1);
      data[i] = (double)val_uint1;
      break;
    case '2' :
      extract_uint2 (handle, rec, offset, &val_uint2);
      data[i] = (double)val_uint2;
      break;
    case '3' :
      extract_uint3 (handle, rec, offset, &val_uint4);
      data[i] = (double)val_uint4;
      break;
    case '4' :
      extract_uint4 (handle, rec, offset, &val_uint4);
      data[i] = (double)val_uint4;
      break;
    case 'f' :
      extract_real4 (handle, rec, offset, &val_float);
      data[i] = (double)val_float;
      break;
    case 'd' :
      extract_real8 (handle, rec, offset, &val_double);
      data[i] = val_double;
      break;
    case 's' :
      extract_sint2 (handle, rec, offset, &val_sint2);
      data[i] = (double)val_sint2;
      break;
    case 'S' :
      extract_sint4 (handle, rec, offset, &val_long);
      data[i] = (double)val_long;
      break;
    default :
      printf ("Illegal type identifier: %c", code);
      *rc = BADPARSETYPE;
      return;
    }
  }
  
  return;
  
}
  

int binex_extract_time(binexHandle *handle, unsigned int record_id,  unsigned char *rec, long *minutes, unsigned short *m_seconds) {

  unsigned int   gpsMinute;
  unsigned int   gpsSeconds;
  unsigned short gpsFracShort;
  unsigned char gpsFracChar;
  unsigned char  uint_data;
  unsigned int offset=0;

  switch (record_id) {
	case 127:
      offset = 1;
      extract_uint4 (handle, rec, &offset, &gpsMinute);
      extract_uint2 (handle, rec, &offset, &gpsFracShort);
      *minutes = gpsMinute;
      *m_seconds = gpsFracShort;
      break;
	case 0:
      extract_uint4 (handle, rec, &offset, &gpsMinute);
      extract_uint1 (handle, rec, &offset, &gpsFracChar);
      *minutes = gpsMinute;
      *m_seconds = (gpsFracChar * 250);
      break;
	case 1:
      extract_uint1 (handle, rec, &offset, &uint_data);
      if ((uint_data >> 5) & 0x01) return(-1);
      extract_uint4 (handle, rec, &offset, &gpsSeconds);
      *m_seconds = gpsSeconds * 1000;
      break;
  }
//  for (; *m_seconds > 60000; *m_seconds -= 60000) {
//    *minutes++;
//  }
  return(0);

}


///---------------------------------------------------------------------------
// Binex record-specific routines
//---------------------------------------------------------------------------

// GNSS Obs record type 0x7f, subtype 0
// data is a 2D matrix: sat id x data type
// data type = ca, p1, p2, l1ca, l1p1, l2p2, l1snr, l2snr, l1rnxsnr, l2rnxsnr, rxchan, satids, error
int read_0x7f00 (binexHandle *handle, unsigned char *rec, unsigned int *offset, int *rc, 
			      double *time, int *nsats, int *rxfmt, char *data) {

  unsigned char  subtype;
  unsigned int   gpsMinute;
  unsigned short gpsFrac;
  unsigned char  nsv_rxfmt;
  unsigned char  satid;
  unsigned char  uint_data;
  double   ca, val, l1ca;
  int id [NSATS]; // satellite id list
  int i,j;
  bool readallerrors = 1;

  // offsets into the ddata array
#define IDX(ARR,X,Y) *(ARR + (X) + (NSATS * Y))
  double * ddata   = (double *)data;
  enum datanames {ca_, p1_, p2_, l1ca_, l1p1_, l2p2_, l1snr_, l2snr_, l1rnxsnr_, l2rnxsnr_, rxchan_, error_};

  // set data to all missing values
  for (i=0;i<NSATS;i++) {
    for (j=0;j<12;j++) {
      IDX(ddata,i,j) = D_NO_DATA;
    }
  }
  
  extract_uint1 (handle, rec, offset, &subtype);
  if (subtype != 0) { 
    fprintf (stderr, "Only handles binex 0x7f subtype 0");
    *rc = BADTYPE;
    return -1;
  }

  extract_uint4 (handle, rec, offset, &gpsMinute);
  extract_uint2 (handle, rec, offset, &gpsFrac);
  extract_uint1 (handle, rec, offset, &nsv_rxfmt);

  *rxfmt = (int)((nsv_rxfmt >> 5) & 0x07);

  *time  = (double)((gpsMinute * 60) + (gpsFrac * 0.001));

  // read in number of satellites and list of sat ids
  *nsats = (int)((nsv_rxfmt & 0x1f) + 0x01);

  // note that the real SV numbers are one greater.
  // this is a zero-based list used in indexing PDLs
  for (i=0;i<*nsats;i++) {
    extract_uint1 (handle, rec, offset, &satid);
    id[i] = satid;
  }

  // loop over satellite, read in data
  for (i=0;i<*nsats;i++) {

    // If the first error byte has its bit zero set, then read in
    // individual error bytes.  Otherwise, use this first error byte
    // for all satellites.
    if (readallerrors) {
      extract_uint1 (handle, rec, offset, &uint_data);
      IDX(ddata,id[i],error_) = (double)uint_data;
      if (((unsigned char)IDX(ddata,id[0],error_) & 0x01) != 1) readallerrors = 0; // only one error byte for all satellites
    } else {
      IDX(ddata,id[i],error_) = IDX(ddata,id[0],error_); 
    }

    extract_uint1 (handle, rec, offset, &uint_data);
    IDX(ddata,id[i],rxchan_) = (double)uint_data;

    binex_extract_mgfz_capi (handle, rec, offset, rc, &ca);
    if (*rc && *rc != NODATAGFZI) { fprintf (stderr, "failed ca read, sat = %d\n", id[i]); return -1; }
    ca *= 0.001;
    IDX(ddata,id[i],ca_) = ca;

    binex_extract_mgfz_capi (handle, rec, offset, rc, &val);
    if (*rc && *rc != NODATAGFZI) { fprintf (stderr, "failed p1 read, sat = %d\n", id[i]); return -1; }
    IDX(ddata,id[i],p1_) = ca - (val * 0.001);

    binex_extract_mgfz_capi (handle, rec, offset, rc, &val);
    if (*rc && *rc != NODATAGFZI) { fprintf (stderr, "failed p2 read, sat = %d\n", id[i]); return -1; }
    IDX(ddata,id[i],p2_) = ca - (val * 0.001);

    binex_extract_mgfz_capi (handle, rec, offset, rc, &val);
    if (*rc && *rc != NODATAGFZI) { fprintf (stderr, "failed l1snr read, sat = %d\n", id[i]); return -1; }
    IDX(ddata,id[i],l1snr_) = val * 0.25;

    binex_extract_mgfz_capi (handle, rec, offset, rc, &val);
    if (*rc && *rc != NODATAGFZI) { fprintf (stderr, "failed l2snr read, sat = %d\n", id[i]); return -1; }
    IDX(ddata,id[i],l2snr_) = val * 0.25;

    extract_uint1 (handle, rec, offset, &uint_data);
    IDX(ddata,id[i],l1rnxsnr_) = (double)uint_data;
    extract_uint1 (handle, rec, offset, &uint_data);
    IDX(ddata,id[i],l2rnxsnr_) = (double)uint_data;

    binex_extract_mgfz_capi (handle, rec, offset, rc, &l1ca);
    if (*rc && *rc != NODATAGFZI) { fprintf (stderr, "failed l1ca read, sat = %d\n", id[i]); return -1; }
    l1ca *= 0.0001;
    IDX(ddata,id[i],l1ca_) = l1ca;

    binex_extract_mgfz_capi (handle, rec, offset, rc, &val);
    if (*rc && *rc != NODATAGFZI) { fprintf (stderr, "failed l1p1 read, sat = %d\n", id[i]); return -1; }
    IDX(ddata,id[i],l1p1_) = l1ca - (val * 0.0001);

    binex_extract_mgfz_capi (handle, rec, offset, rc, &val);
    if (*rc && *rc != NODATAGFZI) { fprintf (stderr, "failed l2p2 read, sat = %d\n", id[i]); return -1; }
    IDX(ddata,id[i],l2p2_) = (l1ca - (val * 0.0001)) * L1_L2;
  }

  return 0;
}

 #ifdef RTCM_ENABLE
/* translation routines */

void setSessionPRN(rtcm_t *rtcm, int index, int SVid)
{
	debug(D_FLOW,"setSessionPRN: Entering...\n");
	for (int j = 0; j < RTCMMAXTYPES; j++) {
		rtcm->msg_data.obs_data[j].obs[index].prn = SVid;
       		rtcm->msg_data.obs_data[j].obs[index].GPS = 0;
		rtcm->msg_data.obs_data[j].obs[index].multiple = 1;
		rtcm->msg_data.obs_data[j].obs[index].usable = false;
	}
	debug(D_FLOW,"setSessionPRN: Exiting...\n");
}

int read_0x7f02(binexHandle * handle, struct rtcm_t *rtcm, unsigned char *rec, size_t in_length)
{
	uint i,n;
	unsigned char subtype;
	unsigned char uc;

	unsigned int   gpsMinute;
	unsigned short gpsFrac;
	unsigned char  nsv_RxFmt;

	unsigned short	bit;
	unsigned int nsats;
	unsigned int offset = 0;

	debug(D_FLOW,"read_0x7f02: Entering...\n");
        extract_uint1 (handle, rec, &offset, &subtype);
        if (subtype != 0x02) {
                fprintf (stderr, "Only handles binex 0x7f subtype 0x02");
                return(-1);
        }       
                        
        extract_uint4 (handle, rec, &offset, &gpsMinute);
        extract_uint2 (handle, rec, &offset, &gpsFrac);
        extract_uint1 (handle, rec, &offset, &nsv_RxFmt);

        // read in number of satellites and list of sat ids
        nsats = (int)((nsv_RxFmt & 0x1f) + 0x01);

        // note that the real SV numbers are one greater.
        // this is a zero-based list used in indexing PDLs
        for (i=0;i<nsats;i++) {
                extract_uint1 (handle, rec, &offset, &uc);
        }

	unsigned char BitFlags;
	double RxClkOff;
	extract_uint1 (handle, rec, &offset, &BitFlags);
	int rc;

	bool getOffset = 0;
	if ((BitFlags & BIT_0) == 0x01) getOffset = 1;
	if (((BitFlags >> 1) & BIT_0) == 0x01) getOffset = 1;
	if (((BitFlags >> 2) & BIT_0) == 0x01) getOffset = 1;

	if (getOffset) {
    		binex_extract_mgfz_capi (handle, rec, &offset, &rc, &RxClkOff);
    		if (rc && rc != NODATAGFZI) { fprintf (stderr, "failed clock offset read\n"); return(1); }
        }

	int readallobspresent = 1;
	unsigned char obs_present[10];
	unsigned char rxchan_flags = 0;
	char haveObs[24];
	memset((void *)haveObs,0,(size_t)24);
	for (i= n= bit = 0; i < nsats; i++, n++) {
		unsigned int prn_offset = 8L+i;
		unsigned char prn;
		extract_uint1(handle, rec, &prn_offset, &prn);
		if (readallobspresent) {
			bool readNext = 1;
			int t,s;
			for (t = 0, s = 0; (t < 10) && readNext; t++) {
				extract_uint1 (handle, rec, &offset, &obs_present[t]);
				if ((t == 0) && ((obs_present[0] & BIT_0) != 0x01)) readallobspresent = 0;
				for (int u = 0; u < 7; u++) {
					if (( t == 0) && (u==0)) u++;
					haveObs[s++] = ((obs_present[t] >> u) & 0x01);
					debug(D_VALUES,"Setting haveObs[%d] to %d\n",s-1,haveObs[s-1]);
				}
				if (((obs_present[t]>>7) & BIT_0) != 0x01) readNext = 0;
			}
		} else {
			debug(D_VALUES, "read_0x7f02: using existing obs_present value\n");
		}

		extract_uint1 (handle, rec, &offset, &rxchan_flags);
		int channel = (rxchan_flags & 0x1f);
		int as = ((rxchan_flags >> 5) & 0x01);
		int lli_l1 = ((rxchan_flags >> 6) & 0x01);
		int lli_l2 = ((rxchan_flags >> 7) & 0x01);

		unsigned char l1snr,l2snr;
		extract_uint1 (handle, rec, &offset, &l1snr);
		extract_uint1 (handle, rec, &offset, &l2snr);

		debug(D_VALUES, "read_0x7f02: channel = %d, as = %d, lli_l1 = %d, lli_l2 = %d\n",channel, as, lli_l1, lli_l2);

		setSessionPRN(rtcm, n, prn + 1);

		double ca;
		bool foundobs = false;
		if (haveObs[0]) {
    			binex_extract_mgfz_capi (handle, rec, &offset, &rc, &ca);
    			if (rc && rc != NODATAGFZI) { fprintf (stderr, "failed ca read, sat = %d\n", prn); return(1); }
			ca *= 0.001;
			debug(D_VALUES,"read_0x7f02: ca = %f\n",ca);
			rtcm->msg_data.obs_data[RTCM_C1].obs[n].value = ca;
			rtcm->msg_data.obs_data[RTCM_C1].obs[n].cacode = 0;
			rtcm->msg_data.obs_data[RTCM_C1].obs[n].quality = 0;
			rtcm->msg_data.obs_data[RTCM_C1].obs[n].usable = true;
			rtcm->msg_data.obs_data[RTCM_C1].obs_avail = true;
			foundobs = true;
		}
		double p1;
		if (haveObs[1]) {
    			binex_extract_mgfz_capi (handle, rec, &offset, &rc, &p1);
    			if (rc && rc != NODATAGFZI) { fprintf (stderr, "failed p1 read, sat = %d\n", prn); return(1); }
			p1 *= 0.001;
			if (haveObs[0]) p1 = ca - p1;
			debug(D_VALUES,"read_0x7f02: p1 = %f\n",p1);
			rtcm->msg_data.obs_data[RTCM_P1].obs[n].value = p1;
			rtcm->msg_data.obs_data[RTCM_P1].obs[n].cacode = 1;
			rtcm->msg_data.obs_data[RTCM_P1].obs[n].quality = 0;
			rtcm->msg_data.obs_data[RTCM_P1].obs[n].usable = true;
			rtcm->msg_data.obs_data[RTCM_P1].obs_avail = true;
			foundobs = true;
		}
		double p2;
		if (haveObs[2]) {
    			binex_extract_mgfz_capi (handle, rec, &offset, &rc, &p2);
    			if (rc && rc != NODATAGFZI) { fprintf (stderr, "failed p2 read, sat = %d\n", prn); return(1); }
			p2 *= 0.001;
			if (haveObs[0]) {
				p2 = ca - p2;
			} else {
				p2 = p1 - p2;
			}
			debug(D_VALUES,"read_0x7f02: p2 = %f\n",p2);
			rtcm->msg_data.obs_data[RTCM_P2].obs[n].value = p2;
			rtcm->msg_data.obs_data[RTCM_P2].obs[n].cacode = 1;
			rtcm->msg_data.obs_data[RTCM_P2].obs[n].quality = 0;
			rtcm->msg_data.obs_data[RTCM_P2].obs[n].usable = true;
			rtcm->msg_data.obs_data[RTCM_P2].obs_avail = true;
			foundobs = true;
		}
		double l1;
		if (haveObs[3]) {
    			binex_extract_mgfz_capi (handle, rec, &offset, &rc, &l1);
    			if (rc && rc != NODATAGFZI) { fprintf (stderr, "failed l1 read, sat = %d\n", prn); return(1); }
			l1 *= 0.0001;
			debug(D_VALUES,"read_0x7f02: l1 = %f\n",l1);
			rtcm->msg_data.obs_data[RTCM_L1].obs[n].value = l1; 
			rtcm->msg_data.obs_data[RTCM_L1].obs[n].cacode = 0;
			rtcm->msg_data.obs_data[RTCM_L1].obs[n].continuityloss = lli_l1;
			rtcm->msg_data.obs_data[RTCM_L1].obs[n].quality = 0;
			rtcm->msg_data.obs_data[RTCM_L1].obs[n].usable = true;
			rtcm->msg_data.obs_data[RTCM_L1].obs_avail = true;
			foundobs = true;
		}
		double l2;
		if (haveObs[4]) {
    			binex_extract_mgfz_capi (handle, rec, &offset, &rc, &l2);
    			if (rc && rc != NODATAGFZI) { fprintf (stderr, "failed l2 read, sat = %d\n", prn); return(1); }
			l2 *= 0.0001;
			debug(D_VALUES,"read_0x7f02: l2 = %f - %f * 77/60\n",l1,l2);
			l2 = l1 - l2 * 77/60;
			debug(D_VALUES,"read_0x7f02: l2 = %f\n",l2);
			rtcm->msg_data.obs_data[RTCM_L2].obs[n].value = l2; 
			rtcm->msg_data.obs_data[RTCM_L2].obs[n].cacode = 0;
			rtcm->msg_data.obs_data[RTCM_L2].obs[n].continuityloss = lli_l2;
			rtcm->msg_data.obs_data[RTCM_L2].obs[n].quality = 0;
			rtcm->msg_data.obs_data[RTCM_L2].obs[n].usable = true;
			rtcm->msg_data.obs_data[RTCM_L2].obs_avail = true;
			foundobs = true;
		}
		double d1;
		if (haveObs[5]) {
    			binex_extract_mgfz_capi (handle, rec, &offset, &rc, &d1);
    			if (rc && rc != NODATAGFZI) { fprintf (stderr, "failed d1 read, sat = %d\n", prn); return(1); }
			d1 *= 0.001;
			debug(D_VALUES,"read_0x7f02: d1 = %f\n",d1);
		}
		if (foundobs) rtcm->msg_data.numobs++;


	}
	debug(D_FLOW,"read_0x7f02: Exiting...\n");
	return(0);
}


int read_0x7f03(binexHandle *handle, struct rtcm_t *rtcm, unsigned char *rec, size_t in_length)
{

//	unsigned long	ul;
	uint i,n;
	unsigned char subtype;
	unsigned int ui;

	unsigned int   gpsMinute;
	unsigned short gpsFrac;
	unsigned char  nsv_flags;
	unsigned short us;

	unsigned short	bit;
	unsigned int nsats;
	unsigned int offset = 0;
	SAT_ID system;

	// Loop variables
	int readallobspresent = 1;
	unsigned char obs_present[10];
	unsigned char rxchan_flags = 0;
	char haveObs[24];
	memset((void *)haveObs,0,(size_t)24);

	debug(D_FLOW,"read_0x7f03: Entering...\n");
        extract_uint1 (handle, rec, &offset, &subtype);
        if (subtype != 0x03) {
                fprintf (stderr, "Only handles binex 0x7f subtype 0x03");
                return(-1);
        }       
                        
        extract_uint4 (handle, rec, &offset, &gpsMinute);
        extract_uint2 (handle, rec, &offset, &gpsFrac);
        extract_uint1 (handle, rec, &offset, &nsv_flags);

        // read in number of satellites and list of sat ids
        nsats = (int)((nsv_flags & 0x1f) + 0x01);

	bool getOffset = 0;
	bool getBitFlags = 0;
	if (((nsv_flags >> 7) & 0x01) == 0x01) getBitFlags = 1;
	if (((nsv_flags >> 6) & 0x01) == 0x01) getOffset = 1;

//	double RxClkOff;
	unsigned char BitFlags;
	if (getBitFlags) {
		do {
			extract_uint1 (handle, rec, &offset, &BitFlags);
		} while (((BitFlags >> 7) & 0x01) == 0x01);
	}

	if (getOffset) {
		extract_uint3 (handle, rec, &offset, &ui);
		int sign = 1;
		if (((ui >> 22) & 0x01) == 0x01) sign = -1;
	}

	// Set offsets for reading during loop
	// prn_offset is fixed offset to get satellite PRN
	// offset is increased by number of satellites so that 
	// records can be found properly
	unsigned int prn_offset = offset;
	offset += nsats;

	for (i= n= bit = 0; i < nsats; i++, n++) {
        	// note that the real SV numbers are one greater.
        	// this is a zero-based list used in indexing PDLs
		unsigned int t_off = prn_offset + i;
		unsigned char prn;
		extract_uint1(handle, rec, &t_off, &prn);
		binex_extract_SV_id(prn, &system);

		debug(D_VALUES, "read_0x7f03: Reading Observations for TYPE %d, SVid %d\n", system.sys, system.SVid);
		switch (system.sys) {
			case NAVSTAR_GPS:
				debug(D_VALUES, "read_0x7f03: Reading Observations for GPS SAT PRN %d\n", system.SVid);
				break;
			case GLONASS:
				debug(D_VALUES, "read_0x7f03: Reading Observations for GLONASS SAT slot# %d, skipping decode\n", system.SVid);
				continue;
			default:
				debug(D_VALUES, "read_0x7f03: Unsupported Satellite System (%d) with SVid %d, skipping decode\n", system.sys, system.SVid);
				continue;
		}		

		if (readallobspresent) {
			bool readNext = 1;
			int t,s;
			for (t = 0, s = 0; (t < 10) && readNext; t++) {
				extract_uint1 (handle, rec, &offset, &obs_present[t]);
				if ((t == 0) && ((obs_present[0] & BIT_0) != 0x01)) readallobspresent = 0;
				for (int u = 0; u < 7; u++) {
					if (( t == 0) && (u==0)) u++;
					haveObs[s++] = ((obs_present[t] >> u) & 0x01);
				}
				if (((obs_present[t]>>7) & BIT_0) != 0x01) readNext = 0;
			}
		} else {
			debug(D_VALUES, "read_0x7f03: using existing obs_present value\n");
		}

		extract_uint1 (handle, rec, &offset, &rxchan_flags);
		double scalePhase = 0.02;
		if ((rxchan_flags >> 6) & 0x01) scalePhase = 0.10;
		int channel   = (rxchan_flags & 0x1f);
		int as        = ((rxchan_flags >> 5) & 0x01);
		int unhealthy = ((rxchan_flags >> 7) & 0x01);

		debug(D_VALUES,"read_0x7f03: channel = %d, as = %d, unhealthy = %d\n",channel,as,unhealthy);
		setSessionPRN(rtcm, n, system.SVid);

		double ca = 0.0;
		double l1 = 0.0;
		double p1 = 0.0;
		double p2 = 0.0;
		double l2 = 0.0;
#pragma pack(1)
#ifndef WORDS_BIGENDIAN
		struct obs1_ {
			unsigned char L1CNo_MSB;
			struct {
				uint _pad:24;
				uint L1range_lo:32;
				uint L1range_hi:4;
				uint P:1;
				uint lli:1;
				int L1CNo_LSB:2;
			} range;
			struct {
				uint _pad:8;
				uint phase_lo:16;
				uint phase_hi:7;
				uint sign:1;
			} phase;
		} obs1 __attribute((aligned(1)));

		struct obs2 {
			unsigned char L2CNo_MSB;
			struct {
				uint _pad:8;
				uint L2range_lo:16;
				uint L2range_hi:3;
				uint sign:1;
				uint L2CNo_LSB:2;
				uint lli:1;
				uint P:1;
			} range;
			struct {
				uint _pad:8;
				uint phase_lo:16;
				uint phase_hi:7;
				uint sign:1;
			} phase;
		} obs2 __attribute((aligned(1)));
#else
		struct obs1_ {
			unsigned char L1CNo_MSB;
			struct {
				int L1CNo_LSB:2;
				uint lli:1;
				uint P:1;
				uint L1range_hi:4;
				uint L1range_lo:32;
				uint _pad:24;
			} range;
			struct {
				uint sign:1;
				uint phase_hi:7;
				uint phase_lo:16;
				uint _pad:8;
			} phase;
		} obs1 __attribute((aligned(1)));

		struct obs2_ {
			unsigned char L2CNo_MSB;
			struct {
				uint P:1;
				uint lli:1;
				uint L2CNo_LSB:2;
				uint sign:1;
				uint L2range_hi:3;
				uint L2range_lo:16;
				uint _pad:8;
			} range;
			struct {
				uint sign:1;
				uint phase_hi:7;
				uint phase_lo:16;
				uint _pad:8;
			} phase;
		} obs2 __attribute((aligned(1)));
#endif
#pragma pack()
		memset((void*)&obs2,0,sizeof(obs2));
		memset((void*)&obs1,0,sizeof(obs1));
		unsigned long long ull;
		bool foundobs = false;

		if (haveObs[0]) {
    			extract_uint1 (handle, rec, &offset, &(obs1.L1CNo_MSB));

    			extract_uint5 (handle, rec, &offset, &ull);
			memcpy((void *)&(obs1.range), &ull, sizeof(ull));
			ull = obs1.range.L1range_hi;
			ull <<= 32;
			ull += obs1.range.L1range_lo;
			ca = (double)ull * 0.001;

    			extract_uint3 (handle, rec, &offset, &ui);
			memcpy((void *)&(obs1.phase), &ui, sizeof(ui));
			ull = obs1.phase.phase_hi;
			ull <<= 16;
			ull += obs1.phase.phase_lo;
			double delta_m = (double)ull;

			double l1dBHz = (double)(((obs1.L1CNo_MSB << 2) + obs1.range.L1CNo_LSB)) * 0.1;

			if (obs1.phase.sign) delta_m *= -1;
			delta_m *= (scalePhase * 0.001);
			l1 = FREQ1 * ((delta_m + ca)/LIGHTC);

			if (ca != 0.0) {
				if (obs1.range.P) {
					p1 = ca;
					rtcm->msg_data.obs_data[RTCM_P1].obs[n].value = p1;
					rtcm->msg_data.obs_data[RTCM_P1].obs[n].cacode = 1;
					rtcm->msg_data.obs_data[RTCM_P1].obs[n].usable = true;
					rtcm->msg_data.obs_data[RTCM_P1].obs_avail = true;
				} else {
					rtcm->msg_data.obs_data[RTCM_C1].obs[n].value = ca;
					rtcm->msg_data.obs_data[RTCM_C1].obs[n].cacode = 0;
					rtcm->msg_data.obs_data[RTCM_C1].obs[n].usable = true;
					rtcm->msg_data.obs_data[RTCM_C1].obs_avail = true;
				}
				rtcm->msg_data.obs_data[RTCM_L1].obs[n].value = l1; 
				rtcm->msg_data.obs_data[RTCM_L1].obs[n].cacode = 0;
				rtcm->msg_data.obs_data[RTCM_L1].obs[n].quality = ((l1dBHz / .25) > 255) ? 255 : (unsigned int)(l1dBHz/.25) ;
				rtcm->msg_data.obs_data[RTCM_L1].obs[n].usable = true;
				rtcm->msg_data.obs_data[RTCM_L1].obs_avail = true;
				foundobs = true;
			}
		}
		if (haveObs[1]) {
			extract_uint1 (handle, rec, &offset, &(obs2.L2CNo_MSB));

			if (haveObs[0]) {
    				extract_uint3 (handle, rec, &offset, &ui);
				memcpy((void *)&(obs2.range), &ui, sizeof(ui));
				ull = obs2.range.L2range_hi;
				ull <<= 16;
				ull += obs2.range.L2range_lo;
				p2 = ((double)ull * 0.001);
				if (obs2.range.sign) p2 *= -1;
				p2 += ca;
			} else {
				// same definitoin as the obs1 record
    				extract_uint5 (handle, rec, &offset, &ull);
				memcpy((void *)&(obs1.range), &ull, sizeof(ull));
				ull = obs1.range.L1range_hi;
				ull <<= 32;
				ull += obs1.range.L1range_lo;
				p2 = (double)ull * 0.001;
			}

			extract_uint3 (handle, rec, &offset, &ui);
			memcpy((void *)&(obs2.phase), &ui, sizeof(ui));
			ull = obs2.phase.phase_hi;
			ull <<= 16;
			ull += obs2.phase.phase_lo;
			double delta_m = (double)ull;

			double l2dBHz = (double)(((obs2.L2CNo_MSB << 2) + obs2.range.L2CNo_LSB)) * 0.1;

			if (obs2.phase.sign) delta_m *= -1;
			delta_m *= (scalePhase * 0.001);
			l2 = FREQ2 * ((delta_m + p2)/LIGHTC);

			if (p2 != 0.0) {
				rtcm->msg_data.obs_data[RTCM_P2].obs[n].value = p2;
				rtcm->msg_data.obs_data[RTCM_P2].obs[n].cacode = 1;
				rtcm->msg_data.obs_data[RTCM_P2].obs[n].usable = true;
				rtcm->msg_data.obs_data[RTCM_L2].obs[n].value = l2; 
				rtcm->msg_data.obs_data[RTCM_L2].obs[n].cacode = 0;
				rtcm->msg_data.obs_data[RTCM_L2].obs[n].usable = true;
				rtcm->msg_data.obs_data[RTCM_L2].obs_avail = true;
				rtcm->msg_data.obs_data[RTCM_L2].obs[n].quality = ((l2dBHz / .25) > 255) ? 255 : (unsigned int)(l2dBHz/.25) ;
				rtcm->msg_data.obs_data[RTCM_P2].obs_avail = true;
				foundobs = true;
			}
		}
		if (haveObs[2]) {
			extract_uint3 (handle, rec, &offset, &ui);
		}
		if (haveObs[3]) {
			extract_uint2 (handle, rec, &offset, &us);
		}
		rtcm->msg_data.numobs++;
#if DEBUG
		debug(D_VALUES,"read_0x7f03(prn = %d): %f %f %f %f %f\n",system.SVid, l1, l2, ca, p2, p1);
#endif

	}
	debug(D_FLOW,"read_0x7f03: Exiting...\n");
	return(0);
}


uint2
binex_7f_05_CNo (uint1 Msnr, uint1 Lsnr)
{
	uint2 CNo;

#if DEBUG
	fprintf(err.fp, "binex_7f_05_CNo()\n");
#endif
	CNo= ((uint2)Msnr << 2);	/* x4 to get units of 0.1 dBHz */
	switch (Lsnr & 0xc0) {		/* the +/-0.1 dBHz correction is in bits 6 and 7 of Lsnr */
	case 0x40: CNo++; break;
	case 0xc0: CNo--; break;
	case 0x80: CNo-= 0x0002; break;
	}

	return(CNo);
}

void binex_7f_05_signal_map (uint1 SV_type, uint1 id, uint1 *Ln, uint1 *Pn, uint1 *Dn, uint1 *Sn, uint1 *as, real8 *mpc, esint1 fcn)
{
#if DEBUG
	fprintf(stderr, "binex_7f_05_signal_map()\n");
#endif
	*Ln= *Pn= *Dn= *Sn= ZZ;
	*as= 0x00;
	*mpc = 0.0;

	// mpc = meters per cycle; for GLONASS varies with the satellite's requency Channel Number (FCN)
	switch (SV_type) {
	case NAVSTAR_GPS:
		switch (id) {
		case 0:	/* L1 unknown */
		case 1:	/* L1C/A */
			*Pn= C1;
			*Ln= L1, *Dn= D1, *Sn= S1;
			*mpc= meters_per_GPS_L1_cycle;
			break;

		case 3:	/* L1P Z-tracking */
			*as= BIT_2;
			/* no break
			*/
		case 2:	/* L1P */
		case 4:	/* L1P(Y) */
			*Pn= P1;
			*Ln= L1, *Dn= D1, *Sn= S1;
			*mpc= meters_per_GPS_L1_cycle;
			break;

		case 17:/* L2P Z-tracking */
			*as= BIT_2;
			/* no break
			*/
		case 10:/* L2 unknown */
		case 12:/* L1C/A+(P2-P1) */
		case 16:/* L2P */
		case 18:/* L2P(Y) */
			*Pn= P2;
			*Ln= L2, *Dn= D2, *Sn= S2;
			*mpc= meters_per_GPS_L2_cycle;
			break;

		case 13:/* L2C (M) in 0x7f-05 */
		case 14:/* L2C (L) in 0x7f-05 */
		case 15:/* L2C (M+L) in 0x7f-05 */
			*Pn= C2;
			*Ln= L2, *Dn= D2, *Sn= S2;
			*mpc= meters_per_GPS_L2_cycle;
			break;

		case 23:/* L5 unknown */
		case 24:/* L5 (I) */
		case 25:/* L5 (Q) */
		case 26:/* L5 (I+Q) */
			*Ln= L5; *Pn= C5; *Dn= D5; *Sn= S5;
			*mpc= meters_per_GPS_L5_cycle;
			break;
		}
		break;

	case GLONASS:
		switch (id) {
		case 0:	/* L1 unknown */
		case 1:	/* L1C/A */
			*Ln= L1; *Pn= C1; *Dn= D1; *Sn= S1;
			switch (fcn) {
			case UNKNOWN_GLONASS_FCN:
				*mpc= 0.;
				break;
			default:
				*mpc= c_WGS84/(GLONASS_f_L1 + (real8)fcn*GLONASS_n_L1);
			}
			break;

		case 2:	/* L1P */
			*Ln= L1; *Pn= P1; *Dn= D1; *Sn= S1;
			switch (fcn) {
			case UNKNOWN_GLONASS_FCN:
				*mpc= 0.;
				break;
			default:
				*mpc= c_WGS84/(GLONASS_f_L1 + (real8)fcn*GLONASS_n_L1);
			}
			break;

		case 10:/* L2 unknown */
		case 11:/* L2C/A */
			*Ln= L2; *Pn= C2; *Dn= D2; *Sn= S2;
			switch (fcn) {
			case UNKNOWN_GLONASS_FCN:
				*mpc= 0.;
				break;
			default:
				*mpc= c_WGS84/(GLONASS_f_L2 + (real8)fcn*GLONASS_n_L2);
			}
			break;

		case 12:/* L2P */
			*Ln= L2; *Pn= P2; *Dn= D2; *Sn= S2;
			switch (fcn) {
			case UNKNOWN_GLONASS_FCN:
				*mpc= 0.;
				break;
			default:
				*mpc= c_WGS84/(GLONASS_f_L2 + (real8)fcn*GLONASS_n_L2);
			}
			break;
		}
		break;


	case SBAS:
		switch (id) {
		case 0:	/* L1 unknown */
		case 1:	/* L1C/A */
			*Ln= L1; *Pn= C1; *Dn= D1; *Sn= S1;
			*mpc= meters_per_GPS_L1_cycle;
			break;

		case 6:	/* L5 unknown */
		case 7:	/* L5 (I) */
		case 8:	/* L5 (Q) */
		case 9:	/* L5 (I+Q) */
			*Ln= L5; *Pn= C5; *Dn= D5; *Sn= S5;
			*mpc= meters_per_GPS_L5_cycle;
			break;
		}
		break;
	case Galileo:
		switch (id) {
		case 0:	/* E1 unknown */
		case 1:	/* E1 A PRS */
		case 2:	/* E1 B INAV OS/CS/SoL */
		case 3:	/* E1 no data */
		case 4:	/* E1 B+C */
		case 5:	/* E1 A+B+C */
			*Ln= L1; *Pn= C1; *Dn= D1; *Sn= S1;
			*mpc= 0.; // *mpc= meters_per_Galileo_L1_cycle;
			break;
		case 6: /* E5a unknown */
		case 7: /* E5a I FNAV OS */
		case 8: /* E5a Q no data */
		case 9: /* E5a I+Q */
			*Ln= L5; *Pn= C5; *Dn= D5; *Sn= S5;
			*mpc= 0.; // *mpc= meters_per_Galileo_L5a_cycle;
			break;
		case 10: /* E5b unknown */
		case 11: /* E5b I INAV OS/CS/SoL */
		case 12: /* E5b Q no data */
		case 13: /* E5b I+Q */
			*Ln= L7; *Pn= C7; *Dn= D7; *Sn= S7;
			*mpc= 0.; // *mpc= meters_per_Galileo_L5b_cycle;
			break;
		case 14: /* E5a+b unknown */
		case 15: /* E5a+b I */
		case 16: /* E5a+b Q */
		case 17: /* E5a+b I+Q */
			*Ln= L8; *Pn= C8; *Dn= D8; *Sn= S8;
			*mpc= 0.; // *mpc= meters_per_Galileo_L5ab_cycle;
			break;
		case 18: /* E6 unknown */
		case 19: /* E6 A PRS */
		case 20: /* E6 B CNAV CS */
		case 21: /* E6 C no data */
		case 22: /* E6 B+C */
		case 23: /* E6 A+B+C */
			*Ln= L6; *Pn= C6; *Dn= D6; *Sn= S6;
			*mpc= 0.; // *mpc= meters_per_Galileo_L6_cycle;
			break;
		}
		break;
	case Beidou:
		switch (id) {
		case 0:  /* B1/E2 unknown */
		case 1:  /* B1/E2 I */
		case 2:  /* B1/E2 Q */
		case 3:  /* B1/E2 I+Q */
			*Ln= L1; *Pn= C1; *Dn= D1; *Sn= S1;
			*mpc= 0.; // *mpc= meters_per_Beidou_B1_cycle;
			break;
		case 4:  /* B2/E5b unknown */
		case 5:  /* B2/E5b I */
		case 6:  /* B2/E5b Q */
		case 7:  /* B2/E5b I+Q */
			*Ln= L7; *Pn= C7; *Dn= D7; *Sn= S7;
			*mpc= 0.; // *mpc= meters_per_Beidou_B2_cycle;
			break;
		case 8:  /* B3/E6 unknown */
		case 9:  /* B3/E6 I */
		case 10: /* B3/E6 Q */
		case 11: /* B3/E6 I+Q */
			*Ln= L6; *Pn= C6; *Dn= D6; *Sn= S6;
			*mpc = 0.; // *mpc= meters_per_Beidou_B3_cycle;
			break;
		case 12: /* B1-2/E1 unknown */
		case 13: /* B1-2/E1 I */
		case 14: /* B1-2/E1 Q */
		case 15: /* B1-2/E1 I+Q */
			*Ln= L2; *Pn= C2; *Dn= D2; *Sn= S2;
			*mpc= 0.; // *mpc= meters_per_Beidou_B1_2_cycle;
			break;
		}
		break;
	case QZSS:
		switch (id) {
		case 0:  /* L1 unknown */
		case 1:  /* L1 C/A */
		case 2:  /* L1C (data) */
		case 3:  /* L1C (pilot) */
		case 4:  /* L1C (data + pilot) */
			*Ln= L1; *Pn= C1; *Dn= D1; *Sn= S1;
			*mpc= meters_per_GPS_L1_cycle;
			break;
		case 7:  /* L2 unknown */
		case 8:  /* L2C M (data) */
		case 9:  /* L2C L (no data) */
		case 10: /* L2C M+L  */
			*Ln= L2; *Pn= C2; *Dn= D2; *Sn= S2;
			*mpc= meters_per_GPS_L2_cycle;
			break;
		case 13: /* L5 unknown */
		case 14: /* L5 I */
		case 15: /* L5 Q */
		case 16: /* L5 I+Q */
			*Ln= L5; *Pn= C5; *Dn= D5; *Sn= S5;
			*mpc= meters_per_GPS_L5_cycle;
			break;
		case 19: /* LEX unknown */
		case 20: /* LEX S (data) */
		case 21: /* LEX L (no data) */
		case 22: /* LEX S+L */
			*Ln= L6; *Pn= C6; *Dn= D6; *Sn= S6;
			*mpc= 0.; // *mpc= meters_per_QZSS_LEX_cycle;
			break;
		}
		break;
	case IRNSS:
		switch (id) {
		case 0:  /* L5 unknown */
		case 1:  /* L5 SPS BPSK(1) */
		case 2:  /* L5 RS BOC(5,2) Data */
		case 3:  /* L5 RS BOC(5,2) Pilot */
		case 4:  /* L5 RS BOC(5,2) Data+Pilot */
			*Ln= L5; *Pn= C5; *Dn= D5; *Sn= S5;
			*mpc= meters_per_GPS_L5_cycle;
			break;
		}
		break;
	default:
		debug(D_VALUES, "binex_7f_05_signal_map: Unsupported Satellite System (%d)\n", SV_type);
		break;
	}
}


int read_0x7f05(binexHandle *handle, struct rtcm_t *rtcm, unsigned char *rec, size_t in_length)
{
	// lli = loss-of-lock indicator (a.k.a cycle-slip) for all carrier frequencies
	// as  = anti-spoofing (a.k.a A/S) staus of each GPS satellite being tracked (on/off)
	bool valid;
	uint1	blk, blks, Code, Flg0, Flg0_ref, Flgn, Ln, Pn, Dn, Sn, Msnr, Lsnr, lli, as;
	esint1	fcn = UNKNOWN_GLONASS_FCN;
	uint2	CNo;
	sint2	s;	
	real8   rcvrClkOffset = 0.0, d, ref_range, range, phase, scale, mpc, snr, doppler;

	unsigned int   gpsMinute, i, nsats,  offset = 0, off, unhealthy=0;
	unsigned short gpsFrac;
	unsigned char  subtype, nsv_flags;

	SAT_ID system;
	signed short int rxMsJump = 0;
	
	debug(D_FLOW,"read_0x7f05: Entering...\n");
        extract_uint1 (handle, rec, &offset, &subtype);
        if (subtype != 0x05) {
                fprintf (stderr, "Only handles binex 0x7f subtype 0x05");
                return(-1);
        }       
                        
        extract_uint4 (handle, rec, &offset, &gpsMinute);
        extract_uint2 (handle, rec, &offset, &gpsFrac);
        extract_uint1 (handle, rec, &offset, &nsv_flags);

        // read in number of SVs
        nsats = (int)((nsv_flags & 0x3f) + 0x01);
	debug(D_VALUES, "read_0x7f05: #satellites=%d\n", nsats);

	// N SVs + Flags : 1 byte
	bool getSysTimeHeader = 0, getOffset = 0;
	if (((nsv_flags >> 6) & 0x01) == 0x01) getSysTimeHeader = 1; // SysTimeHeader present
	if (((nsv_flags >> 7) & 0x01) == 0x01) getOffset = 1;        // rcvr clk offset present

	// FOO - is receiver clock offset used/needed?
	if (getOffset) {
		// offset gets automatically incremented by 3 bytes
		extract_2cNb(handle, rec, &offset, 2, 22, &valid, &rcvrClkOffset);
		rcvrClkOffset /= 1e9; // nanosecs to secs
		if (!valid) rcvrClkOffset = 0.0;

		// get millisecond reset from bits 22-23 of 3-byte RxClkOff
		switch (rec[offset + (handle->little_endian ? 2 : 0)] & 0xc0) {
		case 0x40:	/* "01" == +ms reset has occurred at this epoch */
		  rxMsJump= (esint1)1;
		  break;
		case 0x00:	/* "00" == no ms reset has occurred at this epoch */
		  rxMsJump= (esint1)0;
		  break;
		case 0xc0:	/* 11 == -ms reset has occurred at this epoch */
		  rxMsJump= (esint1)-1;
		  break;
		default:	/* 10 == invalid value */
		  rxMsJump= (esint1)0;
		  break;
		}
	}

	// any SysTimeHeader and SysTimeOffest bytes - skip over
	if (getSysTimeHeader) offset += 1 + 4*(rec[offset] >> 4 & 0xf);
#ifdef FOO
	// do something with SysTimeOffsets?
	unsigned char ref_system_ID = 0; // GPS by default
	unsigned char numSysTimeOffsets = 0; // none by default
	unsigned int SysTimeOffsets[16]; // for each int, bits 0-3 : Sat system ID, bits 4-7 : reserved, bits 8-31 : clock offset
	unsigned char SysTimeHeader;
	if (getSysTimeHeader) {
		extract_uint1 (handle, rec, &offset, &SysTimeHeader);
		ref_system_ID = SysTimeHeader & 0xf;
		numSysTimeOffsets = SysTimeHeader >> 4 & 0xf;
		for (i=0; i < numSysTimeOffsets; i++) {
			extract_uint4 (handle, rec, &offset, &SysTimeOffsets[i]);
		}
	}
#endif

	int n = rtcm->msg_data.numobs = 0; // RTM obs index for current SV (GPS and GLONASS only)

	// loop through the SVs
	for (i= 0; i < nsats; i++) {
		bool haveObs = false;
		debug(D_FLOW, "FOO:1:offset=%d\n", offset);

        	// SVid = PRN for CDMA systems, slot # for FDMA GLONASS
		extract_uint1(handle, rec, &offset, &system.SVid);
		debug(D_FLOW, "FOO:2:offset=%d\n", offset);

		// parse this byte: Satellite system, number of obs blocks, health
		system.sys = rec[offset] & 0xf;
		blks = rec[offset] >> 4 & 0x7;
		unhealthy = rec[offset] >> 7 & 0x1;

		// BINEX stream can have any of these known constellations; otherwise we have a problem!
		debug(D_VALUES, "read_0x7f05: Reading Observations for sat#=%d SVsys=%d, SVid=%3d, #blks=%d, unhealthy=%d\n",
		      i, system.sys, system.SVid, blks, unhealthy);
		switch (system.sys) {
		case NAVSTAR_GPS:
		case GLONASS:
		case SBAS:
		case Galileo:
		case Beidou:
		case QZSS:
		case IRNSS:
			break;			
		default:
		        debug(D_VALUES, "read_0x7f05: Unknown Satellite System (%d) with SVid %d, skipping decode\n", system.sys, system.SVid);
			break;
		}
		offset++; /* advance to obs blocks */
		debug(D_FLOW, "FOO:3:offset=%d\n", offset);

		for (blk= 0; blk < blks; blk++) {
			unsigned int numObsFlags = 0;
			Code = rec[offset++]; // observation code byte; save for later
			debug(D_FLOW, "FOO:4:sat#=%d blk=%d offset=%d\n", i, blk, offset);
			switch (Code & BIT_7) {
			case 0x0:
				switch (blk) {
				case 0:
					Flg0_ref = Flg0 = 0x0;	/* set ObsFlags(0) to all 0 */
					break;
				default:
					Flg0 = Flg0_ref;        /* set ObsFlags(0) from reference block 0 */
				}
				break;
			default:
		        	Flg0 = 0x0;	/* default if ObsFlags(0) is not present */
				do {
				        debug(D_FLOW, "FOO:5:offset=%d\n", offset);
			       		/* advance and read any pertinent parts of any ObsFlags(0) - ObsFlags(3):
			    		*/
					switch (rec[offset] & 0x3) {	/* the ID of the ObsFlags() byte */
					case 0x0:
						Flg0 = rec[offset]; // save ObsFlag(0) for later use
						break;
			        	case 0x2: // ObsFlag(2)
						Flgn = rec[offset] >> 2;
						off = 0;
						debug(D_FLOW, "FOO:5.5:offset=%d\n", offset);
						extract_2cNb(handle, &Flgn, &off, 4, 4, &valid, &d);
						fcn = (esint1)d;
						break;
					default:
						break;
					}
					numObsFlags++;
		                        debug(D_FLOW, "FOO:6:offset=%d\n", offset);
				} while ((rec[offset++] & BIT_7) && numObsFlags < 4);

				switch (blk) {
				case 0:
					Flg0_ref = Flg0;
					break;
				}
			}
			debug(D_FLOW, "FOO:6.5:numObsFlags=%d\n", numObsFlags);

			lli = Code >> 5 & 0x1; // bit 5: 1 = Cycle slip, 0 = otherwise
			binex_7f_05_signal_map(system.sys, Code & 0x1f, &Ln, &Pn, &Dn, &Sn, &as, &mpc, fcn);

			switch (Flg0 & BIT_5) {		/* ReducedPhaseAccuracy; scale in meters/phase-unit */
			case BIT_5:
				scale= 0.10e-3;		/* resolution is 0.10 mm */
				break;
			default:
				scale= 0.02e-3;		/* resolution is 0.02 mm */
				break;
			}

			// carrier-to-noise-density ratio (CNo) MSBs to nearest 0.4dBHz
			debug(D_FLOW, "FOO:7:offset=%d\n", offset);
			extract_uint1(handle, rec, &offset, &Msnr);
			debug(D_FLOW, "FOO:8:offset=%d\n", offset);

			switch (blk) {
			case 0:
				Lsnr = rec[offset + (handle->little_endian ? 4 : 0)];
				CNo = binex_7f_05_CNo(Msnr, Lsnr);
				debug(D_FLOW, "FOO:9:offset=%d\n", offset);
				extract_xNb(handle, rec, &offset, 2, 38, false, &valid, &d);
				debug(D_FLOW, "FOO:10:offset=%d\n", offset);
				ref_range = range = d/1e3;	/* range in meters */
				if (!valid) {
					debug(D_VALUES, "Invalid range=%f(meters) in reference obs blk=%d for SVSys=%d SVid=%d\n",
					   range, blk, system.sys, system.SVid);
				break;
			default:
				switch (Flg0 & BIT_6) {
				case BIT_6:	/* ExpandedDelta */
					Lsnr= rec[offset + (handle->little_endian ? 2 : 0)];
					CNo = binex_7f_05_CNo(Msnr, Lsnr);
					debug(D_FLOW, "FOO:11:offset=%d\n", offset);
					extract_2cNb(handle, rec, &offset, 4, 20, &valid, &d);
					debug(D_FLOW, "FOO:12:offset=%d\n", offset);
					d /= 1e3; /* delta range in meters */
					break;
				default:
					extract_sint2(handle, rec, &offset, &s);
					d = (real8)s/1e3; /* delta range in meters */
					break;
				}
				range = ref_range + d;
				break;
			}
			switch (Flg0 & BIT_6) {
			case BIT_6:	/* ExpandedDelta */
			        debug(D_FLOW, "FOO:13:offset=%d\n", offset);
				extract_2cNb(handle, rec, &offset, 0, 24, &valid, &d);
				debug(D_FLOW, "FOO:14:offset=%d\n", offset);
				phase = scale*d; /* delta phase in meters */
				break;
			default:
				Lsnr= rec[offset + (handle->little_endian ? 2 : 0)];
				CNo = binex_7f_05_CNo(Msnr, Lsnr);
				debug(D_FLOW, "FOO:15:offset=%d\n", offset);
				extract_2cNb(handle, rec, &offset, 2, 22, &valid, &d);
				debug(D_FLOW, "FOO:16:offset=%d\n", offset);
				phase = scale*d; /* delta phase in meters */
				break;
			}
			if (!valid) {
				debug(D_VALUES, "Warning: Invalid expanded phase=%f(meters) computed from reference obs blk=%d bit 6 for SVSys=%d SVid=%d\n",
					   phase, blk, system.sys, system.SVid);
			} else {
				if (mpc > 0.0) {
					phase = (range + phase)/mpc;
				}
			}

			snr = (real8)CNo/10.;

			if (Flg0 & BIT_2) { // Doppler
				debug(D_FLOW, "FOO:17:offset=%d\n", offset);
				extract_2cNb(handle, rec, &offset, 0, 24, &valid, &d);
				debug(D_FLOW, "FOO:18:offset=%d\n", offset);
				doppler = d/256.;  // doppler in Hz
				if (!valid) {
					debug(D_VALUES, "Warning: Invalid doppler=%f(Hz) in reference obs blk=%d for SVSys=%d SVid=%d\n",
					   doppler, blk, system.sys, system.SVid);
				}
			}
			if (Flg0 & BIT_3) { // Slip count
				if (Flg0 & BIT_4) {
				        offset += 2; // 2-byte slip count
				} else {
				        offset++;    // 1-byte slip count
				}
			}
			debug(D_FLOW, "FOO:19:offset=%d\n", offset);
                        } // switch blk

			// translate to RTCM3 (GPS only for now, core dumps with GLONASS)
			if (system.sys != NAVSTAR_GPS) {
			     continue;
			}

                } // for blk

		if (haveObs) {
		        n = rtcm->msg_data.numobs++;
		}

        } // for SV=i

	debug(D_FLOW,"read_0x7f05: Exiting...\n");
	return 0;
}
#endif

#endif
