/* 
#    $Id: Leica.cpp,v 1.2 2018/02/23 04:34:13 gps Exp $
#
#	 Leica.cpp -- This file will be the major interaction with
#                   an leica receiver
#
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>

#include "gpscommon.h"
#include "Leica.h"

SLeicaFATEntry Leica::pFAT[MAX_LEICA_FILES];
SLeicaPortConfig Leica::receiverPorts;
U_1_CHAR Leica::logicalPort;
U_1_CHAR Leica::serialPort;

//  ===================================================================
//  NAME:		Leica::Leica
//
//  PURPOSE:	Constructor. Set several default values
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
Leica::Leica () {

	debug(D_FLOW,"Leica::Leica: Entering...\n");
	pFATentry = NULL;

	usesFAT = TRUE;
	myIndex = -1;

	cimageFile = 0;
	numMessages = 0;
	imageFile = NULL;
	session = ' ';
	debug(D_FLOW,"Leica::Leica: Exiting\n");

}

//  ===================================================================
//  NAME:		Leica::Login
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::Login(Serial *pSerial, int direct, string login, string password)
{
	pSerial->setFlowControl(SERIAL_FLOW_NONE);
	return ReadPortConf(pSerial);
}


//  ===================================================================
//  NAME:		Leica::WriteFiles
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::WriteFiles() {

	U_2_INT length=0;
	ofstream rfile;

	debug(D_FLOW, "Leica::WriteFiles: Entering...\n");
	/* skip out if no messages */
	if(numMessages == 0) return(SUCCESS);

	if (OpenFile(&rfile, GetFileName(myIndex).c_str()) != 0) return(FAILURE);

	/* dump bytes of message and free memory */
	for(int i = 0; i < numMessages; i++) {
		memcpy(&length, &imageFile[i][2], 2);
		if(imageFile[i][4] == 0xA1) {
			rfile.write((const char *)&imageFile[i][10], length-10);
		} else {
			rfile.write((const char *)imageFile[i], length+2);
		}
	}
	debug(D_SERINFO, "Leica::WriteFiles: Wrote %d messages to test\n", numMessages);

	rfile.close();

	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Leica::AddMessage
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::AddMessage(U_1_CHAR *message, int length) {
	/* save data messages */
	debug(D_FLOW,"Leica::AddMessage: Entering...\n");
	imageFile = (unsigned char **) realloc(imageFile, (numMessages+1)*sizeof(char *));
	imageFile[numMessages] = new unsigned char[length+6];
	memcpy((void *)imageFile[numMessages],(void *)message,length+6);
	numMessages++;
	debug(D_FLOW,"Leica::AddMessage: Exiting...\n");
	return(numMessages);
}

//  ===================================================================
//  NAME:		Leica::ReceiverA1File
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::ReceiveA1File(Serial *pSerial, U_4_INT indexFile) {

	U_4_INT seq=0;
	U_1_CHAR *message;
	char adi, id;
	U_1_INT numErrors=0;
	U_2_INT length=0;
	int status=FAILURE;
	U_4_INT fileSize=0;

	debug(D_FLOW, "Leica::ReceiveA1File: Entering...\n");

	/* max number of erros - should be adjustable */
	while(numErrors < 10) {

		/* receive an lb2 message */
		debug(D_RECINFO, "Leica::ReceiveA1File: receiving next lb2...\n");
		ReceiveLB2(pSerial, &message, &length);

		/* sleep and wait for it again */
		debug(D_RECINFO, "Leica::ReceiveA1File: Length is %d\n", length);
		if(length < 2) {
			debug(D_WARNMINOR, "Leica::ReceiveA1File: received a null\n");
			numErrors++;
			sleep(1);
			continue;
		}

		/* check message - reject non-A1 messages */
		id = message[4];
		debug(D_RECINFO, "Leica::ReceiveA1File: Received %2.2X\n", (unsigned char)id);
		adi = message[5];

	/* save first FE for debug output */
		if(id == (char)0xFE) {
			if(message[8] == 0x01) {
				memcpy(&fileSize, &message[11], 4);
				debug(D_RECINFO, "Leica::ReceiveA1File: file size %d bytes\n", fileSize);
				debug(D_RECINFO, "Leica::ReceiveA1File: FOUND an 0x01 FE in ReceiveA1File!\n");
				debug(D_RECINFO, "Leica::ReceiveA1File: there are %d messages already recorded\n", numMessages);
			} else if(message[8] == 0x02) {
				debug(D_RECINFO, "Leica::ReceiveA1File: FOUND an 0x02 FE in ReceiveA1File!\n");
				debug(D_RECINFO, "Leica::ReceiveA1File: there are %d messages already recorded\n", numMessages);
			}
		}

		/* parse message */
		if((id==(char)0xA1) && (adi==31 || adi==21)) {    
			memcpy(&seq, &message[6], 4);
			unsigned char adiResp[4];
			debug(D_RECINFO, "Leica::ReceiveA1File: id %2.2X (adi is %d seq is %u)\n", (unsigned char)id, adi, seq);

			AddMessage(message,length);

			/* prepare response */
			adiResp[0] = (unsigned char) 0xA1;        
			adiResp[1] = (unsigned char) 0xA1;
			adiResp[2] = (unsigned char) 0x00;
			seq++;
			adiResp[3] = (unsigned char) seq%256;

			debug(D_RECINFO, "Leica::ReceiveA1File: sending resp %2.2X %2.2X %2.2X %2.2X\n", 
					(unsigned char)adiResp[0],
					(unsigned char)adiResp[1],            
					(unsigned char)adiResp[2],            
					(unsigned char)adiResp[3]);

			/* send response */
			status = SendLB2(pSerial, adiResp, 4);        
			if(status == FAILURE) {
				debug(D_ERR, "Leica::ReceiveA1File: error sending A1 response\n");
				return(FAILURE);
			} else {
				debug(D_RECINFO, "Leica::ReceiveA1File: sent A1 response\n");
			}

			/* this looks nice */
			memcpy(&length, &message[2], 2);
			reportstruct.bytes_downloaded += length-10;
			reportstruct.elapsed_time = time(NULL) - reportstruct.start_time;
			TransferReport(&reportstruct);

			/* end of download? if so, return */
			if(adi == 21) {
				debug(D_STD_V3, "Leica::ReceiveA1File: end of download, returning\n");
				status = SUCCESS;
				break;
			}
		} else {
			/* non-A1 download message (0xFE) */
			debug(D_RECINFO, "Leica::ReceiveA1File: id %2.2X (adi is %d seq is %d) - not responding\n", (unsigned char)id, adi, seq);
		}
	}

	debug(D_FLOW, "Leica::ReceiveA1File: Exiting\n");

	return(status);
}


//  ===================================================================
//  NAME:		Leica::DownloadFileAtIndex
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::DownloadFileAtIndex(Serial *pSerial, int index) {

	U_4_INT fileSize=0;
	U_1_CHAR dumpRequest[28];
	U_1_CHAR *message;
	U_1_CHAR tries=0;
	U_2_INT messLength=0;
	int status = FAILURE;

	debug(D_FLOW, "Leica::DownloadFileAtIndex: Entering...\n");

	debug(D_VALUES, "Leica::DownloadFileAtIndex: indexFile %d\n", index);

	/* prepare file dump request */
	memset(dumpRequest, 0, sizeof(dumpRequest));
	dumpRequest[0] = 0x93;
	dumpRequest[1] = 0x00;
	dumpRequest[2] =
	dumpRequest[3] = /* avail disk space */
	dumpRequest[4] =
	dumpRequest[5] = 0x00;
	dumpRequest[6] = (U_1_CHAR) index; /* index to dump */
	dumpRequest[7] =
	dumpRequest[8] = /* file size */
	dumpRequest[9] =
	dumpRequest[10] = 0x00;
	dumpRequest[11] = ((U_1_CHAR) (serialPort)) << 3;
	dumpRequest[11] |= 0x04;   /* dump out serial port */
	dumpRequest[12] = 0x00;    
	dumpRequest[13] = 0x00;    /* number of control blocks to follow */
	dumpRequest[14] = 0x00;    /* dump extent */
	dumpRequest[15] = 0x00;    /* dump extent interval */

	/* send dump request */
	status = SendLB2(pSerial, dumpRequest, 28);
	if(status == FAILURE) {
		debug(D_WARN, "Leica::DownloadFileAtIndex: Error %d from SendLB2\n");
		return(FAILURE);
	}
	debug(D_FLOW, "Leica::DownloadFileAtIndex: requested file\n");

	/* receive response to dump request before file */
	tries=0;
	while(1) {
		U_2_INT length;
		ReceiveLB2(pSerial, &message, &length);
		if(length < 2) {
			debug(D_WARN, "Leica::DownloadFileAtIndex: error reading response\n");
			return(FAILURE);
		}

		/* look for response */
		if(message[4] == (U_1_CHAR) 0xFF) {
			if(message[5] == (U_1_CHAR) 0x93) {
				if(message[6] == (U_1_CHAR) 0x00) {
					break;
				} else if(message[6] == (U_1_CHAR) 0x02) {
					debug(D_WARN, "Leica::DownloadFileAtIndex: File is not on receiver\n");
					return(FAILURE);
					break;
				} else {
					debug(D_WARN, "Leica::DownloadFileAtIndex: error %X in response\n", (int) message[6]);
					return(FAILURE);
				}
			}
		}

		if(++tries == MAX_MESSAGES_BEFORE_RESPONSE) {
			debug(D_WARN,"Leica::DownloadFileAtIndex: Too many messages before valid response\n");
			return(FAILURE);
		}
	}

	reportstruct.filename = NULL;
	reportstruct.filesize = ReceiverFAT[index].fileSize;
	reportstruct.blocks_downloaded = 0;
	reportstruct.bytes_downloaded = 0;
	reportstruct.start_time = time(NULL);
	reportstruct.elapsed_time = 0;
	reportstruct.errors = 0;

	TransferReportStart(&reportstruct);
	int receivedFile = FALSE;

	/* receive file */
	U_2_INT length;
	while(ReceiveLB2(pSerial, &message, &length) >= 1) {
		debug(D_RECINFO, "Leica::DownloadFileAtIndex: received message id %2.2X\n", message[4]);
#if 1
		if (message[4] == 0xA1) {
			debug(D_STD_V3, "Leica::DownloadFileAtIndex: receiving A1 file\n");
			status = ReceiveA1File(pSerial, index);
			if(status == FAILURE) {
				debug(D_ERR, "Leica::DownloadFileAtIndex: Error receiving file %d\n", index);
			} else {
				receivedFile = TRUE;
			}
			debug(D_STD_V3, "Leica::DownloadFileAtIndex: received A1 file, breaking\n");
			break;
		} else {
#endif
			debug(D_RECINFO, "Leica::DownloadFileAtIndex: inserting message\n");
			AddMessage(message, length);
			if(message[4] == 0xFE) {
				debug(D_RECINFO, "Leica::DownloadFileAtIndex: message was FE....\n");
				if(message[8] == 0x02) {
					debug(D_RECINFO, "Leica::DownloadFileAtIndex: signifies end of file, breaking\n");
					receivedFile = TRUE;
					break;
				}
				if(message[8] == 0x01) {
					debug(D_RECINFO, "Leica::DownloadFileAtIndex: signifies beginning of file, not breaking\n");
					memcpy(&fileSize, &message[11], 4);
					debug(D_VALUES, "Leica::DownloadFileAtIndex: file size %d bytes\n");
				}
			}
#if 1
		}
#endif

		/* this looks nice */
		memcpy(&messLength, &message[2], 2);
		reportstruct.bytes_downloaded += messLength;
		reportstruct.elapsed_time = time(NULL) - reportstruct.start_time;
		TransferReport(&reportstruct);
	}
	debug(D_FLOW, "Leica::DownloadFileAtIndex: \n");

	/* check received file status */
	if(receivedFile) {
		char buffer[100];
		myIndex = index;
		addToSumFile(" Start Time: %s\n",ReceiverFAT[index].fileStart.chartime("%%G %%g", buffer, 100));
		addToSumFile(" End Time: %s\n",ReceiverFAT[index].fileEnd.chartime("%%G %%g", buffer, 100));
		addToSumFile(" File: %s\n",GetFileName(myIndex).c_str());
		addToSumFile(" Bytes: %d\n",fileSize);
		addToSumFile(" Time: %d seconds\n",reportstruct.elapsed_time);
	} else {
		debug(D_ERR, "Leica::DownloadFileAtIndex: Download Failed\n");
		debug(D_ERR, "Leica::DownloadFileAtIndex: Clearing Transfer\n");
		pSerial->ClearBuffers(); sleep(1); pSerial->ClearBuffers();
	}
	addToSumFile("\n");

	/* write image files */

	return((receivedFile) ? SUCCESS : FAILURE);
}

//  ===================================================================
//  NAME:		Leica::DownloadFileByTime
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::DownloadFileByTime(Serial *pSerial, gpstime *start, gpstime *stop) {

	int numindex, indexes[MAX_LEICA_FILES];
	int status = FAILURE;

	debug(D_FLOW,"Leica::DownloadFileByTime: Entering...\n");

	numindex = FindIndexesByTime(start, stop, indexes, FALSE);

	if (numindex > 0) {
		debug(D_STD_V2,"Leica::DownloadFileByTime: Downloading file at index %d\n", indexes[0]);
		status = DownloadFileAtIndex(pSerial, indexes[0]);
	}
#if 0
	Leica *tmp;
	char buffer[100];
	if (downLoadOK && (numindex > 1)) {
		for (int i = 1; i < numindex; i++) {
			tmp = new Leica;
			debug(D_STD_V2,"Leica::DownloadFileByTime: Downloading file at index %d\n", indexes[i]);
			downLoadOK = tmp->DownloadFileAtIndex(pSerial, indexes[i]);
			if (!downLoadOK) {
				delete tmp;
				break;
			}
			AddImage(tmp);
			delete tmp;
			addToSumFile(" Span: ");
			addToSumFile(start->chartime("%%G/%%g-",buffer,100));
			addToSumFile(stop->chartime("%%G/%%g\n",buffer,100));
		}
	}
#endif
	debug(D_FLOW,"Leica::DownloadFileByTime: Exiting\n");
	return(status);
}

//  ===================================================================
//  NAME:		Leica::DeleteFileAtIndex
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::DeleteFileAtIndex(Serial *pSerial, int index) {

	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Leica::DeleteFileByTime
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::DeleteFileByTime(Serial *pSerial, gpstime *start, gpstime *stop) {

	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Leica::DownloadReceiverID
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
string Leica::DownloadReceiverID(Serial *pSerial) {

	int status=0;
	U_1_CHAR requestID[8];
	U_1_CHAR *respondID;
	U_1_CHAR numBlocks=0, indexBlock=0;
	unsigned tries=0;
	U_2_INT idLen=0;
	S_1_CHAR *idString;

	debug(D_FLOW, "Leica::DownloadReceiverID: Entering...\n");

	pSerial->ClearTXBuffer();
	pSerial->ClearRXBuffer();

	/* prepare request for RX ID message */
	requestID[0] = (U_1_CHAR) 0x82; /* message ID   */
	requestID[1] = (U_1_CHAR) 0x00; /* logical port */
	requestID[2] = (U_1_CHAR) 0x00; /* output port  */
	requestID[3] = (U_1_CHAR) 200;  /* output rate  */
	requestID[4] = (U_1_CHAR) 0x00; /* output ID    */
	requestID[5] = requestID[6] = requestID[7] = (U_1_CHAR) 0x00;

	/* package 'requestID' as LB2 and send */
	while(1) {
	status = SendLB2(pSerial, requestID, 8);
	if(status == FAILURE) {
		debug(D_ERR, "Leica::DownloadReceiverID: error %d requesting ID string\n", status);
		return "";
	}

	/* make sure we get the right message back */
		U_2_INT length;
		ReceiveLB2(pSerial, &respondID, &length);
		if(length < 2) {
			debug(D_ERR, "Leica::DownloadReceiverID: error reading RX ID\n");
			return "";
		}

	/* look for ID message */
		if(respondID[4] == (U_1_CHAR) 0x00) break;

		if(++tries == MAX_MESSAGES_BEFORE_RESPONSE) return "";
		debug(D_STD_V3,"Leica::ReadReceiverID: respondID[4] = %X\n",(int) respondID[4]);
		debug(D_STD_V3,"Leica::ReadReceiverID: Tries = %d\n",tries);
	}

	/* check for valid software */
	if(respondID[5] == '\0') {
		debug(D_ERR, "Leica::DownloadReceiverID: no valid software - system reset required\n");
		return "";
	}

	/* create ID string */
	RxID = respondID[5];
	idLen = strlen((const char *)&respondID[5]);
	numBlocks = respondID[5+idLen+1];
	idString = (char *)&respondID[5+idLen+2];
	for(indexBlock=0; indexBlock<numBlocks; indexBlock++) {
		debug(D_STD_V3, "Leica::DownloadReceiverID: %s", idString);
		idString += strlen(idString) + 1;
		debug(D_STD_V3, "%s\n", idString);
		RxID += idString;
		idString += strlen(idString) + 1;
	}

	/* free memory */
	debug(D_FLOW, "Leica::DownloadReceiverID: Exiting\n");
	return RxID;
}

//  ===================================================================
//  NAME:		Leica::ReadPortConf
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::ReadPortConf(Serial *pSerial) {

	U_1_CHAR *mess80, *mess82;
	U_1_CHAR request80[8], request82[8];
	unsigned tries=0, indexPort=0;
	int status=0;
	U_2_INT length=0;
	U_1_CHAR portsProcessed=0;
	U_1_CHAR portConfBlock[4];
	U_1_CHAR content=0;
	U_1_CHAR logicalPort=0;

	mess80 = mess82 = NULL;
	debug(D_FLOW, "Leica::ReadPortConf: Entering...\n");
	/* prepare request for port config message */
#if 1
	request80[0] = (U_1_CHAR) 0x82; /* message ID   */
	request80[1] = (U_1_CHAR) 0x00; /* logical port */
	request80[2] = (U_1_CHAR) 0x00; /* output port  */
	request80[3] = (U_1_CHAR) 200;  /* output rate  */
	request80[4] = (U_1_CHAR) 0x80; /* output ID    */
#else
	request80[0] = (U_1_CHAR) 0x82; /* message ID   */
	request80[1] = (U_1_CHAR) 0x00; /* logical port */
	request80[2] = (U_1_CHAR) 0x00; /* output port  */
	request80[3] = (U_1_CHAR) 0x00;  /* output rate  */
	request80[4] = (U_1_CHAR) 0x0A; /* output ID    */
#endif
	request80[5] = request80[6] = request80[7] = (U_1_CHAR) 0x00;

	/* package 'requestID' as LB2 and send */
	while(1) {
	status = SendLB2(pSerial, request80, 8);
	if(status == FAILURE) {
		debug(D_ERR, "Leica::ReadPortConf: error %d requesting port config\n", status);
		return(FAILURE);
	}

	/* make sure we get the right message back */
		ReceiveLB2(pSerial, &mess80, &length);
		if(length < 2) {
			debug(D_ERR, "Leica::ReadPortConf: error reading port config\n");
			return(FAILURE);
		}

		/* look for port config message */
		if(mess80[4] == (U_1_CHAR) 0x80) {
			debug(D_FLOW,"Leica::ReadPortConf: found port config message\n");
			break;
		}

		debug(D_VALUES,"Leica::ReadPortConf: mess80[4] = %X\n",(int) mess80[4]);
		debug(D_VALUES,"Leica::ReadPortConf: Tries = %d\n",tries);

		if(++tries == MAX_MESSAGES_BEFORE_RESPONSE) {
			return(FAILURE);
		}
	}

	/* message length */
	memcpy(&length, &mess80[2], 2);
	debug(D_RECINFO, "Leica::ReadPortConf: length %d\n", length);

	/* content flag */
	content = mess80[5];

	/* number of ports and which ones */
	if(content & 0x01) {
		receiverPorts.numPorts++;
		receiverPorts.availablePort[0] = TRUE;
	}
	if(content & 0x02) {
		receiverPorts.numPorts++;
		receiverPorts.availablePort[1] = TRUE;
	}
	if(content & 0x04) {
		receiverPorts.numPorts++;
		receiverPorts.availablePort[2] = TRUE;
	}
	if(content & 0x08) {
		receiverPorts.numPorts++;
		receiverPorts.availablePort[3] = TRUE;
	}
	debug(D_RECINFO, "Leica::ReadPortConf: numPorts %d\n", receiverPorts.numPorts);

	/* parse message for each port */
	for(indexPort=0; indexPort<MAX_RCVR_PORTS; indexPort++) {

		/* skip unavaliable ports */
		if(!receiverPorts.availablePort[indexPort]) {
			continue;
		}

		/* port configuration block */
		memcpy(portConfBlock, &mess80[6 + (portsProcessed++)*4], 4);

		/* parse pcb */
		/* baud rate, parity, etc not implemented yet */
		receiverPorts.outputFormat[indexPort] = portConfBlock[2];
		debug(D_RECINFO, "Leica::ReadPortConf: indexPort %d outputFormat %d\n", indexPort,
		receiverPorts.outputFormat[indexPort]);
		receiverPorts.inputFormat[indexPort] = portConfBlock[3];
		debug(D_RECINFO, "Leica::ReadPortConf: indexPort %d inputFormat %d\n", indexPort,
		receiverPorts.inputFormat[indexPort]);
	}

	/* prepare request for LB2 output message */
	request82[0] = (U_1_CHAR) 0x82; /* message ID   */
	request82[1] = (U_1_CHAR) 0x00; /* logical port */
	request82[2] = (U_1_CHAR) 0x00; /* output port  */
	request82[3] = (U_1_CHAR) 200;  /* output rate  */
	request82[4] = (U_1_CHAR) 0x82; /* output ID    */
	request82[5] = request82[6] = request82[7] = (U_1_CHAR) 0x00;

	/* package 'requestID' as LB2 and send */
	status = SendLB2(pSerial, request82, 8);
	if(status == FAILURE) {
		debug(D_ERR, "Leica::ReadPortConf: error %d requesting LB2 output\n", status);
		return(FAILURE);
	}

	/* make sure we get the right message back */
	while(1) {
		ReceiveLB2(pSerial,&mess82, &length);
		if(length < 2) {
			debug(D_ERR, "Leica::ReadPortConf: error reading LB2 output\n");
			return(FAILURE);
		}

		/* look for ID message */
		if(mess82[4] == (U_1_CHAR) 0x82) {
			break;
		}

		if(++tries == MAX_MESSAGES_BEFORE_RESPONSE) {
			return(FAILURE);
		}
	}

	/* message length */
	memcpy(&length, &mess82[2], 2);
	debug(D_RECINFO, "Leica::ReadPortConf: length %d\n", length);

	/* logical port message received on */
	logicalPort = mess82[5];
	if(logicalPort & 0x01) {
		logicalPort = 1;
	}
	if(logicalPort & 0x02) {
		logicalPort = 2;
	}
	if(logicalPort & 0x04) {
		logicalPort = 3;
	}
	debug(D_RECINFO, "Leica::ReadPortConf: logicalPort %d => %d\n", logicalPort, logicalPort);

	/* serial port message received on */
	if(receiverPorts.outputFormat[0] == logicalPort &&
		receiverPorts.inputFormat[0] == logicalPort) {
		serialPort = 0;
	}
	if(receiverPorts.outputFormat[1] == logicalPort &&
		receiverPorts.inputFormat[1] == logicalPort) {
		serialPort = 1;
	}
	if(receiverPorts.outputFormat[2] == logicalPort &&
		receiverPorts.inputFormat[2] == logicalPort) {
		serialPort = 2;
	}
	if(receiverPorts.outputFormat[3] == logicalPort &&
		receiverPorts.inputFormat[3] == logicalPort) {
		serialPort = 3;
	}
	debug(D_RECINFO, "Leica::ReadPortConf: serial port %d\n", serialPort);

	debug(D_FLOW, "Leica::ReadPortConf: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Leica::SetReceiverSession
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::SetReceiverSession(Serial *pSerial) {
	return(SUCCESS);
}


//  ===================================================================
//  NAME:		Leica::DownloadReceiverFAT
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::DownloadReceiverFAT(Serial *pSerial) {

	int status=0;
	U_1_CHAR setPartition[28], requestFAT[8];
	U_1_CHAR *response;
	unsigned tries=0;
	U_2_INT length=0;
	U_4_INT space=0;
	U_1_CHAR logFileBlock[22];
	U_1_CHAR fileNum=0;
	unsigned indexFile=0;
	unsigned indexPage=0;
	U_2_INT dlcs=0, temp=0;

	debug(D_FLOW, "Leica::DownloadReceiverFAT: Entering...\n");

	/* prepare request for FAT message */
	memset((void *)setPartition, 0, sizeof(setPartition));
	setPartition[0] = (U_1_CHAR) 0x93; /* message ID   */
	setPartition[1] = (U_1_CHAR) 0x00; /* reserved     */
	setPartition[2] = setPartition[3] =
	setPartition[4] = setPartition[5] = 0; /* available disk space */
	setPartition[6] = (U_1_CHAR) 255;        /* 255 => send all data */

	for(indexPage=0; indexPage<MAX_RCVR_PAGES; indexPage++) {

		/* set file partition by using message 0x93 */
		/* data logging / control status */
		dlcs = 7;  /* get all files */
		temp = indexPage; /* partition */
		temp <<= 13;
		dlcs |= temp;
		memcpy(&setPartition[11], &dlcs, 2);

		/* package 'requestID' as LB2 and send */
		status = SendLB2(pSerial, setPartition, 28);
		if(status == FAILURE) {
			debug(D_ERR, "Leica::DownloadReceiverFAT: error %d requesting ID string\n", status);
			return(FAILURE);
		}

		/* look for 0x93 acknowledgement */
		tries=0;
		while(1) {
			ReceiveLB2(pSerial,&response, &length);
			if(length < 2) {
				debug(D_ERR, "Leica::DownloadReceiverFAT: error reading RX FAT\n");
				return(FAILURE);
			}

			/* look for FAT message */
			if(response[4] == (U_1_CHAR) 0xFF) {
				if(response[5] == (U_1_CHAR) 0x93) {
					if(response[6] == 0x00) {
						debug(D_RECINFO, "Leica::DownloadReceiverFAT: partition set to %d\n", indexPage);
						break;
					} else {
						debug(D_ERR, "Leica::DownloadReceiverFAT: error %d in partition set acknowledge\n", response[6]);
						return (int) response[6];
					}
				}
			}

			if(++tries == MAX_MESSAGES_BEFORE_RESPONSE) {
				return(FAILURE);
			}
		}

		/* now make request for message 0x93 */
		requestFAT[0] = 0x82; /* message control */
		requestFAT[1] = 0x00; /* only valid on input */
		requestFAT[2] = 0x00; /* request output to this port */
		requestFAT[3] = 200;  /* output now */
		requestFAT[4] = 0x93; /* request message 0x93 */
		requestFAT[5] = requestFAT[6] = requestFAT[7] = 0x00;

		/* package and send as LB2 */
		status = SendLB2(pSerial, requestFAT, 8);
		if(status == FAILURE) {
			debug(D_ERR, "Leica::DownloadReceiverFAT: error %d requesting ID string\n", status);
			return(FAILURE);
		}

		for(indexFile=0; indexFile<MAX_RCVR_FILES_PER_PAGE; indexFile++) {

			/* look for 0x93 with data */
			tries=0;
			while(1) {
				ReceiveLB2(pSerial,&response,&length);
				if(length < 2) {
					debug(D_ERR, "Leica::DownloadReceiverFAT: error reading RX FAT\n");
					return(FAILURE);
				}

				/* look for FAT message */
				if(response[4] == (U_1_CHAR) 0x93) {
					break;
				}

				if(++tries == MAX_MESSAGES_BEFORE_RESPONSE) {
					return(FAILURE);
				}
	        	}

			/* message length */
			memcpy(&length, &response[2], 2);
			debug(D_RECINFO, "Leica::DownloadReceiverFAT: length %d\n", length);

			/* avaliable disk space */
			memcpy(&space, &response[6], 4);
			debug(D_RECINFO, "Leica::DownloadReceiverFAT: %d kB available\n", space/1024);

			/* parse log file control block */
			memcpy(logFileBlock, &response[10], 22);

			/* file number */
			fileNum = logFileBlock[0];
			debug(D_RECINFO, "Leica::DownloadReceiverFAT: file num %d\n", fileNum);
			pFAT[fileNum].fileNumber = fileNum;

			/* file length */
			memcpy(&(pFAT[fileNum].fileLength), &logFileBlock[1], 4);
			debug(D_RECINFO, "Leica::DownloadReceiverFAT: file length %d\n", pFAT[fileNum].fileLength);
			ReceiverFAT[fileNum].fileSize = pFAT[fileNum].fileLength;

			/* logging / control status */
			memcpy(&(pFAT[fileNum].logContStatus), &logFileBlock[5], 2);
			debug(D_RECINFO, "Leica::DownloadReceiverFAT: logContStatus %4.4X\n", pFAT[fileNum].logContStatus);

			/* number of log file control blocks (LB2 messages)*/
			pFAT[fileNum].numLogFileControlBlocks = logFileBlock[7];
			debug(D_RECINFO, "Leica::DownloadReceiverFAT: %d log file control blocks follow\n", pFAT[fileNum].numLogFileControlBlocks);

			/* logging / dump extent */
			pFAT[fileNum].extent = logFileBlock[8];
			debug(D_RECINFO, "Leica::DownloadReceiverFAT: extent %2.2X\n", pFAT[fileNum].extent);

			/* extent interval */
			pFAT[fileNum].extentInterval = logFileBlock[9];
			debug(D_RECINFO, "Leica::DownloadReceiverFAT: extentInterval %2.2X\n", pFAT[fileNum].extentInterval);

			/* next if this file is empty */
			if(!(pFAT[fileNum].fileLength)) {
				continue;
			}

			/* fill in receiver fat day month year (deal with 2 digit years) */
			gpstime curtime;
			curtime.set_time(); // set time to now
			int tempyear;
			double tempday;
			curtime.getjmt(&tempyear, &tempday);
			tempyear += 1900;
			int YYNOW = tempyear % 100;
			int CENTNOW = 100 * (tempyear / 100);
			int startyear = logFileBlock[12];
			if(startyear < YYNOW) {
				startyear = CENTNOW + 100 + startyear;
			} else if(startyear > YYNOW) {
				startyear = CENTNOW - 100 + startyear;
			} else {
				startyear = CENTNOW + startyear;
			}  

			int endyear = logFileBlock[15];
			if(endyear < YYNOW) {
				endyear = CENTNOW + 100 + endyear;
			} else if(endyear > YYNOW) {
				endyear = CENTNOW - 100 + endyear;
			} else {
				endyear = CENTNOW + endyear;
			}  

			debug(D_FLOW, "Leica::DownloadReceiverFAT: Setting FAT start and stop times\n");
			pFAT[fileNum].fileStart.set_time(startyear, (int)logFileBlock[11],
								(int)logFileBlock[10],
								(int)logFileBlock[16],
								(int)logFileBlock[17],
								(double)logFileBlock[18]);
			ReceiverFAT[fileNum].fileStart = pFAT[fileNum].fileStart;
			pFAT[fileNum].fileEnd.set_time(endyear, (int)logFileBlock[14],
								(int)logFileBlock[13],
								(int)logFileBlock[19],
								(int)logFileBlock[20],
								(double)logFileBlock[21]);
			ReceiverFAT[fileNum].fileEnd = pFAT[fileNum].fileEnd;
			numFATEntries++;
			FindSession(fileNum);
			CreateFileName(fileNum);

		}
	}
	debug(D_FLOW, "Leica::DownloadReceiverFAT: Exiting...\n");

	return SUCCESS;
}

//  ===================================================================
//  NAME:		Leica::ReceiverLB2
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::ReceiveLB2(Serial *pSerial, U_1_CHAR **pmessage, U_2_INT *length) {

	int status=0;
	U_1_CHAR sync[2];
	U_2_INT checksum=0, checksumread=0, checksumsum=0;

	debug(D_FLOW, "Leica::ReceiveLB2: Entering...\n");
	/* sync word */
	status = pSerial->Read((void *)sync, 2, 5*GENERAL_TIMEOUT);
	if(status != 2) {
		debug(D_ERR, "Leica::ReceiveLB2: error receiving sync word\n");
		return(FAILURE);
	}
	if(!((sync[0] == (U_1_CHAR) 0x9C) && (sync[1] == (U_1_CHAR) 0xAE))) {
		debug(D_ERR, "Leica::ReceiveLB2: error reading sync word\n");
		return(FAILURE);
	}

	/* message length */
	status = pSerial->Read((void *)length, 2, GENERAL_TIMEOUT);
	if(status != 2) {
		debug(D_ERR, "Leica::ReceiveLB2: error reading message length\n");
		return(FAILURE);
	}
	debug(D_VALUES, "Leica::ReceiveLB2: read message length %d\n",*length);

	*pmessage = new U_1_CHAR [(time_t)*length + 6];
	if (*pmessage == NULL) {
		debug(D_ERR,"Leica::ReceiveLB2: Cannot allocate memory\n");
	}
	/* copy sync, length to beginning of message */
	memcpy((void *)*pmessage, (void *)sync, 2);
	memcpy((void *)(*pmessage+2), (void *)length, 2);

	if ((*length + 2) > MAXLB2SIZE) {
		debug(D_VALUES, "Leica::ReceiveLB2: length is %d, MAXLB2SIZE is %d\n",*length+2,MAXLB2SIZE);
		debug(D_FLOW, "Leica::ReceiveLB2: length byte 1 = %02X, byte 2 = %02X\n",*pmessage+2, *pmessage+3);
		status = pSerial->Read((void *)(*pmessage+4), MAXLB2SIZE-4, GENERAL_TIMEOUT);
for(unsigned i = 2; i < (unsigned)status; i++) {
	checksum += (U_1_CHAR) *(*pmessage+i);
}
		return(MAXLB2SIZE - (*length+2));
	}

	/* message bytes */
	status = pSerial->Read((void *)(*pmessage+4), *length-2, 2 * GENERAL_TIMEOUT);
	if(status != (*length-2)) {
		debug(D_ERR, "Leica::ReceiveLB2: error read %d bytes of %d\n", status, *length-2);
		return(FAILURE);
	}

	/* compute checksum */
	for(unsigned i = 2; i < (unsigned)*length; i++) {
		checksum += (U_1_CHAR) *(*pmessage+i);
	}

	memcpy((void *)&checksumread, (void *)(*pmessage+*length), 2);
	checksumsum = checksum + checksumread;
	if(checksumsum) {
		debug(D_ERR, "Leica::ReceivelLB2: bad checksum = %hu\n", checksumsum);
		return(FAILURE);
	}

	debug(D_FLOW, "Leica::ReceiveLB2: Exiting\n");
	return(*length);
}


//  ===================================================================
//  NAME:		Leica::SendLB2
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Leica::SendLB2(Serial *pSerial, U_1_CHAR *msg, unsigned short msgLength) {

	int status=0;
	U_1_CHAR sync[2];
	unsigned indexChar;
	U_2_INT totalLength=0;
	U_2_INT checksum=0, checksumnot=0;

	//pSerial->ClearBuffers();
	pSerial->ClearTXBuffer();
	debug(D_FLOW, "Leica::SendLB2: Entering...\n");

	sync[0] = (U_1_CHAR) 0x9C;
	sync[1] = (U_1_CHAR) 0xAE;
	status = pSerial->Write((unsigned char *) sync, 2);
	debug(D_SERINFO,"Sending: %X %X\n",sync[0],sync[1]);
	if(status != 2) {
		debug(D_ERR, "Leica::SendLB2: error %d sending sync word\n", status);
		return(FAILURE);
	}

	totalLength = msgLength + 4; /* msg + length word + cksum word */
	U_1_CHAR * writePtr = (U_1_CHAR *) &totalLength;
	debug(D_SERINFO,"Sending %X %X\n",(int)*writePtr, (int)*(writePtr+1));
	status = pSerial->Write((unsigned char *) writePtr, 2);
	if(status != 2) {
		debug(D_ERR, "Leica::SendLB2: error %d sending message length\n", status);
		return(FAILURE);
	}

	status = pSerial->Write((unsigned char *) msg, msgLength);
	if(status != msgLength) {
		debug(D_ERR, "Leica::SendLB2: error %d sending message\n", status);
		return(FAILURE);
	}

	U_1_CHAR * charPtr = (U_1_CHAR *) &totalLength;
	checksum += (U_1_CHAR) *charPtr; charPtr++;
	checksum += (U_1_CHAR) *charPtr;

	for(indexChar=0; indexChar<msgLength; indexChar++) {
		checksum += (U_1_CHAR) msg[indexChar];
	}
	checksumnot = (unsigned short)((unsigned)65536 - checksum);
	writePtr = (U_1_CHAR *) &checksumnot;
	debug(D_SERINFO,"Sending %X %X\n",(int)*writePtr, (int)*(writePtr+1));
	pSerial->ClearRXBuffer();
	status = pSerial->Write((unsigned char *) writePtr, 2);
	if(status != 2) {
		debug(D_ERR, "Leica::SendLB2: error %d sending message\n", status);
		return(FAILURE);
	}
	debug(D_FLOW, "Leica::SendLB2: Exiting\n");

	return SUCCESS;
}

