/* 
#    $Id: Navcom.cpp,v 1.1 2011/04/04 17:29:34 gps Exp $
#
#	 Navcom.cpp -- This file will be the major interaction with
#                   an Navcom receiver/file
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <iomanip>
#include <time.h>
#include <vector>
#include <sys/time.h>
#include <sys/types.h>

#include "gpscommon.h"
#include "Navcom.h"

// define this globally since we only act on one record at a time.
Navcom_Msg navcom_msg;
vector<int> Navcom::blockRequests;

//  ===================================================================
//  NAME:		Navcom::Navcom
//
//  PURPOSE:	Set some default values
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
Navcom::Navcom () {

	debug(D_FLOW,"Navcom::Navcom: Entering...\n");
	pFATentry = NULL;
	pImageHead = NULL;

	usesFAT = TRUE;
	myIndex = -1;

	// Actual File image
	cimageFile = 0;
	imageFile = NULL;
	header = NULL;
	fatEntry = NULL;
	session = ' ';
	for (int i = 0; i < MAX_RCVR_FILES; i++) indexData[i] = NULL;
	debug(D_FLOW,"Navcom::Navcom: Exiting\n");

}

#define NOAUTOBAUD 1
int Navcom::Login(Serial *pSerial, int direct, string login, string password) {

	long baud, currentspeed;
	debug(D_FLOW,"Navcom::Login: Entering\n");
	if (!direct || NOAUTOBAUD) {
		debug(D_STD_V3,"Navcom::Login: Not direct, no change to baud rate\n");
	} else {

		currentspeed = pSerial->getSpeed();
		baud = AutoBaud(pSerial);

		debug(D_CHECK,"Navcom::Login: Checking %d == %d\n",baud, currentspeed);
		if (baud != currentspeed)
			return(SetPortInfo(pSerial, currentspeed));

	}
	debug(D_FLOW,"Navcom::Login: Exiting\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Navcom::AutoBaud
//
//  PURPOSE:	find the current port and baud rate of the receiver
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
long Navcom::AutoBaud(Serial *pSerial) {

	debug(D_FLOW,"Navcom::AutoBaud: Entering...\n");

	long BaudRates[7];
	BaudRates[0] = 115200;
	BaudRates[1] = 57600;
	BaudRates[2] = 38400;
	BaudRates[3] = 19200;
	BaudRates[4] = 9600;
	BaudRates[5] = 4800;
	BaudRates[6] = 2400;

	long currentspeed = pSerial->getSpeed();

	char port;
	long speed;
	if (GetPortInfo(pSerial,&port,&speed) == SUCCESS) return(currentspeed);
	for (int i = 0; i <= 6; i++) {
		debug(D_STD_V,"Navcom::AutoBaud: Trying speed %d\n",BaudRates[i]);
		pSerial->setSpeed(BaudRates[i]);
		if (GetPortInfo(pSerial,&port,&speed) == 0) {
			debug(D_STD_V,"Navcom::AutoBaud: GetPortInfo succeeded for port %c and speed %d\n",port,speed);
			return(speed);
		}
	}
	debug(D_FLOW,"Navcom::AutoBaud: Exiting...\n");
	return(-1);
}

//  ===================================================================
//  NAME:		Navcom::GetPortInfo
//
//  PURPOSE:	find the current port and baud rate of the receiver
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Navcom::GetPortInfo(Serial *pSerial, char *port, long *speed) {

	string PRT;
	unsigned int ret;
	int status = SUCCESS;
	debug(D_FLOW,"Navcom::GetPortInfo: Entering...\n");
	*port = '0';
	*speed = -1;
	try {
		BuildDataRequest(0x11, RATE_ONCE, PORT_ALL);
		SendQuery(pSerial,"PRT");
		ret = FindNCTPacket(pSerial,3000, 0x11);
		if (ret != 0)
			throw "Navcom::GetPortInfo: No Response from port request\n";
	}
	catch (const char *pmessage)
	{
   		debug(D_WARN, pmessage);
		status = FAILURE;
   	}

	debug(D_FLOW,"Navcom::GetPortInfo: Exiting...\n");
	return(status);

}

//  ===================================================================
//  NAME:		Navcom::SetPortInfo
//
//  PURPOSE:	set the current port and baud rate of the receiver
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Navcom::SetPortInfo(Serial *pSerial, long baud) {

	char portc;
	long curbaud;
	char baudchar;
	int status = SUCCESS;
	debug(D_FLOW,"Navcom::SetPortInfo: Entering...\n");

	try {
		if (GetPortInfo(pSerial, &portc, &curbaud) != 0)
			throw "Navcom::SetPortInfo: Cannot get port info\n";

		debug(D_VALUES,"Navcom::SetPortInfo: Baud to set = %d\n",baud);
		switch (baud) {
			case 300:    baudchar='0'; break;
			case 600:    baudchar='1'; break;
			case 1200:   baudchar='2'; break;
			case 2400:   baudchar='3'; break;
			case 4800:   baudchar='4'; break;
			case 9600:   baudchar='5'; break;
			case 19200:  baudchar='6'; break;
			case 38400:  baudchar='7'; break;
			case 57600:  baudchar='8'; break;
			case 115200: baudchar='9'; break;
			default: throw "Navcom::SetPortInfo: Unrecognized baud rate\n";
		}

		char buffer[100];
		sprintf(buffer,"SPD,%c,%c",portc,baudchar);
		SendCommand(pSerial,buffer,FALSE,TRUE);
		pSerial->setSpeed(baud);
		char port;
		long speed;
		if (GetPortInfo(pSerial,&port,&speed) != 0) {
			AutoBaud(pSerial);
			throw "Navcom::SetPortInfo: The speed has _NOT_ been set\n";
		}
		debug(D_STD_V2,"Navcom::SetPortInfo: The speed has been set\n");
	}
	catch (const char *pmessage) {
		debug(D_WARN,pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"Navcom::SetPortInfo: Exiting...\n");
	return(status);

}

//  ===================================================================
//  NAME:		Navcom::SetImageFile
//
//  PURPOSE:	set the imagefile contents and size.
//
//  RETURNS:	SUCCESS if image correctly copied. FAILURE if
//				memory cannot be allocated.
//
//  HISTORY:
//
//  ===================================================================
int Navcom::setImageFile(unsigned char *myImage, int fileSize) {

	debug(D_FLOW,"Navcom::setImageFile: Entering...\n");

	if (imageFile != NULL)
		delete(imageFile);

	imageFile = new unsigned char[fileSize];
	if (imageFile == NULL) {
		debug(D_FATALERROR,"Unable to allocate memory for imageFile\n");
		return(FAILURE);
	}

	memcpy(imageFile, myImage, fileSize);
	cimageFile = fileSize;
	debug(D_FLOW,"Navcom::setImageFile: Exiting\n");
	return(SUCCESS);

}

//  ===================================================================
//  NAME:		Navcom::CompareSessions
//
//  PURPOSE:	Compare sessions found on receiver with sessions 
//				supplied by command line.
//
//  RETURNS:	SUCCESS on complete match
//				FAILURE otherwise
//
//  HISTORY:
//
//  ===================================================================
int Navcom::CompareSessions(SNavcomSession *tmpsess) {

	addToSumFile("\nSession Programming:\n RefDay = %d, Offset = %d, inuse = %d\n",
				tmpsess->refday, tmpsess->offset, tmpsess->inuse);
	addToSumFile("inuse start      end  interval mask minsv type\n");
	for (int i = 0; i < NUM_SESS; i++) {
		addToSumFile("%2d %c      %02d:%02d    %02d:%02d %03.1f      %03d  %03d    %03d\n",
				i,
				tmpsess->flag[i],
				tmpsess->start[i]/3600, (tmpsess->start[i]%3600)/60,
				tmpsess->end[i]/3600, (tmpsess->end[i]%3600)/60,
				tmpsess->interval[i], (int)tmpsess->elevmask[i],
				(int)tmpsess->minsv[i], (int)tmpsess->type[i]);
	}

	int activeSess = 0; 
	for (int i = 0; i < NUM_SESS; i++) {
		if (tmpsess->flag[i] == 'Y') activeSess++;
		if (tmpsess->flag[i] == 'A') activeSess++;
	}

	bool found = (activeSess == numSessions);
	for (int i = 0; (i < numSessions) && (found); i++) {
		long start    = GetSessionStartTime(i);
		long end      = GetSessionEndTime(i);
		float interval = GetSessionInterval(i);
		int elevmask = GetSessionElevationMask(i);
		int minsv    = GetSessionMinimumSV(i);
		found = FALSE;
		for (int j = 0; (j < NUM_SESS) && (!found); j++) {
			if (tmpsess->flag[j] == 'N') continue;
			if (start != tmpsess->start[j]) continue;
			if (end != tmpsess->end[j]) continue;
			if (interval != tmpsess->interval[j]) continue;
			if (elevmask != tmpsess->elevmask[j]) continue;
			if (minsv != tmpsess->minsv[j]) continue;
			debug(D_STD_V,"Navcom::CompareSessions: Found Match for Session %d\n",j);
			found = TRUE;
		}
	}
	debug(D_FLOW,"Navcom::CompareSessions: Exiting...\n");
	if (found) return(SUCCESS);
	return(FAILURE);

}

//  ===================================================================
//  NAME:		Navcom::WriteFiles
//
//  PURPOSE:	Write the image file to disk. This may not include
//				all of the information the file needs. It will only
//				write the contents of imageFile.
//
//  RETURNS:	SUCCESS on writing the file successfully, FAILURE
//				if write failed.
//
//  HISTORY:
//
//  ===================================================================
int Navcom::WriteFiles() {

	return(WriteFile(NULL, (char *) GetFileName(myIndex).c_str(), imageFile, cimageFile));

}

//  ===================================================================
//  NAME:		Navcom::CloseOpenFile
//
//  PURPOSE:	Close the Open file on the Navcom using pash command
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Navcom::CloseOpenFile(Serial *pSerial) {

	debug(D_FLOW,"CloseOpenFile: Entering...\n");
	debug(D_STD_V2,"CloseOpenFile:Closing Open File on Receiver\n");
	SendCommand(pSerial, "FIL,C", FALSE, FALSE);
	debug(D_FLOW,"CloseOpenFile: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Navcom::DeleteFileByTime
//
//  PURPOSE:	Delete all of the file indexes that are completely contained
//				within the specified time range
//
//  RETURNS:	SUCCESS or FAILURE. This really should return the number of 
//				files deleted.
//
//  HISTORY:
//
//  ===================================================================
int Navcom::DeleteFileByTime(Serial *pSerial, gpstime *start, gpstime *stop) {
	int i,j;
	int numindex, indexes[MAX_RCVR_FILES];
	int deleteOK = FALSE;

	debug(D_FLOW,"Navcom::DeleteFileByTime: Entering...\n");

	numindex = FindIndexesByTime(start, stop, indexes, TRUE);

	for (j = (MAX_RCVR_FILES-1); j >= 0; j--) {
		for (i = 0; i < numindex; i++) {
			if (indexes[i] != j) continue;
			debug(D_STD,"Navcom::DeleteFileByTime: Deleteing file at index %d\n", indexes[i]);
			deleteOK = DeleteFileAtIndex(pSerial, indexes[i]);
		}
	}
	debug(D_FLOW,"DeleteFiles: Exiting\n");
	return(deleteOK);
}


//  ===================================================================
//  NAME:		Navcom::DeleteFileAtIndex
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Navcom::DeleteFileAtIndex(Serial *pSerial, int index) {
	string Check;
	int status = FAILURE;
	int ret;

	debug(D_FLOW,"Navcom::DeleteFileAtIndex: Entering...\n");

	debug(D_STD,"\n\nDeleting file %s",ReceiverFAT[index].fileName.c_str());
	debug(D_FLOW," (fatName %s)", ReceiverFAT[index].fatName.c_str());
	debug(D_STD," -- ");

	ret = BuildFileRequest(0x03, (char *) ReceiverFAT[index].fatName.c_str());
	SendQuery(pSerial,"RID");
	ret = FindNCTPacket(pSerial,3000, 0x7e);
	if (0x7e != ret)
		return(FAILURE);
	if (0x03 != navcom_msg.msg_type.type_0x7e.SubID)
		return(FAILURE);
	if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x03.Result.bad_bits)
		debug(D_WARN,"Navcom::DownloadFileAtIndex: Error Deleting %s (BAD SPECIAL BITS)\n",ReceiverFAT[index].fileName.c_str());
	if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x03.Result.bad_filename)
		debug(D_WARN,"Navcom::DownloadFileAtIndex: Error Deleting %s (BAD FILENAME)\n",ReceiverFAT[index].fileName.c_str());
	if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x03.Result.mnt_fail)
		debug(D_WARN,"Navcom::DownloadFileAtIndex: Error Deleting %s (MOUNT FAILURE)\n",ReceiverFAT[index].fileName.c_str());

	if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x03.num_deleted == 1) {
		status = SUCCESS;
		debug(D_STD," Deleted\n");
	} else {
		debug(D_STD," NOT Deleted\n\n");
	}
	debug(D_FLOW,"Navcom::DeleteFiles: Exiting\n");
	return(status);
}
//  ===================================================================
//  NAME:		Navcom::Stream
//
//  PURPOSE:	Set the Navcom to start streaming
//
//  RETURNS:	SUCCESS/FAILURE
//
//  HISTORY:
//
//  ===================================================================
int Navcom::Stream(Serial *pSerial, int format, double rate, char port) {

	debug(D_FLOW,"Navcom::Stream: Entering...\n");
	int status = SUCCESS;
	int rc;
	if (rate <= 0) {
		BuildDataRequest(0xB0, RATE_ONCE, PORT_DATA);
		BuildDataRequest(0xB1, RATE_ONCE, PORT_DATA, FALSE);
		BuildDataRequest(0x86, RATE_ONCE, PORT_DATA, FALSE);
		BuildDataRequest(0x44, RATE_ONCE, PORT_DATA, FALSE);
		BuildDataRequest(0x81, RATE_ONCE, PORT_DATA, FALSE);
		rc = SendQuery(pSerial,"TNMP");
		return(rc);
	}

	try {

		int t_rate = 0;
		switch (int(rate*100)) {
			case 2: t_rate = RATE_50HZ;
				break;
			case 4: t_rate = RATE_25HZ;
				break;
			case 10: t_rate = RATE_10HZ;
				break;
			case 20: t_rate = RATE_5HZ;
				break;
			case 50: t_rate = RATE_2HZ;
				break;
			default: t_rate = (int)RATE_SECS(rate);
		}
		debug(100,"rate is %x\n",t_rate);
		BuildDataRequest(0xB0, t_rate, PORT_DATA);
		BuildDataRequest(0xB1, t_rate, PORT_DATA, FALSE);
		BuildDataRequest(0x86, t_rate, PORT_DATA, FALSE);
		BuildDataRequest(0x44, RATE_CHANGE, PORT_DATA, FALSE);
		BuildDataRequest(0x81, RATE_CHANGE, PORT_DATA, FALSE);
		rc = SendQuery(pSerial, "TMP");

		debug(D_STD,"Navcom::Stream: Receiver has been set to stream\n");
	}
	catch (const char *pmessage) {
		debug(D_WARN,pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"Navcom::Stream: Exiting...\n");
	return(status);
}

int Navcom::BinexStream(Serial *pSerial, double rate, char port) {

	debug(D_FLOW,"Navcom::BNXStream: Entering...\n");
	int status = SUCCESS;
	int rc;
	if (rate <= 0) {
		char buffer[101];
		sprintf(buffer,"BNX,OFF,%c",port);
		for (int i = 0; i < 10; i++) {
			pSerial->ClearBuffers();
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == SUCCESS) break;
		}
		return(rc);
	}

	try {
		char buffer[101];
		for (char pp = 'A'; pp <= 'B'; pp++) {
			// turn on 0x0 type for port
			sprintf(buffer,"BNX,TYP,%d,%c,%.1f",0x0,pp,rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Error setting streaming for type 0x0\n";

			// turn on 0x1 type for port
			sprintf(buffer,"BNX,TYP,%d,%c,%.1f",0x1,pp,rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Error setting streaming for type 0x1\n";

			// turn on 0x7e type for port
			sprintf(buffer,"BNX,TYP,%d,%c,%.1f",0x7e, pp,rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Error setting streaming for type 0x7e\n";

			// turn on 0x0 type for port
			sprintf(buffer,"BNX,TYP,%d,%c,%.1f",0x7f, pp,rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Error setting streaming for type 0x7f\n";

		}

		if ((RxID.find("CJ12") != string::npos) || (RxID.find("CJ00") != string::npos)) {
			sprintf(buffer,"BNX,INT,%.1f",rate);
			rc = SendCommand(pSerial, buffer, FALSE, TRUE);
			if (rc == FAILURE) throw "Receiver has _NOT_ been set to stream\n";
		}

		sprintf(buffer,"BNX,ON,%c",port);
		rc = SendCommand(pSerial, buffer, FALSE, TRUE);
		if (rc == FAILURE) throw "Receiver has _NOT_ been set to stream\n";
		debug(D_STD,"Navcom::BinexStream: Receiver has been set to stream\n");
	}
	catch (const char *pmessage) {
		debug(D_WARN,pmessage);
		status = FAILURE;
	}
	debug(D_FLOW,"Navcom::BinexStream: Exiting...\n");
	return(status);

}

//  ===================================================================
//  NAME:		Navcom::DownloadReceiverID
//
//  PURPOSE:	Get the ID from the Navcom Receiver. This will also
//				reset the receiver type depending on the ID returned
//
//  RETURNS:	Receiver ID if found, Empty string if not found
//
//  HISTORY:
//
//  ===================================================================
string Navcom::DownloadReceiverID(Serial *pSerial) {

	debug(D_FLOW,"Navcom::DownloadReceiverID: Entering...\n");

	char locbuf[100];
	RxID = "";
	int ret;
	if (! pSerial->isConnected()) {
		debug(D_ERRMAJOR,"Navcom::DownloadReceiverID: Not connected to the receiver\n");;
		return("");
	}

	for (int attempt = 0; (attempt < 3) && (RxID == ""); attempt++) {
		pSerial->ClearBuffers();
		pSerial->ClearBuffers();
		debug(D_STD,"Navcom::DownloadReceiverID: Querying Receiver ID (attempt %d):\n", attempt);
		BuildDataRequest(0xAE, RATE_ONCE, PORT_ALL);
		SendQuery(pSerial, "RID");
		ret = FindNCTPacket(pSerial,3000,0xAE);
		if (0xAE == ret) {
			printf("Navcom::DownloadReceiverID: Found Packet\n");
			sprintf(locbuf,"Version=%d:%d,DigitalCard=%d:%d,RFCard=%d:%d",
					navcom_msg.msg_type.type_0xae.version.Major,
					navcom_msg.msg_type.type_0xae.version.Minor,
					navcom_msg.msg_type.type_0xae.DigitalCardNumber.serial_num,
					navcom_msg.msg_type.type_0xae.DigitalCardNumber.serial_class,
					navcom_msg.msg_type.type_0xae.RFCardNumber.serial_num,
					navcom_msg.msg_type.type_0xae.RFCardNumber.serial_class);
			RxID = locbuf;
		}

	}
	if (RxID == "") {
		debug(D_ERRMAJOR,"Navcom::DownloadReceiverID: Unable to read Receiver ID\n");
		return(RxID);
	}

	addToSumFile(" Receiver ID: %s\n",RxID.c_str());

	debug(D_STD,"Navcom::DownloadReceiverID: Found: %s\n",RxID.c_str());
	debug(D_FLOW,"Navcom::DownloadReceiverID: Exiting\n");
	return(RxID);
}


//  ===================================================================
//  NAME:		Navcom::DownloadReceiverFAT
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Navcom::DownloadReceiverFAT(Serial *pSerial) {

	printf("Navcom::DownloadReceiverFAT: Entering...\n");
	int ret;
	int numfiles;
	int filesfound = 0;
	int Page = 0;

	do {
		BuildFileRequest(0x01, Page, "????????.???\0\0");
		
		SendQuery(pSerial, "RID");
		ret = FindNCTPacket(pSerial,3000,0x7e);
		if (0x7e != ret)
			return(FAILURE);
		if (0x01 != navcom_msg.msg_type.type_0x7e.SubID)
			return(FAILURE);

		debug(D_FLOW,"Navcom::DownloadReceiverFAT: Decoding sub packet ID %d\n",navcom_msg.msg_type.type_0x7e.SubID);
		numfiles = navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.NumEntries;
		debug(D_FLOW,"Navcom::DownloadReceiverFAT: Present DateTime %d-%d-%d,%d:%d:%d\n",
			navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Date.year+1980,
			navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Date.mon,
			navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Date.mday,
			navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Time.hour,
			navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Time.min,
			navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Time.sec);
		
		for (int i = 0; i < numfiles; i++) {
			debug(D_FLOW,"Navcom::DownloadReceiverFAT: Create DateTime %d-%d-%d,%d:%d:%d\n",
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.year+1980,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.mon,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.mday,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.hour,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.min,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.sec);
			debug(D_FLOW,"Navcom::DownloadReceiverFAT: Mod DateTime %d-%d-%d,%d:%d:%d\n",
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.year+1980,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.mon,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.mday,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.hour,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.min,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.sec);
			debug(D_FLOW,"Navcom::DownloadReceiverFAT: FileName = %s, FileSize = %d\n",
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].FileName,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].FileSize);
			ReceiverFAT[filesfound].fileName = (char *) navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].FileName;
			ReceiverFAT[filesfound].fatName = (char *) navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].FileName;
			ReceiverFAT[filesfound].siteName.erase(0,ReceiverFAT[i].siteName.length());
			ReceiverFAT[filesfound].siteName += ReceiverFAT[i].fileName.substr(0,4);
			ReceiverFAT[filesfound].fileStart.set_time(
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.year+1980,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.mon,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.mday,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.hour,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.min,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.sec);
			ReceiverFAT[filesfound].fileEnd.set_time(
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.year+1980,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.mon,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.mday,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.hour,
				navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.min,
				(double)navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.sec);
			ReceiverFAT[filesfound].fileSize = navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].FileSize;
			ReceiverFAT[filesfound].active = FALSE;
			FindSession(filesfound);
			CreateFileName(filesfound);
			filesfound++;
		}
		Page++;
	} while ( ! navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Result.final_page );
	numFATEntries = filesfound;
	
	return(SUCCESS);
}


//  ===================================================================
//  NAME:		Navcom::FindNCTPacket
//
//  PURPOSE:	Search incoming serial port for NCT packet
//				
//
//  RETURNS:	return packet found or -1 for error
//
//  HISTORY:
//
//  ===================================================================
int Navcom::FindNCTPacket(Serial *pSerial, time_t timeout, int packet) {

	unsigned char buffer[NAVCOM_MAX_MSG_SIZE];
	int id;
	U16 msg_size;
	int index;

	debug(D_FLOW,"Navcom::FindNCTPacket: Entering...\n");

	do {
		try {
			memset((void *) buffer, 0, sizeof(buffer));
			index = 0;
	
			if (pSerial->WaitFor(timeout, NC_STX) != 0)
				throw("Navcom::FindNCTPacket: timeout looking for start of packet\n");
			if (pSerial->WaitFor(timeout, NC_PREAMBLE_1) != 0)
				throw("Navcom::FindNCTPacket: timeout looking for first preamble\n");
			if (pSerial->WaitFor(timeout, NC_PREAMBLE_2) != 0)
				throw("Navcom::FindNCTPacket: timeout looking for second preamble\n");
	
			buffer[index++] = NC_STX;
			buffer[index++] = NC_PREAMBLE_1;
			buffer[index++] = NC_PREAMBLE_2;
	
			// Read Command ID
			buffer[index++] = pSerial->Inkey(100);
			id = buffer[index-1];
			debug(D_VALUES,"Navcom::FindNCTPacket: Found CommandID %X\n",(int) id);
	
			// Read Message Length
			buffer[index++] = pSerial->Inkey(100);
			buffer[index++] = pSerial->Inkey(100);
			memcpy((void *) &msg_size, (void *)&buffer[index-2], 2);
			debug(D_VALUES,"Navcom::FindNCTPacket: Read message size %X\n",msg_size);
	
			// Read Message
			for (int i = 0; i < msg_size - 3; i++)
				buffer[index++] = pSerial->Inkey(100);
	
			// Read ETX
			buffer[index] = (char)pSerial->Inkey(timeout);
			if (buffer[index] != NC_ETX)
				throw("Navcom::FindNCTPacket: Incorrect ETX value\n");
	
			if (id == packet)
				DecodeNCTPacket(id,msg_size,buffer,index+1);
	
		}
		catch (const char *pmessage)
		{
   			debug(D_WARN, pmessage);
			id = -1;
		}
	} while((id != packet));
	
	debug(D_FLOW,"Navcom::FindNCTPacket: Exiting...\n");
	return(id);


}


//  ===================================================================
//  NAME:		Navcom::BuildDataRequest
//
//  PURPOSE:	Set up data request to get NCT messages from receiver
//				
//
//  RETURNS:	returns number of queued data requests
//
//  HISTORY:
//
//  ===================================================================
int Navcom::BuildDataRequest(int blockID,  int dataRate, char outputPorts, bool newReq) {

	debug(D_FLOW,"Navcom::BuildDataRequest: Entering...\n");

	if (newReq) {
		debug(D_VALUES,"Navcom::BuildDataRequest: Building new data request\n");
		memset((void *) &navcom_msg, 0, sizeof(navcom_msg));
		blockRequests.clear();
	}

	blockRequests.push_back(blockID);
	int numBlocks = blockRequests.size();

	navcom_msg.STX_ = NC_STX;
	navcom_msg.Preamble1 = NC_PREAMBLE_1;
	navcom_msg.Preamble2 = NC_PREAMBLE_2;
	navcom_msg.CommandID = NC_QUERY_ID;
	navcom_msg.MessageLength = 0; // will be set later
	navcom_msg.msg_type.type_0x20.w1.nvram = 0;
	navcom_msg.msg_type.type_0x20.w1.dflt = 0;
	navcom_msg.msg_type.type_0x20.w1.clear = 1;
	navcom_msg.msg_type.type_0x20.w1.user_nvram = 0;
	navcom_msg.msg_type.type_0x20.numrequests = numBlocks;
	navcom_msg.msg_type.type_0x20.requests[numBlocks-1].blockID = blockID;
// send to all ports
	debug(200,"outputPorts = %x\n",outputPorts);
	navcom_msg.msg_type.type_0x20.requests[numBlocks-1].w2.diag = (outputPorts & PORT_DIAG) ? 1 : 0; //outputPorts & PORT_DIAG;
	debug(200,"diag = %x\n",navcom_msg.msg_type.type_0x20.requests[numBlocks-1].w2.diag);
	navcom_msg.msg_type.type_0x20.requests[numBlocks-1].w2.control = (outputPorts & PORT_CONTROL) ? 1 : 0; //outputPorts & PORT_CONTROL;
	debug(200,"control = %x\n",navcom_msg.msg_type.type_0x20.requests[numBlocks-1].w2.control);
	navcom_msg.msg_type.type_0x20.requests[numBlocks-1].w2.data = (outputPorts & PORT_DATA) ? 1 : 0; //outputPorts & PORT_DATA;
	debug(200,"data = %x\n",navcom_msg.msg_type.type_0x20.requests[numBlocks-1].w2.data);
	navcom_msg.msg_type.type_0x20.requests[numBlocks-1].w2.log = (outputPorts & PORT_LOG) ? 1 : 0; //outputPorts & PORT_LOG;
	debug(200,"log = %x\n",navcom_msg.msg_type.type_0x20.requests[numBlocks-1].w2.log);
	navcom_msg.msg_type.type_0x20.requests[numBlocks-1].dataRate = dataRate;

	navcom_msg.MessageLength = 6+(numBlocks*4);
	debug(D_VALUES,"Navcom::BuildDataRequest: Message length is %d\n",navcom_msg.MessageLength);

	debug(D_FLOW,"Navcom::BuildDataRequest: Exiting(%d)...\n",numBlocks);
	return(numBlocks);
}

//  ===================================================================
//  NAME:		Navcom::BuildFileRequest
//
//  PURPOSE:	Set up data request to get NCT messages from receiver
//				
//
//  RETURNS:	returns number of queued data requests
//
//  HISTORY:
//
//  ===================================================================
int Navcom::BuildFileRequest(int subID) {

	debug(D_FLOW,"Navcom::BuildFileRequest: Entering...\n");

	debug(D_VALUES,"Navcom::BuildFileRequest: Building new data request(subID = %d)\n",subID);
	memset((void *) &navcom_msg, 0, sizeof(navcom_msg));
	blockRequests.clear();

	blockRequests.push_back(0x7e);
	int numBlocks = blockRequests.size();

	navcom_msg.STX_ = NC_STX;
	navcom_msg.Preamble1 = NC_PREAMBLE_1;
	navcom_msg.Preamble2 = NC_PREAMBLE_2;
	navcom_msg.CommandID = 0x7e;
	navcom_msg.MessageLength = 0; // will be set later
	navcom_msg.msg_type.type_0x7e.SubID = subID;
	navcom_msg.msg_type.type_0x7e.Command = 0; // this is a request

	navcom_msg.MessageLength = 8;
	debug(D_VALUES,"Navcom::BuildFileRequest: Message length is %d\n",navcom_msg.MessageLength);

	debug(D_FLOW,"Navcom::BuildFileRequest: Exiting(%d)...\n",numBlocks);
	return(numBlocks);
}

// 0x01
int Navcom::BuildFileRequest(int subID, int Page, const char * fileSpec) {

	
	int len;
	if (subID != 0x01) {
		debug(D_ERRMAJOR,"Incorrect subID passed to BuildFileRequest\n");
		return(FAILURE);
	}
	BuildFileRequest(subID);
	debug(D_FLOW,"Navcom::BuildFileRequest: Setting values for subID = %d\n",subID);
	navcom_msg.msg_type.type_0x7e.subID_type.req_0x01.PageNumber = Page;
	len = strlen((const char*) fileSpec);
	memcpy(navcom_msg.msg_type.type_0x7e.subID_type.req_0x01.FileName,fileSpec,len);

	len += (len % 2) ?  3 : 2; // make even number of nulls at end of string 

	navcom_msg.MessageLength += len + 2; // length + Page
	return(SUCCESS);

}

// 0x08, 0x03
int Navcom::BuildFileRequest(int subID, char * fileSpec) {

	int len = 0;
	BuildFileRequest(subID);
	switch (subID) {
		case 0x03:
			debug(D_FLOW,"Navcom::BuildFileRequest: Setting values for subID = %d\n",subID);
			memcpy(navcom_msg.msg_type.type_0x7e.subID_type.req_0x03.FileName,fileSpec,14);
			navcom_msg.msg_type.type_0x7e.subID_type.req_0x03.special = SPECIAL_0X03;
			len = strlen((const char*) fileSpec);
			len += (len % 2) ?  3 : 2; // make even number of nulls at end of string 
			navcom_msg.MessageLength += len;
			break;
		case 0x08:
			debug(D_FLOW,"Navcom::BuildFileRequest: Setting values for subID = %d\n",subID);
			memcpy(navcom_msg.msg_type.type_0x7e.subID_type.req_0x08.FileName,fileSpec,14);
			len = strlen((const char*) fileSpec);
			len += (len % 2) ?  3 : 2; // make even number of nulls at end of string 
			navcom_msg.MessageLength += len;
			break;
		default:
			debug(D_ERRMAJOR,"Navcom::BuildFileRequest: Incorrect subID passed to BuildFileRequest\n");
			return(FAILURE);
	}
	return(SUCCESS);

}

// 0x09
int Navcom::BuildFileRequest(int subID, U16 fh, U32 sector) {

	BuildFileRequest(subID);
	switch (subID) {
		case 0x09:
			debug(D_FLOW,"Navcom::BuildFileRequest: Setting values for subID = %d\n",subID);
			navcom_msg.msg_type.type_0x7e.subID_type.req_0x09.fh = fh;
			navcom_msg.msg_type.type_0x7e.subID_type.req_0x09.sec_number = sector;
			navcom_msg.MessageLength += 5;
			break;
		default:
			debug(D_ERRMAJOR,"Navcom::BuildFileRequest: Incorrect subID passed to BuildFileRequest\n");
			return(FAILURE);
	}
	return(SUCCESS);

}


//  ===================================================================
//  NAME:		Navcom::SendQuery
//
//  PURPOSE:	Send the specified query to the receiver. Do not wait for
//				a return. Do not request an ACK.
//
//  RETURNS:	returns SUCCESS if Write to serial port OK. FAILURE if
//				serial port is not connected.
//
//  HISTORY:
//
//  ===================================================================
int Navcom::SendQuery(Serial *pSerial, string QueryString) {

	int status = SUCCESS;
	debug(D_FLOW,"Navcom::SendQuery: Entering...\n");
	unsigned char buffer[NAVCOM_MAX_MSG_SIZE];

	try {
		if (! pSerial->isConnected())
			throw "Not connected to the receiver\n";
		
		int sendsize = navcom_msg.MessageLength + 4; // includes header, chksum and ETX

		memset(buffer, 0, sizeof(buffer));
		memcpy((void *) buffer, (void *) &navcom_msg, sendsize - 2);
		
		for (int i = 3; i < sendsize-2; i++) {
			buffer[sendsize - 2] ^= buffer[i];
		}
		buffer[sendsize-1] =  NC_ETX;

		for (int i = 0; i < sendsize; i++) {
			debug(300, "%d %X\n",i, buffer[i]);
		}
		pSerial->Write((void *) buffer, sendsize);
	}
	catch (const char *pmessage)
	{
   		debug(D_WARN, pmessage);
		status = FAILURE;
	}

	debug(D_FLOW,"Navcom::SendQuery: Exiting\n");
	return(status);
}

//  ===================================================================
//  NAME:		Navcom::SendCommand
//
//  PURPOSE:	Send the specified PASH command to the receiver. Add the checksum
//				to the end if needed and wait for an ack if desired.
//
//  RETURNS:	if WaitAck is requested it will return SUCCESS/FAILURE depending
//				on return from receiver. If WaitAck is not requested it will
//				return SUCCESS if connected to the serial port and Write went OK.
//
//  HISTORY:
//
//  ===================================================================
int Navcom::SendCommand(Serial *pSerial, string CommandString, int DoCheckSum, int WaitAck) {

	int status = SUCCESS;

	debug(D_FLOW,"Navcom::SendCommand: Entering...\n");

	try {
		if (! pSerial->isConnected())
			throw "Not connected to the receiver\n";

		char buffer[101];
		string Send = "";
		if (CommandString.find("$PASHS,") != 0) {
			Send += "$PASHS,";
		}
		Send += CommandString;
		int Checksum = 0;

		if (DoCheckSum) {
			for (int i = 1; i < (int)Send.length(); i++) {
				Checksum ^= Send[i];
			}
			sprintf(buffer, "*%02X", Checksum);
			Send += buffer;
		}
		debug(D_FLOW,"Navcom::SendCommand: Sending %s\n", Send.c_str());
		Send += "\r\n";
		pSerial->ClearBuffers();
		pSerial->Write((char *)Send.c_str());

		if (WaitAck) 
			status = CheckAck(pSerial);
	}
	catch (const char *pmessage)
	{
   		debug(D_WARN, pmessage);
		status = FAILURE;
	}

	debug(D_FLOW,"SendCommand: Exiting\n");
	return(status);
}

int Navcom::CheckAck(Serial *pSerial) {
		string Check;
		char buffer[101];
		debug(D_FLOW,"Navcom::WaitAck: Entering...\n");
		for (int ticks = 3000; ticks > 0; ticks -= 500) {
			if (pSerial->Readline(buffer,100,500) != 0) {
				Check += buffer;
				if (Check.find("ACK") != string::npos) return(SUCCESS);
				if (Check.find("NAK") != string::npos) return(FAILURE);
			}
			debug(D_FLOW,"Navcom::CheckAck: Waiting for response - %d\r",ticks);
			ticks -= 500;
		}
		debug(D_WARN,"Navcom::CheckAck: Timeout waiting for ACK/NAK\n");
		debug(D_FLOW,"Navcom::CheckAck: Exiting...\n");
		return(FAILURE);
}

void Navcom::AddToIndexBuffer(int index, unsigned char *in, size_t size) {
	if (index >= MAX_RCVR_FILES) return;
	if (indexData[index] != NULL) return;

	indexData[index] = new unsigned char[size];
	if (indexData[index] != NULL) return;
	memcpy(indexData[index], in, size);
	return;
}

#define PROMPT "Terminal-%s> "
int Navcom::TerminalSession(Serial *pSerial) {

	debug(D_FLOW,"Navcom::TerminalSession: Entering...\n");
	bool quit = false;
	char prompt[100];
	sprintf(prompt,PROMPT,GetSiteName(0).c_str());
	FILE *in = fdopen(1,"r");
	do {
		string myin = "";
		int query = -1;
		std::cout << prompt;
		char input[1024];
		fgets(input,1024,in);
		if (input[0] == '\\') {
			for (int i = 1; i < (int) strlen(input); i++) {
				if ((input[i] == '\r') || (input[i] == '\n')) break;
				myin += input[i];
			}
		} else {
			for (int i = 0; i < (int) strlen(input); i++) {
				if ((input[i] == '\r') || (input[i] == '\n')) break;
				myin += toupper(input[i]);
			}
		}
		printf("Input = %s\n",myin.c_str());
		if (myin.compare("QUIT") == 0) break;
		if (myin.find("$PASHS,") == 0) query = 0;
		if (myin.find("$PASHQ,") == 0) query = 1;
		switch (query) {
			case 0:
				SendCommand(pSerial,myin);
				break;
			case 1:
				SendQuery(pSerial,myin);
				break;
			default:
				printf("Unknown Command or Query: %s\n",myin.c_str());
				continue;
		}
		char buffer[1024];
		int c;
		while ((c = pSerial->Readline(buffer, 1024, 1000)) > 0) {
			std::cout << buffer;
		}
	} while (!quit);
	fclose(in);

	debug(D_FLOW,"Navcom::TerminalSession: Exiting...\n");
	return(SUCCESS);
}

int Navcom::DecodeNCTPacket(int id, U16 msg_size, unsigned char * buffer, int index ) {

	debug(D_FLOW,"Navcom::DecodeNCTPacket: Entering...\n");
	int numfiles;
	int j;
	int len;

	debug(D_FLOW,"Navcom::DecodeNCTPacket: Passed blockID = %X\n", navcom_msg.CommandID);
	memcpy((void *) &navcom_msg, (void *) buffer, index);
	for (int i = 0; i < msg_size + 4; i++) {
		debug(300, "%d %X\n",i, buffer[i]);
	}

	switch (navcom_msg.CommandID) {
		case 0xae:
			debug(D_FLOW,"Navcom::DecodeNCTPacket: config = %d\n",
					navcom_msg.msg_type.type_0xae.config);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: ASIC = %d\n",
					navcom_msg.msg_type.type_0xae.ASIC);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: Version Major = %d, Minor = %d\n",
					navcom_msg.msg_type.type_0xae.version.Major,
					navcom_msg.msg_type.type_0xae.version.Minor);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: DigitalCardNumber Serial number = %d, class = %d\n",
					navcom_msg.msg_type.type_0xae.DigitalCardNumber.serial_num,
					navcom_msg.msg_type.type_0xae.DigitalCardNumber.serial_class);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: RFCardNumber Serial number = %d, class = %d\n",
					navcom_msg.msg_type.type_0xae.RFCardNumber.serial_num,
					navcom_msg.msg_type.type_0xae.RFCardNumber.serial_class);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: IOPVersion Major = %d, Minor = %d\n",
					navcom_msg.msg_type.type_0xae.IOPVersion.Major,
					navcom_msg.msg_type.type_0xae.IOPVersion.Minor);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: softwareTimeMark = %16s\n",
					navcom_msg.msg_type.type_0xae.softwareTimeMark);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: bootString = %16s\n",
					navcom_msg.msg_type.type_0xae.bootString);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: IOPTimeMark = %16s\n",
					navcom_msg.msg_type.type_0xae.IOPTimeMark);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: PicVersion = %d\n",
					navcom_msg.msg_type.type_0xae.PicVersion);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: StarlightBuild = %d\n",
					navcom_msg.msg_type.type_0xae.StarlightBuild);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: IOPBuild = %d\n",
					navcom_msg.msg_type.type_0xae.IOPBuild);
			break;
		case 0x11:
			debug(D_FLOW,"Navcom::DecodeNCTPacket: port = COM%d\n",
					navcom_msg.msg_type.type_0x11.ports[0].w2.port);
			debug(D_FLOW,"Navcom::DecodeNCTPacket: baud = ");
			for (int i = 0; (6 + i*2 + 2) < navcom_msg.MessageLength; i++) {
				switch (navcom_msg.msg_type.type_0x11.ports[i].w2.baud) {
					case 0: debug(D_FLOW,"1200\n"); break;
					case 1: debug(D_FLOW,"2400\n"); break;
					case 2: debug(D_FLOW,"4800\n"); break;
					case 3: debug(D_FLOW,"9600\n"); break;
					case 4: debug(D_FLOW,"19200\n"); break;
					case 5: debug(D_FLOW,"38400\n"); break;
					case 6: debug(D_FLOW,"57600\n"); break;
					case 7: debug(D_FLOW,"115200\n"); break;
				}
				debug(D_FLOW,"Navcom::DecodeNCTPacket: parity = ");
				switch (navcom_msg.msg_type.type_0x11.ports[i].w2.parity) {
					case 0: debug(D_FLOW,"NONE\n"); break;
					case 1: debug(D_FLOW,"ODD\n"); break;
					case 2: debug(D_FLOW,"EVEN\n"); break;
				}
			}
			break;
		case 0x1A:
			for (int i = 0; i < 4; i++) {
				debug(D_FLOW,"Navcom::DecodeNCTPacket: Values for port %d:\n",i);
				debug(D_FLOW,"\t\tDIAG    = %d\n", navcom_msg.msg_type.type_0x1a.port[i].diag);
				debug(D_FLOW,"\t\tCONTROL = %d\n", navcom_msg.msg_type.type_0x1a.port[i].control);
				debug(D_FLOW,"\t\tDATA    = %d\n", navcom_msg.msg_type.type_0x1a.port[i].data);
				debug(D_FLOW,"\t\tNMEA    = %d\n", navcom_msg.msg_type.type_0x1a.port[i].nmea);
				debug(D_FLOW,"\t\tRTCM    = %d\n", navcom_msg.msg_type.type_0x1a.port[i].rtcm);
				debug(D_FLOW,"\t\tCMRIN   = %d\n", navcom_msg.msg_type.type_0x1a.port[i].cmrin);
				debug(D_FLOW,"\t\tCMROUT  = %d\n", navcom_msg.msg_type.type_0x1a.port[i].cmrout);
			}
			break;
		case 0x7E:
			if (navcom_msg.msg_type.type_0x7e.Command != 0xff) {
				debug(D_WARN,"Navcom::DecodeNCTPacket: 7e response packet not 0xff (%X)\n",
					navcom_msg.msg_type.type_0x7e.Command != 0xff);
				return(FAILURE);
			}
			switch (navcom_msg.msg_type.type_0x7e.SubID) {
				case 0x00:
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Decoding sub packet ID %d\n",navcom_msg.msg_type.type_0x7e.SubID);
					break;
				case 0x01:
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Decoding sub packet ID %d\n",navcom_msg.msg_type.type_0x7e.SubID);
					numfiles = navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.NumEntries;
					j = 30;
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Present DateTime %d-%d-%d,%d:%d:%d\n",
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Date.year+1980,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Date.mon,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Date.mday,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Time.hour,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Time.min,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.Time.sec);
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Copying %d file entries (bufstart = %d)\n",
						numfiles,j);
					for (int i = 1; i < numfiles; i++) {
						len = strlen((const char*)&buffer[j]);
						len += (len % 2) ?  3 : 2; // make even number of nulls at end of string 
						j += len;
						debug(D_FLOW,"Navcom::DecodeNCTPacket: Copying from %d index -j = %d\n",j,index-j);
						memcpy((void *) &navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i],
								(void *) &buffer[j],
								index - j);
						j += 14;
					}
					for (int i = 0; i < numfiles; i++) {
						debug(D_FLOW,"Navcom::DecodeNCTPacket: Create DateTime %d-%d-%d,%d:%d:%d\n",
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.year+1980,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.mon,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatDate.mday,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.hour,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.min,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].CreatTime.sec);
						debug(D_FLOW,"Navcom::DecodeNCTPacket: Mod DateTime %d-%d-%d,%d:%d:%d\n",
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.year+1980,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.mon,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModDate.mday,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.hour,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.min,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].ModTime.sec);
						debug(D_FLOW,"Navcom::DecodeNCTPacket: FileName = %s, FileSize = %d\n",
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].FileName,
							navcom_msg.msg_type.type_0x7e.subID_type.res_0x01.DirEntry[i].FileSize);
					}
						
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Decoding sub packet ID %d\n",navcom_msg.msg_type.type_0x7e.SubID);
					break;
				case 0x08:
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Decoding sub packet ID %d\n",navcom_msg.msg_type.type_0x7e.SubID);
					debug(D_FLOW,"Navcom::DecodeNCTPacket: File Name = %s\n",
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.FileName);
					debug(D_FLOW,"Navcom::DecodeNCTPacket: File Handle = %d\n",
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.fh);
					debug(D_FLOW,"Navcom::DecodeNCTPacket: File Sectors = %d, Bytes = %d\n",
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.sec_size,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.byte_size);
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Create DateTime %d-%d-%d,%d:%d:%d\n",
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.CreatDate.year+1980,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.CreatDate.mon,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.CreatDate.mday,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.CreatTime.hour,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.CreatTime.min,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.CreatTime.sec);
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Modified DateTime %d-%d-%d,%d:%d:%d\n",
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.ModDate.year+1980,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.ModDate.mon,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.ModDate.mday,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.ModTime.hour,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.ModTime.min,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.ModTime.sec);
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Access Date %d-%d-%d\n",
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.AccDate.year+1980,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.AccDate.mon,
						navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.AccDate.mday);
					break;
				case 0x09:
					break;
				case 0x0a:
					break;
					
					
				default:
					debug(D_FLOW,"Navcom::DecodeNCTPacket: Unsupported 0x7e sub packet found: %X\n",
						navcom_msg.msg_type.type_0x7e.SubID);
			}
			break;
		default:
			debug(D_WARN,"Navcom::DecodeNCTPacket: Unsupported packet found: %X\n",navcom_msg.CommandID);
			break;
	}

	debug(D_FLOW,"Navcom::DecodeNCTPacket: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
void Navcom::AddImage(Navcom *newNavcom) {
	debug(50,"Navcom::AddImage(Navcom): Entering...\n");

	int fileSize = newNavcom->getImageFileSize();
	unsigned char *image = newNavcom->getImageFile();

	unsigned char *tmp = new unsigned char[cimageFile+fileSize];
	memcpy(tmp, imageFile, cimageFile);
	memcpy(&tmp[cimageFile], image, fileSize);
	setImageFile(tmp, cimageFile+fileSize);
	debug(25,"Navcom::Addimage(Navcom): fileSize is %d\n",cimageFile);
	delete [] tmp;

	debug(50,"MicroZ::AddImage(Navcom): Entering...\n");


}
//  ===================================================================
//  NAME:		Navcom::DownloadFileByTime
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Navcom::DownloadFileByTime(Serial *pSerial, gpstime *start, gpstime *stop) {
	int i;
	int numindex, indexes[MAX_RCVR_FILES];
	int status = FAILURE;
	char buffer[100];

	debug(D_FLOW,"Navcom::DownloadFileByTime: Entering...\n");

	numindex = FindIndexesByTime(start, stop, indexes, FALSE);

	if (numindex > 0) {
		debug(D_STD_V2,"Navcom::DownloadFileByTime: Downloading file at index %d\n", indexes[0]);
		status = DownloadFileAtIndex(pSerial, indexes[0]);
	} else {
		debug(D_STD_V2,"Navcom::DownloadFileByTime: No Files found for time range\n");
		status = SUCCESS_NOFILE;
	}
	if ((status == SUCCESS) && (numindex > 1)) {
		for (i = 1; i < numindex; i++) {
			debug(D_STD_V2,"Navcom::DownloadFileByTime: Downloading file at index %d\n", indexes[i]);
			Navcom *pTmp = new Navcom;
			status = pTmp->DownloadFileAtIndex(pSerial, indexes[i]);
			if (status == FAILURE) {
				delete(pTmp);
				break;
			}
			AddImage(pTmp);
			delete(pTmp);
			addToSumFile(" Span: ");
			addToSumFile(start->chartime("%%G/%%g-",buffer,100));
			addToSumFile(stop->chartime("%%G/%%g\n",buffer,100));
		}
	}
	debug(D_FLOW,"DownloadFileByTime: Exiting\n");
	return(status);
}



//  ===================================================================
//  NAME:		Navcom::DownloadFileAtIndex
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Navcom::DownloadFileAtIndex(Serial *pSerial, int index) {

	debug(D_FLOW,"Navcom::DownloadFileAtIndex: Entering...\n");

	unsigned char *myfile = NULL;
	deque<unsigned char> File;
	int downLoadOK = FALSE;

	debug(D_STD_V2,"Navcom::DownloadFileAtIndex: Download at index %d\n",index);
	if (index >= numFATEntries) {
		debug(25,"Navcom::DownloadFileAtIndex: Index %d does not exist\n",index);
		return(FAILURE);
	}

	addToSumFile("\nDownload:\n");
	U32 fileSize;
	U32 numsectors;
	int ret;
	U16 fh;

	sleep(1);
	pSerial->ClearBuffers();
	U32 expectSize = ReceiverFAT[index].fileSize;
	debug(D_FLOW,"Navcom::DownloadFileAtIndex: Checking for Open files\n");
	ret = BuildFileRequest(0x00);
	SendQuery(pSerial,"RID");
	ret = FindNCTPacket(pSerial,3000, 0x7e);
	if (0x7e != ret)
		return(FAILURE);
	if (0x00 != navcom_msg.msg_type.type_0x7e.SubID)
		return(FAILURE);
	if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x00.Status.read_open) {
		debug(D_FLOW,"Found File open for reading, closing\n");
		ret = BuildFileRequest(0x0a);
		navcom_msg.msg_type.type_0x7e.subID_type.req_0x0a.Control.read_close = 1;
		navcom_msg.MessageLength += 3; // will be set later
		SendQuery(pSerial,"RID");
		ret = FindNCTPacket(pSerial,3000, 0x7e);
		if (0x7e != ret)
			return(FAILURE);
		if (0x0a != navcom_msg.msg_type.type_0x7e.SubID)
			return(FAILURE);
	}
		
	debug(D_FLOW,"Navcom::DownloadFileAtIndex: Present DateTime %d-%d-%d,%d:%d:%d\n",
		navcom_msg.msg_type.type_0x7e.subID_type.res_0x00.Date.year+1980,
		navcom_msg.msg_type.type_0x7e.subID_type.res_0x00.Date.mon,
		navcom_msg.msg_type.type_0x7e.subID_type.res_0x00.Date.mday,
		navcom_msg.msg_type.type_0x7e.subID_type.res_0x00.Time.hour,
		navcom_msg.msg_type.type_0x7e.subID_type.res_0x00.Time.min,
		navcom_msg.msg_type.type_0x7e.subID_type.res_0x00.Time.sec);
		

	debug(D_FLOW,"Navcom::DownloadFileAtIndex: Opening %s for Reading\n",ReceiverFAT[index].fatName.c_str());
	ret = BuildFileRequest(0x08, (char *) ReceiverFAT[index].fatName.c_str());
	SendQuery(pSerial,"RID");
	ret = FindNCTPacket(pSerial,3000, 0x7e);
	if (0x7e != ret)
		return(FAILURE);
	if (0x08 != navcom_msg.msg_type.type_0x7e.SubID)
		return(FAILURE);
	if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.Result.no_file)
		debug(D_WARN,"Navcom::DownloadFileAtIndex: Error Opening %s (NO FILE)\n",ReceiverFAT[index].fileName.c_str());
	if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.Result.open_err)
		debug(D_WARN,"Navcom::DownloadFileAtIndex: Error Opening %s (ERROR OPENING)\n",ReceiverFAT[index].fileName.c_str());
	if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.Result.other_file)
		debug(D_WARN,"Navcom::DownloadFileAtIndex: Error Opening %s (ANOTHER FILE OPEN)\n",ReceiverFAT[index].fileName.c_str());

	numsectors = navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.sec_size;
	fileSize = navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.byte_size;
	fh = navcom_msg.msg_type.type_0x7e.subID_type.res_0x08.fh;

	debug(D_FLOW,"Navcom::DownloadFileAtIndex: Downloading %d sectors (%d bytes)\n",numsectors, fileSize);


	ReportStruct reportstruct;
	memset((void *) &reportstruct, 0, sizeof(reportstruct));
	reportstruct.filename = ReceiverFAT[index].fileName.c_str();
	reportstruct.filesize = fileSize;
	reportstruct.start_time = time(NULL);
	TransferReportStart(&reportstruct);

	for (U32 sector = 0; sector < numsectors; sector++) {
		try {
			ret = BuildFileRequest(0x09, fh, sector);
			SendQuery(pSerial,"RID");
			ret = FindNCTPacket(pSerial,3000, 0x7e);
			if (0x7e != ret)
				throw("Navcom::DownloadFileAtIndex: Error (NO FILE HANDLE)\n");
			if (0x09 != navcom_msg.msg_type.type_0x7e.SubID)
				throw("Navcom::DownloadFileAtIndex: Error (NO FILE HANDLE)\n");
			if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x09.Result.fh_fail)
				throw("Navcom::DownloadFileAtIndex: Error (NO FILE HANDLE)\n");
			if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x09.Result.sec_fail)
				throw("Navcom::DownloadFileAtIndex: Error (NO SUCH SECTOR)\n");
			if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x09.Result.read_fail)
				throw("Navcom::DownloadFileAtIndex: Error (READ FAILURE)\n");
			if (navcom_msg.msg_type.type_0x7e.subID_type.res_0x09.Result.mnt_fail)
				throw("Navcom::DownloadFileAtIndex: Error (MOUNT FAILURE)\n");
			int n = navcom_msg.msg_type.type_0x7e.subID_type.res_0x09.bytes;
			for (char *p = (char *)navcom_msg.msg_type.type_0x7e.subID_type.res_0x09.data; --n >= 0; p++) {
                       		File.push_back(*p);
			}
			reportstruct.blocks_downloaded = sector;
			reportstruct.bytes_downloaded = File.size();
			reportstruct.elapsed_time = time(NULL) - reportstruct.start_time;
			TransferReport(&reportstruct);

		}
		catch (const char *pmessage)
		{
   			debug(D_WARN, pmessage);
			myfile = NULL;
			break;
   		}
	}
	TransferReportStop(&reportstruct);
        fileSize = File.size();
        myfile = new unsigned char[fileSize];
        for (U32 i = 0; i < fileSize; i++) {
                myfile[i] = File[i];
        }

	if ((myfile != NULL) && (fileSize >= (expectSize-1024))) { //within one block

		debug(D_STD_V,"Navcom::DownloadFileAtIndex: Download OK, Copying image to Navcom Class\n");
		setImageFile(myfile, fileSize);
		myIndex = index;
		downLoadOK = TRUE;
		char buffer[100];
		delete [] myfile;

		addToSumFile(" Index: %d\n",index+1);
		addToSumFile(" Start Time: %s\n",ReceiverFAT[index].fileStart.chartime("%%G %%g", buffer, 100));
		addToSumFile(" End Time: %s\n",ReceiverFAT[index].fileEnd.chartime("%%G %%g", buffer, 100));
		addToSumFile(" File: %s\n",GetFileName(myIndex).c_str());
		addToSumFile(" Bytes: %d\n",expectSize);
//		addToSumFile(" Time: %d seconds\n",t_used);

	} else {
		downLoadOK = FALSE;
		pSerial->ClearBuffers(); sleep(1); pSerial->ClearBuffers();
	}
	addToSumFile("\n");
	debug(D_FLOW,"Navcom::DownloadFiles: Exiting\n");

	return((downLoadOK) ? SUCCESS : FAILURE);

}

