/* 
#    $Id: Topcon.cpp,v 1.3 2012/02/10 00:48:57 gps Exp $
#
#	 Topcon.cpp -- This file will be the major interaction with
#                   an Topcon receiver
#
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <algorithm>
#include <vector>
#include <list>
#include <ctype.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>

#include "gpscommon.h"
#include "Topcon.h"

typedef unsigned short Crc16;

enum {
        TPS_WIDTH = 16,                     // Width of poly (number of most significant bits).
        TPS_POLY = 0x1021,                  // Poly. Bit #16 is set and hidden.
        TPS_BYTE_BITS = 8,                  // Number of bits in byte.
        TPS_TABLE_SIZE = 1 << TPS_BYTE_BITS,    // Size of table.
        TPS_MSB_MASK = 1 << (TPS_WIDTH - 1)     // Mask for high order bit in word.
};

// Table (generated by 'crc16init()')

static Crc16 table[TPS_TABLE_SIZE];

// Initializes the table. Should be called once before the first call to
// 'crc16()'

void crc16init(void) {
        Crc16 i;
        for(i = 0; i < TPS_TABLE_SIZE; ++i) {
                Crc16 val = i << (TPS_WIDTH - TPS_BYTE_BITS);
                int j;
                for(j = 0; j < TPS_BYTE_BITS; ++j)
                        val = (val << 1) ^ ((val & TPS_MSB_MASK) ? TPS_POLY : 0);
                table[i] = val;
        }
}
// Calculates CRC16 of 'cnt' bytes from 'src' and returns result.
// Starting value of CRC16 is supplied by caller in 'crc'
Crc16 crc16(Crc16 crc, void const* src, int cnt) {
        unsigned char const* s = (unsigned char const*)src;
        while(cnt--) {
                crc = (crc << TPS_BYTE_BITS) ^ table[(crc >> (TPS_WIDTH - TPS_BYTE_BITS)) ^ *s++];
        }
	return crc;
}

bool fat_sort (SFATEntry i,SFATEntry j) { return (i.fileStart.gettime_t() < j.fileStart.gettime_t()); }


//  ===================================================================
//  NAME:		Topcon::Topcon
//
//  PURPOSE:	Constructor. Set several default values
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
Topcon::Topcon () {

	debug(D_FLOW,"Topcon::Topcon: Entering...\n");

	usesFAT = TRUE;
	myIndex = -1;

	cimageFile = 0;
	numMessages = 0;
	imageFile = NULL;
	session = ' ';
	debug(D_FLOW,"Topcon::Topcon: Exiting\n");
	crc16init();


}

//  ===================================================================
//  NAME:		Topcon::Login
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Topcon::Login(Serial *pSerial, int direct, string login, string password)
{

	debug(D_FLOW,"Topcon::Login: Entering...\n");
	int k = 0;
	bool _login = FALSE;
	if (direct) {
		// Check for login prompt
		const char *pattern[4] = {
			"login:",
			"Password:",
			"incorrect",
			"/dev/"
		};

		while (!_login && (k < 3)) {
			int answer = pSerial->WaitFor(5000, pattern, 4);
			pSerial->ClearBuffers();
			switch (answer) {
				case 0:
					debug(5,"Topcon::Login: Sending Username (%s)\n",login.c_str());
					pSerial->Write(login.c_str());
					pSerial->Write("\n");
					k++;
					break;
				case 1:
					debug(5,"Topcon::Login: Sending Password (%s)\n",password.c_str());
					pSerial->Write(password.c_str());
					pSerial->Write("\n");
					break;
				case 2:
					debug(5,"Topcon::Login: incorrect username or password\n");
					k++;
					sleep(1);
					break;
				case 3:
					debug(5,"Topcon::Login: Found Port Identifier, must be logged in\n");
					_login = TRUE;
					char a[300];
					pSerial->Read((void *)a,300,1000);
					break;
				default:
					debug(5,"Topcon::Login: No prompt found, force print of port\n");
					pSerial->Write("print,/cur/term\n");
					k++;
					break;
			}
			sleep(1);
		}
	} else {
		_login = TRUE; 
		debug(5,"Topcon::Login: Not direct, force print of port\n");
		pSerial->Write("\n\n");
		char a[300];
		pSerial->Read((void *)a,300,1000);
		pSerial->Write("print,/cur/term\n");
		pSerial->Read((void *)a,300,1000);
	}


	debug(D_FLOW,"Topcon::Login: Exiting...\n");
	if (!_login) return(FAILURE);
	return(SUCCESS);
}


//  ===================================================================
//  NAME:               Topcon::WriteFiles
//
//  PURPOSE:    Write the image file to disk. This may not include
//                              all of the information the file needs. It will only
//                              write the contents of imageFile.
//
//  RETURNS:    SUCCESS on writing the file successfully, FAILURE
//                              if write failed.
//
//  HISTORY:
//
//  ===================================================================
int Topcon::WriteFiles() {

        return(WriteFile(NULL, (char *) GetFileName(myIndex).c_str(), imageFile, cimageFile));

}

//  ===================================================================
//  NAME:		Topcon::DownloadFileAtIndex
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Topcon::DownloadFileAtIndex(Serial *pSerial, int index) {

	debug(D_FLOW, "Topcon::DownloadFileAtIndex: Entering...\n");

	debug(D_VALUES, "Topcon::DownloadFileAtIndex: indexFile %d\n", index);

	char buffer[TPS_BLOCKSIZE * 2];
	STPSBlock curblock;
	vector <unsigned char> datafile;

	bool err_found = FALSE;
	int rec;
	int downLoadOK = FALSE;
	unsigned int fileSize = ReceiverFAT[index].fileSize;

	int bs = TPS_BLOCKSIZE;
	U_2_INT bn = 0;
	Crc16 crc = 0;
	Crc16 t_crc;
	bool finished = FALSE;

	memset(&reportstruct, 0, sizeof(reportstruct));
	reportstruct.filename = ReceiverFAT[index].fileName.c_str();
	reportstruct.filesize = fileSize;
	reportstruct.blocks_downloaded = 0;
	reportstruct.bytes_downloaded = 0;
	reportstruct.start_time = time(NULL);
	reportstruct.elapsed_time = 0;
	reportstruct.errors = 0;
	TransferReportStart(&reportstruct);

	if (fileSize >= datafile.max_size()) {
		debug(D_FLOW, "Topcon::DownloadFileAtIndex: File is too large(%d) to download, use FTP?\n",TPS_MAXVECTORS * datafile.max_size());
		return(FAILURE);
	}
	string filename = ReceiverFAT[index].fatName;
	// Read any outstanding characters on interface
	pSerial->Write("\n");
	pSerial->ClearBuffers();
	rec = pSerial->Read((void *)buffer,512,1000);
	
	sprintf(buffer,"get,/log/%s:{100,%d}\n",filename.c_str(),TPS_BLOCKSIZE);

	debug(D_FLOW, "Topcon::DownloadFileAtIndex: requested file %s with command %s\n",filename.c_str(),buffer);
	GRILCommand(pSerial, (const char *) buffer);
//	pSerial->Write(buffer);
	sleep(3);
	pSerial->PutChar((char)TPS_NACK);
	sleep(2);
	pSerial->set_raw(1); // Required for NULL characters in files that are not escaped
	try { // allow for changing raw mode
	while (!finished && (reportstruct.errors < 20)) {
		err_found = false;

		try {
			rec = pSerial->Read((void *)&curblock.type,1,5000);
			debug(D_TXFR,"Topcon::DownloadFileAtIndex:  Read %d of 1 characters(bn=%d)\n",rec,bn);
			if (rec < 1) throw "Topcon::DownloadFileAtIndex: No block type read, NAKing and trying again\n";
			switch (curblock.type) {
				case TPS_ABORT: 
					debug(D_WARN,"Topcon::DownloadFileAtIndex: Aborting transfer, found abort\n");
					throw FAILURE;
					break;
			
				case TPS_LAST: 
					rec = pSerial->Read((void *)&curblock.count,2,5000);
					debug(D_TXFR,"Topcon::DownloadFileAtIndex: TPS_LAST Read %d characters(bn=%d)\n",rec,bn);
					if (curblock.count == -1) {
						debug(D_WARN,"Topcon::DownloadFileAtIndex: Aborting transfer, curblock.count = %d\n",curblock.count);
						throw FAILURE;
					}
					bs = curblock.count;
					bn++;
					finished = TRUE;
					break;

				case TPS_ORDINAL: 
					rec = pSerial->Read((void *)&curblock.number,2,5000);
					debug(D_TXFR,"Topcon::DownloadFileAtIndex: Read %d characters(bn=%d)\n",rec,bn);
					debug(D_TXFR,"Topcon::DownloadFileAtIndex: Block Number %d == %d\n",curblock.number, bn);
					if (curblock.number != bn) throw "Topcon::DownloadFileAtIndex: Block Number mismatch, NACKing";
					bs = TPS_BLOCKSIZE;
					bn = curblock.number + 1;
					break;
				default:
					debug(D_ERR,"Topcon::DownloadFileAtIndex: curblock.type is unknown(%d)\n",curblock.type);
					throw "Topcon::DownloadFileAtIndex: Unknown Block Type, NACKing";
					break;
			}
		}
		catch(const char *pmessage)
		{
			err_found = TRUE;
			debug(D_WARN,"Topcon::DownloadFileAtIndex: error #%d: %s\n",reportstruct.errors, pmessage);
			rec = pSerial->Read((void *)buffer,TPS_BLOCKSIZE + 30,5000);
			pSerial->PutChar((char)TPS_NACK);
			sleep(2);
		}
		catch(int myc)
		{
			throw myc;
		}
		if (err_found) {
			reportstruct.errors++;
			continue;
		}
	
		rec = pSerial->Read((void *)curblock.data,TPS_BLOCKSIZE,10000);
		debug(D_TXFR,"Topcon::DownloadFileAtIndex: Read %d characters\n",rec);

		rec = pSerial->Read((void *)&curblock.cs,2,5000);

		t_crc = crc;
		crc = crc16(crc, (void *) &curblock.type, 1);
		crc = crc16(crc, (void *) &curblock.count, 2);
		crc = crc16(crc, (void *) &curblock.data, TPS_BLOCKSIZE);

		if(crc != curblock.cs) {
			debug(D_ERR,"Topcon::DownloadFileAtIndex: CRC does not match, sending NAK (%d = %d)\n",crc,curblock.cs);
			pSerial->PutChar(TPS_NACK);
			crc = t_crc;
			bn--;
			continue;
		}

		rec = pSerial->Read((void *)&curblock.eob,1,5000);
		if (curblock.eob != TPS_EOB) {
			debug(D_WARN,"Topcon::DownloadFileAtIndex: Aborting transfer, EOB(%x) does not match TPS_EOB\n",curblock.eob);
			throw FAILURE;
		}		
	
		for (int i = 0; i < bs; i++) {
			datafile.push_back(curblock.data[i]);
		}
		pSerial->PutChar(TPS_ACK);

		if (finished) downLoadOK = TRUE;

		reportstruct.bytes_downloaded += bs;
		reportstruct.elapsed_time = time(NULL) - reportstruct.start_time;
		TransferReport(&reportstruct);
	}
	}
	catch (int myc)
	{
		pSerial->set_raw(0);
		return(myc);
	}
		
	debug(D_FLOW, "Topcon::DownloadFileAtIndex: \n");

	/* check received file status */
	if(downLoadOK) {
		setImageFile(&datafile);
		char buffer[100];
		myIndex = index;
		addToSumFile(" Start Time: %s\n",ReceiverFAT[index].fileStart.chartime("%%G %%g", buffer, 100));
		addToSumFile(" End Time: %s\n",ReceiverFAT[index].fileEnd.chartime("%%G %%g", buffer, 100));
		addToSumFile(" File: %s\n",GetFileName(myIndex).c_str());
		addToSumFile(" Bytes: %d\n",fileSize);
		addToSumFile(" Time: %d seconds\n",reportstruct.elapsed_time);
	} else {
		debug(D_ERR, "Topcon::DownloadFileAtIndex: Download Failed\n");
		debug(D_ERR, "Topcon::DownloadFileAtIndex: Clearing Transfer\n");
		pSerial->ClearBuffers(); sleep(1); pSerial->ClearBuffers();
	}
	addToSumFile("\n");


	return((downLoadOK) ? SUCCESS : FAILURE);
}
//  ===================================================================
//  NAME:
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
void Topcon::AddImage(Topcon *newTopcon) {

        debug(50,"Topcon::AddImage(Topcon): Entering...\n");

	vector <unsigned char> datafile;
	for (int i = 0; i < cimageFile; i++) {
		datafile.push_back(imageFile[i]);
	}
	int fs = newTopcon->getImageFileSize();
	unsigned char *tmp = newTopcon->getImageFile();
	for (int i = 0; i < fs; i++) {
		datafile.push_back(tmp[i]);
	}
	setImageFile(&datafile);
	debug(25,"Topcon::Addimage(Topcon): fileSize is %d\n",cimageFile);

	debug(50,"Topcon::AddImage(char, int, FAT): Exiting...\n");

}


//  ===================================================================
//  NAME:		Topcon::DownloadFileByTime
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Topcon::DownloadFileByTime(Serial *pSerial, gpstime *start, gpstime *stop) {

	int numindex, indexes[MAX_TOPCON_FILES];
	int status = FAILURE;
	char buffer[100];

	debug(D_FLOW,"Topcon::DownloadFileByTime: Entering...\n");

	numindex = FindIndexesByTime(start, stop, indexes, FALSE);

	if (numindex > 0) {
		debug(D_STD_V2,"Topcon::DownloadFileByTime: Downloading file at index %d\n", indexes[0]);
		status = DownloadFileAtIndex(pSerial, indexes[0]);
	} else {
		debug(25,"Topcon::DownloadFileByTime: No Files found for time range\n");
		status = SUCCESS_NOFILE;
        }

	if ((status == SUCCESS) && (numindex > 1)) {
		for (int i = 1; i < numindex; i++) {
			debug(25,"Topcon::DownloadFileByTime: Downloading file at index %d\n", indexes[i]);
			Topcon *pTmp = new Topcon;
			status = pTmp->DownloadFileAtIndex(pSerial, indexes[i]);
			if (status == FAILURE) {
				delete(pTmp);
				break;
			}
			AddImage(pTmp);
			delete(pTmp);
			addToSumFile(" Span: ");
			addToSumFile(start->chartime("%%G/%%g-",buffer,100));
			addToSumFile(stop->chartime("%%G/%%g\n",buffer,100));
		}
	}
	debug(D_FLOW,"Topcon::DownloadFileByTime: Exiting\n");
	return(status);
}

//  ===================================================================
//  NAME:		Topcon::DeleteFileAtIndex
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Topcon::DeleteFileAtIndex(Serial *pSerial, int index) {
	string check;
	int status;
	char buffer[100];

	debug(D_FLOW,"Topcon::DeleteFileAtIndex: Entering...\n");
	string filename = ReceiverFAT[index].fatName;

	
	sprintf(buffer,"remove,/log/%s",filename.c_str());
	debug(D_RECINFO,"Topcon::DeleteFileAtIndex:Command: %s\n",buffer);
	debug(D_STD,"\n\nDeleting file %s",filename.c_str());
	debug(D_STD," -- ");
	check = GRILCommand(pSerial, buffer);
	status = check.compare("OK");

	if (status == 0) {
		debug(D_STD," Deleted\n");
	} else {
		debug(D_STD," NOT Deleted\n\n");
	}
	debug(D_FLOW,"Topcon::DeleteFileAtIndex: Exiting\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Topcon::DeleteFileByTime
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Topcon::DeleteFileByTime(Serial *pSerial, gpstime *start, gpstime *stop) {
	int i,j;
	int numindex, indexes[MAX_RCVR_FILES];
	int deleteOK = FALSE;

	debug(D_FLOW,"Topcon::DeleteFileByTime: Entering...\n");

	numindex = FindIndexesByTime(start, stop, indexes, TRUE);

	for (j = (MAX_RCVR_FILES-1); j >= 0; j--) {
		for (i = 0; i < numindex; i++) {
			if (indexes[i] != j) continue;
			debug(D_STD,"Topcon::DeleteFileByTime: Deleteing file at index %d\n", indexes[i]);
			deleteOK = DeleteFileAtIndex(pSerial, indexes[i]);
		}
	}
	debug(D_FLOW,"Topcon::DeleteFilesByTime: Exiting\n");
	return(deleteOK);
}

//  ===================================================================
//  NAME:		Topcon::DownloadReceiverID
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
string Topcon::DownloadReceiverID(Serial *pSerial) {

	debug(D_FLOW, "Topcon::DownloadReceiverID: Entering...\n");

	pSerial->ClearTXBuffer();
	pSerial->ClearRXBuffer();

	// Seems to work OK.
	RxID = GRILCommand(pSerial, (const char *) "print,/par/rcv/ver");

	debug(D_FLOW, "Topcon::DownloadReceiverID: Exiting\n");
	return RxID;
}

//  ===================================================================
//  NAME:		Topcon::SetReceiverSession
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Topcon::SetReceiverSession(Serial *pSerial) {
	return(SUCCESS);
}


//  ===================================================================
//  NAME:		Topcon::DownloadReceiverFAT
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int Topcon::DownloadReceiverFAT(Serial *pSerial) {

	debug(D_FLOW, "Topcon::DownloadReceiverFAT: Entering...\n");

	string filenames[MAX_TOPCON_FILES];
	string tmp = GRILCommand(pSerial,"list");
	int numfiles = stringtokenize(tmp,",",filenames,MAX_TOPCON_FILES);
	list <SFATEntry> fat_list;
	int maxerror = 5;
	char buffer[1025];

	int i = 0;
	for(i = 0; i < numfiles; i++ ) {
		SFATEntry rFAT;
		string file = filenames[i];
		string info[2];
		string command = "print,/log/"; command.append(file);
		int j = stringtokenize(GRILCommand(pSerial,command.c_str()), ",", info, 2);
		if (j != 2) {
			debug(D_WARN,"Unable to get info about file %s\n",file.c_str());
			--maxerror; --i;
			if (maxerror <= 0) return(FAILURE);
			pSerial->Read(buffer, 1024, 1000);
			pSerial->ClearBuffers();
			continue;
		}

		int filesize = atoi(info[0].c_str());
		int year,month,day,hour,min,sec;
		j = sscanf(info[1].c_str(),"%dY%dM%dD%dh%dm%ds",&year,&month,&day,&hour,&min,&sec);
		if(filesize <= 0) continue;

		/* file length */
		debug(D_RECINFO, "Topcon::DownloadReceiverFAT: file length %d\n", filesize);
		rFAT.fileSize = filesize;

		/* filename in FAT */
		debug(D_RECINFO, "Topcon::DownloadReceiverFAT: file name %s\n", file.c_str());
		rFAT.fatName = file.c_str();

		debug(D_FLOW, "Topcon::DownloadReceiverFAT: Setting FAT start and stop times\n");
		rFAT.fileEnd.set_time(year, month, day, hour, min, (double)sec);
// File Formats:
// [N Chars]\d\d\d\d[a-z]
// [N Chars]\d\d\d\d[a-z]\d\d\d
		size_t found = file.find_last_of("0123456789");
		string check;
		if (found == (file.length()-1)) {
			rFAT.siteName = file.substr(0,file.length()-8);
			check = file.substr(file.length() - 8);
		} else {
			rFAT.siteName = file.substr(0,file.length()-5);
			check = file.substr(file.length() - 5);
		}
		debug(D_VALUES,"Topcon::DownloadReceiverFAT: Checking against %s\n",check.c_str());
		month = atoi(check.substr(0,2).c_str());
		day   = atoi(check.substr(2,2).c_str());
		hour  = (int)check[4] - (int)'a';

		rFAT.session = check[4];
			
		/* fill in receiver fat day month year (deal with 2 digit years) */
// This was a mistake. Year should be the same as the end time.
/*
		gpstime curtime;
		curtime.set_time(); // set time to now
		double tempday;
		curtime.getjmt(&year, &tempday);
		year += 1900;
*/

		rFAT.fileStart.set_time(year, month, day, hour, 0,0.0);
		fat_list.push_back(rFAT);
	}

	fat_list.sort(fat_sort);
	list<SFATEntry>::iterator it;
	int fileNum = 0;
	for ( it=fat_list.begin() ; it != fat_list.end(); it++ ) {
		ReceiverFAT[fileNum].siteName = it->siteName;
		ReceiverFAT[fileNum].fileSize = it->fileSize;
		ReceiverFAT[fileNum].fileStart = it->fileStart;
		ReceiverFAT[fileNum].fileEnd = it->fileEnd;
		ReceiverFAT[fileNum].fileName = it->fileName;
		ReceiverFAT[fileNum].fatName = it->fatName;
		ReceiverFAT[fileNum].session = it->session;
		ReceiverFAT[fileNum].active = it->active;

		numFATEntries++;
		CreateFileName(fileNum);
		fileNum++;
	}
	debug(D_FLOW, "Topcon::DownloadReceiverFAT: Exiting...\n");

	return SUCCESS;
}

//  ===================================================================
//  NAME:		Topcon::GRILCommand
//
//  PURPOSE:	Send a GRIL command to the Receiver and return the
//		output
//
//  RETURNS:	string with checksum removed
//		If a set command, it will 
//		
//
//  HISTORY:
//
//  ===================================================================
string Topcon::GRILCommand(Serial *pSerial, const char *command) {

	pSerial->ClearTXBuffer();
	pSerial->ClearRXBuffer();
	debug(D_FLOW, "Topcon::GRILCommand: Entering...\n");

	string outcom = "%SHARC%";
	outcom.append(command);
	outcom.append("\n");
	debug(D_FLOW, "Topcon::GrilCommand: Sending %s\n",outcom.c_str());
	pSerial->Write(outcom.c_str());
//	pSerial->Write("%SHARC%");
//pSerial->Write(command);
//pSerial->Write("\n");

	int line_number = 0;
	bool finished = FALSE;

	string rval = "";

	while (! finished) {
		char buffer[500];
		unsigned int length = 0;
		char delim;

		pSerial->Readline(buffer, 500, 3000);
		pSerial->Inkey(100); // get linefeed
		debug(D_FLOW,"Topcon::GRILCommand: Command %s returned (%s)\n",outcom.c_str(), buffer);
		string ret = buffer;
		if (ret.length() == 0) return "OK";

		ret.erase(ret.length()-1,1);

		if (ret[0] == 'E') {
			debug(D_ERR,"Topcon::GRILCommand: Command Error\n");
			return "ERR";
		}

		// For commands with no return code, this string signifies success
		if (ret.compare("RE007%SHARC%") == 0) {
			debug(D_FLOW,"Topcon::GRILCommand: Found Success response with no string\n");
			return "OK";
		}

		sscanf(ret.c_str(), "RE%x%%SHARC%%%c",&length, &delim);

		if (length != (ret.length() - 5)) {
			debug(D_ERR, "Topcon::GRILCommand: Length mismatch %d vs %d\n", length, ret.length()-5);
			return "ERR";
		}

		// Conditions for finished:
		// first line && delim = ' '
		// not first line and last char = '}'
		if ((line_number == 0) && (delim == ' ')) {
			finished = TRUE;
		}
		if ((line_number == 0) && (delim == '{') && (ret[ret.length()-1] == '}')) {
			ret.erase(ret.length()-1,1); // remove final bracket
			finished = TRUE;
		}
		if ((line_number == 0) && ((delim != ' ') && (delim != '{'))) {
			ret.insert(0," ");
			ret[0] = delim;
			finished = TRUE;
		}
		if ((line_number > 0) && (ret[ret.length()-1] == '}')) {
			ret.erase(ret.length()-1,1); // remove final bracket
			finished = TRUE;
		}
		rval.append(ret.substr(13,ret.length()-1));
		line_number++;
	}
	debug(D_FLOW,"Topcon:GRILCommand: Exiting with %s\n",rval.c_str());
	return rval;
}

//  ===================================================================
//  NAME:               Topcon::SetImageFile
//
//  PURPOSE:    set the imagefile contents and size.
//
//  RETURNS:    SUCCESS if image correctly copied. FAILURE if
//                              memory cannot be allocated.
//
//  HISTORY:
//
//  ===================================================================
int Topcon::setImageFile(vector <unsigned char> *datafile) {

	debug(D_FLOW,"Topcon::setImageFile: Entering...\n");

	if (imageFile != NULL)
		delete(imageFile);

	imageFile = new unsigned char[datafile->size()];
	if (imageFile == NULL) {
		debug(D_FATALERROR,"Unable to allocate memory for imageFile\n");
		return(FAILURE);
	}

	for (unsigned int i = 0; i < datafile->size(); i++) {
		imageFile[i] = datafile->at(i);
	}
	cimageFile = datafile->size();
	debug(D_FLOW,"Topcon::setImageFile: Exiting\n");
	return(SUCCESS);

}

//  ===================================================================
//  NAME:               Topcon::SetImageFile
//
//  PURPOSE:    set the imagefile contents and size.
//
//  RETURNS:    SUCCESS if image correctly copied. FAILURE if
//                              memory cannot be allocated.
//
//  HISTORY:
//
//  ===================================================================

int Topcon::TerminalSession(Serial *pSerial) {

	debug(D_FLOW,"Topcon::TerminalSession: Entering...\n");
	bool quit = false;
	char prompt[100];
	sprintf(prompt,"Terminal-%s> ",GetSiteName(0).c_str());
	FILE *in = fdopen(1,"r");
	string quit_u = "QUIT";
	string quit_l = "quit";
	do {
		string myin = "";
		cout << prompt;
		char input[1024];
		fgets(input,1024,in);
		myin = input;
		myin.substr(0,myin.length()-1);
		printf("Input = %s\n",myin.c_str());
		if (myin.compare(0,4,quit_u) == 0) quit=TRUE;
		if (myin.compare(0,4,quit_l) == 0) quit=TRUE;
		if (!quit) {
			string output = GRILCommand(pSerial,myin.c_str());
			printf("%s\n",output.c_str());
		}
	} while (!quit);
	fclose(in);

	debug(D_FLOW,"Topcon::TerminalSession: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		Topcon::Stream
//
//  PURPOSE:	Set the ashtech to start streaming
//
//  RETURNS:	SUCCESS/FAILURE
//
//  HISTORY:
//
//  ===================================================================
int Topcon::Stream(Serial *pSerial, int format, double rate, char port) {
#ifdef USE_STREAM
	debug(D_FLOW,"Topcon::Stream: Entering...\n");
	if (format == DATATYPE_BINEX) {
		return (BinexStream(pSerial, rate, port));
	}
#endif
	return(FAILURE);

}

int Topcon::BinexStream(Serial *pSerial, double rate, char port) {

	debug(D_FLOW,"Topcon::BNXStream: Entering...\n");

	string lp = "/cur/term";
	switch(port) {
		case 'a': lp = "/dev/ser/a"; break;
		case 'b': lp = "/dev/ser/b"; break;
		case 'c': lp = "/dev/ser/c"; break;
		case 'd': lp = "/dev/ser/d"; break;
		case 'A': lp = "/dev/tcp/a"; break;
		case 'B': lp = "/dev/tcp/b"; break;
		case 'C': lp = "/dev/tcp/c"; break;
		case 'D': lp = "/dev/tcp/d"; break;
		case 'E': lp = "/dev/tcp/e"; break;
	}
	string command;
	string rv;

	if (rate <= 0) {
		for (int i = 0; i < 10; i++) {
			pSerial->ClearBuffers();
			command = "dm," + lp;
			rv = GRILCommand(pSerial,command.c_str());
		}
		if (rv.empty()) return(FAILURE);
		return(SUCCESS);
	}


// Turn on Binex records
	char buffer[100];
	sprintf(buffer,"em,%s,/msg/binex/7F_03:%.1f",lp.c_str(),rate);
	rv = GRILCommand(pSerial,buffer);
	debug(D_STD,"Topcon::BinexStream: Receiver has been set to stream\n");

	debug(D_FLOW,"Topcon::BinexStream: Exiting...\n");
	// if (! rv.empty()) return(FAILURE);
	return(SUCCESS);

}

