/* 
#    $Id: gpstime.cpp,v 1.1 2011/04/04 17:29:34 gps Exp $
#
#     gpstime -- Source code for the time class.  This class handles
#                all possible time conversions.  Stores the date in
#                the standard ctime manner and converts from there.
#                However, it will handle partial seconds which the 
#                regular time is unable to do.  
#    
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define DIFF_GPS_DAYS 3657 
#define DIFF_MJD_DAYS -40587   
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sys/types.h>
#include <string>
#include <stdio.h>
#include <gpstime.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>

#include "gpscommon.h"

double rnd(double in, int place);

//  ===================================================================
//  NAME:		gpstime::gpstime
//
//  PURPOSE:	Constructor. Set values to 0.
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
gpstime::gpstime() 
{  
	partial = 0.0;
	the_time = isset = 0; 
}  

//  ===================================================================
//  NAME:		gpstime::set_time()
//
//  PURPOSE:	Set time to current time. return value of isset.
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time()
{  
	time(&the_time);
	partial = 0.0;
	isset = 1;
	return(isset);
}  

//  ===================================================================
//  NAME:		gpstime::set_time -- double
//
//  PURPOSE:	Set time to supplied modified Julian Day
//
//  RETURNS:	TRUE if set, FALSE if not set
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time(double inMJD)
{  
	double temp;
	if (inMJD < 0)
	{
		isset = 0;
		return(isset);
	}
	temp = (inMJD + DIFF_MJD_DAYS) * 86400;  
	the_time = (int) temp;
	partial = fmod(temp,1.0);
	chkpartial();
	isset = 1; 
	return(isset); 
}  

//  ===================================================================
//  NAME:		gps_time::set_time -- time_t
//
//  PURPOSE:	Set time to number of seconds since 1970
//
//  RETURNS:	TRUE
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time(time_t intime_t) 
{  
	the_time = intime_t;
	partial = 0.0;
	isset = 1; 
	return(isset); 
}  


//  ===================================================================
//  NAME:		gpstime::set_time -- int, int, double
//
//  PURPOSE:	set time to year,month,day supplied
//
//  RETURNS:	TRUE
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time(int inyear,int inmonth,double inday) 
{
	the_time = find_thetime(inyear,inmonth,(int) inday);
	partial = (inday - (int) inday) * 86400.0;
	chkpartial();
	isset = 1;
	return(isset);
}


//  ===================================================================
//  NAME:		gpstime::set_time -- int, double
//
//  PURPOSE:	Set time to GPSweek and GPSSecond supplied
//
//  RETURNS:	TRUE
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time(int gpsweek,double gpssec)   
{  
	partial = fmod(gpssec,1.0);
	the_time = (gpsweek * 7 * 86400) + (int) gpssec;
	the_time += DIFF_GPS_DAYS * 86400; 
	chkpartial();
	isset = 1; 
	return(isset); 
}  


//  ===================================================================
//  NAME:		gpstime::set_time -- int, double, char
//
//  PURPOSE:	Set time to year, day(julian) supplied. c seems silly
//
//  RETURNS:	value of isset
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time(int inyear,double inday,char c)
{
	partial = (inday - (int) inday) * 86400.0;
	the_time = find_thetime(inyear,0,(int) inday);
	chkpartial();
	if (c != 'y') isset = 1;
	return(isset);
}
 
//  ===================================================================
//  NAME:		gpstime::set_time -- int, int, double, char
//
//  PURPOSE:	Set time to hour, minute, second of current day. c seems silly
//
//  RETURNS:	value of isset
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time(int hour, int minute, double second, char c)
{
	int myyear,intday;
	double myday;

	c = 'y';
	this->set_time();
	this->getjmt(&myyear,&myday); intday = (int) myday;
	this->set_time(myyear,(double) intday,c);
	the_time = the_time + (hour * 3600) + (minute * 60) + (int) second;
	partial = fmod(second,1.0);
	chkpartial();
	isset = 1;
	return(isset);
}  

//  ===================================================================
//  NAME:		gpstime::set_time -- int, char, int, int, int, double
//
//  PURPOSE:	Set time to year, month(char), day, hour, minute, second.
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time(int year, char *month, int day, int hour, int minute, double second)
{  
	char inmonth[4];
	int i;

	strncpy(inmonth,month,3);
	for (i = 0;inmonth[i] != '\0';i++)
		inmonth[i] = toupper(inmonth[i]);
	if (strncmp(inmonth,"JAN",3) == 0) i = 1;
	if (strncmp(inmonth,"FEB",3) == 0) i = 2;
	if (strncmp(inmonth,"MAR",3) == 0) i = 3;
	if (strncmp(inmonth,"APR",3) == 0) i = 4;
	if (strncmp(inmonth,"MAY",3) == 0) i = 5;
	if (strncmp(inmonth,"JUN",3) == 0) i = 6;
	if (strncmp(inmonth,"JUL",3) == 0) i = 7;
	if (strncmp(inmonth,"AUG",3) == 0) i = 8;
	if (strncmp(inmonth,"SEP",3) == 0) i = 9;
	if (strncmp(inmonth,"OCT",3) == 0) i = 10;
	if (strncmp(inmonth,"NOV",3) == 0) i = 11;
	if (strncmp(inmonth,"DEC",3) == 0) i = 12;

	return(this->set_time(year, i, day, hour, minute, second));
}  

//  ===================================================================
//  NAME:		gpstime::set_time -- int, int, int, int, int, double
//
//  PURPOSE:	Set time to year, month(int), day, hour, minute, second
//
//  RETURNS:	TRUE
//
//  HISTORY:
//
//  ===================================================================
int gpstime::set_time(int year, int month, int day, int hour, int minute, double second)
{  
	this->set_time(year,month,(double)day);
	the_time = the_time + (hour * 3600) + (minute * 60) + (int) second;
	partial = fmod(second,1.0);
	chkpartial();
	isset = 1;
	return(isset);
}

//  ===================================================================
//  NAME:		gpstime::chkpartial
//
//  PURPOSE:	if partial is more than 1, add seconds to the_time and remove
//				them from partial.
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
void gpstime::chkpartial()
{
	partial = rnd(partial,-5);
	if (((int) partial >= 1.0) || ((int) partial <= -1.0))
	{
		the_time += (int) partial;
		partial -= (int) partial;
	}

	if (partial < 0.0)
	{
		the_time--;
		partial = partial + 1;
	}
}

//  ===================================================================
//  NAME:		gpstime::getMJD()
//
//  PURPOSE:	created MJD from the_time and chkpartial
//
//  RETURNS:	return Modified Julian Day
//
//  HISTORY:
//
//  ===================================================================
double gpstime::getMJD()   
{
	double temp;
	temp = (the_time / 86400) - DIFF_MJD_DAYS;
	temp += ((double)(the_time % 86400) / 86400.0) + (partial / 86400.0);
	return(temp);
}

//  ===================================================================
//  NAME:		gpstime::getjmt()
//
//  PURPOSE:	return year and day of year
//
//  RETURNS:	isset
//
//  HISTORY:
//
//  ===================================================================
int gpstime::getjmt(int *outyear,double *outday)
{
	struct tm *curtime;
	curtime = gmtime(&the_time);
	*outyear = curtime->tm_year; 
	*outday =  curtime->tm_yday+1;
	*outday += ((the_time % 86400) + partial) / 86400.0;
	return(isset);
}


//  ===================================================================
//  NAME:		gpstime::getgps()
//
//  PURPOSE:	created gps week/second
//
//  RETURNS:	return isset
//
//  HISTORY:
//
//  ===================================================================
int gpstime::getgps(int *outweek,double *outsec)
{
	time_t temp;
	temp = the_time - DIFF_GPS_DAYS*86400;
	*outweek = (int) (temp/86400)/7;
	*outsec = temp - (*outweek * 86400.0 * 7.0);
	*outsec += partial;
	return(isset);
}
 
 
 
//  ===================================================================
//  NAME:		gpstime::find_thetime()
//
//  PURPOSE:	create time from year/month/day
//
//  RETURNS:	return number of seconds since 1970
//
//  HISTORY:
//
//  ===================================================================
time_t gpstime::find_thetime(int inyear,int outmonth,int inday)
{
	int leapyr,i,numdays=0;
	static int daytab[2][13] = {
		{0,31,28,31,30,31,30,31,31,30,31,30,31},
		{0,31,29,31,30,31,30,31,31,30,31,30,31}
			}; 
 
 
//make sure year is with century
	if (inyear < 70)
		inyear += 2000;
	else if (inyear < 100)
		inyear += 1900;

	// inday should be one less because we do not want to
	// count the current day
	inday--;
 
	for (i=1970;i<inyear;i++)
	{
		leapyr = ((i%4 == 0) && (i%100 != 0)) || (i%400 == 0);
		numdays += (leapyr) ? 366 : 365;
	}

	leapyr = ((inyear%4 == 0) && (inyear%100 != 0)) || (inyear%400 == 0);

	for (i = 0;i < outmonth; i++)
		inday += daytab[leapyr][i];

	return((numdays + inday) * 86400);
}

//  ===================================================================
//  NAME:		gpstime::chartime()
//
//  PURPOSE:	basically strftime with a few added data types
//
//  RETURNS:	pointer to character string
//
//  HISTORY:
//
//  ===================================================================
char* gpstime::chartime(const char format[],char buffer[],int smax)
{

	struct tm *tp;
	int newmax;
	size_t start;
	string replace, find;

	char seconds[7];
	char cpartial[9];
	char daysecs[6];
	char week[5];

	tp = gmtime(&the_time);
	newmax = strftime(buffer,smax,format,tp);

	if (newmax == smax) return(buffer);
	string tmpbuf = buffer;

	int tmp1;
	double tmp2;
	getgps(&tmp1,&tmp2);
	int tmp3 = (int)tmp2 % 86400;
	sprintf(week, "%04d", tmp1);
	sprintf(seconds, "%06d", (int) tmp2);
	sprintf(daysecs, "%05d", tmp3);
	sprintf(cpartial,".%07d",(int) partial*(10^7));

	REPLACE(tmpbuf,"%%","%");
	REPLACE(tmpbuf,"++","+");

	REPLACE(tmpbuf,"%P",cpartial);
	REPLACE(tmpbuf,"+P",cpartial);

	REPLACE(tmpbuf,"%G",week);
	REPLACE(tmpbuf,"+G",week);

	REPLACE(tmpbuf,"%g",seconds);
	REPLACE(tmpbuf,"+g",seconds);

	REPLACE(tmpbuf,"+J",daysecs);

	strncpy(buffer,tmpbuf.c_str(),smax);
	return(buffer);
}

