// Copyright (C) 1999 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of APE.
// 
// The exception is that, if you link the APE library with other files
// to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the APE library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name APE.  If you copy code from other releases into a copy of
// APE, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for APE, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#ifndef	__APE_SERIAL_H__
#define	__APE_SERIAL_H__

#include <zglobal.h>

void interrupt_transfer(int sig);

void debug(int level, const char *fmt, ...);

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <string>
#include <deque>
#include <termios.h>
using namespace std;

#define TIME_MULT 1
#define ERRORMAX 50
#define timeout_t time_t
//typedef	unsigned long timeout_t;
#define	MAX_SERIAL_TIMEOUT	((timeout_t)10000)


enum
{
	SERIAL_FLOW_NONE,
	SERIAL_FLOW_HARD,
	SERIAL_FLOW_SOFT,
	SERIAL_FLOW_BOTH,
	SERIAL_FLOW_XMODEM

};

enum
{
	SERIAL_FORMAT_7E1,
	SERIAL_FORMAT_7O1,
	SERIAL_FORMAT_7N1,
	SERIAL_FORMAT_7S2,
	SERIAL_FORMAT_8E1,
	SERIAL_FORMAT_8O1,
	SERIAL_FORMAT_8N1,
	SERIAL_FORMAT_8S2
};

typedef struct
{
// set by download routine
	const char *filename;
	int filesize;
	int bytes_skipped;
	size_t blocks_downloaded;
	size_t bytes_downloaded;
	time_t start_time;
	time_t elapsed_time; 
	int errors;
// potentially set as side effect of report program
	int bps;
	int fail;
	int percent;
	int estimated_time;
	int minbps_elapsed;
// values set externally to signify failure conditions
	int minbps;
	int minbps_secs;
} ReportStruct;

class SerialException: public exception
{
public:
	virtual bool isRetryable() = 0;

};

class SerialExceptionRetryable: public SerialException
{
	virtual const char* what() const throw ()
	{
		return "Recoverable Serial Exception occurred\n";
	}
	virtual bool isRetryable() { return TRUE; }
};

class SerialExceptionFail: public SerialException
{
	virtual const char* what() const throw ()
	{
		return "Non-recoverable Serial Exception occurred\n";
	}
	virtual bool isRetryable() { return FALSE; }
};

/**
 * A portable APE class for access to generic serial devices.  The win32
 * version has not yet been written, but the Posix version uses termios
 * services.
 * 
 * @author David Sugar <dyfet@ostel.com>
 * @short Generic serial device access.
 */
class Serial
{
protected:
	SerialExceptionFail	SerialFail;
	SerialExceptionRetryable	SerialRetryable;
	int _fd;
	struct termios _restore;
	struct termios _current;
	struct termios _temp1;
	struct termios _temp2;
	char _lockname[65];
	char _fname[65];
	long _speed;
	int _format;
	int _flow;
	char *resumeDir;
	char *resumeFileName;
	char _nl;
	int _connected;
	ReportStruct reportstruct;
	void (*report)(ReportStruct *prs);
	void (*reportstart)(ReportStruct *prs);
	void (*reportstop)(ReportStruct *prs);

	deque<unsigned char> File;
	int Binary;
	int EOFSeen;
	int MaxErrors;
	int TimeMult;
	int ForceSave;
	int FailedXFers;

	struct {
		int CRCFlag;
		int FirstSector;
		long BytesLeft;
		int BlockLength;
		int Binary;
		int LastRX;
		int Only1EOF;
		int MaxTime;
		int Errors;
		deque<unsigned char> File;
	} _xmodem;

	struct {
/* Globals used by ZMODEM functions */
		unsigned int Rxtimeout; /* Tenths of seconds to wait for something */
		int Rxframeind;     /* ZBIN ZBIN32, or ZHEX type of frame received */
		int Rxtype;     /* Type of header received */
		char Rxhdr[4];      /* Received header */
		char Txhdr[4];      /* Transmitted header */
		long Txpos;     /* Transmitted file position */
		int Txfcs32;        /* TRUE means send binary frames with 32 bit FCS */
		int Crc32t;     /* Display flag indicating 32 bit CRC being sent */
		int Crc32;      /* Display flag indicating 32 bit CRC being received */
		int Znulls;     /* Number of nulls to send at beginning of ZDATA hdr */
		char Attn[ZATTNLEN+1];  /* Attention string rx sends to tx on err */

		char lastsent;   /* Last char we sent */
		int turbo_escape;
		int bytes_per_error;
		char zsendline_tab[256];

		int CRCFlag;
		int FirstSector;
		int EOFSeen;
		long BytesLeft;
		int BlockLength;
		int Binary;
		int LastRX;
		int Only1EOF;
		int MaxTime;
		int Errors;
		int protocol;
		zm_fileinfo zi[MAXFILES];
		deque<unsigned char> File[MAXFILES];
	} _zmodem;

	

public:
	static bool UserCanceled;

	/**
	 * Open a named serial device under a specified set of attributes.
	 * Attributes include the serial device speed (baud rate), the
	 * data format (SERIAL_FORMAT_7E1, SERIAL_FORMAT_7O1, SERIAL_FORMAT_7N1,
         * SERIAL_FORMAT_7S2, SERIAL_FORMAT_8E1, SERIAL_FORMAT_8O1,
	 * SERIAL_FORMAT_8N1, and SERIAL_FORMAT_8S2), as well as the kind of
	 * flow control (SERIAL_FLOW_NONE, SERIAL_FLOW_HARD, SERIAL_FLOW_SOFT and
	 * SERIAL_FLOW_BOTH) to use.  Additional data formats would be easy
	 * enough to add if so desired.
	 * 
	 * @param fname path name of logical serial device.
	 * @speed absolute speed in bits per second.
	 * @format serial data format to use.
	 * @flow serial device flow control method to use.
	 */
	Serial(const char *fname, long speed, int format, int flow);
	Serial();
	virtual void Init();
	virtual int Reconnect(int tries, time_t wait);
	virtual int Open(const char *fname, long speed, int format, int flow);
	virtual int Close();
	
	virtual bool set_speed(long speed, termios *t);
	virtual void set_flow(int flow, termios *t);
	virtual void set_raw(int yes);

	virtual void set_format(int flow, termios *t);
	virtual int get_fd() { return(_fd); }

	virtual void initReportStruct();
	virtual void setReports(void (*a)(ReportStruct *), void (*b)(ReportStruct *), void (*c)(ReportStruct *));
	/**
	 * Release the serial device and restore it to it's previous
	 * attributes.
	 */
	virtual ~Serial();

	/**
	 * Change the flow control mode in effect for the serial device.
	 * This allows post constructor changes.
	 * 
	 * @param flow serial device flow control method to use.
	 */
	virtual void setFlowControl(int flow);
	/**
	 * Change the serial port baud rate (speed in bps) for the serial
	 * device.
	 * 
	 * @param speed absolute speed in bits per second.
	 */
	virtual void setSpeed(long speed);
	virtual long getSpeed();
	/**
	 * Change the serial port data format in use for the serial device.
	 * 
	 * @param format serial data format to use.
	 */
	virtual void setFormat(int format);
	/**
	 * Wait up to a specified timeout for a single character to appear
	 * at the serial device, and return that character.
	 * 
	 * @return character read or -1 if timed out.
	 * @param timer timeout specified in milliseconds.
	 */
	virtual int  Inkey(timeout_t timer);
	/**
	 * Read a specified number of bytes from the serial device but
	 * wait no longer than a specified timeout for data to become
	 * available.  The Posix termios only supports timeout resolution
	 * within 1/10th of a second, hence, specifying a timeout with
	 * an accuracy less than 100 milliseconds is useless.
	 * 
	 * @return number of bytes actually read.
	 * @param buf pointer to store data from the device.
	 * @param len number of bytes to read.
	 * @param timer timeout specified in milliseconds.
	 */
	virtual int Read(void *buf, size_t max, timeout_t timer);
	virtual int isCharAvailable(timeout_t timeout);

	/**
	 * Read a specified number of bytes from the serial device
	 * but wait no longer than a specified timeout, and also stop
	 * if the newline character has been read.
	 * 
	 * @return number of bytes actually read.
	 * @param buf pointer to store data from the device.
	 * @param len number of bytes to read.
	 * @param timer timeout specified in milliseconds.
	 */
	virtual int Readline(char *buf, size_t max, timeout_t timer);
	virtual ssize_t _read(int d, void *buf, size_t nbytes, int flags);
	virtual int ToggleDTR();
	
	/**
	 * Specify the newline character code to use with ReadLine.
	 * 
	 * @param nl newline character code (default 13).
	 */
	inline void setNewline(char nl) {_nl = nl;};
	
	/**
	 * Am I connected?
	 */
	inline int isConnected() { return(_connected); }
	
	/**
	 * set Resume directory
	 */
	virtual void setResumeDir(const char *dir);
	virtual void setMinBPS(int bps, int secs);
	virtual int setResumeFileName(const char *x, int y, int z);
	virtual int writeResumeFile(deque<unsigned char> *File);
	virtual int readResumeFile();

	virtual int tryResume() { return(resumeDir != NULL); }
	inline void setForceSave(int _save) { ForceSave = _save; }

	/**
	 * check for number of failed transfers
	 */
	virtual int failedXFers() { return(FailedXFers); }
	
	/**
	 * set max error and timeout values
	 */
	inline void setTimeoutMultiplier(int _mult) { TimeMult = _mult; }
	inline void setMaxTransferErrors(int _max) { MaxErrors = _max; }
	virtual int WaitFor(int timeout, unsigned char pattern);

	virtual int WaitFor(int timeout, const char **patterns, int numpatterns);

	virtual void ClearRXBuffer();
	virtual void ClearTXBuffer();
	virtual void ClearBuffers();

	/**
	 * XMODEM subroutines
	 *
	 */
	/**
	 * Run Xmodem transfer to temporary file and return
	 * contents of file
	 */
	virtual unsigned char * xmodemreceive(long *fileSize, timeout_t timer, const char *filename);
	virtual int GetXModem(timeout_t timer);
	virtual int GetXModemSector(char *rxbuf);
	virtual int AddSectorToBuffer(char *sector, int n);
	virtual int CancelXModem();
	virtual int ReadXModemBuffer(int timeout);
	virtual void RestartSector(int istimeout);

	/**
	* Run ZModem transfers to malloced memory
	* return using Get* routines
	*/
	virtual unsigned char * zmodemreceive(long *fileSize, timeout_t timer);
	int wcreceive();
	int rzfiles();
	int tryz();
	int procheader(char *name, int index);
	int rzfile(int index);
//	int putsec();
	void zmputs(const char *s);
	void ackbibi();
	int noxrd7(void);
	int zgeth1(void);
	int zgethex(void);
	int zdlread(void);
	int zdlread2(int c);
	void zsendline(int c);
	void zsendline_s(const char *s, size_t count) ;
	void zsbhdr(int type, char *hdr);
	void zsbh32(char *hdr, int type);
	void zshhdr(int type, char *hdr);
	void zsdata(const char *buf, size_t length, int frameend);
	void zsda32(const char *buf, size_t length, int frameend);
	void count_blk(int size);
	void printout_blocksizes(void) ;
	int zrdata(char *buf, int length, size_t *bytes_received);
	int zrdat32(char *buf, int length, size_t *bytes_received);
	int zgethdr(char *hdr, int eflag, size_t *Rxpos);
	int zrbhdr(char *hdr);
	int zrbhdr32(char *hdr);
	int zrhhdr(char *hdr);
	void zputhex(int c, char *pos);
	void zsendline_init(void);
	void stohdr(size_t pos);
	long rclhdr(char *hdr);



	/**
	 * Write a block of data to the serial device.
	 * 
	 * @return number of bytes written on success, -1 on error.
	 * @param buf pointer to data to write.
	 * @param len number of bytes to write.
	 */
	inline virtual int Write(void *buf, size_t len)
		{ return write(_fd, (char *)buf, len);};
	friend inline int write(Serial &s, void *buf, size_t len)
		{ return write(s._fd, (char *)buf, len); };

	inline virtual int Write(unsigned char *buf, size_t len)
		{ return write(_fd, buf, len); };
	friend inline int Write(Serial &s, unsigned char *buf, size_t len)
		{return write(s._fd, buf, len);};
	/**
	 * Write a null terminated C string directly to the serial
	 * device.
	 * 
	 * @return number of bytes written on success, -1 on error.
	 * @param buf pointer to null terminated string to write.
	 */
	inline virtual int Write(const char *buf)
		{return write(_fd, buf, strlen(buf));};
	friend inline int write(Serial &s, char *buf)
		{return write(s._fd, buf, strlen(buf));};

	inline virtual int PutChar(char buf)
		{return write(_fd, &buf, 1);};

};

inline int inkey(Serial &s, timeout_t timer)
	{return s.Inkey(timer);};

inline int read(Serial &s, void *buf, size_t len, timeout_t timer)
	{return s.Read(buf, len, timer);};

inline int readline(Serial &s, char *buf, size_t max, timeout_t timer)
	{return s.Readline(buf, max, timer);};
#endif
