/* 
#    $Id: sharc.cpp,v 1.3 2016/12/01 19:25:04 gps Exp $
#
#    sharc -- main program to download and write out data
#
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* system includes */
#include <iostream>
#include <iomanip>
#include <string>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <signal.h>

/* sharc.cpp */
/* local includes */
#include "gpscommon.h"
#include "Navcom.h"
#include "Z12.h"
#include "MicroZ.h"
#include "Trimble.h"
#include "TRogue.h"
#include "Leica.h"
#include "Topcon.h"
#include "sharc.h"
#include "ArgUtils.h"
#if USE_STREAM
#include "sharcrt.h"
#endif

/* local function declarations */
void DumpPlans(SComm *pComm, SOps *pOps);
int WriteFAT(SOps *pOps);
void cleanup(int sig);
int DownloadFiles(SComm *pComm, SOps *pOps);
int DeleteFiles(SComm *pComm, SOps *pOps);
int HandleFileOps(SOps *pOps);
int SetupFileOps(SOps *pOps);
int InitReceiverType(SOps *pOps);
int InitFunctionPointers(SOps *pOps);

// These globals will specify the default receiver type and
// allow a subclass to override the type during autodetection
// within a Receiver class.
int RECEIVERTYPE = rectype[RX_ASHTECH].type;
string RECEIVERSTRING = rectype[RX_ASHTECH].name;
Serial *pSerial;

string Receiver::RxID         = "";
int Receiver::numFATEntries   = 0;
int Receiver::numSessions     = 0;
int Receiver::memoryused      = 0;
int Receiver::memoryavailable = 0;

SFATEntry Receiver::ReceiverFAT[MAX_RCVR_FILES];
SSession Receiver::ReceiverSessions[MAX_RCVR_SESSIONS];

// Global function variables to make things simpler (I think).
int (*Login) (Serial *pSerial, int direct, string login, string password) = NULL;
int (*Logout) (Serial *pSerial) = NULL;
string (*DownloadReceiverID) (Serial *pSerial) = NULL;
int (*DownloadReceiverFAT) (Serial *pSerial) = NULL;
int (*SetReceiverSession) (Serial *pSerial) = NULL;
int (*Stream) (Serial *pSerial, int format, double rate, char port) = NULL;
int (*TerminalSession) (Serial *pSerial) = NULL;


//  ===================================================================
//  NAME:		main
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int main(int argc, char *argv[]) {

	bool first_try = TRUE;

	SComm *pComm = NULL;	// Communications structure
	SOps *pOps = NULL;		// Operations to be performed

	pComm = new SComm;
	pOps = new SOps;

	signal(SIGINT, cleanup);
	signal(SIGHUP, cleanup);

	/*********************************************/
	/* Initialize structures, read config files, */
	/* and extract command line parameters.      */
	/*********************************************/
	InitCommStruct(pComm);
	InitOpsStruct(pOps);

	//ReadConfig(pComm, pOps);
	DecodeArguments(argc, argv, pOps, pComm);

	InitWithArgs(pOps, pComm);

	pSerial = pComm->pSerial;

	if (pOps->purgeAll && !pOps->downLoadAll) {
		debug(0,"WARNING!!!!! you have selected to purge all files without\n");
		debug(0,"WARNING!!!!! downloading all of them\n");
		debug(0,"WARNING!!!!! \n");
		debug(0,"WARNING!!!!! Hit CTRL-C within 5 seconds to cancel\n");
		sleep(5);
	}

	DumpPlans(pComm, pOps);

	// Set receiver type
	// This will be the default receiver type if nothing specified on the
	// command line.
	pOps->ReceiverType = RECEIVERTYPE;
	pOps->ReceiverTypeString = RECEIVERSTRING;

	InitFunctionPointers(pOps);



	/*************************************************/
	/* Set up Report callbacks for transfers.        */
	/*************************************************/
	pSerial->setReports(TransferReportStart, TransferReport, TransferReportStop);
	if (pComm->minbps > 0) {
		pSerial->setReports(TransferReportStart, TransferReportMinBps, TransferReportStop);
		pSerial->setMinBPS(pComm->minbps, pComm->minbps_secs);
	}

	do {
	try {
		/*************************************************/
		/* Initiate general communication with receiver. */
		/*************************************************/
		if (pOps->getID)
			if (OpenConnection(pComm) == FAILURE)
				throw "Failed opening connection\n";

		pSerial->ClearBuffers();
		debug(50,"Finished OpenConnection block\n");

		/***********************/
		/* Force Terminal Mode */
		/***********************/
		if(pSerial->isConnected() && pOps->forceTerminal) {
			if (Ashtech::TerminalSession(pSerial) == FAILURE)
				throw "Failure returning from terminal session\n";
			else
				throw "Returning from Force terminal session and exiting\n";
		}

#if USE_STREAM
		// bypass receiver setup and go right to getting the data
		if (! pOps->forceStream) {
#endif
		/*******************************************/
		/* Call login routine to finalize connect. */
		/*******************************************/
		if(pOps->getID && pSerial->isConnected())
			if (Login(pSerial, pComm->isDC, pComm->login, pComm->password) == FAILURE)
				throw "Failed Login Procedure, Closing connection and exiting.\n";

		debug(50,"Finished Login block\n");

		/**************************************/
		/* Read receiver id as general check. */
		/**************************************/
		if(pOps->getID && pSerial->isConnected()) {
			string RxID = DownloadReceiverID(pSerial);
			if(RxID == "") throw "Failed Getting Receiver ID, Closing connection and exiting.\n";
		}

		pOps->ReceiverType = RECEIVERTYPE;
		pOps->ReceiverTypeString = RECEIVERSTRING;
		InitFunctionPointers(pOps);
		if (first_try)
			InitReceiverType(pOps);

#if USE_STREAM
		} else {
			// Allow for login if --login is used with --forcestream 
			if(pComm->login.length() > 0 && pComm->password.length() > 0) {
				if(pOps->getID && pSerial->isConnected()) {
					if (Login(pSerial, pComm->isDC, pComm->login, pComm->password) == FAILURE)
						throw "Failed Login Procedure, Closing connection and exiting.\n";
				}
			}
		}
		/***************************************/
		/* go to streaming mode, do not return */
		/***************************************/
		if (pOps->stream) sharcStream(pOps);
#endif

		/**************/
		/* Get rx FAT */
		/**************/
		if(pOps->getFAT  && pSerial->isConnected()) {
			if (DownloadReceiverFAT(pSerial) == FAILURE)
				throw "Failed Getting Receiver FAT, closing connection and exiting.\n";

			if (pOps->writeFAT && (pOps->numFileOps == 0))
				if (WriteFAT(pOps) == FAILURE)
					throw "Could not write FAT\n";

			if (SetupFileOps(pOps) == FAILURE)
				throw "Failure in SetupFileOps\n";
		}

		/********************/
		/* Verify Site Code */
		/********************/
		if(pOps->verifySiteCode && pSerial->isConnected()) {
			int num = Receiver::GetNumFATEntries();
			bool found = false;
			for (int i = 0; i < num; i++) {
				string siteCode = Receiver::GetSiteName(i);
				if (strcasecmp(siteCode.c_str(),pOps->siteCodeToVerify.c_str()) == 0) {
					for ( int j = 0; j < pOps->numFileOps; j++ ) {
						pOps->fileOps[j]->pReceiver->AddToIndexList(i);
					}
					found = true;
				}
			}
			if (!found) {
				debug(-1,"No file matches site code %s\n",pOps->siteCodeToVerify.c_str());
				throw "";
			}
		}

		/********************/
		/* terminal session */
		/********************/
		if(pOps->useTerminal  && pSerial->isConnected() && first_try) {
			if (TerminalSession == NULL)
				throw "Terminal Session not supported by receiver\n";
			if (TerminalSession(pSerial) == FAILURE)
				throw "Failure returning from terminal session\n";
		}

		/*******************/
		/* Set rx Sessions */
		/*******************/
		if(pOps->setSessions  && pSerial->isConnected() && first_try)
			if (SetReceiverSession(pSerial) == FAILURE)
				throw "Failure setting session on receiver\n";

		/**************************************/
		/* Download files                     */
		/**************************************/
		if(pOps->numFileOps)
			if (DownloadFiles(pComm, pOps) == FAILURE)
				throw "Failed return from DownloadFiles, closing connection and exiting.\n";

		/****************************/
		/* Write each Raw File      */
		/****************************/
		if (HandleFileOps(pOps) == FAILURE)
			throw "Failed return from HandleFileOps, exiting.\n";

		/**************************************/
		/* Delete receiver images, reload FAT */
		/**************************************/
		if(pOps->numFileOps && pSerial->isConnected())
			if (DeleteFiles(pComm, pOps) == FAILURE)
				throw "Failed return from DeleteFiles, closing connection and exiting.\n";

		/**************************************/
		/* Download New FAT and               */
		/* Close communications with receiver */
		/**************************************/
		if(pSerial->isConnected()) {
/*			// Don't bother with FAT download, since it should be reloaded after deletion
 *			// This may risk not noticing the newest file on the receiver, but it should 
 *
 *			if (pOps->numFileOps)
 *				if (DownloadReceiverFAT(pSerial) == FAILURE)
 *					throw "Failed return from DownloadReceiverFAT, closing connection and exiting.\n";
 */

			if (Logout(pSerial) == FAILURE) 
				throw "Failed Logout Procedure, Closing connection and exiting.\n";

			if (pSerial->Close() == FAILURE)
				throw "Failed return from Serial Close, exiting.\n";
		}

		/****************************/
		/* Write FAT to output file */
		/****************************/
		if(pOps->writeFAT)
			if (WriteFAT(pOps) == FAILURE)
				throw "Could not write FAT\n";

		cleanup(SUCCESS);
	}
	catch (const char* pmessage) {
		debug(5,pmessage);
		cleanup(FAILURE);
	}
	catch (SerialException& se) {
		debug(5,se.what());
		if (se.isRetryable()) {
			pSerial->CancelXModem();
			pSerial->Close();
			pOps->retries--;
			first_try = FALSE;
		} else {
			cleanup(FAILURE);
		}
	}
	} while (pOps->retries >= 0);

	cleanup(SUCCESS);
}


//  ===================================================================
//  NAME:		DumpPlans
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
void DumpPlans(SComm *pComm, SOps *pOps) {
	int indexOp=0;

	char buffer[100];
	char buffer2[100];
	time_t tmp;
	SFileOps *pFileOps;

	tmp = time(NULL);
	addToSumFile("Creation Time: "); addToSumFile(asctime(gmtime(&tmp))); addToSumFile("\n");
	addToSumFile("Sharc version "); addToSumFile(VERSION); addToSumFile("\n\n");
	addToSumFile("------------------\n");
	addToSumFile("Planned Operations\n");
	addToSumFile("------------------\n");

	if(pOps->getID) {
		addToSumFile("\nCommunications: ");
		if (pComm->isDC) {
			addToSumFile("Direct Connect");
		} else if (pComm->isMC) {
			addToSumFile("Modem");
		} else if (pComm->isFW) {
			addToSumFile("Freewave");
		}
		addToSumFile("\n");
		addToSumFile(" Port: %s\n",pComm->port.c_str());
		addToSumFile(" Baud: %d\n",pComm->baud);
		addToSumFile(" Parity: %c\n",pComm->parity);
		addToSumFile(" Data Bits: %d\n",pComm->dataBits);
		addToSumFile(" Stop Bits: %d\n",pComm->stopBits);
		if(pComm->isMC) {
			addToSumFile(" Modem Init: %s\n",pComm->modemInit.c_str());
			addToSumFile(" Modem Dial: %s\n",pComm->modemDial.c_str());
		}
		if(pComm->isFW) {
			addToSumFile(" Freewave Repeater: %s\n",pComm->fwRepeaterCBEntry.c_str());
			addToSumFile(" Freewave Slave: %s\n",pComm->modemDial.c_str());
		}
	}

	if(pOps->writeFAT) {
		addToSumFile("\nFAT:\n");
		addToSumFile(" Filename: %s\n",pOps->FATFileName.c_str());
	}

	for(indexOp=0; indexOp<pOps->numFileOps; indexOp++) {

		pFileOps = pOps->fileOps[indexOp];

		if(pFileOps->downLoad) {
			addToSumFile("\nDownload:\n");
			if(pFileOps->index >= 0) {
				addToSumFile(" Index: %d\n",pFileOps->index + 1);
			} else {
				addToSumFile(" Span: %s%c%s\n",
					pFileOps->fileStart->chartime("%%G/%%g", buffer, 100),
					'-',
					pFileOps->fileEnd->chartime("%%G/%%g", buffer2, 100));
				addToSumFile(" Auto Join: ");
				if(pOps->joinFiles) {
					addToSumFile("TRUE");
				} else {
					addToSumFile("FALSE");
				}
				addToSumFile("\n");
			}
			if(pFileOps->session != ' ') {
				addToSumFile(" Session: %c\n",pFileOps->session);
			}
		} else if(pFileOps->purge) {
			addToSumFile("Delete:\n");
			if(pFileOps->index >= 0) {
				addToSumFile(" Index: %d\n",pFileOps->index+1);
			} else {
				addToSumFile(" Span: %s%c%s\n",
					pFileOps->fileStart->chartime("%%G/%%g", buffer, 100),
					'-',
					pFileOps->fileEnd->chartime("%%G/%%g", buffer2, 100));
			}
		}
	}

	addToSumFile("\n");
	addToSumFile("-----------------\n");
	addToSumFile("Actual Operations\n");
	addToSumFile("-----------------\n");

	return;
}


//  ===================================================================
//  NAME:		WriteFAT
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int WriteFAT(SOps *pOps) {

	int indexFAT=0;

	ofstream fat;
	char buffer[100];
	time_t tmp;
	int size;

	debug(50,"WriteFAT: Entering...\n");
	fat.open(pOps->FATFileName.c_str(), ios::out);
	if(!fat) {
		debug(1,"Could not open FAT file\n");
		return FAILURE;
	}


	tmp = time(NULL);
	fat << "Creation Time: " << asctime(gmtime(&tmp)) << endl;
	fat << "Sharc version " << VERSION << endl << endl;
	fat << Receiver::GetReceiverID() << endl << endl;
	fat << "Index    Name       DT/S    Start(GMT)          End(GMT)        Size(Bytes)" << endl;
	fat << "===========================================================================" << endl;

	for(indexFAT=0; indexFAT<Receiver::GetNumFATEntries(); indexFAT++) {
		if((size = Receiver::GetFileSize(indexFAT)) > 0) {

			fat << " " << setw(2) << indexFAT+1 << "   ";
			fat << Receiver::GetFileName(indexFAT) << "  ";
			fat << "3" << " ";
			fat << Receiver::GetSession(indexFAT) << "  ";
			fat << (Receiver::GetStartTime(indexFAT)).chartime("%Y/%j/%T", buffer, 100);
			fat << "  ";
			fat << (Receiver::GetEndTime(indexFAT)).chartime("%Y/%j/%T", buffer, 100);
			fat << "     ";
			fat << size;
			if (Receiver::IsFileActive(indexFAT)) {
				fat << " *";
			}
			fat << endl;
		}
	}

	fat.close();
	addToSumFile("FAT:\n Filename: %s\n\n",pOps->FATFileName.c_str());
	debug(50,"WriteFAT: Exiting...\n");
	return SUCCESS;
}

//  ===================================================================
//  NAME:		DownloadFiles
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int DownloadFiles(SComm *pComm, SOps *pOps) {

	SFileOps *op;
	int opindex;

	debug(50,"DownloadFiles: Entering...\n");
	for (opindex = 0; opindex < pOps->numFileOps; opindex++) {
		op = pOps->fileOps[opindex];
		debug(30,"DownloadFiles: Running for fileOp %d, index = %d\n",opindex, op->index);
		if (!(op->downLoad)) continue;
		if (op->downLoadOK == SUCCESS) {
			debug(10, "File is already downloaded (SUCCESS), ignoring\n");
			continue;
		}
		if (op->downLoadOK == SUCCESS_NOFILE) {
			debug(10, "File is already downloaded (SUCCESS_NOFILE), ignoring\n");
			continue;
		}
		if (op->index >= 0) {
			debug(30,"DownloadFiles: Downloading index %d\n",op->index);
			op->downLoadOK = op->pReceiver->DownloadFileAtIndex(pSerial, op->index);
			if (op->downLoadOK == SUCCESS)
				debug(10,"DownloadFiles: DownloadFileAtIndex succeeded for index %d\n",op->index);
			if (op->downLoadOK == SUCCESS_NOFILE)
				debug(10,"DownloadFiles: DownloadFileAtIndex succeeded (WITHOUT FILE) for index %d\n",op->index);
			if (op->downLoadOK == FAILURE)
				debug(10,"DownloadFiles: DownloadFileAtIndex failed for index %d\n",op->index);
		} else if (op->fileStart->set() && op->fileEnd->set()) {
			op->downLoadOK = op->pReceiver->DownloadFileByTime(pSerial, op->fileStart, op->fileEnd);
			if (op->downLoadOK == SUCCESS)
				debug(10,"DownloadFiles: DownloadFileByTime succeeded\n");
			if (op->downLoadOK == SUCCESS_NOFILE)
				debug(10,"DownloadFiles: DownloadFileByTime succeeded (WITHOUT FILE)\n");
			if (op->downLoadOK == FAILURE)
				debug(10,"DownloadFiles: DownloadFileByTime Failed\n");
		}
	}
	debug(50,"DownloadFiles: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		DeleteFiles
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int DeleteFiles(SComm *pComm, SOps *pOps) {

	int indexdone[MAX_RCVR_FILES];
	int i;
	int index;
	int deleted = 0;
	SFileOps *op;
	int status = SUCCESS;

	debug(50,"DeleteFiles: Entering...\n");

	for (i = 0; i < MAX_RCVR_FILES; indexdone[i++] = -1);   
	for (i = 0; i < pOps->numFileOps; i++) {
		op = pOps->fileOps[i];
		if (op->downLoad && (op->downLoadOK == FAILURE)) continue;        
		if (op->purge && (op->index >= 0)) {
			indexdone[op->index] = i;
		}
	}

	addToSumFile("Delete Results\n");
	for (i = 0; i < pOps->numFileOps; i++) {
		op = pOps->fileOps[i];
		if (op->downLoad && (op->downLoadOK == FAILURE)) continue;
		if (!op->purge) continue;
		if (op->purge && (op->index >= 0)) continue;
		if (op->fileStart->set() && op->fileEnd->set()) {
			int tmpindex[MAX_RCVR_FILES];
			int numindex = op->pReceiver->FindIndexesByTime(op->fileStart, op->fileEnd, tmpindex, TRUE);
			for (int j = 0; j < numindex; j++) {
				indexdone[tmpindex[j]] = i;
			}
		}
	}

	for (index = (MAX_RCVR_FILES-1); index >= 0; index--) {
		if (indexdone[index] < 0) continue;
		string result = "FAILURE";
		if (pOps->fileOps[indexdone[index]]->pReceiver->DeleteFileAtIndex(pSerial, index) == SUCCESS) {
			deleted++;
			result = "SUCCESS";
		}
		addToSumFile(" Deleted Index: %d %s\n", index, result.c_str());
	}

	//if FAT has changed because some files have been deleted, download a new FAT
	if (deleted > 0) {
		status = FAILURE;
		for (i = 0; i < 3; i++) {
			status = DownloadReceiverFAT(pSerial);
			if (status == SUCCESS) break;
			pSerial->ClearBuffers();
		}
	}

	for (i = 0; i < pOps->numFileOps; i++) {
		op = pOps->fileOps[i];
		if (!op->purge) continue;
		if (op->purge && (op->index >= 0)) continue;
		if (op->fileStart->set() && op->fileEnd->set()) {
			int tmpindex[MAX_RCVR_FILES];
			int numindex = op->pReceiver->FindIndexesByTime(op->fileStart, op->fileEnd, tmpindex, TRUE);
			char buffer[100];
			char buffer2[100];
			addToSumFile(" DELETE RESULT: %s-%s %s\n",
				op->fileStart->chartime("%%G/%%g", buffer, 100),
				op->fileEnd->chartime("%%G/%%g", buffer2, 100),
				(numindex > 0) ? "FAIL" : "SUCCESS" );
		}
	}

	debug(50,"DeleteFiles: Exiting...\n");
	if (status == FAILURE) return(FAILURE);	// return FAILURE if FAT failed since it will screw
											// things up for the future.
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		HandleFileOps
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int HandleFileOps(SOps *pOps) {
	int i;
	char buffer[100];
	char buffer2[100];

	debug(50,"HandleFileOps: Entering...\n");
	for (i = 0; i < pOps->numFileOps; i++) {
		SFileOps *op;
		op = pOps->fileOps[i];
		if (op->purge && !(op->downLoad)) {
			debug(25,"HandleFileOps: Purge only file at index %d, skipping\n",i);
			continue;
		}
		if (op->downLoad && (op->downLoadOK == FAILURE)) {
			debug(25,"HandleFileOps: File at index %d failed downloading, ignore\n",op->index);
			continue;
		}

		if (op->downLoadOK == SUCCESS_NOFILE) {
			string ONELINE_OUT = "DOWNLOAD RESULT: ";
			if (op->fileStart->set()) {
				op->fileStart->chartime("%%G/%%g-", buffer, 100);
				op->fileEnd->chartime("%%G/%%g", buffer2, 100);
				ONELINE_OUT += (string) buffer + (string) buffer2;
			}
			ONELINE_OUT += " SUCCESS";
			addToSumFile("%s\n",ONELINE_OUT.c_str());
			continue;
		}

// We should only get here if op->downLoadOK == SUCCESS

		gpstime *tmptime = NULL;
		addToSumFile("Output:\n");
		int index = op->pReceiver->getIndex();
		switch (pOps->FileTimeFrom) {
			case REQUESTSTART:
				tmptime = op->fileStart;
				break;
			case REQUESTSTOP:
				tmptime = op->fileEnd;
				break;
			case ACTUALSTART:
				tmptime = new gpstime;
				*tmptime = op->pReceiver->GetStartTime(index);
				break;
			case ACTUALSTOP:
				tmptime = new gpstime;
				*tmptime = op->pReceiver->GetEndTime(index);
				break;
		}
		debug(40,"HandleFileOps: Using %s for file naming\n", tmptime->chartime("%Y/%j/%T", buffer, 100));
		if (op->session != ' ') {
			op->pReceiver->SetSession(index, op->session);
			op->pReceiver->CreateFileName(index, tmptime);
			tmptime = NULL;
		}
		addToSumFile(" Start Time: %s\n",op->pReceiver->GetStartTime(index).chartime("%%G %%g", buffer, 100));
		string ONELINE_OUT = "DOWNLOAD RESULT: ";
		if (op->fileStart->set()) {
			op->fileStart->chartime("%%G/%%g-", buffer, 100);
			op->fileEnd->chartime("%%G/%%g", buffer2, 100);
			ONELINE_OUT += (string) buffer + (string) buffer2;
		}
		string filename = op->pReceiver->GetFileName(index);
		ONELINE_OUT += " " + filename;

		addToSumFile(" Raw File: %s\n", filename.c_str());
		debug(25,"HandleFileOps: Writing Image File -- %s\n",filename.c_str());
		int status = op->pReceiver->WriteFiles();
		if (status == FAILURE) {
			debug(1,"WriteFiles returned failure, Marking downloadOK = FALSE\n");
			op->downLoadOK = FAILURE;
			ONELINE_OUT += " FAILURE";
			addToSumFile(" Download Status: FAILURE\n");
		} else {
			ONELINE_OUT += " SUCCESS";
			addToSumFile(" Download Status: SUCCESS\n");
		}
		addToSumFile("\n");
		addToSumFile("%s\n",ONELINE_OUT.c_str());
	}
	debug(50,"HandleFileOps: Exiting\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		SetupFileOps
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
/* check for duplicate operations, combine purge and delete into one fileop */
int SetupFileOps(SOps *pOps) {
	int i,j,k;
	int index;
	SFileOps *op;

	debug(50,"SetupFileOps: Entering...\n");
	debug(10,"SetupFileOps: numFileOps = %d\n",pOps->numFileOps);
	if (pOps->downLoadAll || pOps->purgeAll) {
		if (pOps->downLoadAll) debug(10,"SetupFileOps: Download ALL selected\n" );
		if (pOps->purgeAll) debug(10,"SetupFileOps: Purge ALL selected\n" );
		debug(10,"SetupFileOps: Any other selections will be ignored!!!!!\n" );
		for (i = 0; i < pOps->numFileOps; i++) {
			delete pOps->fileOps[i];
		}
		pOps->numFileOps = 0;
		for(int indexFAT=0; indexFAT<Receiver::GetNumFATEntries(); indexFAT++) {
			if (Receiver::IsFileActive(indexFAT)) continue;

			pOps->fileOps[indexFAT] = new SFileOps;
			InitFileOpsStruct(pOps->fileOps[indexFAT]);
			if (pOps->downLoadAll) pOps->fileOps[indexFAT]->downLoad = TRUE;
			if (pOps->purgeAll) pOps->fileOps[indexFAT]->purge = TRUE;
			pOps->fileOps[indexFAT]->index = indexFAT;
			pOps->numFileOps++;
		}
		InitReceiverType(pOps);
	} else {
		for (i = 0; i < pOps->numFileOps; i++) {
			op = pOps->fileOps[i];
			if (op->index < 0) continue;
			index = op->index;
			for (j = 0; j < i; j++) {		// reset operations
				if (index == pOps->fileOps[j]->index) {
					debug(10,"SetupFileOps: Found Matching index entry - %d\n",j);
					if (op->downLoad) {
						pOps->fileOps[j]->downLoad = TRUE;
					}
					if (op->purge) {
						pOps->fileOps[j]->purge = TRUE;
					}
					delete op;
					for (k = i; k < pOps->numFileOps - 1; k++) {
						pOps->fileOps[k] = pOps->fileOps[k+1];
					}
					pOps->fileOps[pOps->numFileOps] = NULL;
					pOps->numFileOps--; i--;
				}
			}
		}
	}
	debug(10,"SetupFileOps: numFileOps = %d\n",pOps->numFileOps);
	debug(50,"SetupFileOps: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		cleanup
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
void cleanup(int sig)
{

	debug(50,"cleanup: Entering...\n");
	if (pSerial->isConnected()) {
		debug(10,"Closing serial Port\n");
		FWCleanup(pSerial);
		pSerial->ClearBuffers();
		pSerial->ClearBuffers();
		pSerial->ClearBuffers();
		pSerial->Close();
	}
	debug(10,"Exiting on signal %d\n",sig);
	exit(sig + (pSerial->failedXFers() * 2));
}

//  ===================================================================
//  NAME:		InitFunctionPointers
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int InitFunctionPointers(SOps *pOps) {
	debug(50,"InitFunctionPointers: Entering...\n");
	switch (pOps->ReceiverType) {
		case RX_ASHTECH:
			debug(D_VALUES,"InitFunctiotnPointers: Setting to ASHTECH\n");
			Login = Z12::Login;
			Logout = Z12::Logout;
			DownloadReceiverID = Z12::DownloadReceiverID;
			break;
		case RX_MICROZ:
			debug(D_VALUES,"InitFunctiotnPointers: Setting to ASHTECH MICROZ\n");
			Login = MicroZ::Login;
			Logout = MicroZ::Logout;
			DownloadReceiverID = MicroZ::DownloadReceiverID;
			DownloadReceiverFAT = MicroZ::DownloadReceiverFAT;
			SetReceiverSession = MicroZ::SetReceiverSession;
			Stream = MicroZ::Stream;
			TerminalSession = MicroZ::TerminalSession;
			break;
		case RX_Z12:
			debug(D_VALUES,"InitFunctiotnPointers: Setting to ASHTECH Z12\n");
			Login = Z12::Login;
			Logout = Z12::Logout;
			DownloadReceiverID = Z12::DownloadReceiverID;
			DownloadReceiverFAT = Z12::DownloadReceiverFAT;
			SetReceiverSession = Z12::SetReceiverSession;
			Stream = Z12::Stream;
			TerminalSession = Z12::TerminalSession;
			break;
		case RX_LEICA:
			debug(D_VALUES,"InitFunctiotnPointers: Setting to LEICA\n");
			Login = Leica::Login;
			Logout = Leica::Logout;
			DownloadReceiverID = Leica::DownloadReceiverID;
			DownloadReceiverFAT = Leica::DownloadReceiverFAT;
			SetReceiverSession = Leica::SetReceiverSession;
			Stream = Leica::Stream;
			break;
		case RX_TROGUE:
			debug(D_VALUES,"InitFunctiotnPointers: Setting to TROGUE\n");
			Login = TRogue::Login;
			Logout = TRogue::Logout;
			DownloadReceiverID = TRogue::DownloadReceiverID;
			DownloadReceiverFAT = TRogue::DownloadReceiverFAT;
			SetReceiverSession = TRogue::SetReceiverSession;
			Stream = TRogue::Stream;
			break;
		case RX_TRIMBLE:
			debug(D_VALUES,"InitFunctiotnPointers: Setting to TRIMBLE\n");
			Login = Trimble::Login;
			Logout = Trimble::Logout;
			DownloadReceiverID = Trimble::DownloadReceiverID;
			DownloadReceiverFAT = Trimble::DownloadReceiverFAT;
			SetReceiverSession = Trimble::SetReceiverSession;
			Stream = Trimble::Stream;
			break;
		case RX_NAVCOM:
			debug(D_VALUES,"InitFunctiotnPointers: Setting to NAVCOM\n");
			Login = Navcom::Login;
			Logout = Navcom::Logout;
			DownloadReceiverID = Navcom::DownloadReceiverID;
			DownloadReceiverFAT = Navcom::DownloadReceiverFAT;
// KFS -- pregram set receiver session
//			SetReceiverSession = Navcom::SetReceiverSession;
			Stream = Navcom::Stream;
			break;
		case RX_TOPCON:
			debug(D_VALUES,"InitFunctiotnPointers: Setting to TOPCON\n");
			Login = Topcon::Login;
			Logout = Topcon::Logout;
			DownloadReceiverID = Topcon::DownloadReceiverID;
			DownloadReceiverFAT = Topcon::DownloadReceiverFAT;
			SetReceiverSession = Topcon::SetReceiverSession;
			Stream = Topcon::Stream;
			TerminalSession = Topcon::TerminalSession;
			break;
	}
	debug(50,"InitFunctionPointers: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		InitReceiverType
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int InitReceiverType(SOps *pOps) {
	debug(50,"InitReceiverType: Entering...\n");
	if (pOps->FileTemplate == "") {
		pOps->FileTemplate = rectype[pOps->ReceiverType].FileTemplate;
	}
	Receiver::SetFileTemplate(pOps->FileTemplate);
	Receiver::SetTemplateStrings(pOps->TemplateStrings);
	for (int i = 0; i < pOps->numFileOps; i++) {
		if (pOps->fileOps[i]->pReceiver != NULL) {
			delete pOps->fileOps[i]->pReceiver;
		}
		switch (pOps->ReceiverType) {
			case RX_MICROZ:
				pOps->fileOps[i]->pReceiver = new MicroZ;
				break;
			case RX_ASHTECH:
			case RX_Z12:
				pOps->fileOps[i]->pReceiver = new Z12;
				break;
			case RX_LEICA:
				pOps->fileOps[i]->pReceiver = new Leica;
				break;
			case RX_TROGUE:
				pOps->fileOps[i]->pReceiver = new TRogue;
				break;
			case RX_TRIMBLE:
				pOps->fileOps[i]->pReceiver = new Trimble;
				break;
			case RX_NAVCOM:
				pOps->fileOps[i]->pReceiver = new Navcom;
				break;
			case RX_TOPCON:
				pOps->fileOps[i]->pReceiver = new Topcon;
				break;
		}
		if (pOps->fileOps[i]->fileStart->set()) {
			for (int j = 0; j < pOps->fileOps[i]->indexlistnum; j++) {
				pOps->fileOps[i]->pReceiver->AddToIndexList(pOps->fileOps[i]->indexlist[j]-1);
			}
		}

	}
	debug(50,"InitReceiverType: Exiting...\n");
	return(SUCCESS);
}

//  ===================================================================
//  NAME:		InitOpsStruct
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
void InitOpsStruct(SOps *pOps) {
	debug(50,"InitOpsStruct: Entering...\n");
	pOps->verifySiteCode = FALSE;
	pOps->writeFAT = FALSE;
	pOps->getFAT = FALSE;
	pOps->getID = FALSE;
	pOps->setSessions = FALSE;
	pOps->joinFiles = FALSE;
	pOps->upperCaseNames = TRUE;
	pOps->useTerminal = FALSE;

	pOps->sessionOffset = 0;
	pOps->numFileOps = 0;
	pOps->numFiles = 0;
	pOps->siteCodeToVerify = "";
	pOps->forceTerminal = FALSE;
	pOps->retries = 0;

#if USE_STREAM
	pOps->stream = FALSE;
	pOps->streamSample = 0;
	pOps->streamPort = 'A';
	pOps->stats = 120;
#ifdef USE_LEAPSECONDS_FILE
	pOps->leapsec = leapsecs_since_gps_epoch();
#else
	pOps->leapsec = LEAP_SECONDS;
#endif
	pOps->forceStream = FALSE;

        pOps->isNTRIP = FALSE;
        pOps->ntripMount = "";
        pOps->ntripUser = "";
        pOps->ntripPassword = "";

#ifdef RTCM_ENABLE
	pOps->translateRTCM = FALSE;
	pOps->RTCMVersion = DATATYPE_UNKNOWN;
	pOps->RTCMantennafile = "";
#endif
#endif
	pOps->ReceiverType = -1;
	pOps->ReceiverTypeString = "";

	pOps->FATFileName = "";
	pOps->SUMFileName = "";
	pOps->FileTemplate = "";
	pOps->FileTimeFrom = ACTUALSTART;

	for (int i = 0; i < MAXTEMPLATESTRINGS; i++) {
		pOps->TemplateStrings[i] = "";
	}
	pOps->numTemplateStrings = 0;

	pOps->purgeAll = FALSE;
	pOps->downLoadAll = FALSE;

	debug(50,"InitOpsStruct: Exiting\n");
}

