/* 
#    $Id: sharcrt.cpp,v 1.6 2017/11/06 09:00:28 gps Exp $
#
#    sharc -- main program to download and write out data
#
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* system includes */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string>
#include <iomanip>
#include <math.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/termios.h>
#include <sys/types.h>

#include "gpscommon.h"

#if USE_STREAM

/* sharc.cpp */
/* local includes */
//#include "sharcrt.h"
//#include "ArgUtils.h"
#include "binex.h"
#include "StreamType.h"
#include "StreamDBEN.h"
#include "StreamMBEN.h"
#include "StreamBinex.h"

vector<data_record_t *> data_vector;
input_buffer_t ibuffer;

pthread_mutex_t data_mutex;
pthread_cond_t data_available;
pthread_t thread_serial, thread_read_data, thread_shutdown;

int numwriters;
writerinfo_t *writerThreads[MAXTHREADS];

extern int errno;

void cleanup(int sig);

extern int (*Logout) (Serial *pSerial);
extern int (*Stream) (Serial *pSerial, int format, double rate, char port);
extern Serial *pSerial;

//  ===================================================================
//  NAME:		main
//
//  PURPOSE:
//
//  RETURNS:
//
//  HISTORY:
//
//  ===================================================================
int sharcStream(SOps *pOps) {

	/**************************************/
	/* Stop streaming and exit if sample  */
	/* is <= to 0                         */
	/**************************************/
	debug(D_FLOW,"sharcStream: Entering...\n");
	if(pOps->streamSample <= 0) CloseStreamAndExit(pSerial, pOps);

	int filedes[2];
	int status;
	read_thread_info_t read_info;

	if (pipe(filedes) != 0) {
		debug(D_ERRMAJOR,"sharcStream: Failed creating pipe for process communication\n");
		cleanup(FAILURE);
	} else {
		debug(D_FLOW,"sharcStream: Successfully created pipe for communication\n");
	}

	/**************************************/
	/* Issue command to setup receiver    */
	/* streaming binex                    */
	/**************************************/
	if(pOps->getID && pSerial->isConnected() && !pOps->forceStream) {
		status = Stream(pSerial, pOps->streamType, pOps->streamSample, pOps->streamPort);
		if(status == FAILURE) {
			debug(D_ERRMAJOR,"sharcStream: Failed Setting Streaming, Closing connection and exiting.\n");
			cleanup(FAILURE);
		}
	}
	if (pOps->isNTRIP) {
		status = ntrip_mount(pSerial, pOps->ntripMount, pOps->ntripUser, pOps->ntripPassword);
		if (status == FAILURE) {
			debug(D_ERRMAJOR,"sharcStream: Failed Connecting to NTRIP Mount Point, Closing connection and exiting.\n");
			cleanup(FAILURE);
		}
	}

	/**************************************/
	/* Launch all output handler threads  */
	/* Wait for synchronization           */
	/**************************************/
	cache_init();

	/**************************************/
	/* Start streaming and set all threads*/
	/* to start running                   */
	/**************************************/
	//signal(SIGINT, SIG_IGN);
	//signal(SIGHUP, SIG_IGN);
	sigset_t signals_to_block;
	sigemptyset(&signals_to_block);
	sigaddset(&signals_to_block, SIGINT);
	sigaddset(&signals_to_block, SIGTERM);
	sigaddset(&signals_to_block, SIGUSR1);
	sigaddset(&signals_to_block, SIGHUP);
	sigaddset(&signals_to_block, SIGPIPE);
	pthread_sigmask(SIG_BLOCK, &signals_to_block, NULL);
	pthread_create(&thread_shutdown, NULL, shutdownThread, NULL);

	read_info.filedes = filedes[0];
	read_info.streamType = pOps->streamType;
#ifdef RTCM_ENABLE
	read_info.translateRTCM = pOps->translateRTCM;
	read_info.antennaFile = pOps->RTCMantennafile;
#endif
	createReadThread(pOps->streamType, (void *) &read_info);

	for (int i = 0; (i < MAXTHREADS) && (writerThreads[i] != NULL); i++) {
		writerThreads[i]->dir_template = pOps->DataDir;
		writerThreads[i]->streamType = pOps->streamType;
		writerThreads[i]->stats = pOps->stats;
		writerThreads[i]->leapsec = pOps->leapsec;
		debug(D_VALUES,"sharcStream: Setting dir_template to %s\n",writerThreads[i]->dir_template.c_str());
		debug(D_FLOW,"sharcStream: Creating writerthread #%d\n",i);
		if (pipe(writerThreads[i]->filedes) != 0) {
			debug(D_ERRMAJOR,"sharcStream: Failed creating pipe for process communication\n");
			cleanup(FAILURE);
		} else {
			debug(D_FLOW,"sharcStream: Successfully created pipe for communication\n");
		}
		fcntl(writerThreads[i]->filedes[0],F_SETFL,O_NONBLOCK);
		fcntl(writerThreads[i]->filedes[1],F_SETFL,O_NONBLOCK);
		createWriteThread(pOps->streamType, i);
	}
	pthread_create(&thread_serial, NULL, serialThread, (void *) &filedes[1]);

	void *statusp;
	pthread_join(thread_shutdown, &statusp);
	close(filedes[0]);
	close(filedes[1]);

	debug(D_FLOW,"sharcStream: All threads have been terminated\n");
	/**************************************/
	/* Download New FAT and               */
	/* Close communications with receiver */
	/**************************************/
	CloseStreamAndExit(pSerial, pOps);
	return(1);
}

void CloseStreamAndExit(Serial *pSerial, SOps *pOps)
{

	debug(D_FLOW,"CloseStreamAndExit: Entering...\n");
	if (! pSerial->isConnected()) {
		debug(D_ERRMAJOR,"CloseStreamAndExit: Trying to close stream when not connected to port\n");
		cleanup(FAILURE);
	}

	debug(D_FLOW,"CloseStreamAndExit: pSerial is connected...\n");

	int status = SUCCESS;
	if (!pOps->forceStream) {
		status = Stream(pSerial,pOps->streamType,0,pOps->streamPort);
		status = Logout(pSerial);
	}
	if(status == FAILURE) {
		debug(D_ERRMAJOR,"CloseStreamAndExit: Failed Logout Procedure, Closing connection and exiting.\n");
		cleanup(FAILURE);
	}
	debug(D_STD_V3,"CloseStreamAndExit: Closing Serial Port\n");
	pSerial->Close();
	if(status) {
		debug(D_ERRMAJOR,"CloseStreamAndExit: Failed return from Serial Close, exiting.\n");
		cleanup(FAILURE);
	}
	debug(D_FLOW,"CloseStreamAndExt: Exiting...\n");
	exit(SUCCESS);

}


void * shutdownThread(void *argin) {
	sigset_t signals_to_catch;
	int caught;

	debug(D_FLOW,"shutdownThread: Entering...\n");
	sigemptyset(&signals_to_catch);

	sigaddset(&signals_to_catch, SIGUSR1);
	sigaddset(&signals_to_catch, SIGHUP);
	sigaddset(&signals_to_catch, SIGINT);
	sigaddset(&signals_to_catch, SIGTERM);

	debug(D_FLOW,"shutdownThread: Waiting for signal...\n");
	sigwait(&signals_to_catch, &caught);


	debug(D_WARNMAJOR,"shutdownThread: Caught signal %d\n",caught);
	pthread_cancel(thread_serial);
	pthread_cancel(thread_read_data);
	for (int i = 0; (i < MAXTHREADS) && (writerThreads[i] != NULL); i++) {
		debug(D_STD_V3,"shutdownThread: Canceling writing process %d\n",i);
		pthread_cancel(writerThreads[i]->thread);
		pthread_cancel(writerThreads[i]->thread);
	}

	void *statusp;
	pthread_join(thread_read_data, &statusp);
	pthread_join(thread_serial, &statusp);
	for (int i = 0; (i < MAXTHREADS) && (writerThreads[i] != NULL); i++) {
		debug(D_STD_V3,"shutdownThread: Joining writing process %d\n",i);
		pthread_join(writerThreads[i]->thread,&statusp);
	}
//	cleanup(0);
	pthread_exit(0);
	return(NULL);
}

void * readDataThread(void *argin) {

	read_thread_info_t *tmp = (read_thread_info_t *) argin;
	int *fd = &tmp->filedes;
	int streamType = tmp->streamType;
	
	debug(D_FLOW,"readDataThread: Entering...\n");


	int last_state;
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &last_state);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);


	debug(D_FLOW,"readDataThread: Starting While loop\n");
	StreamType *incoming;
	switch (streamType) {
		case DATATYPE_BINEX:
			incoming = new StreamBinex();
			debug(100,"readDataThread: Setup StreamBinex instance\n");
			break;
		case DATATYPE_DBEN:
			incoming = new StreamDBEN();
			debug(100,"readDataThread: Setup StreamDBEN instance\n");
			break;
		default:
			incoming = new StreamRaw();
			debug(100,"readDataThread: Setup StreamRaw instance\n");
			break;
	}
#ifdef RTCM_ENABLE
	incoming->InitRTCM(tmp->translateRTCM, RTCM_ANTFREQ, tmp->antennaFile);
#endif

	while (1) {
		pthread_testcancel();
		int rc = INCOMPLETE_PACKET;
		data_record_t *myrecord = new data_record_t;
		memset(myrecord, 0, sizeof(data_record_t));

		do {
			pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
			rc = incoming->ProcessInput(*fd, 10);
			debug(100,"readDataThread: return from ProcessInput is %d\n",rc);
			pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
		} while (rc != PACKET_FOUND);

		if (rc != PACKET_FOUND) {
			debug(D_WARN,"readDataThread: Error reading PACKET (shouldn't happen)\n");
			continue;
		}
		debug(D_SERINFO,"readDataThread: PACKET FOUND\n");

		incoming->GetRecord(myrecord->vrecord);

		if (streamType == DATATYPE_BINEX) {
			myrecord->record_id = (unsigned int) incoming->GetRecordInfo(BINEX_RECORD_ID);
			myrecord->reversible = (int) incoming->GetRecordInfo(BINEX_REVERSIBLE);
			myrecord->little_endian = (int) incoming->GetRecordInfo(BINEX_LITTLE_ENDIAN);
		}




		myrecord->record_len = myrecord->vrecord.size();
		myrecord->record = new unsigned char[myrecord->record_len];
		debug(D_SERINFO,"readDataThread: Copying vector\n");
		for (size_t i = 0; i < myrecord->record_len; i++) {
			myrecord->record[i] = myrecord->vrecord[i];
		}

#ifdef RTCM_ENABLE
		incoming->DecomposePacket();
		if (tmp->translateRTCM) {
		  if (incoming->allowRTCM()) {
			debug(D_SERINFO,"readDataThread: Datatype supports RTCM\n");

			incoming->GetRTCMRecords(myrecord->vrtcm, myrecord->vrtcm3);
			myrecord->rtcm_len = myrecord->vrtcm.size();
			myrecord->rtcm = new unsigned char[myrecord->rtcm_len];
			for (size_t i = 0; i < myrecord->rtcm_len; i++) {
				myrecord->rtcm[i] = myrecord->vrtcm[i];
			}
			myrecord->rtcm3_len = myrecord->vrtcm3.size();
			myrecord->rtcm3 = new unsigned char[myrecord->rtcm3_len];
			for (size_t i = 0; i < myrecord->rtcm3_len; i++) {
				myrecord->rtcm3[i] = myrecord->vrtcm3[i];
			}
			myrecord->vrtcm3.clear();
			myrecord->vrtcm.clear();
			debug(D_SERINFO,"readDataThread: Got RTCM 2 Packet length = %d\n",myrecord->rtcm_len);
			debug(D_SERINFO,"readDataThread: Got RTCM 3 Packet length = %d\n",myrecord->rtcm3_len);
		  } else {
			debug(D_SERINFO,"readDataThread: Datatype does _NOT_ support RTCM\n");
		  }
		}		  
#endif
		incoming->ResetPacket();

		incoming->GetRecordTime(&myrecord->gpsMinutes, &myrecord->gpsm_Seconds);

		debug(D_VALUES,"readDataThread:Read %d byte data record, Adding to Stack\n",myrecord->record_len);
		debug(D_VALUES,"readDataThread:\n");
		debug(D_VALUES,"\trecord_len    = %d\n",myrecord->record_len);
		debug(D_VALUES,"\ttime          = %dmin %dmSec\n",myrecord->gpsMinutes, myrecord->gpsm_Seconds);
//		debug(D_VALUES,"\t%s", incoming->GetDebugInfo().c_str());
#ifdef RTCM_ENABLE
		if (tmp->translateRTCM) {
			debug(D_VALUES,"\tRTCM length   = %d\n",myrecord->rtcm_len);
		}
#endif
		myrecord->handled = 0;
		cache_add(myrecord);
	}
	return(NULL);
}


void createWriteThread(int streamType, int i) {
	debug(D_FLOW,"createWriteThread: Entering...\n");
	switch (writerThreads[i]->output_type) {
		case WRITE_FILE:
			pthread_create(&(writerThreads[i]->thread), NULL, writeDataThread, (void *) writerThreads[i]);
			break;
		case WRITE_NET:
		case WRITE_NTRIP:
			pthread_create(&(writerThreads[i]->thread), NULL, writeNetworkThread, (void *) writerThreads[i]);
			break;
	}
}

void createReadThread(int streamType, void *argin) {
	debug(D_FLOW,"createReadThread: Entering...\n");
	pthread_create(&thread_read_data, NULL, readDataThread, argin);
}

void * serialThread(void *argin) {

	int *fd = (int *) argin;
	debug(D_FLOW,"serialThread: Entered new thread\n");
	int last_state;
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &last_state);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	char buffer[1025];
	long bytes = 0;
	time_t t = time(NULL);
	time_t uptime_start = t;
	time_t now = t;
	time_t last_data = t;
	int reconnects = 0;
	int no_data_start = 0;

	while (1) {
		t = time(NULL);
		if (now < t) now = t;

		int c = pSerial->Read(buffer, 1024, 20);
		debug(D_SERINFO+5,"serialThread: Read %d characters\n",c);

// Reconnect if error with connection or no data received in 60 seconds
		if ((c == -1) || pSerial->get_fd() < 0 || ((now - last_data) > 60)) {
		        if (++reconnects > MAXRECONNECT) {
			  debug(D_SERINFO,"#reconnects %d exceeds %d\n", reconnects, MAXRECONNECT);
		          all_exit();
			}
		        if (pSerial->Reconnect(MAXRECONNECT, 1) != 0) all_exit();
			debug(D_USTREAM, "reconnect: now=%ld, uptime_start=%ld, last_data=%ld\n",  now, uptime_start, last_data);
			uptime_start = t;
			pthread_testcancel();
			continue;
		}

		bytes += c;
		if ((now % 120) == 0) {
			debug(D_USTREAM, "stats: uptime=%ld, bytes=%ld\n", now - uptime_start, bytes);
			// now++;
		}

		if (c == 0) {
			// toggle start of no data and set time of last data
			if(no_data_start == 0) {
			    debug(D_SERINFO+5,"serialThread: ***** Data stopped coming in\n");
			    no_data_start = 1;
			    last_data = t;
			}
			pthread_testcancel();
			continue;
		} else {
			if(no_data_start) {
			     debug(D_SERINFO+5,"serialThread: ***** Data coming in again after %d secs\n", (now - last_data));
			     no_data_start = 0;
			}
		}

		last_data = t;

		c = write(*fd, buffer, c);
		debug(D_SERINFO,"serialThread: Wrote %d characters\n",c);
		pthread_testcancel();
	}
	return(NULL);
}

void all_exit() {
 	debug(D_SERINFO+5,"all_exit: Killing process %d with SIGHUP\n", getpid());
	kill(getpid(),SIGHUP);
}

void last_breath(void *message) {
	debug(D_FLOW,"Freeing data_mutex\n");
	pthread_mutex_unlock(&data_mutex);
}

// check to see if writernum needs to do anything
// return -1 if nothing to do,
// otherwise return position to start working on
int needHandling(int writernum) {

	int pos;
	debug(D_FLOW,"needHandling: Entering...\n");

	debug(D_SERINFO,"needHandling: Checking empty status\n");
	if (data_vector.empty()) {
		debug(D_SERINFO, "needHandling: Vector is empty\n");
		return(-1);
	}

	int end = data_vector.size();
	debug(D_SERINFO,"needHandling: Checking handled status for %d elements\n",end);
	for (pos = 0; pos <= (end-1); pos++) {
		debug(D_SERINFO,"needHandling: Checking handled status for pos %d\n",pos);
		debug(D_SERINFO,"\t\thandled = %x, writernum = %x -- ",data_vector[pos]->handled, writernum);
		if (data_vector[pos]->handled & writernum) {
			debug(D_SERINFO,"handled\n");
			continue;
		} else {
			debug(D_SERINFO,"NOT handled\n");
			break;
		}
	}
	if (pos > (end - 1)) {
		debug(D_SERINFO,"needHandling: Writer has handled all of the cache (pos = %d of %d)\n",pos, end -1);
		return(-1);
	}
	return(pos);
}

void computeLatency(time_t rec_time_sec, unsigned short milli_sec, time_t *latency, int *numrecs, time_t *now, int stats = 120, int leapsec = LEAP_SECONDS) {

	debug(D_FLOW,"computeLatency: Entering...\n");
	struct timeval tv;
	gettimeofday(&tv, NULL);
	time_t gpstime = ((tv.tv_sec + leapsec) * 1000) + (int)(tv.tv_usec/1000.0);
	time_t rectime = (rec_time_sec * 1000) + milli_sec;
	*latency += gpstime - rectime;
	(*numrecs)++;

	if (*now <= tv.tv_sec) *now = tv.tv_sec;
	if ((*now % stats) == 0 ) {
		debug(D_USTREAM,"stats: avg_latency=%fms total_latency=%dms records=%d\n",((double)*latency/(double)*numrecs), *latency, *numrecs);
		*latency = 0;
		*numrecs = 0;
		(*now)++;
	}
}

// Check sample rate stuff for BINEX only. We are not decoding DBEN yet.
bool checkSample(writerinfo_t *info, data_record_t *record) {

	debug(D_FLOW,"checkSample: Entering...\n");
	unsigned long rec_time = record->gpsMinutes % 1440;
	rec_time = (rec_time * 60000) + record->gpsm_Seconds;

	debug(D_VALUES,"checkSample: Record time is %d\n",rec_time);
	debug(D_VALUES,"checkSample: Sample rate is %d\n",info->sample_rate);
	if (info->sample_rate <= 0) return(0);
	if ((rec_time % info->sample_rate) == 0) return(0);

	switch (info->streamType) {
		case DATATYPE_DBEN:
		case DATATYPE_RAW:
			return(1);
		case DATATYPE_BINEX:
			if (record->record_id != 127) return(1);

			if ((rec_time % info->sample_rate) == 0) return(0);
			return(1);
	}
	return(1);
}

int getOutputData(writerinfo_t *info, binexHandle *wbh, data_record_t *record, unsigned char ** output, unsigned int *bytes_out)
{

	int streamType = info->streamType;
	int rc;
	debug(D_FLOW,"getOutputData: Entering...\n");

	if (record->record == NULL) return(-1);
#ifdef RTCM_ENABLE              
	if (info->outputRTCM) streamType = info->RTCMVersion;
#endif                                  
	switch (streamType) {
		case DATATYPE_BINEX:
			if (wbh == NULL) return(-1);
			set_endian (wbh, &(record->little_endian));
			*output = write_binex_buf (wbh, record->record,
				&(record->record_len),
				&(record->record_id),
				&(record->reversible),
				bytes_out,
				&rc);
			break;     
#ifdef RTCM_ENABLE
		case DATATYPE_RTCM:
		case DATATYPE_RTCM22:
		case DATATYPE_RTCM24:
			if (record->rtcm_len <= 0) return(-1);
			*output = new unsigned char[record->rtcm_len];
			*bytes_out = record->rtcm_len;
			memcpy((void *) *output, record->rtcm, record->rtcm_len);
			break;
		case DATATYPE_RTCM30:
			if (record->rtcm3_len <= 0) return(-1);
			*output = new unsigned char[record->rtcm3_len];
			*bytes_out = record->rtcm3_len;
			memcpy((void *) *output, record->rtcm3, record->rtcm3_len);
			break;
#endif
		case DATATYPE_DBEN:
		default:
			if (record->record_len <= 0) return(-1);
			*output = new unsigned char[record->record_len];
			*bytes_out = record->record_len;
			memcpy((void *) *output, record->record, record->record_len);
			break;
	}
	return(0);
}


void * writeDataThread(void *argin) {

	writerinfo_t *info = (writerinfo_t *) argin;

	debug(D_FLOW,"writeDataThread: Entering Thread(%d)...\n",info->writernum);
	int last_state;
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &last_state);

	debug(D_FLOW,"writeDataThread: Entering...\n");
	binexHandle *wbh = NULL;
	FILE *fh = NULL;

	string savefilename;
	string savedir;

	long latency = 0;
	int numrecs = 0;
	time_t now = 0;

	while (1) {

		char * tmp = NULL;
		int pos = 0;
		pthread_cleanup_push(last_breath, (void *) tmp);
		pthread_mutex_lock(&data_mutex);

		pos = needHandling(info->writernum);
		if (pos >= 0) {

			if (checkSample(info, data_vector[pos]) != 0) {
				debug(D_WARNMINOR,"writeDataThread: Ignoring record\n");
				data_vector[pos]->handled |= info->writernum;
				pos = -1;
			}

		}
		pthread_testcancel();

		if (pos >= 0) {
			debug(D_FLOW,"writeDataThread: Getting record from cache entry %d\n",pos);
			debug(D_VALUES,"writeDataThread: record of size %d\n",data_vector[pos]->record_len);

			// get time of record or set to now
			time_t rec_time_sec;
			if (data_vector[pos]->gpsMinutes >= 0) {
       				rec_time_sec = (3657*24*60*60) + (data_vector[pos]->gpsMinutes*60);
			} else {
				time(&rec_time_sec);
			}

			time_t file_time = rec_time_sec - (rec_time_sec % info->seconds_in_file);

			// Output Statistics at regular interval
			// Ignore records without a timetag
			if ((data_vector[pos]->gpsMinutes > 0) && (info->stats > 0)) {
#ifdef USE_LEAPSECONDS_FILE
				int leapsec = leapsecs_since_gps_epoch();
#else
				int leapsec = info->leapsec;
#endif
				debug(D_VALUES,"writeDataThread: rec_time_sec = %ld, file_time = %ld\n",rec_time_sec, file_time);
				debug(D_VALUES,"writeDataThread: latency=%ld, numrecs = %d\n",latency, numrecs);
				computeLatency(rec_time_sec, data_vector[pos]->gpsm_Seconds, &latency, &numrecs, &now, info->stats, leapsec);
				debug(D_VALUES,"writeDataThread: latency=%ld, numrecs = %d\n",latency, numrecs);
				debug(D_VALUES,"writeDataThread: file time would be %s\n",asctime(gmtime(&file_time)));
			}

			int rc;
			switch (info->streamType) {
				case DATATYPE_BINEX:
					rc = getSaveFile((void **)&wbh, &savedir, &savefilename, file_time, info->dir_template, info->file_template, 1);
					break;
				case DATATYPE_DBEN:
				default:
					rc = getSaveFile((void **)&fh, &savedir, &savefilename, file_time, info->dir_template, info->file_template, 0);
					break;
			}
			if (rc == 0) 
				debug(D_VALUES,"writeDataThread: Filename not changed\n");
			if (rc == -1) 
				debug(D_ERR,"writeDataThread: Error opening file\n");

			debug(D_VALUES,"writeDataThread: Writing to %s/%s\n",savedir.c_str(),savefilename.c_str());

			// after writing bytes_out should be set to number of bytes written
			// after writing rc should be 0 for success, 1 for failure. Need  to clean up returns
			// from various writes to make consistent
			if (data_vector[pos]->record == NULL) debug(D_ERR,"writeDataThread: record is NULL!!!!\n");

			unsigned int bytes_out = 0;
			rc = 1;
			switch (info->streamType) {
				case DATATYPE_BINEX:
					set_endian (wbh, &(data_vector[pos]->little_endian));
					write_binex (wbh, data_vector[pos]->record,
								&(data_vector[pos]->record_len),
								&(data_vector[pos]->record_id),
								&(data_vector[pos]->reversible),
								&bytes_out,
								&rc);
					if (ftell(wbh->fh) >= 0) {
					  (void)fflush(wbh->fh);
					}
					break;
				case DATATYPE_DBEN:
				default:
					rc = fwrite(data_vector[pos]->record, sizeof (unsigned char), data_vector[pos]->record_len, fh);
					if (ftell(fh) >= 0) {
					  (void)fflush(fh);
					}
					bytes_out = rc; rc = 0;
					if ((unsigned int) bytes_out != data_vector[pos]->record_len) rc = 1;
					break;
			}

			debug(D_VALUES,"writeDataThread: Return from write, rc = %d, bytes_out = %d\n",rc, bytes_out);
			if (rc) {
				debug(D_ERR,"writeDataThread: Error writing data record #%d, (bytes_out = %d)\n",pos,bytes_out);
			} else {
				data_vector[pos]->handled |= info->writernum;
				debug(D_VALUES,"writeDataThread: wrote record of size %d with result %d\n",bytes_out,rc);
			}
// No work found.
		} else {
			debug(D_FLOW,"writeDataThread: Waiting for more data to become available\n");
			pthread_cond_wait(&data_available,&data_mutex);
			debug(D_FLOW,"writeDataThread: writernum %d has been notified of impending data\n",info->writernum);
		}
		pthread_testcancel();
		pthread_mutex_unlock(&data_mutex);
		pthread_cleanup_pop(0);
	}
	return(NULL);
}

void * writeNetworkThread(void *argin) {

	writerinfo_t *info = (writerinfo_t *) argin;

	debug(D_FLOW,"writeNetworkThread: Entering Thread(%d)...\n",info->writernum);
	int last_state;
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &last_state);

	debug(D_FLOW,"writeNetThread: Entering...\n");
	int fd = -1;

	binexHandle *wbh = NULL;

	while (1) {

		int rc = 0;
		char * tmp = NULL;
		int pos = 0;
		for (int i = 0; (fd <= 0) && (i < 10); i++ && sleep(2)) {
			if (wbh != NULL) delete [] wbh;
			wbh = NULL;
			debug(D_CHECK,"writeNetThread: Attempt %d to open network connection to %s:%d\n",i, info->hostname.c_str(), info->port);
			fd = open_net(info->hostname, info->port);
			if (fd <= 0) continue;

			if (info->output_type == WRITE_NTRIP) {
				debug(D_CHECK,"writeNetThread: Logging into NTRIP server with mount %s\n",info->mount.c_str());
				rc = ntrip_login(fd, info->mount, info->password);
				if (rc != 0) {
					debug(D_ERR,"writeNetThread: Unable to login to NTRIP server for mount %s\n",info->mount.c_str());
					close(fd);
					fd = -1;
				}
			}
			if ((info->streamType == DATATYPE_BINEX) && (fd >= 0) && (wbh == NULL)) wbh = new_binex(fd,"w");
		}
		pthread_cleanup_push(last_breath, (void *) tmp);
		pthread_mutex_lock(&data_mutex);

		pos = needHandling(info->writernum);

		if (pos >= 0) {
			if (checkSample(info, data_vector[pos]) != 0) {
				debug(D_WARNMINOR,"writeNetThread: Ignoring record\n");
				data_vector[pos]->handled |= info->writernum;
				pos = -1;
			}
		}
		pthread_testcancel();

		if ((pos >= 0) && (fd >= 0)) {
			debug(D_FLOW,"writeNetThread: Getting record from cache entry %d\n",pos);
			debug(D_VALUES,"writeNetThread: record of size %d\n",data_vector[pos]->record_len);

			unsigned char *output = NULL;
			unsigned int bytes_out;
			int rc = getOutputData(info, wbh, data_vector[pos], &output, &bytes_out);
			if (rc < 0) {
				data_vector[pos]->handled |= info->writernum;
				debug(D_ERR,"writeNetThread: record is NULL!!!!\n");
			} else {
				pthread_mutex_unlock(&data_mutex);
				debug(D_VALUES,"writeNetThread: writing record of size %d at address %lx\n",bytes_out, output);
				rc = send(fd, (const void *) output, (size_t) bytes_out, 0);
				debug(D_VALUES,"writeNetThread: Return from send is %d\n",rc);
				pthread_mutex_lock(&data_mutex);
				delete [] output;
				if ((rc == 0) || ((rc == -1) && (!is_recoverable(errno)))) {
					debug(D_WARN,"writeNetThread: Closing socket to recover from send\n",rc);
					close(fd);
					fd = -1;
				} else {
					if (rc != (int)bytes_out) {
						debug(D_ERR,"writeNetThread: Error writing data record #%d, (bytes_out = %u)\n",pos,bytes_out);
						usleep(100);
					} else {
						data_vector[pos]->handled |= info->writernum;
						debug(D_VALUES,"writeNetThread: wrote record of size %d with result %d\n",bytes_out,rc);
					}
				}
			}
// No work found.
		}

		if ((pos < 0) || (fd < 0)) {
			if (fd < 0)
				debug(D_ERR,"writeNetThread: Network connection not available\n");
			debug(D_FLOW,"writeNetThread: Waiting for more data to become available\n");
			pthread_cond_wait(&data_available,&data_mutex);
			debug(D_FLOW,"writeNetThread: writernum %d has been notified of impending data\n",info->writernum);
		}

		pthread_testcancel();
		pthread_mutex_unlock(&data_mutex);
		pthread_cleanup_pop(0);
	}
	return(NULL);
}

void sigpipe_handler (int signum)
{
	debug(D_WARN,"sigpipe_handler: received broken PIPE (Network?)\n");
	return;
}

int open_net(string hostname, unsigned int port) {

	struct hostent *he;
	struct sockaddr_in addr;

	int fd;
	int size = 2048;
	int rc;

	debug(D_FLOW,"open_net: Entering...\n");
	he = gethostbyname(hostname.c_str());
	if (!he) {
		debug(D_CHECK, "open_ntrip: Unknown hostname %s\n",hostname.c_str());
		return (-1);
	}

	fd = socket(AF_INET, SOCK_STREAM, 0);

	struct sigaction sa;
	sa.sa_handler = sigpipe_handler;
	sa.sa_flags = 0;
	sigset_t pipe_sig;
	sigemptyset(&pipe_sig);
	sigaddset(&pipe_sig, SIGPIPE);
	pthread_sigmask(SIG_UNBLOCK, &pipe_sig, NULL);
	sigaction(SIGPIPE, &sa, NULL);


	if (fd < 0) {
		debug(D_CHECK,"open_ntrip:Could not create socket\n");
		return(-1);
	}

	memset((char *) &addr, 0x00, sizeof(addr));
	memcpy(&addr.sin_addr, he->h_addr, (size_t)he->h_length);
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);

	/* connect to caster */
	debug(D_VALUES, "caster output: host = %s, port = %d\n", inet_ntoa(addr.sin_addr), port );
	rc = connect(fd, (struct sockaddr *) &addr, sizeof(addr));

	debug(D_VALUES, "Connect response is %d\n",rc);
	if (rc == -1) return(-1);

	/* set socket buffer size */
	setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (const char *) &size, sizeof(const char *));
	int on = 1;
	setsockopt (fd, IPPROTO_TCP, TCP_NODELAY, &on, sizeof (on));
//	fcntl(fd,F_SETFL,O_NONBLOCK);

	return(fd);
}

static const std::string base64_chars = 
             "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
             "abcdefghijklmnopqrstuvwxyz"
             "0123456789+/";


/* does not buffer overrun, but breaks directly after an error */
/* returns the number of required bytes */
std::string base64_encode(unsigned char const* bytes_to_encode, unsigned int in_len) {
	std::string ret;
	int i = 0;
	int j = 0;
	unsigned char char_array_3[3];
	unsigned char char_array_4[4];

	while (in_len--) {
		char_array_3[i++] = *(bytes_to_encode++);
		if (i == 3) {
			char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
			char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
			char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
			char_array_4[3] = char_array_3[2] & 0x3f;

			for(i = 0; (i <4) ; i++)
				ret += base64_chars[char_array_4[i]];
			i = 0;
		}
	}

	if (i) {
		for(j = i; j < 3; j++)
		char_array_3[j] = '\0';

		char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
		char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
		char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
		char_array_4[3] = char_array_3[2] & 0x3f;

		for (j = 0; (j < i + 1); j++)
		ret += base64_chars[char_array_4[j]];

		while((i++ < 3))
		ret += '=';
	}
	return ret;
}

int ntrip_mount(Serial *pSerial, string mount, string user, string password) {

	string sBuffer;
	char buffer[1024];
	int i;
//	int j;

	debug(D_FLOW,"ntrip_mount: Entering...\n");
	if (! pSerial->isConnected()) {
		debug(D_VALUES, "ntrip_mount: Serial not connected\n");
		return(FAILURE);
	}
	/* send message to caster */
	sBuffer  = "GET /" + mount + " HTTP/1.0\r\n";
	sBuffer += "User-Agent: NTRIP Sharc/1.9\r\n";
	string auth = user + ":" + password;
	sBuffer += "Authorization: Basic ";
	sBuffer += base64_encode(reinterpret_cast<const unsigned char*>(auth.c_str()), auth.length());
	sBuffer += "\r\n\r\n";
	debug(D_VALUES, "Sending: %s\n", sBuffer.c_str());
	i = pSerial->Write((void *) sBuffer.c_str(), sBuffer.length());
	if ( (i < 0) || ( (unsigned int) i != sBuffer.length()) ) {
		debug(D_CHECK,"ntrip_mount: ERROR: could get data from \n");
		return (FAILURE);
	}
	// check caster's response
	i = pSerial->Readline(buffer, sizeof(buffer), 1000);
	buffer[i] = '\0';
	debug(D_VALUES,"open_ntrip: Found caster response %s\n", buffer);
	sBuffer = buffer;
	debug(D_VALUES,"open_ntrip: Found caster response %s(string)\n", sBuffer.c_str());
	if (sBuffer.compare(0,10,"ICY 200 OK")) {
		debug(D_CHECK,"ntrip_mount: ERROR: %s\n",sBuffer.c_str());
		return(FAILURE);
	}

	return(SUCCESS);
}

int ntrip_login(int fd, string mount, string password) {

	string sBuffer;
	char buffer[1024];

	debug(D_FLOW,"ntrip_login: Entering...\n");
	if (fd <= 0) {
		debug(D_VALUES, "ntrip_login: file descriptor is <= 0\n");
		return(1);
	}
	/* send message to caster */
	sBuffer  = "SOURCE " + password + " /" + mount + "\r\n";
	sBuffer += "Source-Agent: NTRIP Sharc\r\n\r\n";
	int nBufferBytes = sBuffer.length();
	debug(D_VALUES, "Sending: %s\n", sBuffer.c_str());
	if ((send(fd, sBuffer.c_str(), (size_t)nBufferBytes, 0)) != nBufferBytes) {
		debug(D_CHECK,"ntrip_login: ERROR: could not send to caster\n");
		close(fd);
		fd = -1;
		return (1);
	}
	// check caster's response
	nBufferBytes = recv(fd, buffer, sizeof(buffer), 0);
	buffer[nBufferBytes] = '\0';
	debug(D_VALUES,"open_ntrip: Found caster response %s\n", buffer);
	sBuffer = buffer;
	debug(D_VALUES,"open_ntrip: Found caster response %s(string)\n", sBuffer.c_str());
	if ( sBuffer != "OK\r\n" && strncmp(sBuffer.c_str(), "ICY 200 OK", 10) != 0 ) {
		debug(D_CHECK,"ntrip_login: ERROR: %s\n",sBuffer.c_str());
		close(fd);
		fd = -1;
		return(1);
	}

	return(0);
}

//void *writePipeThread(void *argin) {
//
//
//}

// fh is file handle (FILE *) or (binexHandle *)
int getSaveFile(void **fh, string *dir, string *filename, time_t rec_time, string dir_template, string file_template, bool BINEX)
{

	gpstime mytime;
	char fn_buf[1025];
	char dd_buf[1025];
	string tmpfn;
	string tmpdd;
	debug(D_FLOW,"getSaveFile: Entering dir_template=%s  file_template=%s\n",dir_template.c_str(), file_template.c_str());
	debug(D_FLOW,"getSaveFile: Entering dir=%s  filename=%s\n",dir->c_str(), filename->c_str());

	mytime.set_time(rec_time);
	mytime.chartime(file_template.c_str(),fn_buf,1024);
	mytime.chartime(dir_template.c_str(),dd_buf,1024);
	tmpfn = fn_buf;
	tmpdd = dd_buf;

	if (tmpdd == "") tmpdd = ".";
	if (tmpfn == "") tmpfn = "sharc.out";

// check data directory

	bool dirmatch = TRUE;
	if (tmpdd != *dir) {
		if (make_path(tmpdd) == 0) {
			*dir = tmpdd;
			debug(D_VALUES,"getSaveFile: dir is now %s\n",dir->c_str());
		} else {
			*dir = ".";
			debug(D_ERR,"Unable to create dir %s\n",dir->c_str());
		}
		dirmatch = FALSE;
	}

	// If the filename matches, do nothing and return(0);
	if ((tmpfn == *filename) && dirmatch) {
		debug(D_FLOW,"getSaveFile: Exiting... matched file %s\n",filename->c_str());
		return(0);
	}

	*filename = tmpfn;
	string output;
	output.append(dir->c_str());
	if (dir->at(dir->length()-1) != '/') output.append("/");
	output.append(filename->c_str());
	debug(D_FLOW,"getSaveFile: Opening %s\n",output.c_str());
	int outfd = open(output.c_str(),O_WRONLY | O_CREAT | O_APPEND | O_NONBLOCK, 0660);

	if (*fh != NULL) {
		debug(D_FLOW,"getSaveFile: Closing File handle\n");
		(BINEX) ? free_binex((binexHandle *)*fh) : fclose((FILE *) *fh);
	}
	if (BINEX) {
		debug(D_FLOW,"getSaveFile: Opening binex file handle\n");
		*fh = new_binex(outfd,"w");
	} else {
		debug(D_FLOW,"getSaveFile: Opening standard file handle\n");
		*fh = fdopen(outfd, "w");
	}
	if (*fh == NULL) {
		*filename = "";
		close(outfd);
		debug(D_ERR,"getSaveFile: Cannot open output file %s\n",output.c_str());
		return(-1);
	}
	debug(D_FLOW,"getSaveFile: Exiting\n");
	return(0);
}

int cache_init() {
	int rtn;
	if((rtn = pthread_mutex_init(&data_mutex,NULL)) != 0) {
		debug(D_ERR,"cache_init: cannot init mutex\n");
		return(FAILURE);
	}
	if((rtn = pthread_cond_init(&data_available,NULL)) != 0) {
		debug(D_ERR,"cache_init: cannot init conditional variable\n");
		return(FAILURE);
	}
	return(SUCCESS);
}

int cache_add(data_record_t *rec) {

	debug(D_FLOW,"cache_add: Entering...\n");
	cache_removehandled();
	pthread_mutex_lock(&data_mutex);
	data_vector.push_back(rec);
	debug(D_VALUES,"cache_add: cache size is now %d\n",data_vector.size());
	if (data_vector.size() > MAXCACHESIZE) {
		delete [] data_vector[0]->record;
		data_vector[0]->record = NULL;
		delete data_vector[0];
		data_vector[0] = NULL;
		data_vector.erase(data_vector.begin());
	}
	pthread_mutex_unlock(&data_mutex);
	pthread_cond_broadcast(&data_available);

	return(SUCCESS);
}

int cache_removehandled() {
	pthread_mutex_lock(&data_mutex);

	debug(D_FLOW,"cache_removehandled: Entering...\n");
	debug(D_STD,"cache: %04d writers: \r",data_vector.size());
	if (data_vector.empty()) {
		pthread_mutex_unlock(&data_mutex);
		return(SUCCESS);
	}
// build numwriters mask
	unsigned short mask = 0;
	for (int j = 0; j < numwriters; j++) {
		mask |= (1 << j);
	}
	for (unsigned int i = 0; i < data_vector.size(); i++) {
		debug(D_FLOW,"removehandled: comparing %d == %d\n",data_vector[i]->handled, mask);
		if (data_vector[i]->handled == mask) {
			delete [] data_vector[i]->record;
			data_vector[i]->record = NULL;
#ifdef RTCM_ENABLE
			if (data_vector[i]->rtcm != NULL) {
				delete [] data_vector[i]->rtcm;
				data_vector[i]->rtcm = NULL;
			}
			if (data_vector[i]->rtcm3 != NULL) {
				delete [] data_vector[i]->rtcm3;
				data_vector[i]->rtcm3 = NULL;
			}
#endif

//			delete data_vector[i]->data;
//			data_vector[i]->data = NULL;
			delete data_vector[i];
			data_vector[i] = NULL;
			data_vector.erase(data_vector.begin()+i);
			--i;
		}
	}
	pthread_mutex_unlock(&data_mutex);
//	debug(D_STD,"\n");
	return(SUCCESS);
}

int read_rawStream(int fd, void *buf, size_t max, unsigned int * record_len, timeout_t timeout)
{
	size_t len = 0;
	int timer;
	int rc;
	unsigned char *mybuf;
	*record_len = 0;

	mybuf = (unsigned char *) buf;

	while (*record_len == 0) {
		timer = timeout;
		while(len < max)
		{
			timer = isCharAvail(fd, timer);
			if (timer == 0) break;

			rc = read(fd, mybuf, 1);
			if (rc == -1) break;

			len++; mybuf++;
		}
		*mybuf = 0;
		*record_len = len;
	}
	return(SUCCESS);
}


#endif
