/* 
#    $Id: ttylock.cpp,v 1.2 2018/02/23 04:34:13 gps Exp $
#
#    Copyright (C) 2000 US Geological Survey
#    Written by Keith F. Stark
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <iostream>
#include <fstream>

#include "ttylock.h"
using namespace std;

char pid_name[MAXNAMLEN];
char dev_name[MAXNAMLEN];

/* Callable functions, part of library. */
#ifdef __cplusplus
extern "C" {
#endif
int ttylock( char * );
int ttyunlock( char * );
int ttyrelock( char *, int );
#ifdef __cplusplus
};
#endif

/* Internal functions */
int __get_names( char * );
int __write_lock( char *, int );
int __check_lock( char *, int );
#ifdef USE_ALT
void __dumb_op( char *name );
#endif

/* This is not the best implementation of the debugging
 * paradign. */
//#ifdef MY_DEBUG
//extern int debug;
//#else
int debug;
//#endif

/* 
 * This function creates a "LCK..<pid>" file and writes the current
 * pid to it in HDB format.  It then links the "LCK..<dev>" and the
 * "LCK..<alt>" to it.
 *
 * On success, a 0 is returned.
 *
 * If an active pid is found in an existing lockfile, that pid is
 * returned.  If the link fails for any other reason, all the
 * preceding successful links are unlinked and a -1 is returned.
 *
 * The "LCK..<pid>" name is unlinked before leaving whether
 * successful or failure.
 *
 * The argument, *dev, may be a full path name to the device driver
 * or just the name of the file.  That is, either "/dev/ttyS0" or
 * "ttyS0" will suffice.
 */

int
ttylock( char *dev )
{
	int old_mask, ret, pid;

	if( __get_names( dev ) < 0 ) {
		cerr << "locks: Error in setting names." << endl;
		return( -1 );
	}
	if ( debug > 2 )
		cerr << "locks: >" << dev << "< in >" << dev_name << "<\n";

	/*
	 * Set permissions to allow lockfiles to word read/write.
	 */
	old_mask = umask( 0 );
	ret = 0;
	pid = getpid();
	/* 
	 * This is fairly safe.  It is unlikely that our pid
	 * will have been recycled.
	 */
	if ( __write_lock( pid_name, pid ) < 0 ) {
		umask( old_mask );
		return( -1 );
	}
	umask( old_mask );
	if ( debug > 2 )
		cerr << "lock: pid >" << pid_name << "<, dev >" << dev_name << "<, " << pid << endl;

	if ( link( pid_name, dev_name ) < 0 ) {
		if(( ret = __check_lock( dev_name, getpid())) >= 0 )
			/* This is our own lockfile or a process exists */
			goto quit;
		if ( debug > 2 )
			cerr << "ttylock: fail link pid to dev with pid " << ret << endl;
	}
	if ( debug > 2 )
		cerr << "lock: " << pid_name << " linked to " << dev_name << endl;
#ifdef USE_ALT
	if ( debug > 2 )
		cerr << "pid >" << pid_name << "<, alt >" << alt_name << "<, " << pid << endl;
	if ( link( pid_name, alt_name ) < 0 ) {
		if(( ret = __check_lock( alt_name, getpid())) >= 0 ) {
			cerr << "alt link identity" << endl;
			/* This is our own lockfile or a process exists */
			goto quit;
			if ( debug > 2 )
				cerr << "ttylock: link pid to alt with pid " << ret << endl;
		}
		if ( debug > 2 ) {
			cerr << "lock: " << pid_name << " linked to " << alt_name << endl;
		}
	}

#endif	/* USE_ALT */
 quit:
	/*
	 * Don't leave a stray file which will be orphaned
	 * if we crash before we can remove it.
	 */
	if( ret < 0 ) {
		unlink( dev_name );
#ifdef USE_ALT
		unlink( alt_name );
#endif

	}
	unlink( pid_name );
	return( ret );
}

int
ttyunlock( char *dev )
{
	if( __get_names( dev ) < 0 ) {
		cerr << "unlock: Error in setting names." << endl;
		return( -1 );
	}
	if ( debug > 2 )
		cerr << "Unlocking >" << dev << "< in >" << dev_name << "<" << endl;
	if( __check_lock( dev_name, 0 ) == 0 ) unlink( dev_name );
#ifdef USE_ALT
	if( __check_lock( alt_name, 0 ) == 0 ) unlink( alt_name );
#endif
	return( 0 );
}

/* 
 * This is for use by a parent process which wants to hand off
 * the lockfiles to a child after forking.
 * 
 * It returns -1 on failure or 0 on success.
 */
int
ttyrelock( char *dev, int chld_pid )
{
	int my_pid, ret;

	if( __get_names( dev ) < 0 ) {
		cerr << "relock: Error in setting names." << endl;
		return( -1 );
	}
	my_pid = getpid();
	if ( debug > 2 ) {
		cerr << "relock: >" << dev << "< in >" << dev_name << endl;
		cerr << "\tmy " << my_pid << ", chld " << chld_pid << endl;
	}
	if (( ret = __check_lock( dev_name, my_pid )) == 0 )
		__write_lock( dev_name, chld_pid );
	else
		return( -1 );
#ifdef USE_ALT
	/*
	 * If 'alt_name' is hard linked to 'dev_name', this step
	 * is redunant.  If we can make this assumption, this
	 * step can be deleted.
	 */
	if (( ret = __check_lock( alt_name, my_pid )) == 0 )
		__write_lock( alt_name, chld_pid );
	else
		if( ret != chld_pid ) return( -1 );
#endif	/* USE_ALT */
	return( 0 );
}

/*
 * Open a file and write our pid into it.
 */
int
__write_lock( char *dev, int my_pid )
{
	FILE *lock;

	if (( lock = fopen( dev, "w" )) == NULL ) {
		cerr << "lock: Unable to open " << dev << endl;
		return( -1 );
	}
	fprintf( lock, "%10d\n", my_pid );
	fclose( lock );
	return( 0 );
}

/*
 * Read the pid stored in an existing lockfile.  Then determine if
 * it our own or the child pid if relocking.
 * 
 * Return:
 *    -1 - in case of failure of any step.
 *     0 - it is our own or the process is dead.
 *   fpid - the pid found in the lockfile.
 */
int
__check_lock( char *file, int wrk_pid )
{
	char file_pid[12];
	int fd;
	int fpid;

	if (( fd = open( file, O_RDONLY )) < 0 ) {
		/* Shut up extraneous messages. */
		if ( debug > 2 )
			cerr << "check: Unable to open" << file << endl;
		return( -1 );
	}
	if ( read( fd, &file_pid, sizeof( file_pid ))
		 != sizeof( file_pid ) -1 ) {
		/* Shut up extraneous messages. */
		if ( debug > 2 )
			cerr << "check: Unable to read from" << file << endl;
		return( -1 );
	}
	close( fd );
	/*
	 * If this is a child pid, we have already rewritten
	 * a linked lockfile.
	 */
	file_pid[10] = '\0';
	if ( debug > 2 )
		cerr << "lock " << file << ", file " << file_pid << ", wrk " << wrk_pid << ", get " << getpid() << endl;
	fpid = atoi( file_pid );
	if( fpid == wrk_pid )
		return( 0 );

	if ( fpid != getpid()) {
		errno = 0;

		if ( kill( fpid, 0 ) == 0 ) {
			/*
			 * This would be a nice message if the 'getty' didn't
			 * repeat it so often.
			 */
			if ( debug > 2 )
				cerr << "check: " << file << " locked by pid " << fpid << endl;
			return ( fpid );
		}
		/* The process doesn't exist */
		unlink( file );
	}
	return( 0 );
}

/*
 * Generate the names for the lockfiles
 * These will be of three or four types.
 *
 * 1. /path/LCK..<pid>
 * 2. /path/LCK..<dev>
 * 3. /path/LCK..<alt>	(Used if old cua* devices are used)
 *
 * The <pid> is removed before leaving the calling function.
 *
 * These names will be placed in the global char arrays:
 *    pid_name[]
 *    dev_name[]
 *    alt_name[]  if USE_ALT is defined
 *    
 * Return:
 *   -1 - A failure occurred in some step.
 *    0 - Success
 */
int
__get_names( char *dev )
{
	DIR *dev_dir;
	struct stat mystat;
	struct dirent *dir_p;
	int is_sym;
	char *p, *q, *s;
	char lock_name[MAXNAMLEN];

	/*
	 * Mark the names as undefined.
	 */
	pid_name[0] = '\0';
	dev_name[0] = '\0';
#ifdef USE_ALT
	alt_name[0] = '\0';
#endif	/* USE_ALT */

	/* 
	 * Scan past the path portion of 'dev' if it was included in
	 * 'dev'.
	 */
	p = dev;
	q = DEV_PATH;
	s = p + MAXNAMLEN - 1;
	while ( *p == *q && p < s ) p++, q++;

	/*
	 * Step past a trailing '/'.
	 */
	if( !*q && *p == '/' ) p++;

	/*
	 * Have we gone too far?
	 */
	if ( p >= s || !*p ) {
		cerr << "get_names: The the device name is faulty. '" << dev << endl;
		return( -1 );
	}

	/* 
	 * Now attach to the path.
	 */
	strncpy( lock_name, p, MAXNAMLEN - 1 );
	lock_name[MAXNAMLEN-1] = '\0';
	sprintf( dev_name, "%s%s", DEV_PATH, lock_name );

	is_sym = 0;
	if ( lstat( dev_name, &mystat) < 0 ) {
		if ( debug > 2 )
			cerr << "get_names: Unable to lstat " << dev_name << endl;
	}
	else is_sym = S_ISLNK( mystat.st_mode );

	if ( stat( dev_name, &mystat ) < 0 ) {
		if ( debug > 2 )
			cerr << "get_names: Unable to stat" << dev_name << endl;
		return( -1 );
	}

	/*
	 * Reuse this array
	 */
	dev_name[0] = '\0';

	/* 
	 * Create the "LCK..<pid>" file name
	 */
	sprintf( pid_name, "%sLCK..%d", LOCK_PATH, (int)getpid() );

	/*
	 * Open the '/dev' directory so that we can search for the
	 * name of the device.  This also works if the 'dev' name
	 * given to us was a symbolic link.
	 */
	if( is_sym ) {
		if(( dev_dir = opendir( DEV_PATH )) == NULL ) {
			cerr << "get_names: Unable to opendir " << DEV_PATH << endl;
		}
		else {
			while(( dir_p = readdir( dev_dir )) != NULL ) {
					strncpy( lock_name, dir_p->d_name, MAXNAMLEN - 1 );
					lock_name[MAXNAMLEN-1] = '\0';
					break;
			}
			closedir( dev_dir );
		}
	}
	sprintf( dev_name, "%sLCK..%s", LOCK_PATH, lock_name );
#ifdef USE_ALT
	__dumb_op( lock_name );
#endif
	return( 0 );
}

#ifdef USE_ALT
void
__dumb_op( char *name )
{
	char *p, *q;

	p = name;
	q = TTYS;
	while( *p == *q ) p++, q++;
	if( *q != '\0' ) {
		p = name;
		q = ALT;
		while( *p == *q ) p++, q++;
		if( *q != '\0' ) {
			cerr << "dumb_op: No compatible device name found.\n" );
			return;
		}
		q = TTYS;
	}
	else q = ALT;
	sprintf( alt_name, "%sLCK..%s%s", LOCK_PATH, q, p );
}
#endif	/* USE_ALT */
